#ifndef _IMAGEATTRIBUTES_H_INCLUDED_
#define _IMAGEATTRIBUTES_H_INCLUDED_
#include <Base/Common/BIASpragmaStart.hh>

#include <bias_config.h>
#include <Base/Common/W32Compat.hh>
#include <math.h>

#include <Base/Debug/Debug.hh>
#include "ExifTag.hh"
#include <iostream>
#include <vector>
#include <string>
#include <map>



/// return code for success:
#define BIAS_IMAGEATTRIBUTES_SUCCESS 0

#define D_EXIF 1

#ifdef BIAS_HAVE_IMAGEMAGICKLIB
#  include <Magick++.h>
#endif
/// \todo Avoid inclusion of Magick++ header and use forward declaration here to be able to output runtime error on usage (JW).


// Begin namespace BIAS
namespace BIAS {

    /** @class ImageAttributes
    @todo Add Image Attributes write functionality to write meta data directly to images EXIF Tags, comments etc. 
    @brief contains all atribute info and values of e.g. a file.
    @author Jan Woetzel 10/2004 */
    class BIASImageBase_EXPORT ImageAttributes 
        : public std::vector< ExifTag >, public BIAS::Debug 
    {
    public:
        ~ImageAttributes(){};
        ImageAttributes(){};

        /** @brief main entry routine to read all EXIF tags from the header of image file.
        @return  BIAS_IMAGEATTRIBUTES_SUCCESS if OK.
        @param filename image to ping and read header from.
        @author Jan Woetzel */
        int Read(const std::string & filename);

        /** @return  BIAS_IMAGEATTRIBUTES_SUCCESS if reading was OK, negative number in case of error.
        @param filename image file to read attributes form header.
        @author Jan Woetzel */
        int ReadMagickPP(const std::string & filename);

#ifdef BIAS_HAVE_IMAGEMAGICKLIB
        /** @brief direct interface to Magick::IMage to be able to read the XIF tags of an already opened image. 
        @return  BIAS_IMAGEATTRIBUTES_SUCCESS;BIAS_EXIFTAGTABLE_SUCCESS 
        @param filename image to ping and read header from.
        @author Jan Woetzel */
        int ReadMagickPP(Magick::Image & img);
#endif


        /** initialize me vector with all known EXIF tags */
        void InitExifTags();

        /** print the content of this to os 
        @param printAll true prints all tags, false only those whose value is not empty.
        */
        std::ostream& Print(std::ostream& os=std::cout,
            const bool printAll=false) const;

        /** @brief compute focallength in meters 
        @return 0:ok, <0:information missing, >0:missing information guessed
        @author koeser 07/2005 */
        int GetFocalLengthMeter(double &fm);

        /** @brief compute focallength in pixels 
        @return 0:ok, <0:information missing, >0:missing information guessed
        @author koeser 07/2005 */
        int GetFocalLengthXPixel(double &fp);

        /** @brief compute focallength in pixels 
        @return 0:ok, <0:information missing, >0:missing information guessed
        @author koeser 07/2005 */
        int GetFocalLengthYPixel(double &fp);

        /** @brief compute pixel size in meter 
        @return 0:ok, <0:information missing, >0:missing information guessed
        @author koeser 07/2005 */
        int GetPixelSizeXMeter(double &p);

        /** @brief compute pixel size in meter 
        @return 0:ok, <0:information missing, >0:missing information guessed
        @author koeser 07/2005 */
        int GetPixelSizeYMeter(double &p);

        /** @brief compute field of view in degree 
        @return 0:ok, <0:information missing, >0:missing information guessed
        @author koeser 07/2005 */
        int GetFieldOfViewDegree(double &x, double& y);

        /** @brief get width and height of image 
        @return 0:ok, <0:information missing, >0:missing information guessed
        @author koeser 07/2005 */
        int GetImageDimensions(int& width, int& height);

        /** @brief look up principal point 
        @return 0:ok, <0:information missing, >0:missing information guessed
        @author koeser 07/2005 */
        int GetPrincipalPoint(double& x, double& y);

        /** @brief look up camera make and model
        @return 0:ok, <0:information missing, >0:missing information guessed
        @author koeser 07/2005 */
        int GetHardwareName(std::string& Make, std::string& Model);

    protected:
        bool GetString_(const std::string& key, std::string& value);
        bool GetRational_(const std::string& key, double& value);
        bool GetShort_(const std::string& key, int& value);

    private: // avoid DLL export by private
        std::map<std::string, unsigned int> mapValidTags_;

    }; // end of class
} // end namespace


// ------------------

/** @brief output operator for ExifTable 
@author Jan Woetzel */
BIASImageBase_EXPORT std::ostream& operator<<(std::ostream& os, 
                         const BIAS::ImageAttributes &attr);


#include <Base/Common/BIASpragmaEnd.hh>

#endif
