/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ImageConvert.hh"

#include <iostream>

#include "ToPlanar.cpp"
#include "ToInterleaved.cpp"
#include "ToHSV.cpp"
#include "ToHSL.cpp"

using namespace std;
using namespace BIAS;


#ifdef BIAS_HAVE_DV
dv_decoder_t *ImageConvert::DVDecoder_ = NULL;
#endif



//#include <Base/Common/BIASpragma.hh>


//
//
// macros - replacement for CALLFUNC, complete rewrite Jan Woetzel 05/2006
//

/// one of the elemenatry switch case statements
/// @author Jan Woetzel
#define SWIC(biasTypeEnum, cpptype, callfunc, src, dst) \
case ImageBase::biasTypeEnum :\
{\
  Image< cpptype > typedImg, destImg;\
  const ImageBase* psrc =  &src;                      \
  typedImg.StealImage(*const_cast<ImageBase*>(psrc)); \
  callfunc(typedImg, destImg);\
  const_cast<ImageBase*>(psrc)->StealImage(typedImg);  \
  dst.StealImage(destImg);\
}\
  break;


// use only those case whose instances are compiled
// create one macro for each type, use dby CALLFUNC macro 
// @author Jan Woetzel 05/2006 

// always compiled 
#define CASE_UCHAR(func, src, dst)   SWIC(ST_unsignedchar, unsigned char, func, src, dst)
#define CASE_FLOAT(func, src, dst)   SWIC(ST_float,         float,         func, src, dst)

// compiled iff option enabled vias bias_config.h
#ifdef BUILD_IMAGE_CHAR
#  define CASE_CHAR(func, src, dst)  SWIC(ST_char, char, func, src, dst)
#else 
#  define CASE_CHAR(func, src, dst)  
#endif

#ifdef BUILD_IMAGE_SHORT
#  define CASE_SHORT(func, src, dst)  SWIC(ST_shortint, short int, func, src, dst)
#else 
#  define CASE_SHORT(func, src, dst)  
#endif

#ifdef BUILD_IMAGE_USHORT
#  define CASE_USHORT(func, src, dst)  SWIC(ST_unsignedshortint, unsigned short int, func, src, dst)
#else 
#  define CASE_USHORT(func, src, dst)  
#endif

#ifdef BUILD_IMAGE_INT
#  define CASE_INT(func, src, dst)  SWIC(ST_int, int, func, src, dst)
#else 
#  define CASE_INT(func, src, dst)  
#endif

#ifdef BUILD_IMAGE_UINT
#  define CASE_UINT(func, src, dst)  SWIC(ST_unsignedint, unsigned int, func, src, dst)
#else 
#  define CASE_UINT(func, src, dst)  
#endif

#ifdef BUILD_IMAGE_DOUBLE
#  define CASE_DOUBLE(func, src, dst)  SWIC(ST_double, double, func, src, dst)
#else 
#  define CASE_DOUBLE(func, src, dst)  
#endif



///
/// creates all switch cases given a function
/// no pretty implementation but backward compatible to old macro JW
#define CALLFUNC(afunc,source,dest) \
  CASE_UCHAR (afunc,source,dest) \
  CASE_FLOAT (afunc,source,dest) \
  CASE_CHAR  (afunc,source,dest) \
  CASE_SHORT (afunc,source,dest) \
  CASE_USHORT(afunc,source,dest) \
  CASE_INT   (afunc,source,dest) \
  CASE_UINT  (afunc,source,dest) \
  CASE_DOUBLE(afunc,source,dest) 

//
// end of macros
//


// ----- implementation -----

int ImageConvert::Convert(ImageBase &source, ImageBase &dest, 
                          enum ImageBase::EColorModel targetColorModel,
                          bool bPlanar) 
{
  if ( (dest.GetStorageType()!=ImageBase::ST_invalid) &&
    (dest.GetStorageType()!=source.GetStorageType()) ) {
      BIASERR("color model conversion cannot convert image from StorageType "
        <<source.GetStorageType()<<" to an image with StorageType "
        <<dest.GetStorageType());
      return -1;
    }

    if (source==dest) { ///////////// call the In-Place function
      if (source.GetColorModel()!=targetColorModel)
        switch (targetColorModel) 
      {
        case ImageBase::CM_Grey:          
          switch (source.GetStorageType()) 
          {
            //CALLFUNC_IP(IP_ToGrey,source);
            // >> start of former CALLFUNC_IP jw
          case ImageBase::ST_unsignedchar:
            {
              Image<unsigned char> ucImg;
              ucImg.StealImage(source);
              IP_ToGrey(ucImg);
              source.StealImage(ucImg);
            }
            break;
          case ImageBase::ST_float:
            {
              Image<float> fImg;
              fImg.StealImage(source);
              IP_ToGrey(fImg);
              source.StealImage(fImg);
            }
            break;
#ifdef BUILD_IMAGE_CHAR
          case ImageBase::ST_char:
            {
              Image<char> cImg;
              cImg.StealImage(source);
              IP_ToGrey(cImg);
              source.StealImage(cImg);
            }
            break;
#endif // BUILD_IMAGE_CHAR
#ifdef BUILD_IMAGE_SHORT
          case ImageBase::ST_shortint:
            {
              Image<short int> sImg;
              sImg.StealImage(source);
              IP_ToGrey(sImg);
              source.StealImage(sImg);
            }
            break;
#endif // BUILD_IMAGE_SHORT
#ifdef BUILD_IMAGE_USHORT
          case ImageBase::ST_unsignedshortint:
            {
              Image<unsigned short int> usImg;
              usImg.StealImage(source);
              IP_ToGrey(usImg);
              source.StealImage(usImg);
            }
            break;
#endif // BUILD_IMAGE_USHORT
#ifdef BUILD_IMAGE_INT
          case ImageBase::ST_int:
            {
              Image<int> iImg;
              iImg.StealImage(source);
              IP_ToGrey(iImg);
              source.StealImage(iImg);
            }
            break;
#endif // BUILD_IMAGE_INT
#ifdef BUILD_IMAGE_UINT
          case ImageBase::ST_unsignedint:
            {
              Image<unsigned int> uiImg;
              uiImg.StealImage(source);
              IP_ToGrey(uiImg);
              source.StealImage(uiImg);
            }
            break;
#endif // BUILD_IMAGE_UINT
#ifdef BUILD_IMAGE_DOUBLE
          case ImageBase::ST_double:
            {
              Image<double> dImg;
              dImg.StealImage(source);
              IP_ToGrey(dImg);
              source.StealImage(dImg);
            }
            break;
#endif // BUILD_IMAGE_DOUBLE
            // << end of former CALLFUNC_IP jw

          default:
            BIASERR("invalid storage type in source image");
            return -2;
          }          
          break;    
        default: 
          BIASERR("no In-Place-conversion for sourceColorModel:"<<
            source.GetColorModel()<<
            " to targetColorModel:"<<targetColorModel<< endl<<
            "can only In-Place-convert to CM_Grey");
          return -3;
      }

      if (source.IsPlanar() != bPlanar) {
        ImageBase destImgBase;      
        if (source.IsPlanar()) {
          switch (source.GetStorageType()) {
            CALLFUNC(ToInterleaved,source,destImgBase);        
        default:
          BIASERR("invalid storage type in source image");
          return -2;
          }          
        } else { // it is interleaved
          switch (source.GetStorageType()) {
            CALLFUNC(ToPlanar,source,destImgBase);        
        default:
          BIASERR("invalid storage type in source image");
          return -2;
          }   
        }
        source.StealImage(destImgBase);
      }

    } else { ///////////////////////// not In-Place

      switch (targetColorModel) 
      {
      case ImageBase::CM_Grey:          
        switch (source.GetStorageType()) 
        {
          CALLFUNC(ToGrey,source,dest);
        default:
          BIASERR("invalid storage type in source image");
          return -2;
        }
        break;
      case ImageBase::CM_RGB:  
        switch (source.GetStorageType()) 
        {    
          CALLFUNC(ToRGB,source,dest);
        default:
          BIASERR("invalid storage type in source image");
          return -2;
        }  
        break;
      case ImageBase::CM_RGBA:  
        switch (source.GetStorageType()) 
        {
          CALLFUNC(ToRGBA,source,dest);
        default:
          BIASERR("invalid storage type in source image");
          return -2;
        }  
        break;         
      case ImageBase::CM_HSV:  
        switch (source.GetStorageType()) 
        {
          CALLFUNC(ToHSV,source,dest);
        default:
          BIASERR("invalid storage type in source image");
          return -2;
        }
        break;
      case ImageBase::CM_HSL:  
        switch (source.GetStorageType()) 
        {
          CALLFUNC(ToHSL,source,dest);
        default:
          BIASERR("invalid storage type in source image");
          return -2;
        }
        break;
      case ImageBase::CM_hsL:  
        switch (source.GetStorageType()) 
        {
          CALLFUNC(TohsL,source,dest);          
        default:
          BIASERR("invalid storage type in source image");
          return -2;
        }
        break;

      default:
        BIASERR("no conversion available to targetColorModel:"<<
          targetColorModel);
        return -3;
      }
    }
    if (source.IsPlanar() != bPlanar) {
      ImageBase destImgBase;      
      if (source.IsPlanar()) {
        switch (source.GetStorageType()) {
          CALLFUNC(ToInterleaved,dest,destImgBase);        
      default:
        BIASERR("invalid storage type in source image");
        return -2;
        }          
      } else { // it is interleaved
        switch (source.GetStorageType()) {
          CALLFUNC(ToPlanar,dest,destImgBase);        
      default:
        BIASERR("invalid storage type in source image");
        return -2;
        }   
      }
      dest.StealImage(destImgBase);
    }
    return 0;
}

/**
Sony DFW-X700 : 
r = y + 1.4022 v got
b = y + 1.7710 u 
g = y - 0.3475 u - 0.7144 v got
*/
#define YUV2RGB(y, u, v, r, g, b)\
  r = y + ( ((v << 11) + (v << 10) - (v << 8) + (v << 6) - (v << 3) ) >> 11);\
  g = y - ( ((v<<10) + (v<<9) - (v<<6) -  (v<<3) ) >> 11 ) - \
  ( ((u << 9) + (u << 7) + (u << 6) + (u << 3) + u) >> 11 );\
  b = y + ( ((u << 12) - (u << 9) + (u << 5)+(u<<4)) >> 11);\
  r = r < 0 ? 0 : r;\
  g = g < 0 ? 0 : g;\
  b = b < 0 ? 0 : b;\
  r = r > 255 ? 255 : r;\
  g = g > 255 ? 255 : g;\
  b = b > 255 ? 255 : b




////////////////////
// ToGrey Section //
////////////////////
template<class StorageType>
int BIAS::ImageConvert::ToGrey(const Image<StorageType>& source,
                               Image<StorageType>& dest)
{  
  BIASGDOUT(D_CONVERT," ImageConvert::ToGrey()");
  int res = -1;

  if ( dest.GetPixelCount() != source.GetPixelCount() ||
       dest.GetWidth() != source.GetWidth() ||
       dest.GetHeight() != source.GetHeight() ){
    BIASGDOUT(D_CONVERT,
              "ToGrey() should be called with an image of same size");
    dest.Release();
  }
  if ((dest.GetChannelCount() != 1) && 
      (dest.GetImageData() != source.GetImageData())){
    BIASGDOUT(D_CONVERT,"ToGrey() should be called with an 1 channel image");
    dest.Release();
  }

  if (dest.IsEmpty()) {
    dest.Init(source.GetWidth(), source.GetHeight(),1,source.GetStorageType());
  }

  switch (source.GetColorModel()) {
  case ImageBase::CM_Grey:
    dest = source;
    res = 0;
    break;
  case ImageBase::CM_DV:
#ifdef BIAS_HAVE_DV
    res = DVToRGB_(source,dest);
    IP_ToGrey(dest);
#else
    BIASERR("CM_DV not supported because BIAS_HAVE_DV not defined!");
    BIASABORT;
#endif
    break;

  case ImageBase::CM_RGB:
  case ImageBase::CM_BGR:
    res = RGBToGrey_(source,dest);
    break;
  case ImageBase::CM_UYVY422:
  case ImageBase::CM_YUYV422:
    res = YUV422ToGrey_(source,dest);
    break;
  case ImageBase::CM_YUV420P:
    res = YUV420PToGrey_(source,dest);
    break;
  case ImageBase::CM_YUV411:
    res = YUV411ToGrey_(source,dest);
    break;
  case ImageBase::CM_Bayer_RGGB:
  case ImageBase::CM_Bayer_GBRG:
  case ImageBase::CM_Bayer_GRBG:
  case ImageBase::CM_Bayer_BGGR:
    // PtGrToGrey_ is not correct
    res = BayerToGrey_(source,dest);
    break;
  case ImageBase::CM_HSL:
  case ImageBase::CM_hsL:
    res = HSLToGrey_(source,dest);
    break;
  default:
    BIASERR("ImageConvert::ToGrey(): unknown color model:"<<
            source.GetColorModel());
    //BIASBREAK; // do NOT abort, here.
    return -1;
  }
  dest.ChannelCount_=1;
  dest.WidthStep_=dest.Width_*dest.Depth_;
  dest.SetColorModel(ImageBase::CM_Grey);
  // needed for inplace conversion
  dest.UpdateImageDataArrayFromImageData_();
  return res;
}


template<class StorageType>
int BIAS::ImageConvert::GetChannel(Image<StorageType>& image, int channel)
{
  BIASGDOUT(D_CONVERT,"GetChannel()");
  register StorageType *sink = image.GetImageData();
  register StorageType *src  = image.GetImageData();

  int channelCount = image.GetChannelCount();

  if (channelCount < channel) {
    BIASERR("Error: Image has not enough channels.");
    return -1;
  }
  
  // in place not necessary , as ToRGB() created new dist already
  for (unsigned long int i = 0; i < image.GetPixelCount(); i++) {
    for (int j = 0; j < channelCount; j++) {
      sink[j] = src[channel];
    }
    src += channelCount;
    sink+= channelCount;
  }
  return 0;
}

int BIAS::ImageConvert::GetChannel(ImageBase& image, int channel)
{
  BIASGDOUT(D_CONVERT,"GetChannel()");
  register unsigned char *sink = (unsigned char*)image.GetImageData();
  register unsigned char *src  = (unsigned char*)image.GetImageData();

  int channelCount = image.GetChannelCount();

  if (channelCount < channel) {
    BIASERR("Error: Image has not enough channels.");
    return -1;
  }

  int depth = image.GetDepth();
  
  for (unsigned long int i = 0; i < image.GetPixelCount(); i++) {
    for (int j = 0; j < channelCount; j++) {
      for (int k=0; k<depth; k++) {
        sink[j*depth+k] = src[channel*depth+k];
      }
    }
    src += channelCount*depth;
    sink+= channelCount*depth;
  }
  return 0;
}


////////////////////
// ToGrey Section //
////////////////////
template<class StorageType>
int BIAS::ImageConvert::GetChannel(const Image<StorageType>& source,
				   Image<StorageType>& dest,
				   const unsigned int channel)
{ 
  BIASASSERT(source.GetStorageType() == dest.GetStorageType());

    if (dest.GetImageData() == source.GetImageData()){
    BIASERR("GetChannel(src,dst,channel) should not be called with"
	    " the same images. In place does not work, currently.");
    dest.Release();
  } else if ((dest.GetWidth() != source.GetWidth() ||
       dest.GetHeight() != source.GetHeight() ))
  {
    BIASWARN("GetChannel(src,dst,channel) should be called"
            "with an image of same size.\n"
            "src: "<<source.GetWidth()<<"x"<<source.GetHeight()<<"\n"
            "dst: "<<dest.GetWidth()<<"x"<<dest.GetHeight() );
    dest.Release();
  } else if ((dest.GetChannelCount() != 1)
    && (dest.GetImageData() != source.GetImageData()) )
  {
    BIASWARN("GetChannel(src,dst,channel) should be called "
      "with a 1 channel dest image but has "
      <<dest.GetChannelCount()<<" channels" );
    dest.Release();
  };
  
  if (channel >= source.GetChannelCount()){
    BIASERR("GetChannel(src,dst,channel) - src doesn't have enough channels");
    return -1;
  }

  if (dest.IsEmpty()) {
    dest.Init(source.GetWidth(), source.GetHeight(),1,source.GetStorageType());
    dest.SetColorModel( ImageBase::CM_Grey );
  }

  StorageType *s=source.GetImageData();
  StorageType *d=dest.GetImageData();

  if (source.IsPlanar()){
    s += channel*source.GetWidth()*source.GetHeight(); // start offset
    // TODO use memcpy instead
    for (unsigned int i=0; i < dest.GetWidth()*dest.GetHeight(); i++){
      d[0]=s[0];
      d++; 
      s++;  
    }
  } else {
    // interleaved
    const int step = source.GetChannelCount(); // in units of StorageType 
    s += channel;
    for (unsigned int i=0; i < dest.GetWidth()*dest.GetHeight(); i++){
      *d = *s;
      d++;
      s+=step;
    }
  }

  dest.ChannelCount_=1;
  dest.WidthStep_=dest.Width_*dest.Depth_;
  dest.SetColorModel(ImageBase::CM_Grey);

  return 0;
}


int BIAS::ImageConvert::ToGrey(const ImageBase& source,
                               ImageBase& dest)
{  
  int res = 0;
  switch (source.GetStorageType()){
  case ImageBase::ST_unsignedchar:
    {
      BIASASSERT(dest.GetStorageType() == ImageBase::ST_unsignedchar);
      Image<unsigned char> s(source);
      Image<unsigned char> d;
      res = ToGrey(s, d);
      dest.StealImage(d);
    }
    break;
#ifdef BUILD_IMAGE_USHORT
  case ImageBase::ST_unsignedshortint:
    {
      BIASASSERT(dest.GetStorageType() == ImageBase::ST_unsignedshortint);
      Image<unsigned short> s(source);
      Image<unsigned short> d;
      res = ToGrey(s, d);
      dest.StealImage(d);
    }
    break;
#endif // BUILD_IMAGE_USHORT
  default:
    BIASERR("unfinished code");
    BIASABORT;
    break;
  }
  return res;
}

template <class StorageType>
int BIAS::ImageConvert::RGBToGrey_(const Image<StorageType>& source,
                                   Image<StorageType>& dest)
{
  BIASGDOUT(D_CONVERT,"ImageConvert::RGBToGrey_()");
  StorageType *p=dest.GetImageData();
  StorageType *r, *g, *b;

  if (source.IsPlanar()){
    switch (source.GetColorModel()){
    case ImageBase::CM_RGB:
      r=source.GetImageData();
      g=r+(source.GetWidth()*source.GetHeight());
      b=r+(2*source.GetWidth()*source.GetHeight());
      break;
    case ImageBase::CM_BGR:
      b=source.GetImageData();
      g=b+(source.GetWidth()*source.GetHeight());
      r=b+(2*source.GetWidth()*source.GetHeight());
      break;
    default:
      BIASERR("RGBToGrey_ must be called with RGB or BGR ColorModel_");
      return -1; 
    }
    for (unsigned int i=0; i < dest.GetWidth()*dest.GetHeight(); i++){
      // casting using rint is necessray for integer data types to deal 
      // with compile precision
      // this is solved by introducing a templated Cast function
      p[0]=Cast<StorageType>(0.299*(double)(r[0])+
                             0.587*(double)(g[0])+
                             0.114*(double)(b[0]));
      p++; r++; g++; b++; 
    }
  } else {
    switch (source.GetColorModel()) {
    case ImageBase::CM_RGB:
      r=source.GetImageData();
      g=r+1;
      b=r+2;
      break;
    case ImageBase::CM_BGR:
      b=source.GetImageData();
      g=b+1;
      r=b+2;
      break;
    default:
      BIASERR("RGBToGrey_ must be called with RGB or BGR ColorModel_");
      return -1; 
    }
    for (unsigned int i=0; i < dest.GetWidth()*dest.GetHeight(); i++){
      p[0]=Cast<StorageType>(0.299*(double)(r[0])+
                             0.587*(double)(g[0])+
                             0.114*(double)(b[0]));
      //cout << "rgb: "<<(int)r[0]<<" "<<(int)g[0]<<" "<<(int)b[0]<<"\tgrey: "<<(int)p[0]<<endl;
      p++; r+=3; g+=3; b+=3; 
    }
  }
  return 0;
}


template <class StorageType>
int BIAS::ImageConvert::YUV422ToGrey_(const Image<StorageType> &source,
                                      Image<StorageType>& dest)
{

  register StorageType *y = source.GetImageData();
  StorageType *sink = dest.GetImageData();
  register StorageType *sinkp;
  bool IdenticalImages = false;
  BIASGDOUT(D_CONVERT,"ImageConvert::YUV422ToGrey_()");

  if (y == sink) {
    BIASGDOUT(D_MI_CONVGREY,"identical images");
    IdenticalImages = true;
    sink = new StorageType[source.GetWidth() * source.GetHeight()];
  } 

  switch (source.GetColorModel()) {
    case ImageBase::CM_YUYV422:
      break;
    case ImageBase::CM_UYVY422:
      y++;
      break;
    default:
      BIASERR("ImageConvert::YUV422ToGrey_(): unsupported colormodel");
      return -1;
  }

  if (source.IsPlanar()){
    memcpy(sink, y, dest.GetPixelCount());
  } else {
    sinkp = sink;
    for (register unsigned long int i=0; i < dest.GetPixelCount(); i++){
      sinkp[0]=y[0];
      sinkp++; 
      y+=2; 
      // cerr << i << " " << endl;
    }
  }

  if (IdenticalImages) {
    dest.ReleaseImageDataPointer();
    dest.ChannelCount_=1;
    dest.RedirectImageDataPointer(sink);
  }
  return 0;
}



///////////////////////
// IP_ToGrey Section //
/////////////////////// 
template <class StorageType>
int BIAS::ImageConvert::IP_ToGrey(Image<StorageType> &img)
{ 
  int res = -1;

  BIASGDOUT(D_CONVERT,"ImageConvert::IP_ToGrey_()");

  //   if (sizeof(StorageType) != 1){
  //     BIASERR("IP_ToGrey makes only sense with <char> and <unsigned char> images");
  //     return -1;
  //   }

  switch (img.GetColorModel()) {
  case ImageBase::CM_Grey://ImageBase::ImageBase::CM_Grey:
    BIASERR("already in grey");
    break;
  case ImageBase::CM_BGR:
    res = IP_RGBToGrey_(img);
    break;
  case ImageBase::CM_RGB:
    res = IP_RGBToGrey_(img);
    break;
  case ImageBase::CM_YUYV422:  
  case ImageBase::CM_UYVY422:
    res = IP_YUV422ToGrey_(img);
    break;
  default:
    BIASERR("Image::IP_ToGrey(): unknown color model");
    res = -1;
    break;
  }
  if (res >=0) {
    img.ChannelCount_=1;
    img.WidthStep_ = img.Depth_ * img.Width_;
    img.SetColorModel(ImageBase::CM_Grey);
    img.UpdateImageDataArrayFromImageData_();
  }    
  return res;
}



template <class StorageType>
int BIAS::ImageConvert::IP_RGBToGrey_(Image<StorageType> &img)
{
  StorageType *p=img.GetImageData();
  StorageType *r, *g, *b;
  if (img.IsPlanar()){
    switch (img.GetColorModel()){
    case ImageBase::CM_RGB:
      r=img.GetImageData();
      g=r+(img.GetWidth()*img.GetHeight());
      b=r+(2*img.GetWidth()*img.GetHeight());
      break;
    case ImageBase::CM_BGR:
      b=img.GetImageData();
      g=b+(img.GetWidth()*img.GetHeight());
      r=b+(2*img.GetWidth()*img.GetHeight());
      break;
    default:
      BIASERR("IP_RGBToGrey_ must be called with RGB or BGR ColorModel_");
      return -1; 
    }
    for (unsigned int i=0; i < img.GetWidth()*img.GetHeight(); i++){
      p[0]=Cast<StorageType>(0.299*(double)(r[0])+
                             0.587*(double)(g[0])+
                             0.114*(double)(b[0]));
      p++; r++; g++; b++; 
    }
  } else {
    switch (img.GetColorModel()) {
    case ImageBase::CM_RGB:
      r=img.GetImageData();
      g=r+1;
      b=r+2;
      break;
    case ImageBase::CM_BGR:
      b=img.GetImageData();
      g=b+1;
      r=b+2;
      break;
    default:
      BIASERR("IP_RGBToGrey_ must be called with RGB or BGR ColorModel_");
      return -1; 
    }

    for (unsigned long int i=0; i < img.GetPixelCount(); i++){
      p[0]=Cast<StorageType>(0.299*(double)(r[0])+
                             0.587*(double)(g[0])+
                             0.114*(double)(b[0]));
      p++; r+=3; g+=3; b+=3; 
    }
  }
  return 0;
}





template <class StorageType>
int BIAS::ImageConvert::IP_YUV422ToGrey_(Image<StorageType> &img)
{
  StorageType *y;

  BIASGDOUT(D_CONVERT,"ImageConvert::IP_YUV422ToGrey_()");


  if (img.IsPlanar()){
    switch (img.GetColorModel()) {
    case ImageBase::CM_YUYV422:
      img.ChannelCount_=1;
      break;
    default:
      BIASERR("ImageConvert::IP_YUV422ToGrey_()  color model not yet implemented");
      break;
    }
  } else {
    StorageType *dest=img.GetImageData();
    switch (img.GetColorModel()) {
    case ImageBase::CM_YUYV422:
      y = img.GetImageData();
      break;
    case ImageBase::CM_UYVY422:
      y = img.GetImageData() + 1;
      break;
    default:
      BIASERR("ImageConvert::IP_YUV422ToGrey_(): unsupported colormodel");
      return -1;
    }
    for (unsigned int i=0; i < img.GetWidth()*img.GetHeight(); i++){
      dest[0]=y[0];
      dest++; 
      y+=2; 
    }
  }
  return 0;

}

////////////////////
//  ToRGBA Section //
////////////////////


template <class StorageType>
int BIAS::ImageConvert::ToRGBA(const Image<StorageType>& source,
                               Image<StorageType>& dest)
{  
  // only implmented for interleaved, out-of-place conversion so far
 
  BIASWARNONCE("inefficient code: -> RGBA conversion");
  int res = 0;
  Image<StorageType> rgb;
  if (source.GetColorModel()!=ImageBase::CM_RGB) {
    res = ToRGB(source, rgb);
    if (res!=0) return res;
  } else {
    rgb = source;
  }

  BIASASSERT(rgb.IsInterleaved());  
  // add alphachannel
  if (!dest.IsEmpty()) dest.Release();
  dest.Init(source.GetWidth(), source.GetHeight(), 4);
  dest.SetColorModel(ImageBase::CM_RGBA);
  
  StorageType *pD = rgb.GetImageData();
  StorageType *pDA = dest.GetImageData();
  for (unsigned int y = 0; y < dest.GetHeight(); y++) {
    for (unsigned int x = 0; x < dest.GetWidth(); x++) {
      *pDA++ = *pD++;
      *pDA++ = *pD++;
      *pDA++ = *pD++;
      *pDA++ = 0;
    }
  }
  return res;
}
////////////////////
//  ToRGB Section //
////////////////////


template <class StorageType>
int BIAS::ImageConvert::ToRGB(const Image<StorageType>& source,
                              Image<StorageType>& dest)
{  
  int res = -1;
  bool inplace=(source.GetImageData()==dest.GetImageData());
  Image<StorageType>* tmpdest;
  if (inplace) 
    tmpdest = new Image<StorageType>;
  else 
    tmpdest = &dest;


  BIASGDOUT(D_CONVERT,"ImageConvert::ToRGB()");
  if (!tmpdest->IsEmpty()) {
    if (tmpdest->GetChannelCount() != 3 || 
      tmpdest->GetWidth()!=source.GetWidth() || 
      tmpdest->GetHeight() != source.GetHeight() ){
        tmpdest->Release();
      }
  }

  if (tmpdest->IsEmpty()) {
    if (source.GetColorModel() == ImageBase::CM_DV) {
#ifdef BIAS_HAVE_DV
      tmpdest->Init(720,576, 3,source.GetStorageType()); // PAL
#else
      BIASERR("BIAS_HAVE_DV not defined!!!");
      BIASABORT;
#endif
    } else {
      tmpdest->Init(source.GetWidth(), source.GetHeight(), 3,
        source.GetStorageType());
    }
  }

  switch (source.GetColorModel()) {
  case ImageBase::CM_Grey:
    res = GreyToRGB_(source,*tmpdest);
    break;
  case ImageBase::CM_BGR:
    res = BGRToRGB_(source,*tmpdest);
    break;
  case ImageBase::CM_RGB:
    memcpy(tmpdest->GetImageData(), source.GetImageData(), 
      sizeof(StorageType) * 3 * source.GetPixelCount());
    res = 0;
    break;
  case ImageBase::CM_UYVY422:
    res = YUV422ToRGB_(source,*tmpdest);
    break;
  case ImageBase::CM_YUYV422:
    res = YUV422ToRGB_(source,*tmpdest);
    break;
  case ImageBase::CM_YUV420P:
    res = YUV420PToRGB_(source,*tmpdest);
    break;
  case ImageBase::CM_DV:
#ifdef BIAS_HAVE_DV
    res = DVToRGB_(source,*tmpdest);
#else
    BIASERR("BIAS_HAVE_DV not defined!!");
    BIASABORT;
#endif
    break;
  case ImageBase::CM_Bayer_RGGB:
  case ImageBase::CM_Bayer_GBRG:
  case ImageBase::CM_Bayer_GRBG:
  case ImageBase::CM_Bayer_BGGR:
    res = BayerToRGB_(source,*tmpdest);
    break;
  case ImageBase::CM_RGBA:
    res = RGBAToRGB_(source,*tmpdest);
    break;
  case ImageBase::CM_BGRA:
    res = BGRAToRGB_(source,*tmpdest);
    break;
  case ImageBase::CM_PGR_XB3_F7M3_GBRG:  // tri-image
    {
      // workaround template implementation: 
      Image<unsigned char>* ps_uc = 
        const_cast<Image<unsigned char>*>
        (dynamic_cast< const Image<unsigned char>* >(&source));
      Image<unsigned char>* pd_uc = 
        dynamic_cast< Image<unsigned char>* >(&dest);
      if ((ps_uc!=NULL) && (pd_uc!=NULL)){
        // use unsigned char special implemetation
        BIAS::Image<unsigned char> tmp3;
        res = DeinterleaveHorizontal(*ps_uc, tmp3);
        if (res!=0){
          BIASERR("could not Deinterleave the 3 Bayer patterns.");
          return res;
        }
        // 3 horizontal Bayer pattern images, now.
        // convert from Bayer pattern to interpolated RGB
        // BayerToRGB_
        res = BIAS::ImageConvert::ToRGB(tmp3, *pd_uc);
        if (res!=0){
          BIASERR("could not Deinterlace");
          return res;
        }
      } else {
        BIASERR("CM_PGR_XB3_F7M3_GBRG implemented only for unsigned char.");
        return -1;
      }
    }
    break;

  default:
    BIASERR("ImageConvert::ToRGB(): unknown color model:"<<
      source.GetColorModel());
    break;
  }
  tmpdest->SetColorModel(ImageBase::CM_RGB);

  if (inplace){
    dest = *tmpdest;
    delete tmpdest;
  }

  return res;
}



template <class StorageType>
int BIAS::ImageConvert::BGRToRGB_(const Image<StorageType>& source,
                                  Image<StorageType>& dest)
{  

  BIASGDOUT(D_CONVERT,"BGRToRGB_()");
  StorageType *sink = dest.GetImageData();
  StorageType *src = source.GetImageData();

  // in place
  if(src == sink){
    StorageType tmp;
    for (unsigned long int i=0;i<dest.GetPixelCount(); i++) {
      tmp = src[2];
      sink[2] = src[0];
      sink[0] = tmp;
      sink+=3; 
      src+=3;
    } 
  } else { 
    for (unsigned long int i=0;i<dest.GetPixelCount(); i++) {
      sink[2] = src[0];
      sink[1] = src[1];
      sink[0] = src[2];
      sink+=3; 
      src+=3;
    }
  }
  return 0;
}

template <class StorageType>
int BIAS::ImageConvert::RGBAToRGB_(const Image<StorageType>& source,
                                   Image<StorageType>& dest)
{  
  BIASGDOUT(D_CONVERT,"RGBAToRGB()");
  register StorageType *sink = dest.GetImageData();
  register StorageType *src = source.GetImageData();
  // in place not necessary , as ToRGB() created new dist already
  for (unsigned long int i=0;i<dest.GetPixelCount(); i++) {
    sink[0] = src[0];
    sink[1] = src[1];
    sink[2] = src[2];
    sink+=3; 
    src+=4;
  }

  return 0;
}


                  
template <class StorageType>
int BIAS::ImageConvert::BGRAToRGB_(const Image<StorageType>& source,
                                   Image<StorageType>& dest)
{  
  BIASGDOUT(D_CONVERT,"RGBAToRGB()");
  register StorageType *sink = dest.GetImageData();
  register StorageType *src = source.GetImageData();
  // in place not necessary , as ToRGB() created new dist already
  for (unsigned long int i=0;i<dest.GetPixelCount(); i++) {
    sink[2] = src[0];
    sink[1] = src[1];
    sink[0] = src[2];
    sink+=3; 
    src+=4;
  }

  return 0;
}


template <class StorageType>
int BIAS::ImageConvert::YUV422ToRGB_(const Image<StorageType>& source,
                                     Image<StorageType>& dest)
{  


  BIASGDOUT(D_CONVERT,"YUV422ToRGB_()");
  StorageType *sink;
  bool IdenticalImages = false;
  StorageType *sourcedata = source.GetImageData();


  if (dest.GetImageData() == sourcedata) {
    IdenticalImages = true;
    sink = new StorageType[source.GetWidth() * source.GetHeight() * 3];
  }
  else
    sink = dest.GetImageData();


#define FASTYUV

#ifndef FASTYUV   
  float y,u = 0.0,v=0.0;
  float r,g,b;

  int sinkoffset =0;
  int yoffset = 0;
  int uvoffset = 1;



  if (source.ColorModel_ == ImageBase::CM_UYVY422) {
    yoffset = 1;
    uvoffset = 0;
  }


  for (int j = 0; j < (source.GetWidth() * source.GetHeight()) ; j++)
  { 
    //        y = 1.164 *(float) sourcedata[j*2 + yoffset]-16;
    y = (float)sourcedata[j*2 + yoffset];
    if ( j%2 == 0){
      u=  (float)sourcedata[j*2+ uvoffset]- 128;
    }
    else {
      v = (float) sourcedata[j*2+ uvoffset ]- 128;
    }

    b = rint(y              + 1.77 * u );
    g = rint(y - 0.714 * v  - 0.345 * u );
    r = rint(y + 1.4 * v );
    //        b = rint(y              + 2.018 * v );
    //        g = rint(y - 0.813 * u  - 0.391 * v );
    //        r = rint(y + 1.596 * u );
    r = r >255 ? 255: r;
    g = g >255 ? 255: g;
    b = b >255 ? 255: b;

    r =  r <0 ? 0: r;
    g =  g <0 ? 0: g;
    b =  b <0 ? 0: b;


    sink[sinkoffset++ ] = (unsigned char) r;
    sink[sinkoffset++] = (unsigned char) g;
    sink[sinkoffset++] = (unsigned char) b;
  }

#else
  unsigned int i,j;
  register int y0,y1,u,v,r,g,b;
  if (source.GetColorModel() == ImageBase::CM_UYVY422) {
    for (i = 0, j = 0; i < (source.GetWidth()*source.GetHeight() << 1); i += 4, j += 6)     {
      u = (unsigned char) sourcedata[i + 0] - 128;
      y0 = (unsigned char) sourcedata[i + 1];
      v = (unsigned char) sourcedata[i + 2] - 128;
      y1 = (unsigned char) sourcedata[i + 3];
      YUV2RGB (y0, u, v, r, g, b);
      sink[j + 0] = StorageType(r);
      sink[j + 1] = StorageType(g);
      sink[j + 2] = StorageType(b);
      YUV2RGB (y1, u, v, r, g, b);
      sink[j + 3] = StorageType(r);
      sink[j + 4] = StorageType(g);
      sink[j + 5] = StorageType(b);
    }

  } 
  else {
    for (i = 0, j = 0; i < (dest.GetWidth()*dest.GetHeight() << 1); i += 4, j += 6)     {
      u = (unsigned char) sourcedata[i + 1] - 128;
      y0 = (unsigned char) sourcedata[i + 0];
      v = (unsigned char) sourcedata[i + 3] - 128;
      y1 = (unsigned char) sourcedata[i + 2];
      YUV2RGB (y0, u, v, r, g, b);
      sink[j + 0] = StorageType(r);
      sink[j + 1] = StorageType(g);
      sink[j + 2] = StorageType(b);
      YUV2RGB (y1, u, v, r, g, b);
      sink[j + 3] = StorageType(r);
      sink[j + 4] = StorageType(g);
      sink[j + 5] = StorageType(b);
    }
  }
#endif
  if (IdenticalImages) {
    dest.ReleaseImageDataPointer();
    dest.ChannelCount_=3;
    dest.RedirectImageDataPointer(sink);
  }

  return 0;
}


template <class StorageType>
int BIAS::ImageConvert::GreyToRGB_(const Image<StorageType>& source,
                                   Image<StorageType>& dest)
{  
  BIASGDOUT(D_CONVERT,"GreyToRGB_()");

  register StorageType* src_start = source.GetImageData();
  register StorageType* src_stop  = src_start + dest.GetPixelCount();
  register StorageType* pDest = dest.GetImageData();

  for (;src_start < src_stop; src_start++) {      
    *pDest++ = *src_start;
    *pDest++ = *src_start;
    *pDest++ = *src_start;
  }

  return 0;
}

#ifdef BIAS_HAVE_DV
template <class StorageType>
int BIAS::ImageConvert::DVToRGB_(const Image<StorageType>& source,
                                 Image<StorageType>& dest,
                                 unsigned char quality )
{
  BIASERR("DVToRGB_ only makes sense with unsigned char images");
  return -1;
}


namespace BIAS {
template <>
int ImageConvert::DVToRGB_(const Image<unsigned char>& source,
                                 Image<unsigned char>& dest,
                                 unsigned char quality )
{
  BIASGDOUT(D_CONVERT,"DVToRGB_()");

  if (!dest.IsEmpty())
    if (dest.GetWidth() != 720 || dest.GetHeight() != 576 
      || dest.GetChannelCount() != 3)
      dest.Release();


  if (dest.IsEmpty()) dest.Init(720, 576,3);


  unsigned char *sink = dest.GetImageData();
  //        4 ? from  avi2pics
  unsigned char *pixels[3];
  int pitches[3];

  pixels[0] = sink;  //target 
  pixels[1] = NULL;
  pixels[2] = NULL;

  pitches[0] = 720 * 3;
  pitches[1] = 0;
  pitches[2] = 0;
  if (DVDecoder_ == NULL) DVDecoder_ = dv_decoder_new(0,0,0);
  if (DVDecoder_ == NULL) BIASERR("dv_decoder_new() failed.");
  //       else cout <<"Got new DVDecoder"<<endl;
  //     dv_init(0,0);// new style for 0.9.8
  switch (quality) {
  case 0:  DVDecoder_->quality = DV_QUALITY_FASTEST ;
    break;
  case 1: DVDecoder_->quality = DV_QUALITY_FASTEST | DV_QUALITY_AC_1;
    break;
  case 2: DVDecoder_->quality = DV_QUALITY_COLOR;
    break;
  case 3: DVDecoder_->quality = DV_QUALITY_COLOR | DV_QUALITY_AC_1;
    break;
  case 4: DVDecoder_->quality = DV_QUALITY_BEST;
    break;
  }
  //  
  //  decoder->quality = DV_QUALITY_COLOR | DV_QUALITY_DC;
  //decoder->quality = DV_QUALITY_COLOR | DV_QUALITY_AC_1;
  //  decoder->quality = DV_QUALITY_BEST;

  dv_parse_header(DVDecoder_, source.GetImageData());
  dv_decode_full_frame(DVDecoder_, source.GetImageData(), 
    e_dv_color_rgb, pixels, pitches);

  return 0;
}

} // namespace BIAS

#endif // of BIAS_HAVE_DV





template <class StorageType>
int BIAS::ImageConvert::BayerToGrey_(const Image<StorageType>& source ,
                                     Image<StorageType>& dest)   
{
  BIASGDOUT(D_CONVERT," Image::BayerToGreySlow_()");

  StorageType *sink = dest.GetImageData() ;
  StorageType *src = source.GetImageData() ;


  int Mask[8];
  StorageType R=0,G=0,B=0; 
  unsigned int srcskip;
  srcskip = source.GetWidth() ;

  Mask[0] = - (int(source.GetWidth()) +1);
  Mask[1] = - int(source.GetWidth());
  Mask[2] = - (int(source.GetWidth()) -1);
  Mask[3] = -1;
  Mask[4] = +1;
  Mask[5] = int(source.GetWidth()) -1;
  Mask[6] = + int(source.GetWidth());
  Mask[7] = int(source.GetWidth()) +1;

  unsigned int xoffset = 0;
  unsigned int yoffset = 0;
  switch (source.GetColorModel()) {
  case ImageBase::CM_Bayer_RGGB:
    // first pixel in first line 
    R = *src;
    G = (*(src+srcskip)+*(src+1))/2;
    B = *(src+srcskip+1);
    break; 
  case ImageBase::CM_Bayer_GBRG: 
    yoffset = 1; 
    // first pixel in first line 
    G = *src;
    R = *(src+srcskip);
    B = *(src+1);
    break;
  case ImageBase::CM_Bayer_GRBG: 
    xoffset = 1; 
    // first pixel in first line 
    G = *src;
    B = *(src+srcskip);
    R = *(src+1);
    break;
  case ImageBase::CM_Bayer_BGGR: 
    xoffset = 1;yoffset = 1; 
    // first pixel in first line 
    B = *src;
    G = (*(src+srcskip)+*(src+1))/2;
    R = *(src+srcskip+1);
    break;
  default:; }
  *sink++ = Cast<StorageType>(0.299*(double)R+0.587*(double)G+0.114*(double)B);
  src++;
  // first line
  if (0 == yoffset) {
    for (unsigned int x=1; x<(unsigned int)dest.GetWidth()-1; x++){
      if (x%2 == xoffset) {  // staying on Red
        R = *src;
        G = ( *(src+Mask[3])+ *(src+Mask[4]) + *(src+Mask[6]) ) /3;
        B = ( *(src+Mask[5]) + *(src+Mask[7])) /2;
      } 
      else { // staying on Green
        R =  ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
        G = *src;
        B = (*(src+Mask[6]) ) ;
      }
      *sink++ = Cast<StorageType>(0.299*(double)R+
                                  0.587*(double)G+0.114*(double)B);
      src++;
    }
  }
  else {
    for (unsigned int x=1; x<(unsigned int)dest.GetWidth()-1; x++){
      if (x%2 == xoffset) {  // staying on Green
        R =  ( *(src+Mask[6]) ) ;
        G = *src;
        B = ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
      }
      else { // staying on Blue
        R = (  *(src+Mask[5]) + *(src+Mask[7])) /2;
        G = ( *(src+Mask[3]) + *(src+Mask[4]) + *(src+Mask[6]) ) /3;
        B = *src;
      }
      *sink++ = Cast<StorageType>(0.299*(double)R+0.587*(double)G+0.114*(double)B);
      src++;      
    }
  }
  // last pixel in first line
  if (0 == yoffset) {   
    if ((dest.GetWidth()-1)%2 == xoffset) {  // staying on Red
      R = *src;
      G = ( *(src+Mask[3])+  *(src+Mask[6]) ) /2;
      B =  *(src+Mask[5])  ;
    } 
    else { // staying on Green
      R =  *(src+Mask[3]);
      G = *src;
      B = (*(src+Mask[6]) ) ;
    }
  }
  else {
    if ((dest.GetWidth()-1)%2 == xoffset) {  // staying on Green
      R =  ( *(src+Mask[6]) ) ;
      G = *src;
      B =  *(src+Mask[3]) ;
    }
    else { // staying on Blue
      R =  *(src+Mask[5]);
      G = ( *(src+Mask[3]) + *(src+Mask[6]) ) /2;
      B = *src;
    }
  }
  *sink++ = Cast<StorageType>(0.299*(double)R+0.587*(double)G+0.114*(double)B);
  src++;      


  //loop
  for (unsigned int y=1; y<(unsigned int)source.GetHeight()-1; y++) {
    // first pixel in line
    if (y%2 == yoffset) 
      if (0%2 == xoffset) {  // staying on Red
        R = *src;
        G = (*(src+Mask[1])  + *(src+Mask[4]) + *(src+Mask[6]) ) /3;
        B = ( *(src+Mask[2]) + *(src+Mask[7])) /2;
      } else { // staying on Green
        R =  *(src+Mask[4]);
        G = *src;
        B = ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
      }
    else
      if (0 == xoffset) {  // staying on Green
        R =  ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
        G = *src;
        B = *(src+Mask[4]);
      }
      else { // staying on Blue
        R = ( *(src+Mask[0]) + *(src+Mask[7])) /2;
        G = (*(src+Mask[1])  + *(src+Mask[4]) + *(src+Mask[6]) ) /3;
        B = *src;
      }
      *sink++ = Cast<StorageType>(0.299*(double)R+0.587*(double)G+0.114*(double)B);
      src++;      
      // main part of line
      for (unsigned int x=1; x<(unsigned int)source.GetWidth()-1; x++) {
        if (y%2 == yoffset) 
          if (x%2 == xoffset) {  // staying on Red
            R = *src;
            G = (*(src+Mask[1]) + *(src+Mask[3]) 
              + *(src+Mask[4]) + *(src+Mask[6]) ) /4;
            B = ( *(src+Mask[0]) + *(src+Mask[2]) 
              + *(src+Mask[5]) + *(src+Mask[7])) /4;
          } else { // staying on Green
            R =  ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
            G = *src;
            B = ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
          }
        else
          if (x%2 == xoffset) {  // staying on Green
            R =  ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
            G = *src;
            B = ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
          }
          else { // staying on Blue
            R = ( *(src+Mask[0]) + *(src+Mask[2]) 
              + *(src+Mask[5]) + *(src+Mask[7])) /4;
            G = (*(src+Mask[1]) + *(src+Mask[3]) 
              + *(src+Mask[4]) + *(src+Mask[6]) ) /4;
            B = *src;
          }
          *sink++ = Cast<StorageType>(0.299*(double)R+0.587*(double)G+0.114*(double)B);
          src++;      
      }
      // last pixel in line
      if (y%2 == yoffset) 
        if ((srcskip-1)%2 == xoffset) {  // staying on Red
          R = *src;
          G = (*(src+Mask[1]) + *(src+Mask[3]) + *(src+Mask[6]) ) /3;
          B = ( *(src+Mask[0]) + *(src+Mask[5])) /2;
        } else { // staying on Green
          R =  *(src+Mask[3]) ;
          G = *src;
          B = ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
        }
      else
        if ((srcskip-1)%2 == xoffset) {  // staying on Green
          R =  ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
          G = *src;
          B = *(src+Mask[3]);
        }
        else { // staying on Blue
          R = ( *(src+Mask[0]) + *(src+Mask[5])) /2;
          G = (*(src+Mask[1]) + *(src+Mask[3])+ *(src+Mask[6]) ) /3;
          B = *src;
        }
        *sink++ = Cast<StorageType>(0.299*(double)R+0.587*(double)G+0.114*(double)B);
        src++;      


  }
  // first pixel in last line
  if ((source.GetHeight()-1)%2 == yoffset) {   
    if (0 == xoffset) {  // staying on Red
      R = *src;
      G = ( *(src+Mask[1])+  *(src+Mask[4]) ) /2;
      B =  *(src+Mask[2])  ;
    } 
    else { // staying on Green
      R =  *(src+Mask[4]);
      G = *src;
      B = (*(src+Mask[1]) ) ;
    }
  }
  else {
    if (0 == xoffset) {  // staying on Green
      R =  ( *(src+Mask[1]) ) ;
      G = *src;
      B =  *(src+Mask[4]) ;
    }
    else { // staying on Blue
      R =  *(src+Mask[2]);
      G = ( *(src+Mask[1]) + *(src+Mask[4]) ) /2;
      B = *src;
    }
  }
  *sink++ = Cast<StorageType>(0.299*(double)R+0.587*(double)G+0.114*(double)B);
  src++;
  // last line
  if ((source.GetHeight()-1)%2 == yoffset) {
    for (unsigned int x=1; x<(unsigned int)dest.GetWidth()-1; x++){
      if (x%2 == xoffset) {  // staying on Red
        R = *src;
        G = ( *(src+Mask[3])+ *(src+Mask[1]) + *(src+Mask[4]) ) /3;
        B = ( *(src+Mask[0]) + *(src+Mask[2])) /2;
      } 
      else { // staying on Green
        R =  ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
        G = *src;
        B = (*(src+Mask[1]) ) ;
      }
      *sink++ = Cast<StorageType>(0.299*(double)R+0.587*(double)G+0.114*(double)B);
      src++;
    }
  }
  else {
    for (unsigned int x=1; x<(unsigned int)dest.GetWidth()-1; x++){
      if (x%2 == xoffset) {  // staying on Green
        R =  ( *(src+Mask[1]) ) ;
        G = *src;
        B = ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
      }
      else { // staying on Blue
        R = (  *(src+Mask[0]) + *(src+Mask[2])) /2;
        G = ( *(src+Mask[1]) + *(src+Mask[3]) + *(src+Mask[4]) ) /3;
        B = *src;
      }
      *sink++ = Cast<StorageType>(0.299*(double)R+0.587*(double)G+0.114*(double)B);
      src++;      
    }
  }
  // last pixel in last line
  if ((source.GetHeight()-1)%2 == yoffset) {   
    if ((dest.GetWidth()-1)%2 == xoffset) {  // staying on Red
      R = *src;
      G = ( *(src+Mask[3])+  *(src+Mask[1]) ) /2;
      B =  *(src+Mask[0])  ;
    } 
    else { // staying on Green
      R =  *(src+Mask[3]);
      G = *src;
      B = (*(src+Mask[1]) ) ;
    }
  }
  else {
    if ((dest.GetWidth()-1)%2 == xoffset) {  // staying on Green
      R =  ( *(src+Mask[1]) ) ;
      G = *src;
      B =  *(src+Mask[3]) ;
    }
    else { // staying on Blue
      R =  *(src+Mask[0]);
      G = ( *(src+Mask[3]) + *(src+Mask[1]) ) /2;
      B = *src;
    }
  }
  *sink++ = Cast<StorageType>(0.299*(double)R+0.587*(double)G+0.114*(double)B);
  src++;
  return 0; 
}



template <class StorageType>
int BIAS::ImageConvert::BayerToRGB_(const Image<StorageType>& source, 
                                    Image<StorageType>&  dest )   
{
  BIASGDOUT(D_CONVERT," Image::BayerToRGB_()");

  StorageType *sink = dest.GetImageData() ;
  StorageType *src = source.GetImageData() ;

  BIASASSERT(sink!=NULL);
  BIASASSERT(src!=NULL);


  int Mask[8];
  StorageType R=0,G=0,B=0; 
  unsigned int srcskip;
  srcskip = source.GetWidth() ;

  Mask[0] = - (int(source.GetWidth()) +1);
  Mask[1] = - int(source.GetWidth());
  Mask[2] = - (int(source.GetWidth()) -1);
  Mask[3] = -1;
  Mask[4] = +1;
  Mask[5] = int(source.GetWidth()) -1;
  Mask[6] = + int(source.GetWidth());
  Mask[7] = int(source.GetWidth()) +1;

  unsigned int xoffset = 0;
  unsigned int yoffset = 0;
  switch (source.GetColorModel()) {
  case ImageBase::CM_Bayer_RGGB:
    // first pixel in first line 
    R = *src;
    G = (*(src+srcskip)+*(src+1))/2;
    B = *(src+srcskip+1);
    break; 
  case ImageBase::CM_Bayer_GBRG: 
    yoffset = 1; 
    // first pixel in first line 
    G = *src;
    R = *(src+srcskip);
    B = *(src+1);
    break;
  case ImageBase::CM_Bayer_GRBG: 
    xoffset = 1; 
    // first pixel in first line 
    G = *src;
    B = *(src+srcskip);
    R = *(src+1);
    break;
  case ImageBase::CM_Bayer_BGGR: 
    xoffset = 1;yoffset = 1; 
    // first pixel in first line 
    B = *src;
    G = (*(src+srcskip)+*(src+1))/2;
    R = *(src+srcskip+1);
    break;
  default:; }
  sink[0]=R; sink[1]=G; sink[2]=B; 
  sink+=3; src++;
  // first line
  if (0 == yoffset) {
    for (unsigned int x=1; x<(unsigned int)dest.GetWidth()-1; x++){
      if (x%2 == xoffset) {  // staying on Red
        R = *src;
        G = ( *(src+Mask[3])+ *(src+Mask[4]) + *(src+Mask[6]) ) /3;
        B = ( *(src+Mask[5]) + *(src+Mask[7])) /2;
      } 
      else { // staying on Green
        R =  ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
        G = *src;
        B = (*(src+Mask[6]) ) ;
      }
      *sink++ =R;
      *sink++ =G;
      *sink++ =B;
      src++;
    }
  }
  else {
    for (unsigned int x=1; x<(unsigned int)dest.GetWidth()-1; x++){
      if (x%2 == xoffset) {  // staying on Green
        R =  ( *(src+Mask[6]) ) ;
        G = *src;
        B = ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
      }
      else { // staying on Blue
        R = (  *(src+Mask[5]) + *(src+Mask[7])) /2;
        G = ( *(src+Mask[3]) + *(src+Mask[4]) + *(src+Mask[6]) ) /3;
        B = *src;
      }
      *sink++ =R;
      *sink++ =G;
      *sink++ =B;
      src++;      
    }
  }
  // last pixel in first line
  if (0 == yoffset) {   
    if ((dest.GetWidth()-1)%2 == xoffset) {  // staying on Red
      R = *src;
      G = ( *(src+Mask[3])+  *(src+Mask[6]) ) /2;
      B =  *(src+Mask[5])  ;
    } 
    else { // staying on Green
      R =  *(src+Mask[3]);
      G = *src;
      B = (*(src+Mask[6]) ) ;
    }
  }
  else {
    if ((dest.GetWidth()-1)%2 == xoffset) {  // staying on Green
      R =  ( *(src+Mask[6]) ) ;
      G = *src;
      B =  *(src+Mask[3]) ;
    }
    else { // staying on Blue
      R =  *(src+Mask[5]);
      G = ( *(src+Mask[3]) + *(src+Mask[6]) ) /2;
      B = *src;
    }
  }
  *sink++ =R;
  *sink++ =G;
  *sink++ =B;
  src++;      


  //loop
  for (unsigned int y=1; y<(unsigned int)source.GetHeight()-1; y++) {
    // first pixel in line
    if (y%2 == yoffset) 
      if (0%2 == xoffset) {  // staying on Red
        R = *src;
        G = (*(src+Mask[1])  + *(src+Mask[4]) + *(src+Mask[6]) ) /3;
        B = ( *(src+Mask[2]) + *(src+Mask[7])) /2;
      } else { // staying on Green
        R =  *(src+Mask[4]);
        G = *src;
        B = ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
      }
    else
      if (0 == xoffset) {  // staying on Green
        R =  ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
        G = *src;
        B = *(src+Mask[4]);
      }
      else { // staying on Blue
        R = ( *(src+Mask[0]) + *(src+Mask[7])) /2;
        G = (*(src+Mask[1])  + *(src+Mask[4]) + *(src+Mask[6]) ) /3;
        B = *src;
      }
      *sink++ =R;
      *sink++ =G;
      *sink++ =B;
      src++;      
      // main part of line
      for (unsigned int x=1; x<(unsigned int)source.GetWidth()-1; x++) {
        if (y%2 == yoffset) 
          if (x%2 == xoffset) {  // staying on Red
            R = *src;
            G = (*(src+Mask[1]) + *(src+Mask[3]) 
              + *(src+Mask[4]) + *(src+Mask[6]) ) /4;
            B = ( *(src+Mask[0]) + *(src+Mask[2]) 
              + *(src+Mask[5]) + *(src+Mask[7])) /4;
          } else { // staying on Green
            R =  ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
            G = *src;
            B = ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
          }
        else
          if (x%2 == xoffset) {  // staying on Green
            R =  ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
            G = *src;
            B = ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
          }
          else { // staying on Blue
            R = ( *(src+Mask[0]) + *(src+Mask[2]) 
              + *(src+Mask[5]) + *(src+Mask[7])) /4;
            G = (*(src+Mask[1]) + *(src+Mask[3]) 
              + *(src+Mask[4]) + *(src+Mask[6]) ) /4;
            B = *src;
          }
          *sink++ =R;
          *sink++ =G;
          *sink++ =B;
          src++;      
      }
      // last pixel in line
      if (y%2 == yoffset) 
        if ((srcskip-1)%2 == xoffset) {  // staying on Red
          R = *src;
          G = (*(src+Mask[1]) + *(src+Mask[3]) + *(src+Mask[6]) ) /3;
          B = ( *(src+Mask[0]) + *(src+Mask[5])) /2;
        } else { // staying on Green
          R =  *(src+Mask[3]) ;
          G = *src;
          B = ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
        }
      else
        if ((srcskip-1)%2 == xoffset) {  // staying on Green
          R =  ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
          G = *src;
          B = *(src+Mask[3]);
        }
        else { // staying on Blue
          R = ( *(src+Mask[0]) + *(src+Mask[5])) /2;
          G = (*(src+Mask[1]) + *(src+Mask[3])+ *(src+Mask[6]) ) /3;
          B = *src;
        }
        *sink++ =R;
        *sink++ =G;
        *sink++ =B;
        src++;      


  }
  // first pixel in last line
  if ((source.GetHeight()-1)%2 == yoffset) {   
    if (0 == xoffset) {  // staying on Red
      R = *src;
      G = ( *(src+Mask[1])+  *(src+Mask[4]) ) /2;
      B =  *(src+Mask[2])  ;
    } 
    else { // staying on Green
      R =  *(src+Mask[4]);
      G = *src;
      B = (*(src+Mask[1]) ) ;
    }
  }
  else {
    if (0 == xoffset) {  // staying on Green
      R =  ( *(src+Mask[1]) ) ;
      G = *src;
      B =  *(src+Mask[4]) ;
    }
    else { // staying on Blue
      R =  *(src+Mask[2]);
      G = ( *(src+Mask[1]) + *(src+Mask[4]) ) /2;
      B = *src;
    }
  }
  *sink++ =R;
  *sink++ =G;
  *sink++ =B;
  src++;
  // last line
  if ((source.GetHeight()-1)%2 == yoffset) {
    for (unsigned int x=1; x<(unsigned int)dest.GetWidth()-1; x++){
      if (x%2 == xoffset) {  // staying on Red
        R = *src;
        G = ( *(src+Mask[3])+ *(src+Mask[1]) + *(src+Mask[4]) ) /3;
        B = ( *(src+Mask[0]) + *(src+Mask[2])) /2;
      } 
      else { // staying on Green
        R =  ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
        G = *src;
        B = (*(src+Mask[1]) ) ;
      }
      *sink++ =R;
      *sink++ =G;
      *sink++ =B;
      src++;
    }
  }
  else {
    for (unsigned int x=1; x<(unsigned int)dest.GetWidth()-1; x++){
      if (x%2 == xoffset) {  // staying on Green
        R =  ( *(src+Mask[1]) ) ;
        G = *src;
        B = ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
      }
      else { // staying on Blue
        R = (  *(src+Mask[0]) + *(src+Mask[2])) /2;
        G = ( *(src+Mask[1]) + *(src+Mask[3]) + *(src+Mask[4]) ) /3;
        B = *src;
      }
      *sink++ =R;
      *sink++ =G;
      *sink++ =B;
      src++;      
    }
  }
  // last pixel in last line
  if ((source.GetHeight()-1)%2 == yoffset) {   
    if ((dest.GetWidth()-1)%2 == xoffset) {  // staying on Red
      R = *src;
      G = ( *(src+Mask[3])+  *(src+Mask[1]) ) /2;
      B =  *(src+Mask[0])  ;
    } 
    else { // staying on Green
      R =  *(src+Mask[3]);
      G = *src;
      B = (*(src+Mask[1]) ) ;
    }
  }
  else {
    if ((dest.GetWidth()-1)%2 == xoffset) {  // staying on Green
      R =  ( *(src+Mask[1]) ) ;
      G = *src;
      B =  *(src+Mask[3]) ;
    }
    else { // staying on Blue
      R =  *(src+Mask[0]);
      G = ( *(src+Mask[3]) + *(src+Mask[1]) ) /2;
      B = *src;
    }
  }
  *sink++ =R;
  *sink++ =G;
  *sink++ =B;
  src++;

  return 0;
}








// 2 pixel border in dest is black
template <class StorageType>
int BIAS::ImageConvert::BayerToRGBSlow_(const Image<StorageType>& source, 
                                        Image<StorageType>& dest )   
{
  BIASGDOUT(D_CONVERT," Image::BayerToRGBSlow_()");

  StorageType *sink = dest.GetImageData() + 2 * dest.GetWidthStep() + 6;
  StorageType *src = source.GetImageData() + 2 * source.GetWidth() + 2;

  int Mask[12];
  StorageType R,G,B, d1, d2; 

  Mask[0] = - (int(source.GetWidth()) +1);
  Mask[1] = - int(source.GetWidth());
  Mask[2] = - (int(source.GetWidth()) -1);
  Mask[3] = -1;
  Mask[4] = +1;
  Mask[5] = int(source.GetWidth()) -1;
  Mask[6] = + int(source.GetWidth());
  Mask[7] = int(source.GetWidth()) +1;
  Mask[8] = - 2 * int(source.GetWidth());
  Mask[9] = - 2 ;
  Mask[10] = + 2 ;
  Mask[11] = + 2 * int(source.GetWidth());

#define D1 (*(src+Mask[8])>*(src+Mask[11]))?\
  (*(src+Mask[8])-*(src+Mask[11])):\
  (*(src+Mask[11])-*(src+Mask[8]))
#define D2 (*(src+Mask[9])>*(src+Mask[10]))?\
  (*(src+Mask[9])-*(src+Mask[10])):\
  (*(src+Mask[10])-*(src+Mask[9]))

  unsigned int xoffset = 0;
  unsigned int yoffset = 0;
  switch (source.GetColorModel()) {
  case ImageBase::CM_Bayer_GBRG: yoffset = 1; break;
  case ImageBase::CM_Bayer_GRBG: xoffset = 1; break;
  case ImageBase::CM_Bayer_BGGR: xoffset = 1;yoffset = 1; break;
  default:; }

  for (unsigned int y=2; y<(unsigned int)source.GetHeight()-2; y++) {
    for (unsigned int x=2; x<(unsigned int)source.GetWidth()-2; x++) {
      if (y%2 == yoffset) 
        if (x%2 == xoffset) {  // staying on Red
          R = *src;
          d1 = D1;
          d2 = D2;
          if (d1<d2){
            G = (*(src+Mask[1]) + *(src+Mask[6]) ) /2;
          } else if (d2<d1){
            G = (*(src+Mask[3]) + *(src+Mask[4]) ) /2;
          } else {
            G = (*(src+Mask[1]) + *(src+Mask[3]) 
              + *(src+Mask[4]) + *(src+Mask[6]) ) /4;
          }
          B = ( *(src+Mask[0]) + *(src+Mask[2]) 
            + *(src+Mask[5]) + *(src+Mask[7])) /4;
        }
        else { // staying on Green
          R =  ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
          G = *src;
          B = ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
        }
      else
        if (x%2 == xoffset) {  // staying on Green
          R =  ( *(src+Mask[1]) + *(src+Mask[6]) ) /2;
          G = *src;
          B = ( *(src+Mask[3]) + *(src+Mask[4]) ) /2;
        }
        else { // staying on Blue
          R = ( *(src+Mask[0]) + *(src+Mask[2]) 
            + *(src+Mask[5]) + *(src+Mask[7])) /4;
          d1 = D1;
          d2 = D2;
          if (d1<d2){
            G = (*(src+Mask[1]) + *(src+Mask[6]) ) /2;
          } else if (d2<d1){
            G = (*(src+Mask[3]) + *(src+Mask[4]) ) /2;
          } else {
            G = (*(src+Mask[1]) + *(src+Mask[3]) 
              + *(src+Mask[4]) + *(src+Mask[6]) ) /4;
          }
          B = *src;
        }
        *sink++ =R;
        *sink++ =G;
        *sink++ =B;
        src++;      
    }
    sink += 12;
    src += 4;
  }
  dest.SetROI(2,2,dest.GetWidth()-2,dest.GetHeight()-2);
  return 0;
}



template <class StorageType>
int BIAS::ImageConvert::YUV420PToRGB_(const Image<StorageType>& source, 
                                      Image<StorageType>&  dest ) 
{
  StorageType *sink = dest.GetImageData() ;
  StorageType *src = source.GetImageData() ;
  float y,u = 0.0,v=0.0;
  float r,g,b;
  int uoffset=0 ;
  int voffset=0 ;
  int width = source.GetWidth();
  int height = source.GetHeight();

  for (int z = 0; z < height ; z++) {
    if (z%2==0) {
      voffset = width*height +z/2*width/2;
      uoffset = width*height + width*height/4+z/2*width/2;
    }
    for (int x=0; x< width; x++) {
      y = 1.164f * (float) (src[z*width + x] -16) ;
      if (  x%2==0 ) {
        v = (float) (src[voffset + x/2] -128);
        u = (float) (src[uoffset + x/2 ] -128);
      }
      b =  float(rint( y + 2.018f*v ));
      g =  float(rint( y - 0.813f*u  - 0.391f*v));
      r =  float(rint( y + 1.596f*u));
      if (r>255) r=255;
      if (g>255) g=255;
      if (b>255) b=255;
      if (r<0) r = 0;
      if (g<0) g = 0;
      if (b<0) b = 0;

      sink[(z*width + x)*3 ] = (unsigned char) r;
      sink[(z*width + x)*3 +1] =(unsigned char)g;
      sink[(z*width + x)*3 +2] =(unsigned char) b;
    }
  }
  return 0;
}

template <class StorageType>
int BIAS::ImageConvert::YUV420PToGrey_(const Image<StorageType>& source, 
                                       Image<StorageType>&  dest ) 
{
  StorageType *sink = dest.GetImageData() ;
  StorageType *src = source.GetImageData() ;

  int width = source.GetWidth();
  int height = source.GetHeight();

  for (int z = 0; z < height ; z++) 
    for (int x=0; x< width; x++) 
      sink[(z*width + x) ] =  src[z*width + x];

  return 0;
}


template <class StorageType>
int BIAS::ImageConvert::YUV411ToGrey_(const Image<StorageType>& source, 
                                       Image<StorageType>&  dest ) 
{
  StorageType *sink = dest.GetImageData() ;
  StorageType *src = source.GetImageData() ;

  int width = source.GetWidth();
  int height = source.GetHeight();

  for (int z = 0; z < height ; z++) 
    for (int x=0; x< width; x++) 
      sink[(z*width + x) ] =  src[z*width*2 + x*2];

  return 0;
}




int ImageConvert::FloatToUC_(const ImageBase &source, ImageBase &dest)
{
  unsigned int values = source.GetPixelCount() * source.GetChannelCount();

  unsigned char *d = (unsigned char*)dest.GetImageData();
  float *s =(float*) source.GetImageData();
  for (unsigned int i=0; i<values; i++){
    d[i] = (unsigned char) *s;
    s ++;
  }  

  return 0;
}



//  Now storage type conversion
int ImageConvert::ConvertST(const ImageBase &source, ImageBase &dest, 
                            ImageBase::EStorageType targetST)
{
  if (!dest.IsEmpty()) dest.Release();
  dest.Init(source.GetWidth(), source.GetHeight(), source.GetChannelCount(), 
            targetST);
  if (targetST==source.GetStorageType()){
    dest=source;
    return 0;
  }

  dest.SetColorModel(source.GetColorModel());

  // this is NumPixels * Channels
  unsigned int values = source.GetPixelCount() * source.GetChannelCount();

  switch(targetST) {
	case ImageBase::ST_unsignedchar: 
    {   //target
	    unsigned char *d = (unsigned char*)dest.GetImageData();
	    switch (source.GetStorageType()){
      case ImageBase::ST_float: 
		    return FloatToUC_(source,dest); break;
#ifdef BUILD_IMAGE_INT
      case ImageBase::ST_int:
        {
          int *s =(int*) source.GetImageData();
          for (unsigned int i=0; i<values; i++)
            d[i] = (unsigned char) s[i];
        }
        break;  
#endif // BUILD_IMAGE_INT
#ifdef BUILD_IMAGE_USHORT
      case ImageBase::ST_unsignedshortint:
        { 
          unsigned short int *s =(unsigned short int*) source.GetImageData();
          for (unsigned int i=0; i<values; i++)
            d[i] = (unsigned char) s[i];
        }
        break;  
#endif // BUILD_IMAGE_SHORT
#ifdef BUILD_IMAGE_CHAR
      case ImageBase::ST_char:
        {
          char *s = (char*)source.GetImageData();
          for (unsigned int i=0; i<values; i++)
            d[i] = (unsigned char) s[i];
        }
        break;
#endif // BUILD_IMAGE_CHAR
#ifdef BUILD_IMAGE_SHORT
      case ImageBase::ST_shortint:
        {
          short *s=(short*)source.GetImageData();
          for (unsigned int i=0; i<values; i++)
            d[i] = (unsigned char) s[i];
        }
        break;
#endif // BUILD_IMAGE_SHORT
      default:
		    BIASERR("ConvertST() not fully implemented "<<source.GetStorageType());
        BIASABORT;
		    return -1;
	    }

    }    
    break;
	case ImageBase::ST_int: 
    {
	    int *d = (int*)dest.GetImageData();
	    switch (source.GetStorageType()){
      case ImageBase::ST_unsignedchar:
        {
          unsigned char *s =(unsigned char*) source.GetImageData();
          for (unsigned int i=0; i<values; i++) d[i] = (int) s[i];
        }
        break;  
      case ImageBase::ST_char:
        {
          char *s =(char*) source.GetImageData();
          for (unsigned int i=0; i<values; i++) d[i] = (int) s[i];
        }
        break;  
      case ImageBase::ST_unsignedshortint:
        {
          unsigned short *s =(unsigned short*) source.GetImageData();
          for (unsigned int i=0; i<values; i++) d[i] = (int) s[i];
        }
        break;  
      case ImageBase::ST_float:
	{
          float *s =(float*) source.GetImageData();
          for (unsigned int i=0; i<values; i++) d[i] = (int) s[i];
	}
	break;  
	  
      default:
		    BIASERR("ConvertST() not fully implemented");  return -1;
	    }
    } break;
	case ImageBase::ST_unsignedshortint: 
    {
	    unsigned short int *d = (unsigned short int*)dest.GetImageData();
	    switch (source.GetStorageType())
        {
        case ImageBase::ST_unsignedchar:
          {
            unsigned char *s =(unsigned char*) source.GetImageData();
            for (unsigned int i=0; i<values; i++) 
              d[i] = (unsigned short int) s[i];
          }
          break;  
        default:
          BIASERR("ConvertST() not fully implemented");  return -1;
        }
    } break;
	case ImageBase::ST_float:
    {
	    float *d = (float*)dest.GetImageData();
	    switch (source.GetStorageType())
        {
        case ImageBase::ST_unsignedchar:
          {
            unsigned char *s =(unsigned char*) source.GetImageData();
            unsigned char *e = s+values;
            while (s<e) 
              *d++=(float)*s++;
          }
          break; 
        case ImageBase::ST_char:
          {
            char *s =(char*) source.GetImageData();
            char *e = s+values;
            while (s<e) *d++=(float)*s++;
          }
          break;   
        case ImageBase::ST_unsignedshortint:
          {
            unsigned short *s =(unsigned short*) source.GetImageData();
            for (unsigned int i=0; i<values; i++) d[i] = (float) s[i];
          }
          break;  
#ifdef BUILD_IMAGE_INT
	case ImageBase::ST_int:
	  {
            int *s =(int*) source.GetImageData();
            for (unsigned int i=0; i<values; i++) d[i] = (float) s[i];
	  }
	  break;  
#endif // BUILD_IMAGE_INT

        default:
          BIASERR("ConvertST() not fully implemented "<<source.GetStorageType());  
          return -1;
        }
    } 
    break;
	default:
    BIASERR("ConvertST() not implemented for this target ST"); return -1;
    break;
  }
  int tlx, tly, brx, bry;
  source.GetROI()->GetCorners(tlx, tly, brx, bry);
  dest.GetROI()->SetCorners(tlx, tly, brx, bry);

  return 0;
}


template <class StorageType>
int BIAS::ImageConvert::HSLToGrey_(const Image<StorageType>& source,
                                   Image<StorageType>& dest)
{
  BIASGDOUT(D_CONVERT,"ImageConvert::hsLToGrey_()");
  StorageType *p=dest.GetImageData();
  StorageType *h, *s, *L;

  if (source.IsPlanar()){
    switch (source.GetColorModel()){
    case ImageBase::CM_hsL:    
    case ImageBase::CM_HSL:
      h=source.GetImageData();
      s=h+(source.GetWidth()*source.GetHeight());
      L=s+(source.GetWidth()*source.GetHeight());
      break;
    default:
      BIASERR("HSLToGrey_ must be called with HSL or hsL ColorModel_");
      return -1; 
    }
    for (unsigned int i=0; i < dest.GetWidth()*dest.GetHeight(); i++){
      *p=(StorageType)*L;
      L++; p++;
    }
  } else { // interleaved
    switch (source.GetColorModel()) {
    case ImageBase::CM_hsL:
    case ImageBase::CM_HSL:
      h=source.GetImageData();
      s=h+1;
      L=h+2;
      break;
    default:
      BIASERR("HSLToGrey_ must be called with HSL or hsL ColorModel_");
      return -1; 
    }
    for (unsigned int i=0; i < dest.GetWidth()*dest.GetHeight(); i++){
      *p=(StorageType)(*L);
      p++; L+=3;
    }
  }
  return 0;
}

#ifdef BIAS_HAVE_OPENCV
template <class StorageType>
int ImageConvert::BIAS2ipl(const Image<StorageType>& source, IplImage*  &dest) 
{
  //#ifdef USE_MY_OPENCV
  //BIASDOUT(D_IMCONV_IPL, "using own CreateIplImage");
  //  cerr << "using own CreateIplImage\n";
  //dest = CreateIplImage(source.GetWidth(), source.GetHeight(), 
  //                      sizeof(StorageType) * 8 , source.GetChannelCount());
   //#else
  //BIASDOUT(D_IMCONV_IPL, "using official CreateIplImage");
  //cerr << "using official CreateIplImage\n";
  dest = cvCreateImage(cvSize(source.GetWidth(), source.GetHeight()), 
    sizeof(StorageType) * 8, source.GetChannelCount());
  //#endif

  StorageType* newImageData = (StorageType*) dest->imageDataOrigin;

  //
  // IplImages in OpenCV are always:
  //   - GREY 
  //   - RGB  (with byte Order BGR due to Windows DIB format) 
  //   - RGBA (with byte Order BGRA)
  //
  // As an limitation of opencv compared to IPL, color images are ALWAYS 
  //   interleaved!

  //cerr << "source.GetColorModel() :"<<source.GetColorModel()<<endl;
  switch (source.GetColorModel()) {

  case ImageBase::CM_Grey: 
    //cerr <<"grey\n";
    memcpy(newImageData, source.GetImageData(), source.GetSizeByte() );
    break;

  case ImageBase::CM_BGR : 
    //cerr <<"BGR\n";
    if (source.IsPlanar()) {
      //cerr <<"Planar\n";
      // planar BGR image
      for(int channel = 0; channel < 3; channel++) {
        register StorageType* src  = source.GetImageData() + 
          channel*source.GetPixelCount();
        register StorageType* act = newImageData + channel;
        register StorageType* stop = src + source.GetPixelCount();
        for(;src < stop; src++) {
          *act = *src;
          act += 3;
        }
      }
    } else {
      //cerr <<"Interleaved\n";
      // interleaved BGR image - the same format as IplImage itself
      memcpy((void*)newImageData, (void*)source.GetImageData(), source.GetSizeByte());
    };
    break;

  case ImageBase::CM_RGB : 
    //cerr <<"RGB\n";
    if (source.IsPlanar()) {
      //cerr <<"Planar\n";
      // planar RGB image
      for(int channel = 0; channel < 3; channel++) {
        register StorageType* src  = source.GetImageData() + (2 - channel) * 
          source.GetPixelCount(); // RGB instead of BGR order!
        register StorageType* act = newImageData + channel;
        register StorageType* stop = src + source.GetPixelCount();
        for(;src < stop; src++) {
          *act = *src;
          act += 3;
        }
      }
    } else { 
      //cerr <<"Interleaved\n";
      // interleaved RGB image
      register StorageType* src  = source.GetImageData();
      register StorageType* stop = src + 3 * source.GetPixelCount();
      register StorageType* act = newImageData;
      while (src < stop) {
        // swap RGB because OpenCV uses BGR as native channel order
        // only the color conversion routines need this, I fixed highgui to handle channelSeq correctly (jw)
        act[0]=src[2];
        act[1]=src[1];
        act[2]=src[0];
        act+=3; 
        src+=3;
      }
    }    
    break;

  default: 
    BIASERR("Only Grey, BGR and RGB images can be converted to IplImages yet."
      <<" Aborting!");
    return -1;
  }

  return 0;
}

template <class StorageType>
int ImageConvert::Ipl2BIAS(IplImage* source, Image<StorageType>& dest) 
{
  // Does the color depth match?
  if (8*sizeof(StorageType) != source->depth) {
    BIASERR("Cannot convert an IplImage into an Image of different color "
      <<"depth!");
    return -1;
  }

  if (source->width <=0 || source->height <= 0) {
    BIASERR("IplImage to be converted to Image has no size!");
    return -1;
  }
  if ( !dest.IsEmpty() )
    dest.Release();     
  // What kind of image do we have?
  switch (source->nChannels) {
  case 1: // gray image
    // Release the old image data and allocate new memory
    dest.Init(source->width, source->height, 1);
    memcpy(dest.GetImageData(), source->imageDataOrigin, dest.GetSizeByte());
    break;
  case 3:
    if (strcmp(source->colorModel, "RGB")==0 && 
      strcmp(source->channelSeq, "BGR")==0 &&
      source->dataOrder==IPL_DATA_ORDER_PIXEL &&
      (source->depth>>3)==sizeof(StorageType) ){
        dest.Init(source->width, source->height, 3);
        register unsigned char* src = (unsigned char*) source->imageDataOrigin;
        register unsigned char* stop = src + 3 * source->width * source->height;
        register StorageType* dst = dest.GetImageData();
        while (src < stop) {
          dst[0]=src[2];
          dst[1]=src[1];
          dst[2]=src[0];
          dst+=3; src+=3;
        }
      } else {
        BIASERR("Ipl2BIAS not implemented for this color model or inavlid storage type");
        return -1;
      }
      break;
  default:
    BIASERR("Only gray IplImages can be converted to Images yet!");
    return -1;      
  }
  return 0;
}
#endif // BIAS_HAVE_OPENCV



int ImageConvert::DeinterleaveHorizontal(const BIAS::ImageBase & raw,
                                         BIAS::ImageBase & out)
{ 
  if (raw.GetColorModel()!= vidap_DEFAULT_PACKED_COLORMODEL)
    return -1;
  if (raw.GetChannelCount()!= 3)
    return -2;
  if ( ImageBase::GetSizeByte(raw.GetStorageType()) != 1 ){
    BIASERR("storagetypes size must be 1 for now due to simple pointer arithmetics.\n"
      "Other pointer sizes need to be implemented.");
    return -3;
  }
  if (raw.GetStorageType()!= out.GetStorageType()){
    BIASERR("storagetypes must match");
    return -4;
  }

  BIASASSERT(raw.GetColorModel()== vidap_DEFAULT_PACKED_COLORMODEL );
  BIASASSERT(raw.GetChannelCount()== 3);
  BIASASSERT(ImageBase::GetSizeByte(raw.GetStorageType()) == 1);

  // check if we need re-allocation
  if (!out.IsEmpty())
    out.Release();

  // triple width for appeding 3 camera images horizontally
  out.Init( 3* raw.GetWidth(), raw.GetHeight(), 1 );
  out.SetColorModel( vidap_DEFAULT_BAYER_COLORMODEL );

  // abcABC --> aAbBcC
  unsigned char *s =NULL, *d0=NULL, *d1=NULL, *d2=NULL;

  for (unsigned int y=0; y<raw.GetHeight(); y++){
    s  = (unsigned char*)raw.GetImageDataArray()[y];
    d0 = (unsigned char*)out.GetImageDataArray()[y];
    d1 = (unsigned char*)out.GetImageDataArray()[y] + raw.GetWidth();
    d2 = (unsigned char*)out.GetImageDataArray()[y] + 2*raw.GetWidth();
    for (unsigned int x=0; x<raw.GetWidth(); x++){
      *d0 = *s; s++; d0++;
      *d1 = *s; s++; d1++;
      *d2 = *s; s++; d2++;
    }
  }

  return 0;
}

//
// solve explicit instantiation of this class
// must be at the end of the file JW
#include "ImageConvertInstantiation.cpp"

