/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// define it for extra debugging output (JW)
//#define ImageIO_DEBUG_IO

#include "ImageIO.hh"
#include <Base/Common/FileHandling.hh>

// ImageMagick support: Jan Woetzel 10/2004
#ifdef BIAS_HAVE_IMAGEMAGICKLIB 
#  include <Base/Common/BIASpragmaStart.hh>
#    include <Magick++.h>
#  include <Base/Common/BIASpragmaEnd.hh>
#endif


// required in header (not only cpp)
// due to public EXR PrintHeader interface (JW)
#ifdef BIAS_HAVE_OPENEXR
#  ifdef WIN32
#    pragma warning( push, 1)
#    define PLATFORM_WINDOWS
#  endif // WIN32
#    include <half.h>
#    include <ImfRgbaFile.h>
#    include <ImfRgba.h>
#    include <ImfArray.h>
#    include <ImathBox.h>
#  ifdef WIN32
#    pragma warning( pop)
#  endif // WIN32
#endif


#ifdef BIAS_HAVE_OPENCV
#  include "WrapBias2Ipl.hh"
#  include <cv.h>
#  include <highgui.h>
#endif


#ifdef BIAS_HAVE_IMLIB
#  include <Imlib.h>
#endif

// use DevIL il with ilu (if DevIL is used at all)
#define USE_DEVIL_ILU

#ifdef BIAS_HAVE_DEVIL
#  include <IL/il.h>
#  ifdef USE_DEVIL_ILU
#    include <IL/ilu.h>
#  endif // USE_DEVIL_ILU
//#  include <IL/ilut.h>
#endif


#ifdef BIAS_HAVE_TIFF
#  include <tiffio.h>
// check TIFF version from tiffvers.h
#  if TIFFLIB_VERSION < 20051230
#    error You need at least TIFF version 3.8.0 (TIFFLIB_VERSION >= 20051230) for muli plane tiff.
#  endif
#endif


#include <Base/Debug/DebugSimple.hh>

#ifdef BIAS_DEBUG
#  include <Base/Debug/TimeMeasurement.hh>
#endif // BIAS_DEBUG

#include <fstream>
#include <sstream>

#include "ImageConvert.hh"
#include "ViffHeader.hh"

// for transfrom:
#include <algorithm>
#include <cctype>

// for O_APPEND
#include <fcntl.h>

// for perror
#include <stdio.h>

using namespace BIAS;
using namespace std;

char *ImageIO::MemAlignedBuffer_ = NULL;
unsigned long int ImageIO::MemAlignedBufferSize_ = 0;



/// helper function for ExtensionName type casting @author Jan Woetzel
char myTolower(char c) {
  return char(tolower(c));
}



// JW try to Read MIP image and if that fails try all import functions. 
int ImageIO::Load(const string& FileNameConst, ImageBase &img)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  string FileName; // (FileNameConst); // non const copy for tilde expansion
  BIAS::FilenameExpansion(FileNameConst, FileName); // tilde

  int result = 0;
  //ifstream is(FileName.c_str());
  ifstream is;
  is.open( FileName.c_str() );
  char identifier[IDENTIFIER_LENGTH];

  if (!is || !is.good()){
    // file does not exist/is not readable!
#if defined(BIAS_DEBUG) || defined(ImageIO_DEBUG_IO)
    BIASWARN("could not open fstream for file "<<FileName<<" with cwd="<<BIAS::GetCwd());
#endif
    result = -1;
  } else {
    is.read(identifier, IDENTIFIER_LENGTH);
    if (!is.good()){
      result = -2;
      is.close();
    } else {
      if (identifier[0] == 'M' && identifier[1] == 'I' && 
        identifier[2] == 'P'){ // mipimage
          is.close();
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
          COUT("MIP format detected\n");
#endif

          result = Read(FileName, img);
      } else { 
        // foreign image format
        is.close();
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
        COUT("image not in mip format\n");
#endif
        result = ImageIO::ImportImage(FileName, img);
      }
    }
  }
  return result;
}


#ifdef BIAS_HAVE_IMAGEMAGICKLIB
int ImageIO::ImportMagickPP(const std::string& FileName, 
                            ImageBase& result, 
                            const bool & /*dummy*/,
                            const bool & readMetaData )
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

#ifdef WIN32
  //initialize DLL loaded library (not required on Linux but once on Windows)
  Magick::InitializeMagick(NULL);
#endif

  try {

    Magick::Image image; 
    image.read( FileName );

    // now extract image info from Magick::Image
    unsigned width=image.image()->columns;
    unsigned height=image.image()->rows;
    unsigned bitdepth=image.image()->depth;
    unsigned depth=0;
    // bitdepth may ba a fractional part of the Byte depth, eg. 12/15 bits are filled.
    while (depth<bitdepth) 
      depth+=8;

    depth /= 8; // bit To Byte
    //#ifdef BIAS_DEBUG
    //    COUT("depth: "<<depth<<"\tbitdepth: "<<bitdepth<<endl);
    //#endif

#ifdef BIAS_DAIMLERCHRYSLER
    // heuristic to get the correct bitdepth in dc images
    // tiff image from daimler have 12 bit bitdepth
    string::size_type pos = FileName.find(".tiff");
    if (pos != string::npos){
      bitdepth = 12;
    }
#endif

    unsigned cc=0; // ChannelCount
    ImageBase::EColorModel cm=ImageBase::CM_invalid;
    ImageBase::EStorageType st=ImageBase::ST_invalid;
    Magick::StorageType imst=MagickLib::UndefinedPixel;

    /// LibMagick: imcs is RGB even for one channel images, like imlib does, too. (bug?)
    Magick::ColorspaceType imcs=image.image()->colorspace;
    std::string cm_map="";

    // choose appropriate storage type
    switch (depth)
    {
    case sizeof(unsigned char):
      st=ImageBase::ST_unsignedchar;
      imst=MagickLib::CharPixel;
      break;
    case sizeof(unsigned short):
      st=ImageBase::ST_unsignedshortint;
      imst=MagickLib::ShortPixel;
      break;
    case sizeof(float): // could this also be int?
      st=ImageBase::ST_float;
      imst=MagickLib::FloatPixel;
      break;
    default:
      BIASERR("funny depth "<<depth<<", maybe just unfinished code");
      return -4;
      break;
    }

    // choose appropriate color model
    switch (imcs)
    {
    case MagickLib::GRAYColorspace:
      cm=ImageBase::CM_Grey;
      cm_map="K";
      cc=1;
      break;

    case MagickLib::RGBColorspace:
      {
        // TODO:  or 4, e.g. for RGBA
        if (image.channelDepth(MagickLib::AlphaChannel)>1){
          cm=ImageBase::CM_RGBA;
          cm_map="RGBA";
          cc=4;
        } else {
          cm=ImageBase::CM_RGB;
          cm_map="RGB";
          cc=3; 
        }
        break;
      }

    default:
      // look in ImageMagick.h for further ColorspaceType's
      BIASERR("color model not implemented. You may want to add a colorspace/map handler.");
      return -3;
      break;
    }
    //#ifdef BIAS_DEBUG
    //    COUT("cm_map: "<<cm_map<<endl);
    //#endif



    //#ifdef BIAS_DEBUG
    //    COUT("MagickLib::Image: Read "<<image.columns()<<"x"<<image.rows()<<" "
    //         <<image.magick()<<" image "<<FileName<<" using ImageMagick++"<<endl);
    //    COUT("MagickLib::Image: depth "<<image.depth()<<" bit."<<endl );
    //    COUT("MagickLib::Image: comment "<<endl<<image.comment()<<endl );
    //    COUT("MagickLib::Image: depth "<<depth<<endl);
    //    COUT("MagickLib::Image: size "<<width<<"x"<<height<<endl);
    //    COUT("MagickLib::Image: storage class "<<image.image()->storage_class
    //         <<endl);
    //    COUT("MagickLib::Image: color space "<<image.image()->colorspace<<endl);
    //    COUT("MagickLib::Image: channel count "<<cc<<endl);
    //#endif

    if (!result.IsEmpty()) result.Release();

    // do error checking if result has already a specific StorageType
    if (result.GetStorageType()!=ImageBase::ST_invalid){
      if (result.GetDepth()!=depth){
        BIASERR("error: you tried to load a "<<depth<<" byte "
          <<"("<<bitdepth<<"bit) "
          <<"image into an image of storage type "
          <<result.GetStorageType()<<" which has depth "
          <<result.GetDepth()<<" byte");
        result.PrintHeader(cout);
        //BIASABORT;
        return -2;
      }
    } 

    if (!result.IsEmpty()) result.Release();

    result.Init(width, height, cc, st, true);
    //result.PrintHeader(cout);
    result.SetBitDepth(bitdepth);

    { // scope
      if (readMetaData)
      {
        //#ifdef BIAS_DEBUG
        //BIAS::TimeMeasurement tm1("DBG get meta data inside ImportMagickPP ");
        //#endif // BIAS_DEBUG
        // TODO: crashes due to problem with const reference streaming !
        // e.g. on loading ibak "M34/Front/image00000000.jpg" (jw)
        // get the meta data from the (ASCII) comment field
        stringstream strstr;
        // string copy required because you canot directly stream a const references!
        // - postpones the crash, but on leaving the scope it crashes, too. MetData operator<< problem?
        // TODO, HACK, FIXME fw: 28.2.2006: fw  TODO, HACK, FIXME: Only crash under windows in certain special cases: Needs to be compiled outside the bias library and needs to be run in the debugger of visual studio .net 2003  to reproduce crash. Maybe different libraries are used in this case? Even in that case it only crashes when the argument result  is of type Image<unsigned char>. When result is of type ImageBase, it does not crash.  In all other cases (linux and windows without .net) it works correctly. Neither valgrind nor efence (linux) find any suspicous  activity here. MetaData.operator<< is *NOT* present in callstack. Maybe, buggy stl under windows? TODO, HACK, FIXME, fw:  28.2.2006: fw  TODO, HACK, FIXME
        string sTmp = image.comment();       
        strstr << sTmp;
        //COUT("DBG a sTmp         xxx"<<sTmp<<"xxx"<<endl);
        //COUT("DBG b strstr.str() xxx"<<strstr.str()<<"xxx"<<endl );
        BIASASSERT(result.GetMetaData() != NULL);
        strstr >> *(result.GetMetaData());
        //result.GetMetaData()->Dump();
      }
    }

    BIASASSERT(result.GetImageData()!=NULL);
    image.write(0, 0, width, height, cm_map, imst, result.GetImageData());

    // verbose?
  }

#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  catch ( Magick::Exception & error_ ) {
    BIASERR("Reading "<<FileName<<" with Magick++ failed."<<endl
      <<"  Caught exception: "<<endl
      <<"  "<<error_.what()<<endl );
    return -1;
  }
#else // verbose
  catch ( Magick::Exception &) {
    // be silent about the reason why Magick couldn't load the file.
    return -1; // error or unsupported format
  }
#endif //verbose
  catch ( std::exception & ) {
    return -3;
  }
  catch (...) {
    return -4;
  }

  return BIAS_IO_SUCCESS; //OK
}
#endif // BIAS_HAVE_IMAGEMAGICKLIB




#ifdef BIAS_HAVE_OPENCV
int ImageIO::ImportOpenCV(const std::string& FileName, ImageBase& result)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif  

  IplImage * p_cv = cvLoadImage( FileName.c_str(), 1);
  if (p_cv==NULL)
  {
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR("Reading "<<FileName<<" with OpenCV failed.");
#endif
    return -1;
  }

  // copy IPL image into BIAS data area
  int res = BIAS::WrapBias2Ipl::CreateBiasImageCopy( p_cv, result );
  cvReleaseImage(&p_cv);
  return res; // OK
}
#endif // BIAS_HAVE_OPENCV


#ifdef BIAS_HAVE_OPENEXR
/** Print header information to stream 
helper function
@author Jan Woetzel 09/2005 */
int PrintInfoEXR(const Imf::RgbaInputFile & f, std::ostream & os)
{
  using namespace Imf;
  using namespace Imath;
  os<<"OpenEXR header information: "<<endl;
  os<<"version:         "<<f.version()<<endl;
  //os<<"isComplete:      "<<f.isComplete()<<endl;

  // (meta) header
  const Imf::Header * h = &f.header();
  BIASASSERT(h!=NULL);
  // data windows
  Box2i dw = f.dataWindow();

  os<<"hasPreviewImage: "<<h->hasPreviewImage()<<endl;
  os<<"pixelAspectRatio:"<<h->pixelAspectRatio()<<endl;
  // OpenEXR can handle special image coordinate systems.
  os<<"dw dim: "<<dw.max.x - dw.min.x + 1
    <<" x "
    <<dw.max.y - dw.min.y + 1 
    <<endl;
  os<<"  dw.x: "<<dw.min.x<<".."<<dw.max.x <<endl;
  os<<"  dw.y: "<<dw.min.y<<".."<<dw.max.y <<endl;

  return BIAS_IO_SUCCESS;
}


// Jan Woetzel 09/2005, Patrick Fittkau 10/2005
int ImageIO::ExportOpenEXR(const std::string& FileName, const ImageBase& img)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  try 
  {
    using namespace Imf;
    // OpenEXR data struct:
    Imf::Rgba *exrPixels;
    // create exr-format-structure for saving
    int exrwidth  = img.GetWidth();
    int exrheight = img.GetHeight();
    exrPixels = new Rgba[exrheight*exrwidth];

    // copy 
    float **pptr = (float**)(img.GetImageDataArray());
    for (int x = 0;x<exrwidth;x++) 
      for (int y = 0;y<exrheight;y++) 
      {
        exrPixels[y*exrwidth + x].r = (half)pptr[y][x*3  ];
        exrPixels[y*exrwidth + x].g = (half)pptr[y][x*3+1];
        exrPixels[y*exrwidth + x].b = (half)pptr[y][x*3+2];
        exrPixels[y*exrwidth + x].a = (half)1.0;
      }

      RgbaOutputFile file(FileName.c_str(), exrwidth, exrheight, WRITE_RGBA);
      file.setFrameBuffer (exrPixels, 1, exrwidth);
      file.writePixels (exrheight);

      delete [] exrPixels;
      return BIAS_IO_SUCCESS;
  } catch (Iex::BaseExc &error_ ) 
  {   
    BIASERR("Reading "<<FileName<<" with OpenEXR failed."<<endl
      <<"  Caught exception: "<<endl
      <<"  "<<error_.what()<<endl 
      );
    return -2; // error
  }
}


// fittkau, jw 2005/08
int ImageIO::ImportOpenEXR(const std::string& FileName, ImageBase& result)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  try 
  {
    using namespace Imf;
    using namespace Imath;
    // OpenEXR data struct:
    Array2D<Rgba> exrPixels;
    // read OpenEXR-Image into array using OpenEXR lib
    // TODO: catch return value/errors/exceptions 
    // TODO: test format and storagteype of read imge

    // see example in "ReadingAndWritingImageFiles.pdf", chapter 2.4 on page 5 
    // in OpenEXR 1.2.2 documentation (JW)
    RgbaInputFile exrfile ( FileName.c_str() );
    // pixel data area window
    Box2i dw = exrfile.dataWindow();
    // OpenEXR can handle special image coordinate systems.
    int exrwidth  = dw.max.x - dw.min.x + 1;
    int exrheight = dw.max.y - dw.min.y + 1;
    BIASASSERT(exrwidth>0);
    BIASASSERT(exrheight>0);

#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    //exrfile.header
    PrintInfoEXR(exrfile, std::cout);
#endif

    // allocate memory:
    exrPixels.resizeErase(exrheight, exrwidth);
    exrfile.setFrameBuffer (
      &exrPixels[0][0] - dw.min.x - dw.min.y * exrwidth, 
      1, 
      exrwidth
      );

    // copy the pixel data from file into buffer 
    // for Y-lines min..max (to support partial read)
    // we read the complete image. 
    exrfile.readPixels (dw.min.y, dw.max.y);

    // done with exr read image
    // convert to BIAS Image Base, now.

    if (!result.IsEmpty())  result.Release();

    // HACK: fixed format float RGB for now. (fittkau, jw 2005/08)
    result.Init(exrwidth, exrheight, 3, BIAS::ImageBase::ST_float, true );
    float **pptr = (float**)(result.GetImageDataArray());

    for (int x = 0;x<exrwidth;x++) 
      for (int y = 0;y<exrheight;y++) 
      {
        pptr[y][x*3  ] = (float)exrPixels[y][x].r;
        pptr[y][x*3+1] = (float)exrPixels[y][x].g;
        pptr[y][x*3+2] = (float)exrPixels[y][x].b;
      }

      return BIAS_IO_SUCCESS;

  } 
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  catch (Iex::BaseExc &error_ ) {
    BIASWARN("Reading "<<FileName<<" with OpenEXR failed."<<endl
      <<"  Caught exception: "<<endl
      <<"  "<<error_.what()<<endl 
      );
#else 
  catch (Iex::BaseExc &/*error_*/ ) {
#endif
    return -1; // error
  }
  catch (...){
    return -2;
  }
}
#endif // BIAS_HAVE_OPENEXR



// JW: major rewrite 2005
int ImageIO::ImportImage(const string& filename, ImageBase &result ) 
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  int res=-1; // init as file not read
  ifstream ifs;
  unsigned char tmpline[512];

#ifdef   WIN32
  ifs.open(filename.c_str(),ios::binary);
#else  //not WIN32
  ifs.open(filename.c_str());
#endif //WIN32

  if (!ifs) {
    BIASERR("ImportImage(): open of file \""<<filename<<"\" failed");
    perror(filename.c_str());
    ifs.close();
    return -1;
  }

  // invalidate old uid, since it may refer to a different image
  result.InvalidateUID();

  // read first two chars to get a hint what this file is
  ifs >>tmpline[0]>>tmpline[1];
  if (!ifs){
    res=-1;
    BIASERR("unable to read from "<<filename);
    ifs.close();
    return res;
  }

  // Check for famous P5/P6 in PNM, PGM, PBM to determine the fileformat
  // and call our own reader if appropriate.
  if (tmpline[0] =='P' && tmpline[1] == '5') {
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    COUT("reading single channel pnm/pgm/pbm image using ImportImagePnm_\n");
#endif
    res = ImportImagePnm_(ifs, result, 1); 
    if (res!=BIAS_IO_SUCCESS) 
      BIASERR("error reading "<<filename<<" although P5 format detected.");
    result.SetColorModel(ImageBase::CM_Grey);
  } else if (tmpline[0] =='P' && tmpline[1] == '6') {
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    COUT("reading 3 channel pnm/pgm/pbm image using ImportImagePnm_\n");
#endif
    res = ImportImagePnm_(ifs, result, 3); 
    if (res!= BIAS_IO_SUCCESS) 
      BIASERR("error reading "<<filename<<" although P6 format detected.");
    result.SetColorModel(ImageBase::CM_RGB);
  } 

  // other loaders do not use file stream but operate on filename directly 
  // and open theirselves.
  ifs.close();

  //
  // Try to read using external image loaders.
  //

  // try EXR first because imlib is too verbose on unsupported formats liek exr
#ifdef BIAS_HAVE_OPENEXR
  if (res!=BIAS_IO_SUCCESS) 
    res=ImportOpenEXR(filename, result);
#endif

#ifdef BIAS_HAVE_IMAGEMAGICKLIB
  if (res!=BIAS_IO_SUCCESS) 
    res=ImportMagickPP(filename, result); 
#endif

#ifdef BIAS_HAVE_DEVIL
  if (res!=BIAS_IO_SUCCESS) 
    res=ImportDevIL(filename, result);
#endif

#ifdef BIAS_HAVE_OPENCV
  if (res!=BIAS_IO_SUCCESS) 
    res=ImportOpenCV(filename, result);
#endif

#ifdef BIAS_HAVE_IMLIB
  if (res!=BIAS_IO_SUCCESS) 
    res=ImportImlib(filename, result);
#endif


  //#ifdef BIAS_HAVE_RADIANCE
  if (res!=BIAS_IO_SUCCESS) 
    res=ImportRADIANCE(filename, result);
  //#endif

  // JW: try to read RAW4096/RAW04096/RAW0nnnn format 
  if (res!=BIAS_IO_SUCCESS) 
    res=ImportRAWwithHeader(filename, result);

#ifdef WIN32 // winapi
  if (res!=BIAS_IO_SUCCESS) 
    res=Import_BITMAP_winapi(filename, result);
#endif

  // FIXME: incorrectly loads .bmp occasionally 
  // due to too simple header parsing 
  // e.g. Ibak Beton1200_Front47.bmp ! JW
  // JW: try to read VIFF if anything else failed
  // this is last try because the routine is buggy. jw
  if (res!=BIAS_IO_SUCCESS) {
    // open fstream 
    std::ifstream ifs(filename.c_str(), std::ios::binary );
    if (!ifs) {
      BIASERR("could not open "<<filename);
    } else {
      // OK
      res=ImportImageViff_(ifs, result );
    }
  };


  if (res==0){
    // --- handle meta data ----------------
    AppData ad;
    if (result.GetMetaData()->Find(AppData::MD_UUID,"#[UUID]", ad)>=0){
      if (ad.tag==AppData::MD_USE_ASCII){
        //cerr << "before |"<< ad.sdata <<"|"<< endl ;
        while (ad.sdata.length()>0 && (ad.sdata[0]=='#' || ad.sdata[0]==' ' ||
          ad.sdata[0]=='\n'))
          ad.sdata.erase(0, 1);
        if (ad.sdata.length()>36)
          ad.sdata.erase(36, ad.sdata.length()-36);
        //cerr<<"after |"<<ad.sdata<<"|"<<" length: "<<ad.sdata.length()<<endl;
        UUID TempID(ad.sdata);
        result.SetUID(TempID);
      } else {
        BIASERR("UUID is no meta datum in binary images");
        BIASERR("possibly two concurrent UUIDs in one image");
        //result.SetUID(*((UUID*)ad.data));
        //BIASASSERT(false);
        BIASABORT;
      }
    } else {
#ifdef BIAS_DEBUG
#  ifdef BIAS_EXTRA_WARN
      BIASERR("warning, no valid UID read, UUID must be set manually");
#  endif
#endif
      result.InvalidateUID();
    }
    //cerr <<"header in ImportImage"<<endl;
    //result.PrintHeader(cout);
  }
  return res;
} // ImportImage





#ifdef BIAS_HAVE_IMLIB
///JW extracted from ImportImage 10/2004 
int ImageIO::ImportImlib(const string& filename, ImageBase &result, 
                         bool assumeGrey)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif
  // Data structures
  ImlibImage *img;
  //Static since I have not been able to find a function that frees Imlibdata
  //If we read one Image then it is reasonable to think that we will call 
  //this function again
  static ImlibData  *iData=NULL;
  static Display    *disp=NULL;

  if(disp==NULL)
    // Imlib really needs a pointer to the current display (on Unix)
    // connect to default X server
    disp=XOpenDisplay(NULL);
  if (disp==NULL)
    BIASERR("disp is NULL ater XOpenDisplay call.");
  if(iData==NULL)
    iData=Imlib_init(disp);

  img = Imlib_load_image(iData, (char*) filename.c_str());

  if (img == NULL) {
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR("Could not read or decode image file "<<filename<<" with Imlib. Nothing read!");
#endif
    return -1;
  }

  bool isGrey = true;
  // Do I have to test for grey at all?
  if (assumeGrey) {
    isGrey = true;
  } else {

    // Go through the image and test for equal R,G,B values for each pixel
    register unsigned char* start = img->rgb_data;
    register unsigned char* stop  = start + 
      ( img->rgb_width * img->rgb_height * 3);

    while (start < stop) {    
      if ( (*(start) == *(start+1)) && *(start)  == *(start+2)) {
        start+=3;
      } else {
        isGrey = false;
        break;
      }
    }
  }

  // must be empty before new image information can be inserted
  if (!result.IsEmpty()){
    result.Release();
  }

  // copy the raw data in
  register unsigned char* sourceImageData = img->rgb_data;
  register unsigned char* imageData;
  register unsigned char* imageDataStop;

  /// Friso said: imlib seems to read three channel images even for 
  // single channel grey images (JW 10/2004)
  if (isGrey) {
    // Grey image
    result.Init(img->rgb_width, img->rgb_height, 1);
    result.SetColorModel(ImageBase::CM_Grey);
    imageData     = (unsigned char*)result.GetImageData();
    imageDataStop = imageData + result.GetPixelCount();

    // Take every third value (r=g=b) to create an image with grey values
    for (; imageData < imageDataStop; imageData++, sourceImageData += 3) {
      *imageData = *sourceImageData;
    }

  } else {
    // Color image
    result.Init(img->rgb_width, img->rgb_height, 3);
    result.SetColorModel(ImageBase::CM_RGB);
    imageData     =  (unsigned char*)result.GetImageData();
    imageDataStop = imageData + 3 *  result.GetPixelCount();
    for (; imageData < imageDataStop; imageData++, sourceImageData++) {
      *imageData = *sourceImageData;
    }
  }

  Imlib_kill_image(iData,img); 
  //XCloseDisplay(disp);
  return BIAS_IO_SUCCESS;
}
#endif  // BIAS_HAVE_IMLIB



#ifdef BIAS_HAVE_IMLIB
// jw
int ImageIO::ExportImlib( const string& c_filename, 
                         const ImageBase &img,
                         enum TFileFormat FileFormat, 
                         bool sync,
                         int jpeg_quality, 
                         bool forceNewID,
                         const bool & writeMetaData) 
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif
  int result=-1;
  // first get copy of image in img, which is unsigned char
  // because most formats based on unsigned char
  Image<unsigned char> outpic;
  outpic=img;
  // Imlib needs an RGB image, interleaved with unsigned char as data type
  Image<unsigned char> RGB_version;
  ImageConvert::ToRGB(outpic, RGB_version);

  // create Imlib data structures
  ImlibImage imgImLib;
  struct _ImlibSaveInfo saveinfo;
  // scale from 0..100 to 0..255
  /// \todo: is this imlib scale 2.56 correct, shouldn't it be 2.55 *100 (JW)?
  saveinfo.quality = (int) (2.56 * (double) jpeg_quality); 

  /// static because we didn't find free for xopendisplay... (jw)
  static ImlibData  *idata=NULL;
  static Display    *imgDisp=NULL;

  /// \todo is this the source of the DISPLAY errors (jw)?
  if (imgDisp==NULL) imgDisp = XOpenDisplay(NULL);
  if (idata==NULL) idata = Imlib_init(imgDisp);

  // create an Imlib image and take the Image image data for this purpose
  imgImLib.rgb_width  = RGB_version.GetWidth();
  imgImLib.rgb_height = RGB_version.GetHeight();
  imgImLib.rgb_data   = RGB_version.GetImageData();
  //avoid const char* to char* cast-warning
  imgImLib.filename   = (char*)(string("dummy_file_name").c_str());
  // call the Imlib saving routine
  if (Imlib_save_image(idata, &imgImLib, (char*)c_filename.c_str(), &saveinfo) == 0) {
    BIASERR("Error saving " <<c_filename << " with imlib occured. Nothing saved!");
    result = -1;
  } else {
    // OK, successful saved 
    result = BIAS_IO_SUCCESS;
  };
  return result;
}
#endif




int ImageIO::ImportImagePnm_(ifstream& ifs, ImageBase &result, int depth)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  int xres=-1, yres=-1, maxval=-1, read=0;
  int res=0;
  char c;

  if (!result.IsEmpty())
    result.Release();
  // start from the beginning and read line with P5 or P6
  ifs.seekg(0,ios::beg);

  ifs >> c; 
  if (c!='P'){
    BIASERR("invalid magic cookie");
    return -1;
  }
  ifs >> c; 
  if (c!='5' && c!='6'){
    BIASERR("invalid magic cookie");
    return -1;
  }

  // this allows an arbitary amount of whitespace in between the header
  while (read<3){
    ifs >> c; 
    ifs.putback(c);
    if (c=='#'){
      ifs >> *result.GetMetaData();
      if (!ifs){
        BIASERR("error reading meta data");
        return -1;
      }
    } else if (c>29 || c<40){ // number 0-9
      switch (read){
    case 0:
      ifs >> xres;
      //cerr << "xres: "<<xres<<endl;
      read++;
      break;
    case 1:
      ifs >> yres;
      //cerr << "yres: "<<yres<<endl;
      read++;
      break;
    case 2:
      ifs >> maxval;
      //cerr << "maxval: "<<maxval<<endl;
      read++;
      break;
    default:
      BIASERR("funny error: "
        <<"there should be only 3 numerical values in pgm header");
      return -1;
      break;
      }; // switch
    } else { // if (c>29 || c<40){
      BIASERR("corrupted image header "<<c);
    }
  }
  ifs.read(&c, 1*sizeof(char)); // last whitespace

  //cerr <<"image has size "<<xres<<"x"<<yres<<"  with maxval "<<maxval<<endl;
  if (maxval==255){
#ifdef BIAS_DEBUG
    if (result.GetStorageType()!=ImageBase::ST_unsignedchar &&
      result.GetStorageType()!=ImageBase::ST_invalid){
        BIASERR("warning trying to load an unsigned char image into an image of"
          <<" storage type "<<result.GetStorageType());
        //BIASASSERT(false);
        BIASABORT;
    }
#endif
    result.Init(xres, yres, depth, ImageBase::ST_unsignedchar);
    ifs.read(reinterpret_cast<char *>(result.GetImageData()), xres * yres * depth);
  } else if (maxval==4095){ // 12 bit images
#ifdef BIAS_DEBUG
    if (result.GetStorageType()!=ImageBase::ST_unsignedshortint &&
      result.GetStorageType()!=ImageBase::ST_invalid){
        BIASERR("warning trying to load an unsigned short int image into an "
          <<"image of storage type "<<result.GetStorageType());
    }
#endif
    result.Init(xres, yres, depth, ImageBase::ST_unsignedshortint);  
    result.SetBitDepth(12);
    ifs.read(reinterpret_cast<char *>(result.GetImageData()), 
      sizeof(unsigned short int) * xres * yres * depth);
    // check if bytes need swapping
    unsigned short int *p=(unsigned short int *)result.GetImageData(), *pe;
    pe=p+result.GetPixelCount()*result.GetChannelCount();
    // check if pixel > maxval exist
    bool swap=false;
    while (p<pe) { swap = (swap || (*p > maxval)); p++; }
    // swap bytes if pixel > maxval existed
    if (swap){
      BIASERR("maxval > "<<maxval<<" : swapping LSB and MSB");
      char *cp=(char *)result.GetImageData(), *cpe;
      cpe=(char *)((unsigned short int *)cp + result.GetPixelCount()*
        result.GetChannelCount());
      char c;
      while (cp<cpe){
        c=cp[0]; cp[0]=cp[1]; cp[1]=c;
        cp+=2;
      }
    }
#ifdef BIAS_DAIMLERCHRYSLER
    // check image, since bosch fills last line with values above 4095
    p=((unsigned short int *)result.GetImageData());
    pe=p+result.GetPixelCount()*result.GetChannelCount();
    while (p<pe) {
      if (*p>maxval) *p=maxval; 
      p++; 
    }
#endif
  }

  if (!ifs.good()){
    BIASERR("error reading from ifs");
    res=-1;
  }
  return res;
}


int ImageIO::Save(const std::string& FileName,
                  const  ImageBase &img,
                  const enum TFileFormat FileFormat,
                  const bool sync,
                  const int jpeg_quality,
                  const bool forceNewID,
                  const bool & writeMetaData)
{
  return ExportImage(FileName, 
    img,
    FileFormat, 
    sync,
    jpeg_quality, 
    forceNewID,
    writeMetaData );
}



int ImageIO::ExportImage(const string& c_filename, 
                         const ImageBase &img_const,
                         const enum TFileFormat c_FileFormat, 
                         const bool sync,
                         const int c_jpeg_quality, 
                         const bool forceNewID,
                         const bool & writeMetaData) 
{ 
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif
  bool TriedIt = false;

  ImageBase img(img_const);
  // do NOT scale or shift in this routine. 
  // Do it before calling this routine if you need it. (JW 04/2003)

  int result = -1;
  int quality = c_jpeg_quality;

  // non const to change for auto detection: 
  TFileFormat FileFormat=c_FileFormat;
  if (FileFormat==FF_auto)
  {
    // determine by filename extension: 
    FileFormat=GetFileFormat(c_filename);
    // if we couldn't determine BIAS FileFormat 
    // give the external libs their own "low level" try
    if (FileFormat==FF_unknown)
      FileFormat=FF_auto;
  };

  // non const for adding suffix
  string filename(c_filename);
  BIAS::FilenameExpansion (c_filename, filename); // tilde


  if (img.GetStorageType()==ImageBase::ST_invalid){
    BIASERR("cannot save images with StoragType ST_invalid. ");
    return -1;
  }


  // -----------------------------------------------------
  // -- UUID handling following the idea
  // -- if the image does not have a UUID we generate one that is written to disk. 
  // Q: Shall the ID be assigned to the orig. image "by ref." too 
  //    which results in a *non const image?
  // A: Yes, we want the side effect of assigning the ID by reference,
  //    and not only to the local copy for disk only. (koeser, jw 12/Dec/2005)
  //    However, this behavior will change when we adopt the calling code 
  //    to make img const.
  // changed, woelk 09/2006
  // -----------------------------------------------------
  //
  // force a fresh UID
  if (forceNewID) {
    img.SetUID(UUID::GenerateUUID());
  }

  // append to metadataonly if newID forced
  // or metadata is really written
  // to avoid side effect in the long run
  if (writeMetaData && (!img.GetUID().IsValid()))
  {
    // the image to be saved will become const some day
    BIASWARN("UUID assignment required. Please fix your code and assign a UID in the calling function if desired.");
    img.SetUID(UUID::GenerateUUID());
  }
  // write uuid to meta data if valid (only temporary to utilize metadata i/o)
  if (writeMetaData)
  {
    BIASASSERT( img.GetUID().IsValid() );
    string suuid;
    img.GetUID().GetString(suuid);
    img.GetMetaData()->Add("#[UUID]", suuid);    
  }


  // -------------------------------------------------------------
  // ---  try to write with our internal format handlers, first.
  // -------------------------------------------------------------

  if (result!=BIAS_IO_SUCCESS) {
    // construct valid filename with appropriate extension
    switch (FileFormat) 
    {
    case FF_mip:
      filename =ImageIO::ExtensionName( filename, GetExtension(FileFormat) );
      result = Write(filename, img, sync, forceNewID);
      TriedIt = true;
      break;
    case FF_pgm:
    case FF_ppm:
      {
        filename = ExtensionName( filename, GetExtension(FileFormat) );
        // use our internal impl. only for 8bit, 16 bit is impl. incorrectly
        if (img.GetDepth()==1) /* Byte per Channel */
        {
          if ((FileFormat==FF_pgm) && (img.GetChannelCount()!=1)) 
          {
            // we need to convert: 
            Image<unsigned char> greyIMG(img);
            ImageConvert::IP_ToGrey(greyIMG);
            result = ExportImagePnm_(filename,greyIMG,sync,writeMetaData);
            TriedIt = true;
          } else {
            // no slow conversion, is already correct fmt 
            result = ExportImagePnm_(filename,img,sync,writeMetaData);
            TriedIt = true;
          };
        }
      }
      break; 
    case FF_jpg:
      filename = ImageIO::ExtensionName( filename, GetExtension(FileFormat) );
      if (c_jpeg_quality < 0 || c_jpeg_quality > 100) 
      {
        BIASERR("The quality "<<c_jpeg_quality
          <<" is out of valid range 0..100 % falling back to 100% full quality." );
        quality=100;
      }
      break;
    case FF_raw4096:
      BIASERR("writing raw4096 is not (yet) implemented in BIAS, only on WIN32 scorpion grab application.");
      return -7;
      break;

      //  // we could add our internal handlers for pnm, here.
      //  // however - the IMageMagicj and DevIL versions 
      //  // are more reliable, in particular for hdr formats. 

    case FF_auto: 
    case FF_unknown:
      // leave the filename as it is for fmt detection by filename.
      break;
    case FF_bmp:
    case FF_pbm:
    case FF_tif:
    case FF_png:
    case FF_viff:
    case FF_exr:
    case FF_dds:
      filename = ExtensionName( filename, GetExtension(FileFormat) );
      break;
    default:
      BIASERR("Unknown image file format for saving "<<c_filename
        <<"Nothing saved!"
        );
      return -1;
    } // switch
  };

  // --------------------------------------------------
  // ---   Try to write using external libraries.   ---
  // --------------------------------------------------

#ifdef BIAS_HAVE_OPENEXR
  if (result!=BIAS_IO_SUCCESS) {
    // handle only those formats OpenEXR supports 
    if (FileFormat==FF_exr) {
      result=ExportOpenEXR(filename, img );
      TriedIt = true;
    }
  }
#endif

#ifdef BIAS_HAVE_IMAGEMAGICKLIB
  // try to use ImageMagick++ library. 
  if (result!=BIAS_IO_SUCCESS) {
    // skip the formats ImageMagick cannot handle, 
    // at least those it claims to handle but is dpoing wrong things (e.g. dds)
    if ( 
      (FileFormat!=FF_dds) &&
      (FileFormat!=FF_exr) &&
      (FileFormat!=FF_mip) &&
      (FileFormat!=FF_raw4096) 
      ) {
        result=ExportMagickPP (filename, img, quality, writeMetaData);
        TriedIt = true;
    }
  }
#endif

#ifdef BIAS_HAVE_DEVIL
  // try to use DevIL library
  if (result!=BIAS_IO_SUCCESS) {
    if (
      (FileFormat!=FF_exr) &&
      (FileFormat!=FF_mip) &&
      (FileFormat!=FF_raw4096) 
      ) {
        result=ExportDevIL(filename, img);
        TriedIt = true;
    }
  }
#endif

  //#ifdef BIAS_HAVE_OPENCV
  //  if (res!=BIAS_IO_SUCCESS)
  //    res=ExportOpenCV(filename, result);
  //#endif

#ifdef BIAS_HAVE_IMLIB
  // try to use imlib library
  // special saving information, only needed for jpeg images
  //struct _ImlibSaveInfo saveinfo;
  //saveinfo.quality = (int) (2.56 * (double) jpeg_quality);
  if (result!=BIAS_IO_SUCCESS) {
    // skip unsupported fmts. 
    if (
      (FileFormat!=FF_dds) &&
      (FileFormat!=FF_exr) &&
      (FileFormat!=FF_mip) &&
      (FileFormat!=FF_raw4096) 
      ) 
    {
      result = ExportImlib( filename, img,
        FileFormat, 
        sync,
        quality, 
        forceNewID,
        writeMetaData);
      TriedIt = true;
    }
  }
#endif 

  if (!TriedIt) 
    BIASERR("Image not even tried to be written");

  // final result:
  return result;
}



int ImageIO::ExportImagePnm_(const string &FileName, 
                             const ImageBase &img, 
                             bool sync,
                             const bool & writeMetaData)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif
  ofstream ofs;
  // bool sync=false;

#ifdef   WIN32
  ofs.open(FileName.c_str(),ios::binary);
#else  //not WIN32
  ofs.open(FileName.c_str());
#endif //WIN32

  if (!ofs) {
    BIASERR("ExportImagePnm_() : open of file "<<FileName<<" failed");
    perror(FileName.c_str());
    return -1;
  }

#ifdef BIASDEBUG
  if (img.GetDepth()!=1)
    BIASERR(FUNCNAME <<" supports only 8 bit per Channel images. Please call "
    <<"ExportImage or Save instead.");
#endif // BIASDEBUG

  switch (img.GetChannelCount()) {
  case 3: // 8bit ppm
    ofs <<"P6"<<endl;
    break;

  case 1: // 8bit pgm
    ofs << "P5"<<endl;
    break;
  default:
    BIASERR("ExportImagePnm_() : unsupported number of channels: "<<
      img.GetChannelCount());
    return -1;
  }

  if (writeMetaData){
    img.GetMetaData()->WriteAscii(ofs);
    if (!ofs){
      BIASERR("writing MetaData failed");
      return -2;
    } 
  };

  ofs << img.GetWidth()<<" "<<img.GetHeight()<<endl;
  ofs << "255"<<endl;

  ofs.write(reinterpret_cast<char *>(img.GetImageData()), img.GetSizeByte());
  if(!ofs) {
    BIASERR("ExportImagePnm_() : write failed");
    return -1;
  }

  if (sync) {
    // flush streambuf from c++
    ofs.flush();
    // flush system buffers
#ifndef _WIN32
#  if __GNUC__ < 3
    fsync(ofs.rdbuf()->fd());
#  else
    ofs.rdbuf()->pubsync();
#  endif
#else // _WIN32
    ofs.rdbuf()->pubsync();
#endif //_WIN32
  }

  ofs.close();
  return 0;
}


std::string ImageIO::ExtensionName(const std::string & oldName, 
                                   const std::string & newExtension) 
{
  string name    = oldName; // full name
  string ext     = BIAS::Extension( oldName );
  string newExtLC= BIAS::LowerCase( newExtension );

  // catch empty file name: 
  if (name.size()<1)
    name="out___"; // default

  // append extension if it is not the correct case insensitive extension: 
  if (ext!=newExtLC) {
    BIASASSERT(name.size()>0);
    const char c = name[name.size()-1];
    if (c!='.')
      name+= ".";
    name+=newExtLC;
  }
  return name;
}


int ImageIO::Write(const string& FileName, const ImageBase &img_const, 
                   bool sync, bool forceNewID) 
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  int res = 0;
  ofstream os;

  ImageBase img(img_const);

  // assign new id if forced or there is none yet
  if ((forceNewID) || (!img.GetUID().IsValid())) {
    img.SetUID(UUID::GenerateUUID());
  }

  // avoid ".mip.mip" (JW) 
  string file = ImageIO::ExtensionName( FileName, "mip");

#ifdef   WIN32
  os.open(file.c_str(),ios::binary);
#else  //not WIN32
  os.open(file.c_str());
#endif //WIN32

  if (!os){
    string Errorstring="Error opening "+file+" writing";
    perror(Errorstring.c_str());
    return -1;
  }

  os << img;

  res = (os.good())?(0):(-5);

  if (res==0){
    //cerr <<"Writing MetaData\n";
    //img.GetMetaData()->Dump();
    img.GetMetaData()->WriteBinary(os);
    if (!os){
      BIASERR("writing MetaData failed");
      res=-6;
    }
  }

  if (sync) {
    // flush streambuf from c++
    os.flush();
    // flush system buffers
#ifndef _WIN32
#  if __GNUC__ < 3 // in gcc 2.9x we can get a POSIX file descriptor 
    fsync(os.rdbuf()->fd());
#  else   // in gcc 3.x this is prohibited :-(
    os.rdbuf()->pubsync(); 
    //     ::sync(); // this is systemcall sync() to sync all buffers    
#  endif
#else // _WIN32
    /** @todo Sync written files and buffers to disk under WIN32 
    (evers 2004-10-14) */
    // 	os.rdbuf()->pubsync();
    //     ::sync();
#endif // _WIN32
  }
  os.close();

#ifndef _WIN32 // alternative way to sync, reopen file to get a handle
  int fd = open(file.c_str(),O_APPEND);
  fsync(fd); // and now fsync()
  close(fd);
#endif 
  return res;
}


int ImageIO::Read(const string& FileName, ImageBase &img)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  int res=0;
  ifstream is;

#ifdef   WIN32
  is.open(FileName.c_str(),ios::binary);
#else  //not WIN32
  is.open(FileName.c_str());
#endif //WIN32

  if (!is){
    string Errorstring="Error opening "+FileName+" for reading";
    perror(Errorstring.c_str());
    return -1;
  }
  is >> img;


  res = (is.good())?(0):(-5);

  if (res==0){
    //cerr << "ImageIO::Read() : "<<is.tellg()<<endl;
    is >> *img.GetMetaData();
    if (!is && !is.eof()){
      BIASERR("error reading MetaData");
      res = -6;
    }
  }

  //img.GetMetaData()->Dump();

  is.close();


  return res;
}



int ImageIO::ImportImageViff_(std::ifstream& ifs, ImageBase &result)
{  
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  /*class*/ xvimage viff;
  if (!ifs) 
    return -1;
  if (!result.IsEmpty())  
    result.Release();
  ifs.seekg(0, ios::beg);  

  // read header struct (jw)
  ifs.read((char*)&viff,sizeof(viff));
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  cout <<"read viff size "<<viff.row_size<<" x "<<viff.col_size<<endl;
#endif
  // too big viff image size usually means it is no viff image
  // and would blow memory (jw 08/2005)
  const unsigned long maxUsefulDim=4096*2;
  const unsigned long minUsefulDim=8;
  if ( viff.row_size>maxUsefulDim 
    || viff.col_size>maxUsefulDim
    || viff.row_size<minUsefulDim 
    || viff.col_size<minUsefulDim
    )
  {
    // corrupt image?
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR(
      "viff image dim "<<viff.row_size<<"x"<<viff.col_size<<" is not useful(>"<<maxUsefulDim<<"),"<<endl
      <<"handling it as error. Skipping viff read. Adapt maxUsefulDim if you feel this behavior is wrong. "
      );
#endif
    return -1; // error
  }

  //   if (xres %4 >0) xresaligned = xres - xres%4;
  //   else xresaligned = xres;

  //result.Init(xresaligned, yres, depth, ImageBase::ST_unsignedchar);
  result.Init(viff.row_size, viff.col_size, 
    1, ImageBase::ST_float); // 1 is the number of channels
  //  cerr << "initialized with " << xres << " " << yres 
  // << " unsigned char" << endl;

  // read data area (jw)
  ifs.read(reinterpret_cast<char*>(result.GetImageData()), result.GetSizeByte());
  result.SetColorModel(ImageBase::CM_Grey);

  return 0;  // ok
}


#ifdef BIAS_HAVE_IMAGEMAGICKLIB
/** @brief: This routine is working inefficiently, 
because it doesn't know what to do about conversion if read and requested 
types don't match (JW 10/2004)
However it works and assumes RGB in case of doubt.
\todo rewrite by making more use of type Image<..> instead of ImageBase
because lossless conversions and warnings for unwanted lossy conversions
are easier, then (JW).
@author Jan Woetzel 10/2004 */
int ImageIO::ImportMagickPPAutoconvert(const std::string& FileName,
                                       ImageBase& result,
                                       const bool &/*assumeGrey*/,
                                       const bool &readMetaData )
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

#ifdef WIN32
  // initialize DLL loaded library (not required on Linux but once on Windows)
  Magick::InitializeMagick(NULL);
#endif

  try {

    Magick::Image image; // temporary converter for lib usage

    // Read a file into image object using Magick++ library function:
    image.read( FileName );

#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    COUT(
      "DBG Read "<<image.columns()<<"x"<<image.rows()<<" "
      <<image.magick()<<" image "
      <<FileName<<" using ImageMagick++"<<endl
      <<"  depth: "<<image.depth()<<" bit."<<endl
      <<"  comment: "<<endl<<image.comment()<<endl 
      );
#endif

    // determine if import image format/channels and bit depth match:
    Magick::StorageType outputST=Magick::CharPixel; // output storage type

    /** keep the output storagetype
    \todo TODO HACK add dynamic cast test for templated Image<> to convert loaded float imaeg automatically to uc etc.. (JW)
    \todo add more knowledge about conversion and the option of errors if requested and read type don't match (JW).    
    check if at least StorageType and bit depth match:
    */

    //
    // first switch: ouput storagetype 
    //
    switch (result.GetStorageType()) 
    {
    case BIAS::ImageBase::ST_invalid:
      BIASERR("invalid Storagetype ST_invalid FileName="<<FileName<<endl);
      break;

    case BIAS::ImageBase::ST_unsignedchar:
    case BIAS::ImageBase::ST_char:
      // 8 bit
      if (image.depth() != 8*sizeof(char)){
        BIASERR("StorageType is "<<8*sizeof(char)<<" bit but data has "<<image.depth()<<" bit depth. FileName="<<FileName<<endl);
        /** \todo handle conversion in extra routine 
        with extra warnings in case of scaling or lossy conversion.
        -Add a 'normalized scale' per conversion for each Storagetype.
        -Add a flag to return errors instead of run-time warning. (discussion 26.Oct.2004) (JW)  */
        return -1; 
      } else {
        outputST=Magick::CharPixel;
        COUT("is an "<<8*sizeof(char)<<" bit (u)char image\n");
      };
      break;

    case BIAS::ImageBase::ST_unsignedshortint:
    case BIAS::ImageBase::ST_shortint:
      // 16 bit
      if (image.depth()!=8*sizeof(short)){
        BIASERR("StorageType is "<<8*sizeof(short)<<" bit "
          <<"but data has "<<image.depth()<<" bit depth. "
          <<"FileName="<<FileName<<endl);
        return -2; 
      } else {
        outputST=Magick::ShortPixel;
        COUT("is a "<<8*sizeof(short)<<" bit (u)short image\n");
      };
      break;

    case BIAS::ImageBase::ST_unsignedint:
    case BIAS::ImageBase::ST_int:
      // 32 bit int
      if (image.depth()==8*sizeof(int)){
        BIASERR("StorageType is "<<8*sizeof(int)<<" bit but data has "<<image.depth()<<" bit depth. FileName="<<FileName<<endl);
        return -3; 
      } else {
        outputST=Magick::IntegerPixel;
        COUT("is a "<<8*sizeof(int)<<" bit (u)int image\n");
      };
      break;

      /**  \todo float and double pixel are 0..1 normalized range in ImageMagick++. Min, max meta data required (JW) */
    case BIAS::ImageBase::ST_float:
      // 32 bit float
      if (image.depth()!=8*sizeof(float)) {
        BIASERR("StorageType is "<<8*sizeof(float)<<" bit but data has "<<image.depth()<<" bit depth. FileName="<<FileName<<endl);
        return -4;
      } else {
        outputST=Magick::FloatPixel;
        COUT("is a "<<8*sizeof(float)<<" bit float image\n");
        BIASERR("warning: You are using ImageMagick to read a float image. Be aware that ImageMagick float values are normalized to 0..1 !!!"<<endl);
      };
      break;

    case BIAS::ImageBase::ST_double:
      // 64 bit double 
      if (image.depth()!=8*sizeof(double)){
        BIASERR("StorageType is "<<8*sizeof(double)<<" bit but data has "<<image.depth()<<" bit depth. FileName="<<FileName<<endl);
        return -4; 
      } else {
        outputST=Magick::DoublePixel;
        COUT("is a "<<8*sizeof(double)<<" bit float image\n");
        BIASERR("warning: You are using ImageMagick to read a double image. Be aware that ImageMagick float values are normalized to 0..1 !!!"<<endl);
      };
      break;

    default:
      BIASERR("StorageType "<<result.GetStorageType()<<" not supported. FileName="<<FileName<<endl);
      return -9; // error
    };


    //
    // second switch for channels: number and color model
    //
    string magickMap="RGB"; // default
    unsigned int nChannels=3;
    //determine import image format/channels:
    switch (result.GetColorModel()) {
  case BIAS::ImageBase::CM_invalid:
    COUT("DBG CM_invalid. Setting default map RGB\n");
    nChannels=3;
    magickMap="RGB";
    break;

  case BIAS::ImageBase::CM_Grey:
    COUT("DBG CM_Grey\n");
    magickMap="K"; // blacK=grey
    nChannels=1;
    break;
  case BIAS::ImageBase::CM_GreyA:
    //COUT("DBG CM_GreyA\n");
    magickMap="KA";
    nChannels=1;
    break;
  case BIAS::ImageBase::CM_RGB:
    //COUT("DBG CM_RGB\n");
    magickMap="RGB";
    nChannels=3;
    break;
  case BIAS::ImageBase::CM_RGBA:
    //COUT("DBG CM_RGBA\n");
    magickMap="RGBA";
    nChannels=4;
    break;
  case BIAS::ImageBase::CM_BGR:
    //COUT("DBG CM_BGR\n");
    magickMap="BGR";
    nChannels=3;
    break;

  default:
    BIASERR("unknown result Color Model "<<result.GetColorModel()<<" "
      <<"reading "<<FileName<<endl);
    };

    // init/allocate result color image data: 
    if (!result.IsEmpty()) result.Release();
    result.Init(image.columns(), image.rows(), nChannels);
    result.SetColorModel(ImageBase::CM_RGB);


    // get the pointer to our result BIAS image data: 
    unsigned char* p_imageData     =  (unsigned char*)result.GetImageData();
    BIASASSERT( p_imageData != NULL);
    // write data array from read in ImageMagick++ image 
    // to our own ImageData Array.
    // details (JW): http://studio.imagemagick.org/Magick++/Image.html#write 
    image.write(0,0, // upper left corner of region to extract 
      result.GetWidth(), result.GetHeight(), // region size to extract
      "RGB", // map=output format in our pixel data, e.g RGBA
      outputST, 
      p_imageData // pointer to our data array of type
      );
    if (readMetaData){
      BIASERR("reading of MetaData is not implemented for IMageMagick++ but could be done using BIAS::ImageAttributes class (JW 10/2004)\n");
    }
  } 
  catch( Magick::Exception &error_ ) {
    BIASERR("Reading "<<FileName<<" with Magick++ failed."<<endl
      <<"  Caught exception: "<<endl
      <<"  "<<error_.what()<<endl );
    return -1; // error
  }
  catch (...) {
    BIASERR("Reading "<<FileName<<" with Magick++ failed."<<endl);
    return -2; // error
  }
  return  BIAS_IO_SUCCESS;// OK
}


// JW
int ImageIO::ExportMagickPP(const std::string& FileNameStr, 
                            const ImageBase& input,
                            const int &quality,
                            const bool &writeMetaData,
                            const bool &forceBinaryPNMformat,
                            const bool &losslesJpgMode,
                            const bool &writeGZipped,
                            const string & comment
                            )
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  // binding local namespace scope
  using namespace Magick;
  using namespace MagickLib;	

  // const to var. conversio to add .gz suffix (later);
  string FileName(FileNameStr);

#ifdef WIN32
  // initialize DLL loaded library (not required on Linux but once on Windows)
  Magick::InitializeMagick(NULL);
#endif

  try {

    // temporary used for saving to disk
    Magick::Image image; 
    // manipulate image information, e.g. compression

    image.colorSpace(RGBColorspace); // default 

    // determine export image format/channels:
    std::string pixmap;
    switch (input.GetChannelCount()) 
    {
    case 1:
      pixmap="K"; 
      image.colorSpace(GRAYColorspace); // default
      break;
    case 2:
      pixmap="KA"; break;
    case 3:
      pixmap="RGB"; break;
    case 4:
      pixmap="RGBA"; break;
    default:
      BIASERR("unsupported channelcount. Could not write image with Magick++");
      return -1;
    }


    // refine the color space used 
    // there is a difference between color model and channel order
    switch (input.GetColorModel())
    {				
    case BIAS::ImageBase::CM_RGB :
      // color model
      image.colorSpace(RGBColorspace);
      // channel order (within colro model)
      pixmap="RGB";
      break;
    case BIAS::ImageBase::CM_BGR :
      image.colorSpace(RGBColorspace);
      pixmap="BGR";
      break;
    case BIAS::ImageBase::CM_RGBA :
      image.colorSpace(RGBColorspace);
      pixmap="RGBA";
      break;
      //case BIAS::ImageBase::CM::CM_BGBA :
      //	pixmap="BGRA";
      //	//image.colorSpace(RGBColorspace);
      //	break;
    case BIAS::ImageBase::CM_Grey :
      pixmap="K";
      image.colorSpace(GRAYColorspace);
      image.type(GrayscaleType);
      break;
    case BIAS::ImageBase::CM_GreyA :
      pixmap="KA";
      //image_info->colorspace=GRAYColorspace;
      image.colorSpace(UndefinedColorspace);
      image.type(GrayscaleMatteType);
      break;
    case BIAS::ImageBase::CM_HSV :
      // HSV=HSB (which is different from HSI and HSL!)
      image.colorSpace(HSBColorspace);
      break;
    case BIAS::ImageBase::CM_HSL :
      // HSL=HSI (which is different from HSV nor HSB!)
      image.colorSpace(HSLColorspace);
      break;
    default:
      BIASERR("unsupported color model. Please add switch for "<<input.GetColorModel());
    }


    // determine per pixel input/output storagetype: 
    Magick::StorageType magick_ST=Magick::CharPixel; // default
    image.depth(8); // default depth in bit 

    // CharPixel, ShortPixel, IntegerPixel, FloatPixel, or DoublePixel
    switch (input.GetStorageType())
    {
    case BIAS::ImageBase::ST_unsignedchar:
    case BIAS::ImageBase::ST_char:
      magick_ST=Magick::CharPixel; 
      image.depth(8*sizeof(char)); // usually 8 bit
      break;
    case BIAS::ImageBase::ST_unsignedshortint:
    case BIAS::ImageBase::ST_shortint:
      magick_ST=Magick::ShortPixel; 
      image.depth(8*sizeof(short)); // usually 16bit
      break;
    case BIAS::ImageBase::ST_unsignedint:
    case BIAS::ImageBase::ST_int:
      magick_ST=Magick::IntegerPixel; 
      image.depth(8*sizeof(int)); // usually 32bit
      break;
    case BIAS::ImageBase::ST_float:
      magick_ST=Magick::FloatPixel; 
      image.depth(8*sizeof(float)); // usually 32bit
      break;
    case BIAS::ImageBase::ST_double:
      magick_ST=Magick::DoublePixel; 
      image.depth(8*sizeof(double)); // usually 64 bit
      break;
    default:
      BIASERR("unsupported input image BIAS StorageType "<<input.GetStorageType() );
    };

    // input depth and switch case depth should match
    // check if ImageMagick has enough bits to store our data
    if (8*input.GetDepth() != image.depth()){
      BIASERR("ImageMagick bitdepth="<<image.depth()
        <<" seems to be smaller than requested "<<8*input.GetDepth()<<" bit input"<<endl
        <<"Do you use a bit-limited ImageMagick version , e.g. ImageMagick-16bit?"<<endl
        <<"You cannot save this image to "<<FileNameStr<<" with your version of ImageMagick due to dynamic range problems!"
        );
      return -1;
    }
    BIASASSERT( 8*input.GetDepth() == image.depth() );


    // for jpeg compression:
    image.quality(quality);

    // ASCII / binary data format: 
    if (forceBinaryPNMformat) 
    {
      string ext = BIAS::LowerCase( BIAS::Suffix(FileName) );
      // useful option only for pnm images JW
      if (   (ext==".pnm")
        || (ext==".pbm")
        || (ext==".pgm")
        || (ext==".ppm")
        )
      {	
        // to write Binary data (instead of ASCII) 
        // set any compression except "none" and "unknown", e.g. RLE works
        // no matter if this compression is really used by coder... (JW)
        //image.compressType(UndefinedCompression); // makes binary data, too. 
        //image.compressType(NoCompression);
        //image.compressType(BZipCompression);
        //image.compressType(FaxCompression);
        //image.compressType(Group4Compression);
        //image.compressType(JPEGCompression);
        //image.compressType(JPEG2000Compression);
        //image.compressType(LosslessJPEGCompression);
        //image.compressType(LZWCompression);
        image.compressType(RLECompression);
        //image.compressType(ZipCompression);

      } else {
        // ASCII PNM - blows size up compared to binary writing
        image.compressType(NoCompression); 
      };
    } // textInsteadOfBinaryMode

    if (losslesJpgMode)
    {
      image.compressType(LosslessJPEGCompression);
      /// quality and compession are two different parameters in jpeg
      /// but the user usually wants lossles *and* no compression on lossless flag,
      /// so force 100% quality, here: 
      image.quality(100);
    }

    // Read a file into image object using Magick++ library function:
    unsigned char* p_imageData     =  (unsigned char*)input.GetImageData();
    BIASASSERT(p_imageData!=NULL);

    // non const copy: add suffix for zipped format
    // useful for ASCII text format liek ASCII pgm --> .pgm.gz
    if (writeGZipped)
      if (BIAS::Suffix(FileName)!=".gz")
        FileName+=".gz";

#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    COUT("DBG writing "<<input.GetWidth()<<"x"<<input.GetHeight()<<" "
      <<pixmap<<" image \""<<FileName
      <<"\" using ImageMagick++ "<<endl);
#endif // BIAS_DEBUG

    // create ImageMagick pixel buffer object from our image data: 
    image.read(input.GetWidth(), // dimension
      input.GetHeight(),
      pixmap, // output map type. e.g. RGBA
      magick_ST, // desired storagetype
      p_imageData  // pointer to our own data array to be written to disk
      );

    if (writeMetaData) {
      /// JW: MetaData could theoretically be written to image.comment 
      /// with BIAS ASCII representation
      /// However, comment header is limited to 4096 Bytes for some formats
      /// and ASCII BIAS::MetaData representation is too long to do so! (JW)
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
      BIASERR("writing MetaData using ImageMagick IO is not implemented.");
#endif
    }
    // write supplied comment/description data
    if (comment.size()>=4096){
      BIASWARN("comment size "<<comment.size()<<" is >= 4096 characters. This may be too long for some image formats.");
    }
    image.comment( comment ); //< replaces

    // start *really* writing to disk using Magick++ 
    // output file type is automatically determined by magick file extension 
    image.write(FileName);
  } 
  catch( Magick::Exception &error_ ) {
    BIASERR("Writing "<<FileName<<" with Magick++ failed."<<endl
      <<"  Caught exception: "<<endl
      <<"  "<<error_.what()<<endl );
    return -1; // error
  }
  catch(...) {
    BIASERR("Writing "<<FileName<<" with Magick++ failed."<<endl);
    return -2; // error
  }

  return  BIAS_IO_SUCCESS;// OK
}

#endif // BIAS_HAVE_IMAGEMAGICKLIB  


// JW
int ImageIO::ImportRAWwithHeader(const std::string & filename, 
                                 BIAS::ImageBase & img)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  // header data size:
  unsigned int hdrSZ=0;
  // tags to find
  unsigned int Cols=0;
  unsigned int Rows=0;


  { // scope
    ifstream is;
    is.open(filename.c_str()); 
    //is.open(filename.c_str(), ios::binary);
    //is.open(filename.c_str(), ios_base::in |ios_base::binary );
    if (!is){
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
      BIASERR("could not open image file: "<<filename);
#endif
      is.close();
      return -1;
    }
    // OK
    string line;
    getline(is, line);
    //cout<<"DBG "<<FUNCNAME<<"line: "<<line<<endl;
    if (line.size()<3) return -2; // too small. 

    // parse header: 
    string ref7("RAW4096"); // 7 Bytes instead of std. 8 for backward compatibility 
    string ref("RAW04096");
    if ( (0==line.compare(0, ref.size(), ref))
      || (0==line.compare(0, ref7.size(), ref7)) )
    {
      COUT("RAW04096/RAW4096 found."<<endl);
      hdrSZ=4096;
    } else {
      ref="RAW08192";
      if (0==line.compare(0, ref.size(), ref)){
        hdrSZ=8192;
      } else {
        ref="2048";
        if (0==line.compare(0, ref.size(), ref)){
          hdrSZ=2048;
        } else {
          ref="1024";
          if (0==line.compare(0, ref.size(), ref)){
            hdrSZ=1024;
          } else {
            ref="0512";
            if (0==line.compare(0, ref.size(), ref)){
              hdrSZ=0512;
            } else {
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
              BIASERR(filename<<" is no RAW4096/RAWnnnn image. aborting.");
#endif
              return -3;
            }
          }
        }
      }
    }


    // create string that can be used as char array:
    string hdr(hdrSZ,' ');
    is.seekg(ios::beg);
    // read data as a block: 
    is.read(&(hdr[0]), hdrSZ);

    bool ok=true;
    //ok = ok && (0==Parse(hdr, "Cols=", Cols));
    ok = ok && (0==ParseRAWwithHeader(hdr, "RowInc=", Cols));
    ok = ok && (0==ParseRAWwithHeader(hdr, "Rows=", Rows));

    if (ok) cout<<"found dim "<<Cols<<"x"<<Rows<<endl;
    is.close();
  }

  if (!img.IsEmpty()) img.Release();
  // HACK: content of RAW is fixed for scorpion 1600x1200 
  // as RAW 1 plane Bayer pattern, 8 bit. (JW)  
  img.Init(Cols, Rows, 1);
  img.SetColorModel(BIAS::ImageBase::CM_Bayer_RGGB);

  //
  // reopen file with C fopen intead of C++ fstream
  // USE C because of higher performance (measured 15%)
  //
  //TODO 
  // FIXME: Linux reads only one line - thu sit is not using binary mode! (jw)
  FILE * fp=NULL;
  //fp = fopen(filename.c_str(), "rb");
  fp = fopen(filename.c_str(), "rt");
  if (fp==NULL){
    cout<<"could not open "<<filename<<" for C reading"<<endl;
    return -3;
  }
  //OK
  unsigned int dataSZ=Cols*Rows;
  //memset(&(img.GetImageData()[0]), 0, dataSZ);
  {
    char * p_data = new char[dataSZ];
    memset(p_data, 0, dataSZ);

    unsigned int readBytes=0;
    // skip header 
    readBytes=fread( p_data, hdrSZ, 1, fp );

    // read data:
    readBytes=fread( p_data, dataSZ, 1, fp );

    img.CopyIn_NoInit(p_data);
    //tmp.CopyIn_NoInit(p_data);

    fclose(fp);
    delete p_data;
  }

  return 0; // OK
}


// JW
int ImageIO::ParseRAWwithHeader(const std::string & buf, 
                                const std::string & tag, 
                                unsigned int & retVal)
{ 
  string::size_type posStart, posEnd;
  string valueStr;
  posStart=buf.find(tag, 0);
  if (posStart!=string::npos){
    // found tag
    posEnd = buf.find( string("\n"), posStart);
    if (posEnd!=string::npos){
      valueStr=buf.substr(posStart+tag.size(), posEnd-posStart-tag.size());
      stringstream ss;
      ss.str(valueStr);
      ss>>retVal;
      cout<<"valueStr="<<valueStr<<" : "<<retVal<<endl;
    }
  } else {
    cout<<"not found!"<<endl;
    return -1;
  }
  // OK.
  return 0;
}


#ifdef WIN32
#  pragma warning (push)
#  pragma warning (disable: 4702) // unreachable code
int ImageIO::WriteUnbuffered(const std::string& /*FileName*/, 
                             const ImageBase &/*img*/)
{
  BIASERR("ImageIO::WriteUnbuffered() not implemented in Win32");
  BIASABORT;
  return -1;
}
#  pragma warning (pop)

#else
int ImageIO::WriteUnbuffered(const std::string& FileName, 
                             const ImageBase &img_const)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif
  ImageBase img(img_const);

  if (!img.GetUID().IsValid()) 
    img.SetUID(UUID::GenerateUUID());

  std::ostringstream FlatImage;
  FlatImage << img;
  img.GetMetaData()->WriteBinary(FlatImage);
  FlatImage.flush();
  unsigned long int is = FlatImage.str().size();
  //   cout <<"FlatImage.size: "<<is<<endl;;
  if (MemAlignedBufferSize_ < is) {
    if (MemAlignedBuffer_ != NULL) free(MemAlignedBuffer_);
    MemAlignedBufferSize_ = 512;
    while (MemAlignedBufferSize_ < is) 
      MemAlignedBufferSize_ +=512;

    // TODO: warning: dereferencing type-punned pointer will break strict-aliasing rules

    //     cout <<"New MemAlignedBuffer_ with size "<<MemAlignedBufferSize_ <<endl;
    char** bufferPointer = &MemAlignedBuffer_;
    posix_memalign((void**)bufferPointer, 512, MemAlignedBufferSize_);

    if (MemAlignedBuffer_ == NULL) {
      BIASERR("Can not posix_memalign()"<<MemAlignedBufferSize_<<" bytes");
      return -1;
    }
  }
  // copy flattened image into Buffer
  memcpy(MemAlignedBuffer_, FlatImage.str().c_str(), is);

  int fd = open(FileName.c_str(),O_WRONLY | O_CREAT | O_DIRECT);
  if (fd <0) {
    BIASERR("Can not open/create file ");
    perror(FileName.c_str());
    return -1;
  }
  fchmod(fd,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
  int res = write(fd,MemAlignedBuffer_, MemAlignedBufferSize_);
  if (res <0) {
    BIASERR("Can not write with O_DIRECT to file ");
    perror(FileName.c_str());
    return -1;
  }

  close(fd);

  return 0;
}
#endif // WIN32



#ifdef BIAS_HAVE_DEVIL
// JW

void ImageIO::InitDevIL()
{
  static bool il_initialized=false;
  // once
  if (!il_initialized) 
  {
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    // Check if the shared lib's version matches the executable's version.
    if (ilGetInteger(IL_VERSION_NUM) < IL_VERSION) {
      BIASERR("DevIL/OpenIL: "<<endl
        <<"IL versions of run-and compiletime are different. "<<endl
        <<"compile time IL_VERSION    ="<<IL_VERSION<<endl
        <<"runtime      IL_VERSION_NUM="<<ilGetInteger(IL_VERSION_NUM)<<endl
        <<"There is a risk that the headers dont match the shared library/DLL. ");
    };
#  ifdef USE_DEVIL_ILU
    if (iluGetInteger(ILU_VERSION_NUM) < ILU_VERSION) {
      BIASERR("DevIL/OpenIL: "
        <<"ILU versions of run-and compiletime are different. "
        <<" There is a risk that the headers dont match the DLL. ");
    };
#  endif // USE_DEVIL_ILU
#endif // BIAS_DEBUG && ImageIO_DEBUG_IO

    // global state initialization, required once: 
    ilInit();
#ifdef USE_DEVIL_ILU
    iluInit();
#endif 

    // done: 
    il_initialized=true;

#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    COUT("initialized DevIL(OpenIL) runtime version: "
      <<ilGetInteger(IL_VERSION_NUM)<<endl );
#endif
  }
}


void ImageIO::PrintInfoDevIL(std::ostream & os)
{
  os
    <<"dim:         "<<ilGetInteger(IL_IMAGE_WIDTH)<<"x"
    <<ilGetInteger(IL_IMAGE_HEIGHT)<<" pixel"<<endl
    <<"depth=planes "<<ilGetInteger(IL_IMAGE_DEPTH)<<" plane(s) ";
  if (ilGetInteger(IL_IMAGE_DEPTH)>1)
    os<<"(tex3D)";
  os<<endl;
  os<<"size of data "<<ilGetInteger(IL_IMAGE_SIZE_OF_DATA)<<" Byte total"<<endl
    <<"BytePerPixel "<<ilGetInteger(IL_IMAGE_BPP)<<endl
    <<"Bytes/Pixel  "<<ilGetInteger(IL_IMAGE_BYTES_PER_PIXEL)<<endl
    <<"Bits/Pixel   "<<ilGetInteger(IL_IMAGE_BITS_PER_PIXEL)<<endl
    <<"num images   "<<ilGetInteger(IL_NUM_IMAGES)<<endl
    <<"num mipmaps  "<<ilGetInteger(IL_NUM_MIPMAPS)<<endl	

    <<"origin_set   "<<ilGetInteger(IL_ORIGIN_SET)<<" "
    <<"image_origin "<<ilGetInteger(IL_IMAGE_ORIGIN)<<" "
    ;
  switch (ilGetInteger(IL_IMAGE_ORIGIN)) {
  case IL_ORIGIN_SET: 
    os<<"IL_ORIGIN_SET"; break;
  case IL_ORIGIN_LOWER_LEFT: 
    os<<"IL_ORIGIN_LOWER_LEFT"; break;
  case IL_ORIGIN_UPPER_LEFT: 
    os<<"IL_ORIGIN_UPPER_LEFT"; break;
  case IL_ORIGIN_MODE: 
    os<<"IL_ORIGIN_MODE"; break;
  };
  os<<endl;

  os<<"offset       "
    <<ilGetInteger(IL_IMAGE_OFFX)<<" "<<ilGetInteger(IL_IMAGE_OFFY)<<endl
    ;
}

int ImageIO::ImportDevIL(const std::string& FileName, 
                         ImageBase& result, 
                         const unsigned int & tex3DplaneZ )
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  if (!result.IsEmpty()) 
    result.Release();

  InitDevIL();
  ILuint id(0), err(0);
  ilGenImages(1, &id);
  ilBindImage( id );

  bool ok = (IL_TRUE==ilLoadImage( (char*)(FileName.c_str()) ));

  err = ilGetError();
  if (err!=IL_NO_ERROR || !ok) {
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    stringstream ss;
    ss<<"could not load "<<FileName<<" ";
#  ifdef USE_DEVIL_ILU
    // there is no erro msg of !ok is the reason for failure and no err
    if (err!=IL_NO_ERROR)
      ss<<": "<<iluErrorString(err);
#  endif
    BIASERR(ss.str());
#endif // BIAS_DEBUG && ImageIO_DEBUG_IO
    ilDeleteImages(1, &id);
    return -1;
  }

#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  PrintInfoDevIL();
#endif

  // multiple pages/planes per image, useful for tiff etc.
  const int nDepthPlanes = int( ilGetInteger(IL_IMAGE_DEPTH) );
  if (nDepthPlanes>1){

    BIASERR("detected tex3D with "<<nDepthPlanes<<" planes. It is a volume, not a 2D texture. NOT implemented.");
    BIASBREAK;
    ilDeleteImages(1, &id);
    return -1;
  }

  // tex3D access?
  if (nDepthPlanes<(int)tex3DplaneZ ){
    BIASERR("Image has only "<<nDepthPlanes<<" z planes, but you requested tex3DplaneZ="<<tex3DplaneZ);
    BIASBREAK;
    ilDeleteImages(1, &id);
    return -1;
  }

  // determine pixel data type: 
  BIAS::ImageBase::EStorageType  st = BIAS::ImageBase::ST_invalid;
  const int typeIL = (int)ilGetInteger(IL_IMAGE_TYPE);
  switch (typeIL) 
  {
  case IL_BYTE:
    st = ImageBase::ST_char; break;
  case IL_UNSIGNED_BYTE:
    st = ImageBase::ST_unsignedchar; break;
  case IL_SHORT:
    st = ImageBase::ST_shortint; break;
  case IL_UNSIGNED_SHORT:
    st = ImageBase::ST_unsignedshortint; break;
  case IL_INT:
    st = ImageBase::ST_int; break;
  case IL_UNSIGNED_INT:
    st = ImageBase::ST_unsignedint; break;
  case IL_FLOAT:
    st = ImageBase::ST_float; break;
  case IL_DOUBLE:
    st = ImageBase::ST_double; break;
  default: 
    BIASERR("unsupported storage type");
  }

  unsigned int channels=3;
  BIAS::ImageBase::EColorModel cm=BIAS::ImageBase::CM_invalid;
  // switch fmt to determine nr. of channels, channel order and seq.
  int formatIL = (int)ilGetInteger(IL_IMAGE_FORMAT);
  switch (formatIL)
  {
  case IL_RGB:
    channels=3;
    cm=BIAS::ImageBase::CM_RGB;
    break;
  case IL_RGBA:
    channels=4;
    cm=BIAS::ImageBase::CM_RGBA;
    break;
  case IL_BGR:
    channels=3;
    cm=BIAS::ImageBase::CM_BGR;
    break;
  case IL_BGRA:
    channels=4;
    cm=BIAS::ImageBase::CM_BGRA;
    break;
  case IL_LUMINANCE:
    channels=1; 
    cm=BIAS::ImageBase::CM_Grey;
    break;
  case IL_LUMINANCE_ALPHA:
    channels=2;
    cm=BIAS::ImageBase::CM_GreyA;
    break;
  default:
    BIASERR("unsupported format");
  }

  // always interleaved
  const bool interleaved = true;
  // dim in pixel
  const unsigned int w = ilGetInteger(IL_IMAGE_WIDTH);
  const unsigned int h = ilGetInteger(IL_IMAGE_HEIGHT);

  // allocate new image
  result.Init(
    w,h,
    channels,
    st,
    interleaved );
  result.SetColorModel(cm);

#ifdef BIAS_DEBUG
  const int bytePerPixIL = int(ilGetInteger(IL_IMAGE_BYTES_PER_PIXEL));
  const int bitPerPixIL  = int(ilGetInteger(IL_IMAGE_BITS_PER_PIXEL ));
  BIASASSERT(int(result.GetDepth()*result.GetChannelCount())    == bytePerPixIL);
  BIASASSERT(int(result.GetBitDepth()*result.GetChannelCount()) >= bitPerPixIL );

  {
    const int ilSZ = ilGetInteger(IL_IMAGE_SIZE_OF_DATA);
    BIASASSERT(int(result.GetPixelCount()*result.GetDepth()*result.GetChannelCount()) == ilSZ);
  }
#endif

  if (ilGetData() == NULL){
    BIASERR("ilGetData NULL");
    ilDeleteImages(1, &id);
    return -1;
  };

  // copy the data area
  //result.CopyIn_NoInit( ilGetData() );
  //result.CopyIn_NoInit( ilGetData() );
  ilCopyPixels(0,0,0, /* x y z offset */
    w,h, 
    tex3DplaneZ, /* depth planes, 1 for tex2D */
    formatIL,    /* format, e.g. IL_RGBA */
    typeIL,      /* type, e.g. UL_UNSIGNED_BYTE */
    result.GetImageData() /* dest ptr */
    );
  result.UpdateImageDataArrayFromImageData_();

  // free temporary devIL image 
  ilDeleteImages(1, &id);

  return BIAS_IO_SUCCESS; // OK
}


int ImageIO::ExportDevIL(const std::string& FileName, 
                         const ImageBase& img,
                         const bool &fileOverwrite,
                         const bool &allowPadding )
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  InitDevIL();

  ILuint id=0;
  ILuint err=0;
  ilGenImages(1, &id);
  ilBindImage( id );

  // fill attributes and data 
  unsigned int w=img.GetWidth();
  unsigned int h=img.GetHeight();
  unsigned int depth=img.GetDepth();
  ILenum ilFormat=IL_RGB; //< defaut
  ILenum ilType  =IL_BYTE; //< default

  // overwrite file if exists? 
  if (fileOverwrite) {
    ilEnable(IL_FILE_OVERWRITE);
  } else {
    ilDisable(IL_FILE_OVERWRITE);
  }


  // native storage data type:
  switch (img.GetStorageType())
  {
  case ImageBase::ST_char:
    ilType=IL_BYTE;  break;
  case ImageBase::ST_unsignedchar:
    ilType=IL_UNSIGNED_BYTE;  break;
  case ImageBase::ST_shortint:
    ilType=IL_SHORT;  break;
  case ImageBase::ST_unsignedshortint:
    ilType=IL_UNSIGNED_SHORT;  break;
  case ImageBase::ST_int:
    ilType=IL_INT;  break;
  case ImageBase::ST_unsignedint:
    ilType=IL_UNSIGNED_INT;  break;
  case ImageBase::ST_float:
    ilType=IL_FLOAT;  break;
  case ImageBase::ST_double:
    ilType=IL_DOUBLE;  break;
  default: 
    BIASERR("unsupported storage type");
  }

  //#ifdef BIAS_DEBUG
  //  const unsigned int channels=img.GetChannelCount();
  //#endif 

  // format --> native nr+seq of channels and color model  
  switch (img.GetColorModel())
  {
  case BIAS::ImageBase::CM_RGB:
    ilFormat=IL_RGB;
    BIASASSERT(img.GetChannelCount()==3);
    break;
  case BIAS::ImageBase::CM_RGBA:
    ilFormat=IL_RGBA;
    BIASASSERT(img.GetChannelCount()==4);
    break;
  case BIAS::ImageBase::CM_BGR:
    ilFormat=IL_BGR;
    BIASASSERT(img.GetChannelCount()==3);
    break;
  case BIAS::ImageBase::CM_BGRA:
    ilFormat=IL_BGRA;
    BIASASSERT(img.GetChannelCount()==4);
    break;
  case BIAS::ImageBase::CM_Grey:
    ilFormat=IL_LUMINANCE;
    BIASASSERT(img.GetChannelCount()==1);
    break;
  case BIAS::ImageBase::CM_GreyA:
    ilFormat=IL_LUMINANCE_ALPHA;
    BIASASSERT(img.GetChannelCount()==2);
    break;
  default:
    BIASERR("unsupported color model/channelcount/ DevIL type");
  }

  // Byte per Pixel, e.g. 3 for "unsigned char RGB" image
  unsigned int BytePerPixel = img.GetDepth()*img.GetChannelCount();

  // create temporary DevIL image
  // setting attributes and copy data
  if (!ilTexImage( 
    w,h,
    depth,
    ILubyte (BytePerPixel),
    ilFormat, 
    ilType, 
    //(ILvoid*)(img.GetImageData()) 
    NULL
    ))
  {
    BIASERR("could not set DevIL attributes for saving "<<FileName);
    ilDeleteImages(1, &id);
    return -1;
  }

  // BIAS image coordinate system has origin in upper left corner
  // (in contrast to OpenGLs lower left)
  ilRegisterOrigin(IL_ORIGIN_UPPER_LEFT);

  BIASASSERT(img.GetImageData()!=NULL);
  // memcopy the image data: 
  ilSetData((ILvoid*)(img.GetImageData()) );

  // check status
  err = ilGetError();
  if (err!=IL_NO_ERROR) {
    stringstream ss;
    ss<<"could not set DevIL for saving "<<FileName<<" ";
#ifdef USE_DEVIL_ILU   
    ss<<": "<<iluErrorString(err);
#endif
    BIASERR(ss.str());
    ilDeleteImages(1, &id);
    return -1;
  }

  BIASASSERT(int(img.GetDepth())==int(ilGetInteger(IL_IMAGE_DEPTH)) );
#ifdef BIAS_DEBUG
  {
    const int ilSZ = ilGetInteger(IL_IMAGE_SIZE_OF_DATA);
    BIASASSERT(
      int(img.GetPixelCount()*img.GetDepth()*img.GetChannelCount())==ilSZ
      );
  }
#endif

  err = ilGetError();
  if (err!=IL_NO_ERROR) {
    stringstream ss;
    ss<<"could not save to file "<<FileName<<" ";
#ifdef USE_DEVIL_ILU
    ss<<": "<<iluErrorString(err);
#endif
    PrintInfoDevIL();
    BIASERR(ss.str());
    ilDeleteImages(1, &id);
    return -2;
  }

  // write to disk: 
  bool ok = (IL_TRUE==ilSaveImage( (char*)(FileName.c_str()) )); 

  // check status again: 
  err = ilGetError();
  if (err!=IL_NO_ERROR || !ok) 
  {
    stringstream ss;
    ss<<"could not save to file "<<FileName<<" ";
#ifdef USE_DEVIL_ILU
    ss<<": "<<iluErrorString(err);
#endif
    if (err==IL_BAD_DIMENSIONS && allowPadding) {
      // try if padding helps: 
      // fallback tmp padded image 
      BIAS::ImageBase tmpIMG(img);
      tmpIMG.PadToPowerOfTwo();
      // do not allow padding again or we will get an endless loop!
      int resultPadded=ExportDevIL(
        FileName, 
        tmpIMG,
        fileOverwrite,
        false /*allowPading?*/ );
      if (resultPadded==0)
      {
        // everything is fine,  give msg that we paded.
        ss<<endl<<"Some formats (e.g. DDS) require power of two size (and depth) "<<endl
          <<"Padding from dim "<<w<<"x"<<h<<" "
          <<"(depth="<<depth<<" ChannelCount="<<tmpIMG.GetChannelCount()<<") "
          <<"to "<<tmpIMG.GetWidth()<<"x"<<tmpIMG.GetHeight()<<" "
          <<"(depth="<<tmpIMG.GetDepth()<<" channels="<<tmpIMG.GetChannelCount()<<") "<<endl
          <<"was required to save "<<FileName;
        BIASWARN(ss.str());
      } else {
        // padding did not help
        BIASERR(ss.str());
        ilDeleteImages(1, &id);
        return -3;
      }
    } else {
      // error we could do nothing about
      BIASERR(ss.str());
      ilDeleteImages(1, &id);
      return -4;
    }
  }

  // OK: 
  // free temporary devIL image 
  ilDeleteImages(1, &id);

  return BIAS_IO_SUCCESS;
}
#endif  // BIAS_HAVE_DEVIL



std::string 
ImageIO::GetExtension(const BIAS::ImageIO::TFileFormat & fmt)
{
  string s;
  switch (fmt)
  {
  case FF_mip:
    s="mip"; break;
  case FF_ppm:
    s="ppm"; break;
  case FF_pgm:
    s="pgm"; break;
  case FF_pbm:
    s="pbm"; break;
  case FF_jpg:
    s="jpg"; break;
  case FF_tif:
    s="tif"; break;
  case FF_png:
    s="png"; break;
  case FF_viff:
    s="vif"; break;
  case FF_exr:
    s="exr"; break;
  case FF_dds:
    s="dds"; break;
  case FF_raw4096:
    s="raw"; break;
  case FF_bmp:
    s="bmp"; break;
    break;
  case FF_Real32:
    s="real32"; break;

  case FF_auto:
  case FF_unknown:
  default:
    BIASERR("Unknown FileFormat "<<fmt<<" Please add switch case");
  }
  return s;
}



BIAS::ImageIO::TFileFormat 
ImageIO::GetFileFormat(const std::string &str)
{
  // suffix = "dot"+extension 
  string ext=BIAS::LowerCase( BIAS::Extension(str) );
  // compare ext to all FF_* fixed formats: 
  if (ext=="mip") {
    return FF_mip;
  } else if (
    (ext=="ppm") ||
    (ext=="pnm") /* pNm can be pbm, pgm, ppm. assume color for pnm. */
    )
  {
    return FF_ppm;
  } else if (ext=="pgm") {
    return FF_pgm;
  } else if (ext=="pbm") {
    return FF_pbm;
  } else if (
    (ext=="jpg") ||
    (ext=="jpe") ||
    (ext=="jpeg") 
    )
  {
    return FF_jpg;
  } else if (
    (ext=="tif") ||
    (ext=="tiff") 
    )
  {
    return FF_tif;
  } else if (ext=="png") {
    return FF_png;
  } else if (
    (ext=="vif") || 
    (ext=="vff") ||
    (ext=="viff")) 
  {
    return FF_viff;
  } else if (ext=="exr") {
    return FF_exr;
  } else if (
    (ext=="dds")  ||
    (ext=="dxt")  ||
    (ext=="dxt1") ||
    (ext=="dxt3") ||
    (ext=="dxt5") ||
    (ext=="dxtc") ||
    (ext=="s3tc") 
    )
  {
    // S3's S3TC format and DXTC are essentially the same
    return FF_dds;
  } else if (
    (ext=="raw") ||
    (ext=="raw4096") ||
    (ext=="scorpion") 
    )
  {
    return FF_raw4096;
  } else if ((ext=="real32") || (ext=="f32")){
    // Mathematica Real32 raw JW
    return FF_Real32;
  } else {
    // unknown ext. 
    // give the external imaeg writer libsa try,
    // they may support a wider raneg of extension.
    // Thsu use "FF_auto" on returned "FF_unknown"
    return FF_unknown;
  }
}


//#ifdef BIAS_HAVE_RADIANCE
// JW
int BIAS::ImageIO::ImportRADIANCE(const std::string& FileName, ImageBase& result)
{
  ifstream is;
  is.open(FileName.c_str(), ios::binary);
  if (!is){
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR("Could not open file "<<FileName);
#endif
    return -1;
  }
  //is.seekg(0,ios::beg);
  string buf;
  unsigned int line=0;
  getline(is, buf); line++;
  //cout<<line<<". = "<<buf<<endl;

  const string signatureRAD("#?RADIANCE");
  if (buf!= signatureRAD )
  {
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR("signatureRAD "<<signatureRAD<<" not found in file "<<FileName);
#endif
    return -2;
  };
  // OK, is a RADIANCE file. 

  bool flipY=false, flipX=false;
  unsigned int width=0, height=0;
  //float exposure=1.0;

  // parse for FORMAT
  while (is){    
    getline(is, buf); line++;
    //cout<<line<<". = "<<buf<<endl;
    // skip comments: 
    if (0==buf.compare(0,1,"#")) {
      //cout<<"found comment: "<<buf<<endl;
      continue;
    }
    if (0==buf.compare(0,6,"FORMAT")) {
      // check format: 
      if (0==buf.compare("FORMAT=32-bit_rle_rgbe")){
        break; // OK
      } else {
        BIASERR("Only 32-bit_rle_rgbe FORMAT supported but "<<buf<<" found. Failed to load.");
        return -5;
      }
    };
  }

  // parse for -Y or +Y 
  while (is){    
    getline(is, buf); line++;
    //cout<<line<<". = "<<buf<<endl;
    // skip comments: 
    if (0==buf.compare(0,1,"#")) {
      //cout<<"found comment: "<<buf<<endl;
      continue;
    }
    if (0==buf.compare(0,2,"-Y") 
      || 0==buf.compare(0,2,"-Y")) 
    {
      // parse x and y value: 
      stringstream ss(buf);
      string dummy;
      ss>>dummy; // Y
      if (dummy=="+Y"){
        flipY=true;
      } else if (dummy=="-Y") {
        // OK, no flip
      } else {
        BIASERR("expected +Y or -Y but failed.");
        return -8;
      }
      // read Y value
      ss>>height;

      ss>>dummy;// X
      if (dummy=="+X"){
        // OK, no flip
      } else if (dummy=="-X") {
        flipX=true;
      } else {
        BIASERR("expected +X or -X but failed.");
        return -9;
      }
      // read X value
      ss>>width;
      break;
    }
  };
  cout<<FUNCNAME<<" found "<<width<<"x"<<height<<" RADIANCE rle RGBE image. "<<FileName<<endl;
  // get RGBE data:
  if (!result.IsEmpty()) result.Release();
  result.Init(width, height, 4, BIAS::ImageBase::ST_unsignedchar, true );

  // get data
  char col[4]={0};
  unsigned int y=0;
  //unsigned int x=0;
  //int pos=0;
  //int start=0;
  //int step=1;
  for (y=0; y<height; y++) {
    //start = bFlippedY ? (height-y-1)*width : y*width;
    //step = bFlippedX ? -1 : 1;

    // read 4 Bytes of line:
    is.read(col,4);
    // is the image /line RLE compressed? 
    if ( (col[0]==0x2) 
      && (col[1]==0x2) 
      && (col[2]==0x0) )
    {
      BIASERR("unsupported new RADIANCE compression detected in line y="<<y);
      // clear the line: 
      memset( result.GetImageDataArray()[y], 0, 4*width);

    } else if ( (col[0]==0x1) 
      && (col[1]==0x1) 
      && (col[2]==0x1) )
    {
      BIASERR("unsupported old RADIANCE run scheme in line y="<<y);
      memset( result.GetImageDataArray()[y], 0, 4*width);

    } else {

      BIASERR("sorry, RADAINCE not completely implemented TODO: RLE (un-)compression is missing. (jw)");
      return -99;
      /// \todo: decode the runlength encode line, map the RGBE to float or do it in ColorConverion?

      /*
      // OK, assume line contains only raw data
      //start = flipY ? (height-y-1)*width : y*width;
      //step  = flipX ? -1 : 1;
      // Do the run
      int nPixel = ((int) col[3])&0xFF;
      unsigned char r = pBuffer[(pos-step)*4];
      unsigned char g = pBuffer[(pos-step)*4+1];
      unsigned char b = pBuffer[(pos-step)*4+2];
      unsigned char e = pBuffer[(pos-step)*4+3];

      for( int i=0; i<num; i++ ) {
      pBuffer[pos*4] = r;
      pBuffer[pos*4+1] = g;
      pBuffer[pos*4+2] = b;
      pBuffer[pos*4+3] = e;
      pos += step;
      }

      x += num-1;
      */
    }
  }
  return 0; // OK
}
//#endif // BIAS_HAVE_RADIANCE 



// JW
int BIAS::ImageIO::ImportReal32(const std::string& FileName, ImageBase& result)
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  ifstream is;
  is.open(FileName.c_str(), ios::binary);
  if (!is){
    BIASERR("Could not open file "<<FileName);
    is.close();
    return -1;
  }

  // header: 4 float
  // format, width, height, nchannels
  vector<float> header(4, 0);
  unsigned int version=0;
  if (is)
    is.read( (char*)(&(header[0])), header.size()*sizeof(float) );

  if (!is)
    return -101;

  // supported format? simple check for now. 
  const float fmt=header[0];
  if ((1.0f <= fmt) && (fmt<=1.01f)) {
    version=1;
  } else {
    return -1; // unsupported
  }
  unsigned int width     = (unsigned int)( header[1] );
  unsigned int height    = (unsigned int)( header[2] );
  unsigned int nchannels = (unsigned int)( header[3] );

  // only small GPU suitable images are written, currently: 
  // check useful header values:
  if (!(0<width)) 
    return -2;
  if (!(width<=4096)) 
    return -3;
  if (!(0<height)) 
    return -4;
  if (!(height<4096)) 
    return -5;
  if (!(0<nchannels)) 
    return -6;
  if (!(nchannels<=4)) 
    return -7;

  // OK

  // (re-)allocate
  if (!result.IsEmpty())
    result.Release();
  result.Init(width, height, nchannels, ImageBase::ST_float, true);

  // read the raw data:
  BIASASSERT(is);
  is.read( (char*)(result.GetImageData()), 
    sizeof(float)*width*height*nchannels );

  is.close();
  return 0;
}



#ifdef WIN32
// implement WINAPI specific calls for BMP loading via disk or memory

int BIAS::ImageIO::Import_BITMAP_winapi(const std::string& FileName, 
                                        ImageBase& result,
                                        const bool flipY )
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif

  // the WINAPI interface
  HBITMAP hbm=NULL;

  // convert from stl string to WINAPI string
  // TODO: what about UNICODE strings: use LPTSTR ?
  LPCSTR szFileName = FileName.c_str();

  // load the image from disk into a DIBSection
  hbm = (HBITMAP)
    LoadImage( NULL, szFileName, IMAGE_BITMAP, 0, 0,
    LR_LOADFROMFILE | LR_CREATEDIBSECTION | LR_DEFAULTSIZE );

  // check success
  if( hbm == NULL ){
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR("could not load "<<FileName );
#endif 
    return -1; // error, could not load
  }

  // we have a valid HBITMAP, now.
  // create object and copy  in, now.
  return BIAS::ImageIO::Import_HBITMAP_winapi(hbm, 
    result,
    flipY );
}


int BIAS::ImageIO::Import_HBITMAP_winapi(const HBITMAP & hbm, 
                                         ImageBase& result,
                                         const bool flipY )
{
  // create object from handle
  BITMAP bm;
  const int nBytesHeader = 
    GetObject( hbm, sizeof(BITMAP), &bm );
  if (nBytesHeader==0) {
    BIASERR("could not GetObject");
    return -1;
  }
  //cout<<"created object with "<<nBytes<<" Bytes."<<endl; // typically 24

  // we have a valid BITMAP, now, 
  // load pixel data
  return BIAS::ImageIO::Import_BITMAP_winapi(bm, 
    result,
    flipY );
}


int BIAS::ImageIO::Import_BITMAP_winapi(const BITMAP & bm, 
                                        ImageBase& result,
                                        const bool flipY )
{
  // If  DIBSection is 256 color or less --> color table
  if( ( bm.bmBitsPixel * bm.bmPlanes ) <= 8 ){
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR("Could not load winapi BITMAP because it contains a color table (not implemented)" );
#endif 
    return -2;
  }

  // tested only with 3 channel(BGR) interleaved images with one plane
  if (bm.bmPlanes != 1){
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
    BIASERR("Could not load winapi BITMAP because it contains more than one plane." );
#endif 
    return -3;
  }

  //
  // copy pixel data into BIAS layout
  // assume 3 channel uc (3*8=24bit) interleaved DIB for now.
  //
  {
    BIAS::Image<unsigned char> img(bm.bmWidth, bm.bmHeight, 3 );
    img.SetColorModel(BIAS::ImageBase::CM_BGR); // Windows DIB = BGR

    const int nBytes = bm.bmWidthBytes*bm.bmHeight;
    memcpy(img.GetImageData(), (unsigned char*)bm.bmBits, nBytes );
    if (flipY){
      // Windows BMP has down-up vertical order like OpenGL
      img.Flip();
    }

    // Image<uc> --> ImageBase
    result = img;
  }

  return EXIT_SUCCESS;
}


#endif // WIN32





#ifdef BIAS_HAVE_TIFF
// see http://www.ibm.com/developerworks/linux/library/l-libtiff/   JW
int BIAS::ImageIO::ExportTIFFLIB(const std::string& FileName, 
                                 const ImageBase& img,
                                 const unsigned int rowsPerStripArg,
                                 const unsigned int compressionAlgo,
                                 const string & comment )
{
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  CALLINFO;
#endif
  try {

    TIFF* tif = TIFFOpen(FileName.c_str(), "w");
    if (tif==NULL){
#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
      BIASERR("Could not open image for TIFFLIB for writing:" << FileName);
#endif
      return -1;
    }
    BIASASSERT(tif!=NULL);
    // work:

    TIFFSetField(tif, TIFFTAG_IMAGEWIDTH,      img.GetWidth() );
    TIFFSetField(tif, TIFFTAG_IMAGELENGTH,     img.GetHeight() );
    TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, img.GetChannelCount() );
    const unsigned short bitspersample = 8* (unsigned short)img.GetDepth(); // img.GetBiTDepth()
    TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, bitspersample ); // more than enough
    TIFFSetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT); // origin 0,0
    // assume planar tiff cfg., needs one channel or interleaved pixel data
    BIASASSERT((img.GetChannelCount()==1) || (img.IsInterleaved()) ); 
    TIFFSetField(tif, TIFFTAG_PLANARCONFIG,    PLANARCONFIG_CONTIG);


    //TIFFSetField(tif, TIFFTAG_COMPRESSION,     COMPRESSION_NONE); //(300 msec)
    //TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_LZW); // works, lossless Lempel Ziff Welch  (650msec)
    //TIFFSetField(tif, TIFFTAG_COMPRESSION,     COMPRESSION_DEFLATE); // works,lossless compression known as Zip-in-TIFF (1750 msec)
    TIFFSetField(tif, TIFFTAG_COMPRESSION, compressionAlgo); // TODO: check if this is a valid compressoinalgorithm

    // size of strips = nr. of lines taht form a strip
    unsigned int rowsPerStrip  = rowsPerStripArg;
    if (rowsPerStrip <= 0)
      rowsPerStrip = img.GetHeight();
    if (rowsPerStrip>img.GetHeight()) {
      BIASERR("rowsPerStrip="<<rowsPerStrip<<" exceed limit. resetting to img height="<<img.GetHeight());
      rowsPerStrip=img.GetHeight();
    }
    /// \todo TODO: do we need to check that rowsPerStrip is modulo image height?
    BIASASSERT(rowsPerStrip >= 0 );
    BIASASSERT(rowsPerStrip <= img.GetHeight() );
    TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, rowsPerStrip);

    //TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
    //TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
    //TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
    switch (img.GetChannelCount()){
      case 1:
        TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK); break;
        //case 2:
      case 3:
        TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB); break;
        //case 4:
      default: 
        {
          BIASERR("unhandled nr. of channels.");
          return -1;
        }
    }

    //// optional tags
    TIFFSetField(tif, TIFFTAG_IMAGEDESCRIPTION, "written Jan Woetzel");
    const string cvsVersion = "CVS $Revision: 2.157 $";
    stringstream swSS;
    swSS <<"BIAS::ImageIO "
      <<"BIAS-"<<BIAS_VERSION_MAJOR<<"."<<BIAS_VERSION_MINOR<<"."<<BIAS_VERSION_PATCH
      <<" "<<cvsVersion<<" (Jan Woetzel)";
    TIFFSetField(tif, TIFFTAG_SOFTWARE, swSS.str().c_str() );

    //TIFFSetField(tif, TIFFTAG_XRESOLUTION, 150.0);
    //TIFFSetField(tif, TIFFTAG_YRESOLUTION, 150.0);
    //TIFFSetField(tif, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH); // RESUNIT_CENTIMETER

    // write TIFF in scanlines
    int res=0;
    //if (rowsPerStrip==1)
    //{
    //  // use scanline based API
    //  0000000000000000 these are scanlines
    //  1111111111111111
    //  2222222222222222
    //  for (unsigned int y=0; y<img.GetHeight(); y+=rowsPerStrip ){
    //    void* pd = img.GetImageDataArray()[y];
    //    // 1is returned on success
    //    res = TIFFWriteScanline(tif, pd, y, 0);
    //    BIASASSERT(res==1);
    //  }
    //} 
    //else 
    {
      // use strip based API, possibly encoded
      // 0000000000  these are strips
      // 0000000000
      // 1111111111
      // 1111111111
      // TODO must rowsPerStrip be module height?
      const unsigned int nBytes = rowsPerStrip * img.GetWidth()*img.GetChannelCount()*img.GetDepth() ;
      int strip =0;
      for (unsigned int y=0; y<img.GetHeight(); y+=rowsPerStrip){
        // the number of written Bytes is returned
        res = TIFFWriteEncodedStrip(tif, strip, img.GetImageDataArray()[y], nBytes );
        BIASASSERT((int)res==(int)nBytes);
        strip++;
      }
    }

    TIFFClose(tif);
    return EXIT_SUCCESS;

  } catch (...){
    BIASERR("catched tifflib exception");
    return -1;
  }
}


int BIAS::ImageIO::ImportTIFFLIB(const std::string& FileName, ImageBase& img )
{
  //  /*
  //#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  //  CALLINFO;
  //#endif
  //  try {
  //
  //    TIFF* tif = TIFFOpen(FileName.c_str(), "r");
  //    if (tif==NULL){
  //#if defined(BIAS_DEBUG) && defined(ImageIO_DEBUG_IO)
  //      BIASERR("Could not open image for TIFFLIB for reading:" << FileName);
  //#endif
  //      return -1;
  //    }
  //    BIASASSERT(tif!=NULL);
  //
  //    // work:
  //
  //    int width=0, height=0, channelcount=0, bitpersample=0;
  //
  //    // Check that it is of a type that we support
  //
  //
  //
  //      if((== 0) || (bps != 1)){
  //        fprintf(stderr, "Either undefined or unsupported number of bits per sample\n");
  //        exit(42);
  //      }
  //
  //      if((TIFFGetField(image, TIFFTAG_SAMPLESPERPIXEL, &spp) == 0) || (spp != 1)){
  //        fprintf(stderr, "Either undefined or unsupported number of samples per pixel\n");
  //        exit(42);
  //      }
  //
  //    TIFFGetField(image, TIFFTAG_BITSPERSAMPLE, &bitpersample) ;
  //
  //
  //    TIFFGetField(tif, TIFFTAG_IMAGEDESCRIPTION, "written by BIAS::ImageIO (Jan Woetzel)");
  //    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH,      img.GetWidth() );
  //    TIFFGetField(tif, TIFFTAG_IMAGELENGTH,     img.GetHeight() );
  //    TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, img.GetChannelCount() );
  //    TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE,   8*img.GetDepth() ); // more than enough
  //    //TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE,   img.GetBitDepth() ); // exactly enough
  //    TIFFGetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT); // origin 0,0
  //    // assume planar tiff cfg., needs one channel or interleaved pixel data
  //    BIASASSERT((img.GetChannelCount()==1) || (img.IsInterleaved()) ); 
  //    TIFFGetField(tif, TIFFTAG_PLANARCONFIG,    PLANARCONFIG_CONTIG);
  //
  //
  //    // read TIFF in scanlines
  //    int res=0;
  //    //if (rowsPerStrip==1)
  //    //{
  //    //  // use scanline based API
  //    //  0000000000000000 these are scanlines
  //    //  1111111111111111
  //    //  2222222222222222
  //    //  for (unsigned int y=0; y<img.GetHeight(); y+=rowsPerStrip ){
  //    //    void* pd = img.GetImageDataArray()[y];
  //    //    // 1is returned on success
  //    //    res = TIFFReadScanline(tif, pd, y, 0);
  //    //    BIASASSERT(res==1);
  //    //  }
  //    //} 
  //    //else 
  //    {
  //      // use strip based API, possibly encoded
  //      // 0000000000  these are strips
  //      // 0000000000
  //      // 1111111111
  //      // 1111111111
  //      // TODO must rowsPerStrip be module height?
  //      const unsigned int nBytes = rowsPerStrip * img.GetWidth()*img.GetChannelCount()*img.GetDepth() ;
  //      int strip =0;
  //      for (unsigned int y=0; y<img.GetHeight(); y+=rowsPerStrip){
  //        // the number of written Bytes is returned
  //        res = TIFFReadEncodedStrip(tif, strip, img.GetImageDataArray()[y], nBytes );
  //        BIASASSERT((int)res==(int)nBytes);
  //        strip++;
  //      }
  //    }
  //
  //    TIFFClose(tif);
  //    return EXIT_SUCCESS;
  //
  //  } catch (...){
  //    BIASERR("catched tifflib exception");
  //    return -1;
  //  }
  //  */
  return -1;
}
#endif // BIAS_HAVE_TIFF

