# - Find Nvidia Compute Unified Driver Architecture (CUDA) tools.
# See http://developer.nvidia.com/cuda
# Based on code from http://www.sci.utah.edu/~abe/FindCuda.html
# 
# The script defines the following variables:
#
#  CUDA_FOUND: Is cuda ready to use, i.e. headers, libs found?  
#  CUDA_NVCC The nvcc .cu compiler executable
#  CUDA_LIBRARY
#  CUDA_CUT_LIBRARY
#  CUDA_CUDART_LIBRARY
#  CUDA_CUBLAS_LIBRARY
#  CUDA_CUBLASEMU_LIBRARY
#  CUDA_CUFFT_LIBRARY
#  CUDA_CUFFTEMU_LIBRARY
#  CUDA_NVCC_INCLUDE
#  CUDA_CUT_INCLUDE
#  CUDA_NVCC_INCLUDE
#
#
# AUTHOR 
# -- Abe Stephens SCI Institute -- http://www.sci.utah.edu/~abe/FindCuda.html
# -- Jan Woetzel                -- http://www.mip.informatik.uni-kiel.de/jw

# The docu must be above to assist CMakes automated help generation.

# working configurations tested with cmake 2.4.6: 
#  WinXP 32 bit:    MS Visual Studio 8 IDE poject
#  WinXP 32 bit:    nmake makefiles,  CL 14.00.50727.762 80x86
#  Linux Suse 10.2: gnu makefiles, gcc 4.1.2

###############################################################################
#  The MIT License
#
#  Copyright (c) 2007
#  Scientific Computing and Imaging Institute, University of Utah
#
#  License for the specific language governing rights and limitations under
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
###############################################################################

# NOTES
#
#  For more information, please see: http://software.sci.utah.edu
#
# The script will prompt the user to specify CUDA_INSTALL_PREFIX.
#
#
# Set CUDA_BUILD_CUBIN to "ON" or "OFF" to enable and extra compilation pass
# with the -cubin option in Device mode. 
#
# The output is parsed and register, shared memory usage is printed during 
# build. Default ON.
# 
# The script creates the following macros:
# CUDA_INCLUDE_DIRECTORIES( path0 path1 ... )
# -- Sets the directories that should be passed to nvcc 
#    (e.g. nvcc -Ipath0 -Ipath1 ... ). These paths usually contain other .cu 
#    files.
# 
# CUDA_ADD_LIBRARY( cuda_target file0 file1 ... )
# -- Creates a shared library "cuda_target" which contains all of the source 
#    (*.c, *.cc, etc.) specified and all of the nvcc'ed .cu files specified.
#    All of the specified source files and generated .c files are compiled 
#    using the standard CMake compiler, so the normal INCLUDE_DIRECTORIES, 
#    LINK_DIRECTORIES, and TARGET_LINK_LIBRARIES can be used to affect their
#    build and link.
#
# CUDA_ADD_EXECUTABLE( cuda_target file0 file1 ... )
# -- Same as CUDA_ADD_LIBRARY except that an exectuable is created.
#
# The script defines the following variables:
#
# ( Note CUDA_ADD_* macros setup cuda/cut library dependencies automatically. 
# These variables are only needed if a cuda API call must be made from code in 
# a outside library or executable. )
#
# CUDA_INCLUDE         -- Include directory for cuda headers.
# CUDA_TARGET_LINK     -- Cuda RT library. 
# CUDA_CUT_INCLUDE     -- Include directory for cuda SDK headers (cutil.h).   
# CUDA_CUT_TARGET_LINK -- SDK libraries.
#
###############################################################################

# FindCuda.cmake

SET(CMAKE_BACKWARDS_COMPATIBILITY 2.2)


# debuggung messages, e.g. to screen or logfile
MACRO(DBG_MSG _MSG)
  MESSAGE(
    # STATUS 
    "${CMAKE_CURRENT_LIST_FILE}(${CMAKE_CURRENT_LIST_LINE}):"  ${ARGV} )
ENDMACRO(DBG_MSG)

IF    (NOT CMAKE_CURRENT_LIST_DIR)
  # extract the directory where this file resides
  GET_FILENAME_COMPONENT(CMAKE_CURRENT_LIST_DIR ${CMAKE_CURRENT_LIST_FILE} PATH)
ENDIF (NOT CMAKE_CURRENT_LIST_DIR)

# resides in same directory
INCLUDE(${CMAKE_CURRENT_LIST_DIR}/CudaDependency.cmake)


###############################################################################
###############################################################################
# Locate CUDA, Set Build Type, etc.
###############################################################################
###############################################################################

# NVCC options to choose a set of flags
# for debugging kernel on CPU or if no cuda capable HW is present
OPTION(CUDA_EMULATION "Use Cuda SW Emulation instead of hardware device mode?" OFF)

SET(CUDA_NVCC_FLAGS_EXTRA "" CACHE STRING "extra flags passed to nvcc, e.g. --use_fast_math")

# Emulation if the card isn't present.
SET(CUDA_NVCC_FLAGS "${CUDA_NVCC_FLAGS_EXTRA}" )
IF (CUDA_EMULATION)
  # slow software  Emulation
  SET(CUDA_NVCC_FLAGS ${CUDA_NVCC_FLAGS} --device-emulation -D_DEVICEEMU -g )
ENDIF(CUDA_EMULATION)

SET(CUDA_BUILD_CUBIN TRUE CACHE BOOL "Generate and parse .cubin files in Device mode.")

# Search for the cuda distribution.
IF (NOT CUDA_INSTALL_PREFIX)
  SET(cuda_path "$ENV{CUDA_BIN_PATH}")
  IF (cuda_path) 
    STRING(REGEX REPLACE "[/\\\\]?bin[/\\\\]?$" "" cuda_path ${cuda_path})
  ENDIF(cuda_path)

  SET(CUDA_INSTALL_PREFIX ${cuda_path} CACHE PATH "Prefix used during install")
    
  IF    (NOT EXISTS ${cuda_path})
    # TODO - make package friendly 
    IF    (NOT CUDA_FIND_QUIETLY)
      MESSAGE(FATAL_ERROR "Specify CUDA_INSTALL_PREFIX, e.g. CUDA_BIN_PATH env. variable")
    ENDIF (NOT CUDA_FIND_QUIETLY)
  ENDIF (NOT EXISTS ${cuda_path})
ENDIF (NOT CUDA_INSTALL_PREFIX)

# CUDA_NVCC
IF (NOT CUDA_NVCC)
  FIND_PROGRAM(CUDA_NVCC 
    nvcc
    PATHS ${CUDA_INSTALL_PREFIX}/bin $ENV{CUDA_BIN_PATH} )
  
#  IF(NOT CUDA_NVCC)
#    # TODO - make package friendly 
#    MESSAGE(FATAL_ERROR "Could not find nvcc")
#  ENDIF(NOT CUDA_NVCC)
ENDIF(NOT CUDA_NVCC)


# CUDA_NVCC_INCLUDE_ARGS
# IF (NOT CUDA_NVCC_INCLUDE)
FIND_PATH(CUDA_NVCC_INCLUDE
  device_functions.h
  PATHS ${CUDA_INSTALL_PREFIX}/include $ENV{CUDA_INC_PATH}
  )
  

IF(NOT CUDA_NVCC_INCLUDE)
  # TODO - make package friendly 
  MESSAGE(FATAL_ERROR "Could not find Cuda headers")
ELSE(NOT CUDA_NVCC_INCLUDE)
  # Set the initial include dir.
  SET (CUDA_NVCC_INCLUDE_ARGS "-I"${CUDA_NVCC_INCLUDE})
  SET (CUDA_INCLUDE ${CUDA_NVCC_INCLUDE})
ENDIF(NOT CUDA_NVCC_INCLUDE)

# ENDIF(NOT CUDA_NVCC_INCLUDE)


# CUDA_TARGET_LINK
IF (NOT CUDA_TARGET_LINK)

  FIND_LIBRARY(CUDA_CUDART_LIBRARY
    NAMES cudart PATHS ${CUDA_INSTALL_PREFIX}/lib 
    DOC "CUDA runtime API library" )

  FIND_LIBRARY(CUDA_CUBLAS_LIBRARY
    NAMES cublas PATHS ${CUDA_INSTALL_PREFIX}/lib )
  FIND_LIBRARY(CUDA_CUBLASEMU_LIBRARY
    NAMES cublasemu PATHS ${CUDA_INSTALL_PREFIX}/lib )

  FIND_LIBRARY(CUDA_CUFFT_LIBRARY
    NAMES cufft PATHS ${CUDA_INSTALL_PREFIX}/lib )
  FIND_LIBRARY(CUDA_CUFFTEMU_LIBRARY
    NAMES cufftemu PATHS ${CUDA_INSTALL_PREFIX}/lib )
    
  
  IF(WIN32)
    # Check to see if cuda libraries were found.
    IF(CUDA_CUDART_LIBRARY)
      SET(CUDA_TARGET_LINK
        ${CUDA_CUDART_LIBRARY} )
    ELSE(CUDA_CUDART_LIBRARY)
      # TODO make package friendly 
      MESSAGE(FATAL_ERROR "Could not find cuda library (cudart)")
    ENDIF(CUDA_CUDART_LIBRARY)  
  ELSE(WIN32)

    # TODO Windows SDK doesn't appear to need the separate library.
    FIND_LIBRARY(CUDA_LIBRARY
      cuda
      PATHS ${CUDA_INSTALL_PREFIX}/lib
      NO_DEFAULT_PATH
      NO_CMAKE_ENVIRONMENT_PATH
      NO_CMAKE_PATH
      NO_SYSTEM_ENVIRONMENT_PATH
      NO_CMAKE_SYSTEM_PATH
      )

    # Check to see if cuda libraries were found.
    IF(CUDA_LIBRARY AND CUDA_CUDART_LIBRARY)
      SET(CUDA_TARGET_LINK
        ${CUDA_LIBRARY}
        ${CUDA_CUDART_LIBRARY} )
    ELSE(CUDA_LIBRARY AND CUDA_CUDART_LIBRARY)
      # TODO make package friendly
      MESSAGE(FATAL_ERROR "Could not find cuda libraries (cuda and cudart)")
    ENDIF(CUDA_LIBRARY AND CUDA_CUDART_LIBRARY)
  ENDIF(WIN32)
  
ENDIF(NOT CUDA_TARGET_LINK)

# CUDA_CUT_INCLUDE
IF(NOT CUDA_CUT_INCLUDE)
  FIND_PATH(CUDA_CUT_INCLUDE
    cutil.h
    PATHS ${CUDA_INSTALL_PREFIX}/local/NVSDK0.2/common/inc
    ${CUDA_INSTALL_PREFIX}/NVSDK0.2/common/inc
    ${CUDA_INSTALL_PREFIX}/NV_CUDA_SDK/common/inc
    ${NVSDKCUDA_ROOT}/common/inc
    $ENV{NVSDKCUDA_ROOT}/common/inc
    "$ENV{ProgramFiles}/NVIDIA Corporation/NVIDIA CUDA SDK/common/inc"
    "$ENV{ProgramFiles}/NVIDIA Corporation/NVIDIA SDK 10/NVIDIA CUDA SDK/common/inc"    
    )
ENDIF(NOT CUDA_CUT_INCLUDE)

# CUDA_CUT_TARGET_LINK
IF(NOT CUDA_CUT_TARGET_LINK)  
  FIND_LIBRARY(CUDA_CUT_LIBRARY
    cutil
    cutil32
    PATHS ${CUDA_INSTALL_PREFIX}/local/NVSDK0.2/lib
    ${CUDA_INSTALL_PREFIX}/NVSDK0.2/lib
    ${CUDA_INSTALL_PREFIX}/NV_CUDA_SDK/lib
    ${NVSDKCUDA_ROOT}/common/lib
    $ENV{NVSDKCUDA_ROOT}/common/lib
    ${NVSDKCUDA_ROOT}/lib
    $ENV{NVSDKCUDA_ROOT}/lib
    "$ENV{ProgramFiles}/NVIDIA Corporation/NVIDIA CUDA SDK/common/lib"
    "$ENV{ProgramFiles}/NVIDIA Corporation/NVIDIA SDK 10/NVIDIA CUDA SDK/common/lib"
    DOC "CUDA util library"
    NO_DEFAULT_PATH
    NO_CMAKE_ENVIRONMENT_PATH
    NO_CMAKE_PATH
    NO_SYSTEM_ENVIRONMENT_PATH
    NO_CMAKE_SYSTEM_PATH
    )
  IF(CUDA_CUT_LIBRARY)
    SET(CUDA_CUT_TARGET_LINK ${CUDA_CUT_LIBRARY})
  ENDIF(CUDA_CUT_LIBRARY)
ENDIF(NOT CUDA_CUT_TARGET_LINK)


###############################################################################
# Add include directories to pass to the nvcc command.
MACRO(CUDA_INCLUDE_DIRECTORIES)
  FOREACH(dir ${ARGN})
    SET(CUDA_NVCC_INCLUDE_ARGS ${CUDA_NVCC_INCLUDE_ARGS} -I${dir})
  ENDFOREACH(dir ${ARGN})
ENDMACRO(CUDA_INCLUDE_DIRECTORIES)


##############################################################################
##############################################################################
# This helper macro populates the following variables and setups up custom commands and targets to
# invoke the nvcc compiler. The compiler is invoked once with -M to generate a dependency file and
# a second time with -cuda to generate a .c file
# ${target_srcs}
# ${cuda_cu_sources}
##############################################################################
##############################################################################

# this is where the generated files go: 
# SET(CUDA_INTERMEDIATE_DIR "src/cuda")
SET(CUDA_INTERMEDIATE_DIR "src_cuda_generated")



MACRO(CUDA_add_custom_commands)

  SET(target_srcs "")
  SET(cuda_cu_sources "")

  # Iterate over the macro arguments and create custom
  # commands for all the .cu files.
  FOREACH(file ${ARGN})
    IF(${file} MATCHES ".*\\.cu$")

      # Add a custom target to generate a cpp file.
      SET(generated_file  "${CMAKE_BINARY_DIR}/${CUDA_INTERMEDIATE_DIR}/${file}_generated.c")
      SET(generated_target "${file}_target")
      
      FILE(MAKE_DIRECTORY ${CMAKE_BINARY_DIR}/${CUDA_INTERMEDIATE_DIR})

      SET(source_file ${CMAKE_CURRENT_SOURCE_DIR}/${file})

      # Note that -cuda generates a .c file not a c++ file.
      SET_SOURCE_FILES_PROPERTIES(${source_file} PROPERTIES CPLUSPLUS ON)
      
      # DBG_MSG("${CUDA_NVCC} ${source_file} ${CUDA_NVCC_FLAGS} -cuda -o ${generated_file} ${CUDA_NVCC_INCLUDE_ARGS}")
      
      # Bring in the dependencies.  Creates a variable CUDA_NVCC_DEPEND
      SET(cmake_dependency_file "${generated_file}.depend")
      CUDA_INCLUDE_NVCC_DEPENDENCIES(${cmake_dependency_file})
      SET(NVCC_generated_dependency_file "${generated_file}.NVCC-depend")

      # Build the NVCC made dependency file
      IF (NOT CUDA_EMULATION AND CUDA_BUILD_CUBIN)
	SET(NVCC_generated_cubin_file "${generated_file}.NVCC-cubin.txt")
	ADD_CUSTOM_COMMAND(
	
	  # Generate the .cubin output.
	  OUTPUT ${NVCC_generated_cubin_file}
	  COMMAND ${CUDA_NVCC}
	  ARGS ${source_file} 
	  ${CUDA_NVCC_FLAGS}
	  -DNVCC
	  -cubin
	  -o ${NVCC_generated_cubin_file} 
	  ${CUDA_NVCC_INCLUDE_ARGS}

	  # Execute the parser script.
	  COMMAND  ${CMAKE_COMMAND}
	  ARGS 
	  -D input_file="${NVCC_generated_cubin_file}"
	  -P "${CMAKE_CURRENT_LIST_DIR}/CudaParse_cubin.cmake"

	  # MAIN_DEPENDENCY ${source_file}
	  DEPENDS ${source_file}
	  DEPENDS ${CUDA_NVCC_DEPEND}

	  COMMENT "Building NVCC -cubin File: ${NVCC_generated_cubin_file}\n"
	  )
      ELSE (NOT CUDA_EMULATION AND CUDA_BUILD_CUBIN)
	# Depend on something that will exist.
	SET(NVCC_generated_cubin_file "${source_file}")
      ENDIF (NOT CUDA_EMULATION AND CUDA_BUILD_CUBIN)

      
      # DBG_MSG("\nCUDA targets\n"
	#  "file=${file}\n"
	#  "source_file=${source_file}\n"
	#  "generated_file=${generated_file}\n"
	#  "generated_target=${generated_target}\n"	
	#  "cmake_dependency_file=${cmake_dependency_file}\n"
	#  "NVCC_generated_dependency_file=${NVCC_generated_dependency_file}\n"
	#  "NVCC_generated_cubin_file=${NVCC_generated_cubin_file}\n"
	# )

      # Build the NVCC made dependency file
      ADD_CUSTOM_COMMAND(
	OUTPUT ${NVCC_generated_dependency_file}
	COMMAND ${CUDA_NVCC}
	ARGS ${source_file} 
        ${CUDA_NVCC_FLAGS}
        -DNVCC
        -M
        -o ${NVCC_generated_dependency_file} 
        ${CUDA_NVCC_INCLUDE_ARGS}
	# MAIN_DEPENDENCY ${source_file}
	DEPENDS ${source_file}
	DEPENDS ${CUDA_NVCC_DEPEND}
	COMMENT "Building NVCC Dependency File: ${NVCC_generated_dependency_file}\n"
	)
      
      # Build the CMake readible dependency file
      ADD_CUSTOM_COMMAND(
	OUTPUT ${cmake_dependency_file}
	COMMAND ${CMAKE_COMMAND}
	ARGS 
	-D input_file="${NVCC_generated_dependency_file}"
	-D output_file="${cmake_dependency_file}"
	-P "${CMAKE_CURRENT_LIST_DIR}/CudaMake2cmake.cmake"
	MAIN_DEPENDENCY ${NVCC_generated_dependency_file}
	COMMENT "Converting NVCC dependency to CMake (${cmake_dependency_file})"
	)
      
      #      DBG_MSG("\nCUDA NVCC compilation\n"
	#	"CUDA_NVCC_FLAGS=${CUDA_NVCC_FLAGS}\n"
	#	"CUDA_NVCC_INCLUDE_ARGS=${CUDA_NVCC_INCLUDE_ARGS}")
      
      ADD_CUSTOM_COMMAND(
	OUTPUT ${generated_file}
	MAIN_DEPENDENCY ${source_file} 
	DEPENDS ${CUDA_NVCC_DEPEND}
	DEPENDS ${cmake_dependency_file}
	DEPENDS ${NVCC_generated_cubin_file}
	COMMAND ${CUDA_NVCC} 
	ARGS ${source_file} 
        ${CUDA_NVCC_FLAGS}
        -DNVCC
        --keep
        -cuda -o ${generated_file} 
        ${CUDA_NVCC_INCLUDE_ARGS}
	COMMENT "Building NVCC ${source_file}: ${generated_file}\n"
	)
      
      SET(cuda_cu_sources ${cuda_cu_sources} ${source_file})

      # Add the generated file name to the source list.
      SET(target_srcs ${target_srcs} ${generated_file})
      
    ELSE(${file} MATCHES ".*\\.cu$")
      
      # Otherwise add the file name to the source list.
      SET(target_srcs ${target_srcs} ${file})
      
    ENDIF(${file} MATCHES ".*\\.cu$")
  ENDFOREACH(file)

ENDMACRO(CUDA_add_custom_commands)

###############################################################################
###############################################################################
# ADD LIBRARY
###############################################################################
###############################################################################
MACRO(CUDA_ADD_LIBRARY cuda_target)

  # Create custom commands and targets for each file.
  CUDA_add_custom_commands( ${ARGN} )  
  
  # Add the library.
  ADD_LIBRARY(${cuda_target}
    ${target_srcs}
    ${cuda_cu_sources}
    )

  TARGET_LINK_LIBRARIES(${cuda_target}
    ${CUDA_TARGET_LINK}
    )

ENDMACRO(CUDA_ADD_LIBRARY cuda_target)


###############################################################################
###############################################################################
# ADD EXECUTABLE
###############################################################################
###############################################################################
MACRO(CUDA_ADD_EXECUTABLE cuda_target)
  
  # Create custom commands and targets for each file.
  CUDA_add_custom_commands( ${ARGN} )
  
  # Add the library.
  ADD_EXECUTABLE(${cuda_target}
    ${target_srcs}
    ${cuda_cu_sources}
    )

  TARGET_LINK_LIBRARIES(${cuda_target}
    ${CUDA_TARGET_LINK}
    )

ENDMACRO(CUDA_ADD_EXECUTABLE cuda_target)


# #####################################################
# ##  mark advanvced variables
MARK_AS_ADVANCED( 
  CUDA_BUILD_CUBIN
  CUDA_LIB
  CUDA_LIBRARY
  CUDA_LIBRARIES
  CUDA_CUT_LIBRARY
  CUDA_CUBLAS_LIBRARY
  CUDA_CUBLASEMU_LIBRARY
  CUDA_CUFFT_LIBRARY
  CUDA_CUFFTEMU_LIBRARY
  CUDA_CUDART_LIBRARY 
  CUDA_NVCC
  #CUDA_NVCC_FLAGS_EXTRA
  CUDA_NVCC_INCLUDE
  CUDA_CUT_INCLUDE
  CUDA_NVCC_INCLUDE_ARGS
  CUDA_TARGET_LINK 
)    


# #####################################################
# ## check result - complete ?
SET (CUDA_FOUND 1)

IF(NOT CUDA_NVCC)
  SET (CUDA_FOUND 0)
  MESSAGE(FATAL_ERROR "Could not find nvcc")
ENDIF(NOT CUDA_NVCC)

IF(NOT CUDA_NVCC_INCLUDE)
  SET (CUDA_FOUND 0)
  MESSAGE(FATAL_ERROR "Could not find Cuda headers")
ENDIF(NOT CUDA_NVCC_INCLUDE)

#IF   (WIN32 AND NOT CUDA_CUDART_LIBRARY)
#    MESSAGE(FATAL_ERROR "Could not find cuda library (cudart)")
#ENDIF(WIN32 AND NOT CUDA_CUDART_LIBRARY)




# #####################################################
# ## debugging output
FOREACH(VAR 
  CUDA_FOUND
  CUDA_INSTALL_PREFIX
  CUDA_NVCC 
  CUDA_NVCC_FLAGS
  CUDA_LIBRARY
  CUDA_CUT_LIBRARY
  CUDA_CUDART_LIBRARY
  CUDA_NVCC_INCLUDE
  CUDA_CUT_INCLUDE
  CUDA_NVCC_INCLUDE
  )
#  DBG_MSG("\n${VAR}=${${VAR}}")
ENDFOREACH(VAR)

#DBG_MSG("CUDA_INCLUDE_DIRS: ${CUDA_INCLUDE_DIRS}")
#DBG_MSG("CUDA_LIBRARY_DIRS: ${CUDA_LIBRARY_DIRS}")
#DBG_MSG("CUDA_LIBRARIES   : ${CUDA_LIBRARIES}")


# #####################################################
# ## make FIND_PACKAGE friendly
IF(NOT CUDA_FOUND)
  IF(NOT CUDA_FIND_QUIETLY)
    IF(CUDA_FIND_REQUIRED)
      MESSAGE(FATAL_ERROR
        "CUDA required, please specify it's location.")
    ELSE(CUDA_FIND_REQUIRED)
      MESSAGE(STATUS "ERROR: CUDA was not found.")
    ENDIF(CUDA_FIND_REQUIRED)
  ENDIF(NOT CUDA_FIND_QUIETLY)
ENDIF(NOT CUDA_FOUND)

