/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "FMatrixEstimation.hh"
#include "MathAlgo/SVD.hh"
#include "Parametrization.hh"
#include "MathAlgo/Minpack.hh"
#include "Base/Geometry/KMatrix.hh"

#include <Base/Common/BIASpragma.hh>

using namespace BIAS;
using namespace std;

FMatrixEstimation *pCurrentObject = NULL;
Parametrization *Para = NULL;


void FErrorFunction(int *m, int *n, double *x, 
                    double *fvec, int *iflag) {
  vector<double> params;
  params.reserve(7);
  for (unsigned int i=0; i<7; i++) params.push_back(x[i]);
  BIAS::FMatrix F;
  Para->ParamsToFMatrix(F, params);
  
  const unsigned int uiDataSize = pCurrentObject->p1_->size();
  const vector<BIAS::HomgPoint2D> &p1 = *pCurrentObject->p1_;
  const vector<BIAS::HomgPoint2D> &p2 = *pCurrentObject->p2_;

  for (unsigned int j=0; j<uiDataSize; j++) {
    fvec[j] = F.GetEpipolarErrorHomogenized(p1[j], p2[j]);
  }
}



int FMatrixEstimation::
SevenPoint(std::vector<BIAS::FMatrix>& vecF,
           const std::vector<BIAS::HomgPoint2D> &p1, 
           const std::vector<BIAS::HomgPoint2D> &p2) 
{
#ifdef BIAS_DEBUG
  if ((p1.size()!=7) || (p2.size()!=7)) {
    BIASERR("Must provide 7 point correspondences for 7-Point algorithm"
            <<"p1.size="<<p1.size()<<"p2.size="<<p2.size());
    BIASABORT;
  }
#endif

  const std::vector<BIAS::HomgPoint2D> *pp1 = &p1;
  const std::vector<BIAS::HomgPoint2D> *pp2 = &p2;
  std::vector<BIAS::HomgPoint2D> p1norm;
  std::vector<BIAS::HomgPoint2D> p2norm;
  KMatrix K1, K2;
  if (NormalizeHartley_) {
    // for (unsigned int i=0; i<7; i++) cout <<" " <<p1[i];
    N_.Hartley(p1, p1norm, K1);
    // for (unsigned int i=0; i<7; i++) cout <<" " <<p2[i];
    N_.Hartley(p2, p2norm, K2);
    pp1 = &p1norm;
    pp2 = &p2norm;
  }

  // determine rank of equation system
  BIAS::Matrix<double> A(7, 9);
  
  for (unsigned int i=0; i<7; i++) {
    // run over all rows of A (and all point correspondences)
    A[i][0] = (*pp1)[i][0] * (*pp2)[i][0];
    A[i][1] = (*pp1)[i][1] * (*pp2)[i][0];
    A[i][2] =                (*pp2)[i][0];
    A[i][3] = (*pp1)[i][0] * (*pp2)[i][1];
    A[i][4] = (*pp1)[i][1] * (*pp2)[i][1];
    A[i][5] =                (*pp2)[i][1];
    A[i][6] = (*pp1)[i][0];
    A[i][7] = (*pp1)[i][1];
    A[i][8] = 1;
  }
  
  BIAS::SVD A_SVD(A);
  Vector<double> f1(9),f2(9);
  if (!A_SVD.GetNullvector(f1, 0)) {
    BIASERR("Error retrieving first F !");
  }
  if (!A_SVD.GetNullvector(f2, 1)) {
    BIASERR("Error retrieving second F !");
  }

  // f1 and f2 span a 2dimensional null space of A
  // find matrix with rank<3 in this subspace:
  // impose constraint that det(alpha*f1+(1-alpha)*f2)=0)
  std::vector<POLYNOMIALSOLVE_TYPE> AlphaPolynomial, Alphas;
  GetDetPolynomial(f1, f2, AlphaPolynomial);

  // solve for alpha
  PolynomialSolve Solver;
  Solver.CheckCoefficients(AlphaPolynomial);
  const unsigned int uiNumberSolutions = 
    Solver.Analytic(AlphaPolynomial, Alphas);

  vecF.resize(uiNumberSolutions);
  if (uiNumberSolutions==0) return -1;

  // compute (1 or 3) solutions from weighting with alphas  
  for (unsigned int i = 0; i<uiNumberSolutions; i++) {
    vecF[i].SetFromVector(f1*Alphas[i] + f2*(1.0-Alphas[i]));
    if (NormalizeHartley_)
      vecF[i] = K2.Transpose() * vecF[i] * K1;
  }
  
  return 0;
}

int FMatrixEstimation::Linear(BIAS::FMatrix& F, 
                              const std::vector<BIAS::HomgPoint2D> &p1, 
                              const std::vector<BIAS::HomgPoint2D> &p2) {

  const std::vector<BIAS::HomgPoint2D> *pp1 = &p1;
  const std::vector<BIAS::HomgPoint2D> *pp2 = &p2;
  std::vector<BIAS::HomgPoint2D> p1norm;
  std::vector<BIAS::HomgPoint2D> p2norm;
  KMatrix K1, K2;
  if (NormalizeHartley_) {
    N_.Hartley(p1, p1norm, K1);
    N_.Hartley(p2, p2norm, K2);
    pp1 = &p1norm;
    pp2 = &p2norm;
  }

  // determine rank of equation system
  BIAS::Matrix<double> A(p1.size(), 9);
  
  for (unsigned int i=0; i<p1.size(); i++) {
    // run over all rows of A (and all point correspondences)
    A[i][0] = (*pp1)[i][0] * (*pp2)[i][0];
    A[i][1] = (*pp1)[i][1] * (*pp2)[i][0];
    A[i][2] =                (*pp2)[i][0];
    A[i][3] = (*pp1)[i][0] * (*pp2)[i][1];
    A[i][4] = (*pp1)[i][1] * (*pp2)[i][1];
    A[i][5] =                (*pp2)[i][1];
    A[i][6] = (*pp1)[i][0];
    A[i][7] = (*pp1)[i][1];
    A[i][8] = 1;
    //  cout<<"P1:"<<p1[i]<<p2[i]<<endl;
  }

  BIAS::SVD A_SVD(A);
  //  cout<<"Linear FEst S:"<<A_SVD.GetS()<<endl;
  
  // get decomposition of F matrix
  BIAS::Matrix<double> U,VT;
  BIAS::Matrix<double> S(3,3,0.0);
  BIAS::Vector<double> s;

  // solve equation system and compute least squares F usually with rank 3
  VT = A_SVD.GetVT();

  for (unsigned int i=0; i<9; i++)  
    F[i/3][i%3] = VT[8][i];

  // re-add hartley normalization, so that f is valid for original coordinates
  if (NormalizeHartley_) F = K2.Transpose() * F * K1;

  // now F is a matrix probably of rank 3, limit to rank 2:
  EnforceMaxRank2(F);

  return 0;
}

int FMatrixEstimation::Optimize(BIAS::FMatrix& F,
                                const std::vector<BIAS::HomgPoint2D> &p1, 
                                const std::vector<BIAS::HomgPoint2D> &p2) {
#ifdef BIAS_DEBUG
  if (p1.size()!=p2.size()) {
    BIASERR("Vectors of corresponding points have different sizes !!!");
    BIASABORT;
  }
  if (p1.size()<7) {
    BIASERR("Too few points to optimize: "<<p1.size());
    return -1;
  }
#endif


  // we need to call an external function from levenberg-marquardt,
  // prepare data structs
  // warning: this section is not thread safe !
  p1_ = &p1;
  p2_ = &p2;
  pCurrentObject = this;

  vector<double> vecFParamsStart, vecFParamsResult; 
  Vector<double> FParamsStart(7), FParamsResult(7);
  Para = new Parametrization;
  Para->FMatrixToParams(F, vecFParamsStart);
  for (unsigned int i=0; i<7; i++)  FParamsStart[i] = vecFParamsStart[i];
  int res;
  if ((res = LevenbergMarquardt(&FErrorFunction, p1.size(),7, 
                         FParamsStart, FParamsResult, 
                         MINPACK_DEF_TOLERANCE, LM_DEF_MAX_ITER))!=0) {
    // BIASERR("Error in Levenberg-Marquardt-Optimization of F: "<<res);
    return -1;
  }
  vecFParamsResult.resize(7);
  for (unsigned int i=0; i<7; i++)  vecFParamsResult[i] = FParamsResult[i];
  Para->ParamsToFMatrix(F, vecFParamsResult);
  delete Para;
  Para = NULL;
#ifdef BIAS_DEBUG
  pCurrentObject = NULL;
  p1_ = NULL;
  p2_ = NULL;
#endif

  return 0;
}

void
FMatrixEstimation::GetDetPolynomial(const BIAS::Vector<double> &F1, 
                                    const BIAS::Vector<double> &F2,
                                    std::vector<POLYNOMIALSOLVE_TYPE> &v) {
#ifdef BIAS_DEBUG
  if (F1.size()!=9) {
    BIASERR("Vector has wrong size "<<F1.size());
    BIASABORT;
  }
  if (F2.size()!=9) {
    BIASERR("Vector has wrong size "<<F2.size());
    BIASABORT;
  }
  if (v.size()!=0) {
    BIASERR("Vector v is not empty: "<<v.size());
    BIASABORT;
  }
#endif

  // We have two FMatrices in vector form F1, F2.
  // We want to find a linear combination of these two which has rank 2 
  
  // set up determinant equation using sarrus' rule for 3x3 matrices
  // where our combined matrix F = alpha F1 + (1-alpha) F2
  // we are looking for alpha, such that det(F)=0

  // first we need the differences
  const double d[9] = { (F1[0]-F2[0]), (F1[1]-F2[1]), (F1[2]-F2[2]), 
                       (F1[3]-F2[3]), (F1[4]-F2[4]), (F1[5]-F2[5]), 
                       (F1[6]-F2[6]), (F1[7]-F2[7]), (F1[8]-F2[8]) };

  // init coeffs with zero
  v.resize(4,0);
 
  // accessing matrix according to sarrus' rule
  // better compute the determinant using short code from MAPLE or something 
  // like that, as it is done in v_x_l::FMatrixCompute7Point
  // this here works definitely, but is quite slow
  unsigned int x,y,z;
  for (unsigned int i=0; i<3; i++) {
    // + diagonal from top to bottom
    x = i;
    y = 3 + ((i+1)%3);
    z = 6 + ((i+2)%3);
    v[3] += d[x] * d[y] * d[z];
    v[2] += d[x] * d[y] * F2[z] + d[x] * F2[y] * d[z] + F2[x] * d[y] * d[z];
    v[1] += d[x] * F2[y] * F2[z] + F2[x] * d[y] * F2[z] + F2[x] * F2[y] * d[z];
    v[0] += F2[x] * F2[y] * F2[z];

    // - diagonal from bottom to top
    x = 6+i;
    // y is the same
    z = (i+2)%3;
    v[3] -= d[x] * d[y] * d[z];
    v[2] -= d[x] * d[y] * F2[z] + d[x] * F2[y] * d[z] + F2[x] * d[y] * d[z];
    v[1] -= d[x] * F2[y] * F2[z] + F2[x] * d[y] * F2[z] + F2[x] * F2[y] * d[z];
    v[0] -= F2[x] * F2[y] * F2[z];
  }
}


void FMatrixEstimation::EnforceMaxRank2(BIAS::FMatrix &F) {
  // get decomposition of F matrix
  BIAS::Matrix<double> U,VT;
  BIAS::Matrix<double> S(3,3,0.0);
  BIAS::Vector<double> s;
  // enforce rank 2 of F using SVD
  BIAS::SVD F_SVD(F);

  // get singular values of F
  s = F_SVD.GetS();
  
  // compose S matrix
  S[0][0] = s[0];
  S[1][1] = s[1];
  // leave smallest singular value at zero
  
  U  = F_SVD.GetU();
  VT = F_SVD.GetVT();

  // recompose F
  F = U * S * VT ;
}

