/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_Projection_hh__
#define __BIAS_Projection_hh__

#include <bias_config.h>
#include <Base/Common/W32Compat.hh>
#include <math.h>

#include <Base/Debug/Debug.hh>
#include <Base/Math/Matrix4x4.hh>
#include <Base/Math/Vector4.hh>
#include <Base/Geometry/HomgPoint3D.hh>
#include <Base/Geometry/Quaternion.hh>
#include <Geometry/Pose.hh>
#include <Geometry/PMatrix.hh>
#include <Geometry/ProjectionParametersBase.hh>
#include <Geometry/ProjectionParametersPerspective.hh>
#include <Geometry/ProjectionParametersPerspectiveDepth.hh>
#include <Geometry/ProjectionParametersSpherical.hh>
#include <Geometry/ProjectionParametersCylindric.hh>
#include <Geometry/ProjectionParametersIO.hh>

#ifdef BIAS_HAVE_XML2
#  include <Base/Common/XMLBase.hh>
#endif // BIAS_HAVE_XML2


namespace BIAS {


  /** @class Projection
   *  @ingroup g_geometry
   This class hides the underlying projection model, like projection matrix,
   spherical camera, lens distortion, ...

  */
  class BIASGeometry_EXPORT Projection
#ifdef BIAS_HAVE_XML2
    : public XMLBase
#endif // BIAS_HAVE_XML2
  {
  public:
    Projection() { 
      vecParameters_.clear();
    }


    explicit Projection(const ProjectionParametersBase& p) {
      BIASASSERT(vecParameters_.empty());
      vecParameters_.push_back(p.Clone());
    }

    Projection(const Projection& p) {
      BIASASSERT(vecParameters_.empty());
      (*this) = p;
    }

    const Projection& operator=(const Projection& p) {
      // delete old paramters if any
      for (unsigned int i=0; i<vecParameters_.size(); i++) {
        delete vecParameters_[i];
      }
      // copy new parameters
      vecParameters_.resize(p.vecParameters_.size());
      for (unsigned int i=0; i<p.vecParameters_.size(); i++) 
        vecParameters_[i] = p.vecParameters_[i]->Clone();
      return *this;
    }

    void CreatePerspective(const BIAS::Pose& pose, 
                           const BIAS::KMatrix& K,
                           int width, int height,
                           const BIAS::Vector4<double>&
                           UndistortionCoefficients = 
                           BIAS::Vector4<double>(0.0, 0.0, 0.0, 0.0));

    /**
       @brief Add a camera to the projection,the coordinates of the pose 
       are given in absolute Coordinates, they are transformed to 
       relative coordinates to the first camera internally              
       @param ppB Pointer to ProjectionParametersBase, 
       the Object is copied and stored in vecParameters_
       @author ischiller
       @date 07/07
    */
    void AddAbsoluteCamera(ProjectionParametersBase *ppb);

    /**
       @brief Add a camera to the projection,the coordinates of the pose 
       are given in relative Coordinates, first cam is origin
       @param ppB Pointer to ProjectionParametersBase, 
       the Object is copied and stored in vecParameters_
       @author ischiller
       @date 07/07
    */
    void AddRelativeCamera(ProjectionParametersBase *ppb);


    /** @bief push_back a new sperical cam to this
        @author Jan Woetzel */
    void CreateSpherical( const Pose& pose, 
                          const unsigned int width, const int height,
                          const double principalX, const double principalY,
                          const double radius,
                          const double aspect=1.0 );

    ///** @bief push_back a new cylindrical cam to this
    //@author Jan Woetzel */
    //void CreateCylindric( const Pose& pose, 
    //  const unsigned int width,
    //  const unsigned int height,
    //  const double zMin,
    //  const double zMax ,
    //  const double phiMin,
    //  const double phiMax );


#ifdef BIAS_HAVE_XML2
    /** 
     *  @brief Reads a camera data XML file and constructs the 
     *  ProjectionParametersBase-pointer, using the ProjectionParametersIO
     *  and the CameraParameter file format.
     *  @param filename Name of the XML parameter file
     */
    int ReadFromCamParamFile(const std::string& filename,
                             const bool silent = false);
  
#endif // BIAS_HAVE_XML2

    /** @brief read BBC's FreeD data from file and constructs a perspective
        projection (may have radial distortion)
        @author koeser 11/2006
    */
    int ReadFromBBC(const std::string& filename,
                    const bool silent = false);

    /** @brief initializes projection from cameradata object
        @author streckel (separated by koeser 11/06) */
    int InitFromCameraData(const std::vector<CameraData>& vcd);


    /** @brief convenience wrapper which tries to read different formats
        @author bartczak, koeser */
    virtual int Load(const std::string& filename);

    bool IsEmpty() const { return (vecParameters_.size()==0);};

    inline unsigned int Size() const { return vecParameters_.size(); } 

    /** @brief parameter read function */
    const ProjectionParametersBase* GetParameters(unsigned int cam = 0) const {
      BIASASSERT(cam < vecParameters_.size());
      BIASASSERT(vecParameters_.size()>cam && vecParameters_[cam]);
      return vecParameters_[cam];
    } 

    /** @brief parameter write function has to invalidate cache ! */
    ProjectionParametersBase* GetParameters(unsigned int cam = 0) {
      BIASASSERT(cam < vecParameters_.size());
      BIASASSERT(vecParameters_.size()>cam && vecParameters_[cam]);
      return vecParameters_[cam];
    }

    /** Allocates memory for parameters base type and returns pointer to it!
     * \attention The responsiblity to free associated memory is passed to 
     *   recipient!
     */
    ProjectionParametersBase* GetParameterCloneWithAbsolutePose(unsigned int cam = 0) const {
      BIASASSERT(cam < vecParameters_.size());
      BIASASSERT(vecParameters_[cam] && vecParameters_[0]);
      ProjectionParametersBase* clone = vecParameters_[cam]->Clone();
      if(cam!=0) {
        Pose rigPose = vecParameters_[0]->GetPose();
        Pose absolutePose;
        rigPose.ConcatenateLocalTransform(clone->GetPose(), absolutePose);
        clone->SetPose(absolutePose);
      }
      return clone;

    }

    /** @brief return Center of camera with index cam
        if cam>0 this is a relative pose in the coordinate system of cam 0
    */
    inline Vector3<double> GetC(unsigned int cam = 0) const {
      BIASASSERT(cam<vecParameters_.size() && vecParameters_[cam]);
      return vecParameters_[cam]->GetC();
    };
    
    /** @brief return RMatrix of camera with index cam
        if cam>0 this is a relative pose in the coordinate system of cam 0 
    */
    inline RMatrix GetR(unsigned int cam = 0) const {
      BIASASSERT(cam<vecParameters_.size() && vecParameters_[cam]);
      return vecParameters_[cam]->GetR();
    };

    inline RMatrix GetAbsoluteR(unsigned int cam = 0) const {
      BIASASSERT(cam<vecParameters_.size());
      BIASASSERT(vecParameters_.size()>cam 
                 && vecParameters_[0]
                 && vecParameters_[cam]);
      if(cam>0)
        return vecParameters_[0]->GetR() * vecParameters_[cam]->GetR();
      else
        return vecParameters_[cam]->GetR();
    }
    
    inline Vector3<double> GetAbsoluteC(unsigned int cam = 0) const {
      BIASASSERT(cam<vecParameters_.size());
      BIASASSERT(vecParameters_.size()>cam 
                 && vecParameters_[0]
                 && vecParameters_[cam]);
      if(cam>0)
        return vecParameters_[0]->GetC() + 
          vecParameters_[0]->GetR() *vecParameters_[cam]->GetC();
      else
        return vecParameters_[0]->GetC();
    }

    inline Quaternion<double> GetAbsoluteQ(unsigned int cam = 0) const {
      BIASASSERT(cam<vecParameters_.size());
      BIASASSERT(vecParameters_.size()>cam 
                 && vecParameters_[0]
                 && vecParameters_[cam]);
      if(cam>0) {
//         Pose rigPose = vecParameters_[0]->GetPose();
//         Pose localCamPose = vecParameters_[cam]->GetPose();
//         Pose globalCamPose;
//         rigPose.ConcatenateLocalTransform(localCamPose, globalCamPose); 
//         return globalCamPose->GetQ();
//    short cut :
        return vecParameters_[0]->GetQ() * vecParameters_[cam]->GetQ();
      }
      else
        return vecParameters_[0]->GetQ();
    }
    
 

    /** @brief return rotation quaternion of camera with index cam 
        if cam>0 this is a relative pose in the coordinate system of cam 0 
    */
    BIAS::Quaternion<double> GetQ(unsigned int cam = 0) const;
   

    /** @brief return const ref to Pose
        @author koeser 02/2006
    */
    inline Matrix3x4<double> GetExternals(unsigned int cam = 0) const {
      BIASASSERT(vecParameters_.size()>cam && vecParameters_[cam]);
      BIASASSERT(cam<vecParameters_.size());
      if (cam==0) { 
        return vecParameters_[cam]->GetExternals();
      } else {
        BIASERR("Not yet defined behaviour!");
        BIASABORT;
        return vecParameters_[0]->GetExternals();
      }
    }

    /** @brief Get the Pose covariance
     */
    virtual Matrix<double> GetCov(unsigned int cam = 0) const {
      BIASASSERT(cam<vecParameters_.size() && vecParameters_[cam]);
      if (cam==0) return vecParameters_[cam]->GetCov();
      else {
        BIASERR("Not yet defined behaviour!");
        BIASABORT;
        return Matrix<double>(7,7);
      }
    }

    /** @brief return complete pose object
        if cam>0 this is a relative pose in the coordinate system of cam 0
    */
    virtual const Pose& 
    GetPose(unsigned int cam = 0) const {
      BIASASSERT(cam<vecParameters_.size() && vecParameters_[cam]);
      return vecParameters_[cam]->GetPose();
    }
  
    /** @brief Set the Pose covariance
     */
    virtual void SetCov(const Matrix<double>& Cov, unsigned int cam = 0) {
      BIASASSERT(cam<vecParameters_.size() && vecParameters_[cam]);
      vecParameters_[cam]->SetCov(Cov);
      //    if (cam==0) vecParameters_[cam]->SetCov(Cov);
      //    else {
      //      BIASERR("Not yet defined behaviour!");
      //      BIASABORT;
      //    }
    }

    /** @brief set center of camera with index cam 
        if cam>0 this is a relative pose in the coordinate system of cam 0 
    */
    void SetC(const Vector3<double>& C, unsigned int cam = 0) {
      BIASASSERT(vecParameters_.size()>cam && vecParameters_[cam]);
      vecParameters_[cam]->SetC(C);
      //    if (cam==0) vecParameters_[cam]->SetC(C);
      //    else {
      //      BIASERR("Not yet defined behaviour!");
      //      BIASABORT;
      //    }
    }

    /** @brief set rotation quaternion of camera with index cam 
        if cam>0 this is a relative pose in the coordinate system of cam 0 
    */
    void SetQ(const Quaternion<double>& Q, unsigned int cam = 0) {
      BIASASSERT(vecParameters_.size()>cam && vecParameters_[cam]);
      vecParameters_[cam]->SetQ(Q);
      //    if (cam==0) vecParameters_[cam]->SetQ(Q);
      //    else {
      //      BIASERR("Not yet defined behaviour!");
      //      BIASABORT;
      //    }
    } 

    /** @brief set rotation matrix of camera with index cam 
        if cam>0 this is a relative pose in the coordinate system of cam 0 
    */
    void SetR(const RMatrix& R, unsigned int cam = 0) {
      BIASASSERT(vecParameters_.size()>cam && vecParameters_[cam]);
      vecParameters_[cam]->SetR(R);
      //    if (cam==0) vecParameters_[cam]->SetR(R);
      //    else {
      //      BIASERR("Not yet defined behaviour!");
      //      BIASABORT;
      //    }
    }

    /** @brief Set the Pose 
        if cam>0 this is a relative pose in the coordinate system of cam 0 
    */
    virtual void SetPose(const BIAS::Pose pose, unsigned int cam = 0) {
      BIASASSERT(vecParameters_.size()>cam && vecParameters_[cam]);
	  vecParameters_[cam]->SetPose(pose);
      //    if (cam==0) vecParameters_[cam]->SetPose(pose);
      //    else {
      //      BIASERR("Not yet defined behaviour!");
      //      BIASABORT;
      //    }
    }

    inline bool PoseValid(unsigned int cam = 0) const {
      if (vecParameters_.size()<=cam || !vecParameters_[cam]) return false;
      return vecParameters_[cam]->PoseValid();
    }

    /** @brief returns the 2d projection of X in camera cam,
        where X is given in the global coordinate frame (not the rig)    
        (need HomgPoint3D to map points at infinity) 
    */
    HomgPoint2D Project(const HomgPoint3D &X, unsigned int cam = 0,
                        bool IgnoreDistortion = false) const;

    /** Checks if 3D point projects into specified image and returns
     *  belonging 2D image point.
     *  @param X assumes homogenized point!
     *  @author bartczak
     */
    virtual bool DoesPointProjectIntoImage(const BIAS::HomgPoint3D& X, 
                                           BIAS::HomgPoint2D& x, 
                                           unsigned int cam = 0,
                                           bool IgnoreDistortion = false) const;

    /**
     * \returns false if all params in argument and all params (*this) 
     * have same extrinsics.
     */
    virtual const bool DoExtrinsicsDiffer(const BIAS::Projection &p) const;
    
    /**
     * \returns false if all params in argument and all params in (*this) 
     * have same intrinsics
     */
    virtual const bool DoIntrinsicsDiffer(const BIAS::Projection &p) const;


    /** @brief calculates the viewing ray in the global coordinate frame 
        (not the rig) from the camera center which belongs to the given image 
        position. The resulting vector3 may have any length. 
    */
    Vector3<double> UnProjectToRay(const HomgPoint2D& pos,
                                   unsigned int cam = 0,
                                   bool IgnoreDistortion = false) const;

    /** @brief calculates a 3D point in
        the global (not the rig) coordinate system, which 
        belongs to the image position pos with distance depth to the
        camera center of cam.
    */
    Vector3<double> UnProjectToPoint(const HomgPoint2D& pos,
                                     double depth,
                                     unsigned int cam = 0,
                                     bool IgnoreDistortion = false) const;

    /** @brief calculates a 3D point in
        a local camera coordinate system specified by camSystem, which 
        belongs to the image position pos in cam with distance depth to the
        camera center cam.                
        @author koeser  */
    Vector3<double> UnProjectToPointLocal(const HomgPoint2D& pos,
                                          const double& depth,
                                          unsigned int camSystem = 0,
                                          unsigned int cam = 0,
                                          bool IgnoreDistortion = false) const;


    /** @brief calculate 3D point and associated covariance by unprojection
        @author woelk 03/2006 */
    Matrix3x3<double> UnProjectToCovLocal(const HomgPoint2D& pos,
                                          const Matrix3x3<double>& cov2D, 
                                          const double& depth,
                                          unsigned int camSystem = 0,
                                          unsigned int cam = 0,
                                          bool IgnoreDistortion = false,
                                          bool Normalize = false) const;


    

    /** @brief calculates a 3D ray in
        a local camera coordinate system specified by camSystem, which 
        belongs to the image position pos in cam.                
    */
    Vector3<double> UnProjectToRayLocal(const HomgPoint2D& pos,
                                        unsigned int camSystem = 0,
                                        unsigned int cam = 0,
                                        bool IgnoreDistortion = false) const;
    
    virtual ~Projection() {
      for (unsigned int i=0; i<vecParameters_.size(); i++)
        delete vecParameters_[i];
    }

#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int XMLGetClassName(std::string& TopLevelTag,
                                double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject); 
#endif // BIAS_HAVE_XML2
    
    /** @brief adapt internal params to resampled image 
        @param ratio 2.0 = downsample, 0.5 = upsample
    */
    inline void Rescale(float ratio, unsigned int cam = 0) {
      BIASASSERT(cam < vecParameters_.size());
      BIASASSERT(vecParameters_[cam]!=NULL);
      vecParameters_[cam]->Rescale(ratio);
    }


    inline void SetIdentifier(const std::string& identString) {
      identifier_ = identString;
    }
    
    inline std::string GetIdentifier() const {
      return identifier_;
    }

    friend std::ostream& operator<<(std::ostream &os, const Projection& p);
    

  protected:

    // exports for dll (cmenk)
    // BIAS_EXPORT_STL_VECTOR( BIASGeometry_EXPORT, ProjectionParametersBase* )

    /// the intrinsics (like K, ...) of the camera usually only one
    /// only in rig mode there may be many, in that case vecIntrinsics[0]
    /// holds the external pose while the others hold the internal pose 
    /// (pose in the coordinate system of the first camera)
    std::vector<ProjectionParametersBase* > vecParameters_; 

    /** Multifunctional identifier.
     */
    std::string identifier_;

  };

  inline std::ostream& operator<<(std::ostream &os, const Projection& p) {
    os <<"Identifier: """<<p.identifier_<<"""";
    os <<"vecParameters_.size(): "<<p.vecParameters_.size()<<std::endl;
    std::vector<ProjectionParametersBase* >::const_iterator it;
    for (it = p.vecParameters_.begin(); it!=p.vecParameters_.end(); it++){
      os<< **it;
      if (ProjectionParametersPerspective *pp 
          = dynamic_cast<ProjectionParametersPerspective *>(*it)) 
        os <<*pp;
      else if (ProjectionParametersPerspectiveDepth *ppd 
               = dynamic_cast<ProjectionParametersPerspectiveDepth *>(*it)) 
        os <<*ppd;
      else if (ProjectionParametersSpherical *pps 
               = dynamic_cast<ProjectionParametersSpherical*>(*it))
        os << *pps;
    }
    return os;
  }


} // end namespace 


#endif //__BIAS_Projection_hh__

