/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_ProjectionParametersBase_hh__
#define __BIAS_ProjectionParametersBase_hh__

#include <bias_config.h>
#include <Base/Common/W32Compat.hh>
#include <math.h>

#include <Geometry/Pose.hh>

#include <Base/Common/BIASpragmaStart.hh>

#include <Base/Common/CompareFloatingPoint.hh>
#include <Base/Math/Vector3.hh>
#include <Base/Math/Matrix3x4.hh>
#include <Base/Math/Matrix4x4.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Geometry/Quaternion.hh>
#include <Geometry/PMatrix.hh>
#include <MathAlgo/UnscentedTransform.hh>

#include <Base/Debug/Debug.hh>
#ifdef BIAS_HAVE_XML2
#  include <Base/Common/XMLBase.hh>
#endif

#include <Geometry/RMatrix.hh>


namespace BIAS {


  /** @class PixelIterator
      @brief can be used to run along the image border
      @author koeser 09/2007
  */
  class PixelIterator {
  public:
    /// coordinates of the pixel in the image (0,0) is center of upper left
    /// (w-1, h-1) is center of lower right pixel
    double x,y;
    PixelIterator():x(-1.0),y(-1.0){};
    PixelIterator(const double& xx, const double& yy):x(xx),y(yy){};
  };
  

  /** @class ProjectionParametersBase
      @brief camera parameters which define the mapping between rays in the 
      camera coordinate system and pixels in the image as well as camera pose
      @author koeser/evers/streckel
  */
  class BIASGeometry_EXPORT ProjectionParametersBase : 
    protected UnscentedTransform
#ifdef BIAS_HAVE_XML2
                                                     , public XMLBase
#endif
  {
  public:


    ProjectionParametersBase();
    
    virtual ProjectionParametersBase& operator=(const 
                                                ProjectionParametersBase& p) {
	    Pose_ = p.Pose_;
	    width_= p.width_;
	    height_= p.height_;
	    principalX_= p.principalX_;
	    principalY_= p.principalY_;
	    aspectratio_= p.aspectratio_;
	    QValid_ = p.QValid_;
	    CValid_ = p.CValid_;
	    return *this;
    }

    virtual ~ProjectionParametersBase();

    /** polymorphic member for output
     */
    virtual void Print() const;

    /** @brief call this to start a run at the outer boundary of an image. 
     *
     *  You get the first boundary position in it, e.g. (0,0) for a perspective
     *  image. Feed this into GetNextBorderPixel to get the next position.
     *  @author koeser 09/2007 */
    virtual void GetFirstBorderPixel(PixelIterator& it);

    /** @brief call this iteratively to run at the outer boundary of an image. 
     *
     *  All returned coordinates must have valid local rays in the camera 
     *  coordinate system (e.g. in fisheye cams, we run at the fov circle) and 
     *  must be in the image. Two subsequent coordinates must not be more
     *  distant than 1 pixel. 
     *  @param it must be initialized by GetFirstBorderPixel
     *  @return false if the returned position closes the loop around the image
     *  @author koeser 09/2007 */
    virtual bool GetNextBorderPixel(PixelIterator& it);

    
    /** Determines the maximal and minimal viewing range in means of
     *  spherical coordinates. The spherical coordinates are calculated
     *  relatively to the passed coordinate frame.
     *  The method is using the class SphericalCoordinates, see also the 
     *  documentation there in order to understand the meaning of the 
     *  sphericalReferenceFrame.
     *     
     * \attention in order to make sense the sphericalReferenceFrame
     *  must have the same origin as the ProjectionParameters.
     *
     * \param sphericalReferenceFrame reference frame for the calculation of 
     * spherical coordinates belonging to optical rays, must be defined in the 
     * same coordinate frame like the local frame of the ProjectionParameters.
     * 
     * \returns minPhi phi angle from (-M_PI, M_PI] with smallest value within viewing range.
     * \returns maxPhi phi angle from (-M_PI, M_PI] with largest value within viewing range.
     * \returns centerPhi phi angle from (-M_PI, M_PI] belonging to the optical ray passing 
     *    through the image center, this angle allows to determine in which direction the 
     *    camera was facing in the sphericalReferenceFrame.
     * \returns minTheta theta angle from [0, M_PI] with smallest value within viewing range.
     * \returns maxTheta theta angle from [0, M_PI] with largest value within viewing range.
     *
     * \author bartczak 10/2007  
     **/
    virtual int GetSphericalViewingRange(const CoordinateTransform3D& sphericalReferenceFrame, 
                                         double& minPhi, double& maxPhi, double& centerPhi,
                                         double& minTheta, double& maxTheta ) ;


    /** Delivers the assumed minimal angular distance between two optical rays
     * belonging to integer image coordinates within the FoV.
     * \author bartczak 10/2007  
     **/
    virtual int GetMinimalAngularSamplingStep(double& minAngleStep)
    {
      BIASERR("to be implemented in specialized child class!");
      return -1;
    }


    virtual const bool 
    DoExtrinsicsDiffer(const ProjectionParametersBase* p) const; 
    
    virtual const bool 
    DoIntrinsicsDiffer(const ProjectionParametersBase* p) const;

    /** Checks if 3D point projects into specified image and returns
     *  belonging 2D image point.
     *  @param X assumes homogenized point!
     *  @author bartczak
     */
    bool DoesPointProjectIntoImage(const HomgPoint3D& X, HomgPoint2D& x, 
                                   bool IgnoreDistortion = false) const;
    
    /** Checks if 3D point projects into specified image and returns
     *  belonging 2D image point.
     *  @param localX assumes homogenized point!
     *  @author bartczak
     */
    virtual bool DoesPointProjectIntoImageLocal(const Vector3<double>& localX, 
                                                HomgPoint2D& x,                
                                                bool IgnoreDistortion = false) 
      const
    {
      BIASERR("to be implemented in specialized child class!");
      return false;
    }


    /** @brief calculates the projection of a point 
        in the world coordinate system  
        to a pixel in the image plane of the camera 
        Was in old calss Projection before
        @author woelk 05/2006 */
    inline HomgPoint2D Project(const HomgPoint3D &X,
                               bool IgnoreDistortion = false) const {
	    Vector3<double> x;
	    GetExternals().Mult(X, x);
	    return ProjectLocal(x, IgnoreDistortion);
    }

    /** @brief calculates the projection of a point 
        in the local camera coordinate system  
        to a pixel in the image plane of the camera
        
        In the simplest case perspective pinhole projection x = K * point  
        where point is transformed of X using  point = (R^T | -R^T C) X */
    virtual HomgPoint2D ProjectLocal(const Vector3<double>& point,
                                     bool IgnoreDistortion = false) const = 0;
    
    /** @brief calculates the unit size viewing ray from the camera center 
        (in the camera coordinate system) which belongs to the image position
        pos e.g. pos=principal point results in (0,0,1)=optical axis */
    virtual Vector3<double> UnProjectLocal(const HomgPoint2D& pos,
                                           bool IgnoreDistortion = false)
	    const = 0;

    /** @brief map points from image onto unit diameter image plane in 3D.
     */
    virtual HomgPoint3D UnProjectToImagePlane(const HomgPoint2D& pos,
                                              const double& depth = 1.0,
                                              bool IgnoreDistortion = false) 
	    const =0;
        
    /** @brief unprojects the covariance matrices to K=Identity camera
        
    This uses the unscented transform to propagate covariance. Use a faster
    or better specialization in derived class if you want.
    @author koeser 06/2006 */
    virtual Matrix3x3<double> 
    UnProjectCovLocal(const HomgPoint2D& pos, const Matrix3x3<double>& cov2D,
                      bool IgnoreDistortion = false, bool Normalize = false);

    /** @brief calculates the viewing ray in the global coordinate frame 
        from the camera center which belongs to the given image 
        position. The resulting vector3 may have any length. 
    
        @author woelk 05/2006 */
    inline Vector3<double> UnProjectToRay(const HomgPoint2D& pos,
                                          bool IgnoreDistortion=false) const {
	    Vector3<double> x, rx;
	    x = UnProjectLocal(pos, IgnoreDistortion);
	    RMatrix R = GetR();
	    R.Mult(x, rx);
	    return rx;
    }


    /** @brief calculates a 3D point in
        the global (not the rig) coordinate system, which 
        belongs to the image position pos with distance depth to the
        camera center.
        @return: (0,0,0,0) is returned on failure
    */
    HomgPoint3D UnProjectToPoint(const HomgPoint2D& pos,
                                 double depth,
                                 bool IgnoreDistortion = false) const;


    /** @brief calculates a 3D point in
        the local camera coordinate system, which 
        belongs to the image position pos in cam with distance depth to the
        camera center cam.                
        @author koeser  */
    virtual Vector3<double> 
    UnProjectToPointLocal(const HomgPoint2D& pos, const double& depth,
                          bool IgnoreDistortion = false) const;

    /** Interface defintion for lens distortion function, implemented by 
        derived classes */
    virtual void Distort(BIAS::HomgPoint2D& point2d) const = 0;

    /** Interface defintion for lens undistortion function, implemented by 
        derived classes */
    virtual void Undistort(BIAS::HomgPoint2D& point2d) const = 0;

    // ---- Getter methods  -------------------------------------------
      
    /** @brief get projection center */
    virtual Vector3<double> GetC() const 
    { 
      if(!CValid_)	      
		    BIASWARN("Projection is not valid");	  
      return Pose_.GetC();
    };

    /** @brief return Q of rotation*/
    virtual Quaternion<double> GetQ() const 
    {
      if(!CValid_)	      
		    BIASWARN("Projection is not valid");	
      return Pose_.GetQ();
    };

    /** @brief return rotation matrix R*/
    virtual BIAS::RMatrix GetR() const;
    
    /** @brief return pose covariance as 7x7 matrix C,Q */
    virtual Matrix<POSE_TYPE> GetCov() const { return Pose_.GetCovariance(); }

    /** @brief return complete pose object */
    virtual const Pose& GetPose() const 
    { 
      if(!QValid_ || !CValid_)
		    BIASWARN("Projection is not valid");
      return Pose_;
    };

    inline PoseParametrization GetPoseParametrization() const 
    { 
      if(!QValid_ || !CValid_)
		    BIASWARN("Projection is not valid");
      return Pose_.GetPoseParameters();
    };
    
    inline void SetPoseParametrization(const PoseParametrization& pp)
    { Pose_.Set(pp);  QValid_ = CValid_ = true; }
   
    /** @brief return cached 3x4 RC representation */
    virtual Matrix3x4<double> GetExternals() const 
    { 
      if(!QValid_ || !CValid_)
		    BIASWARN("Projection is not valid");
      return Pose_.GetMatrix3x4();
    };

    /** @brief return image dimensions */
    inline int GetImageSize(unsigned int& Width, unsigned int& Height) const {
	    Width  = width_;
	    Height = height_;
	    return 0;
    }

    /** @brief return aspectratio = cellSizeX_ / cellSizeY_;*/
    inline double GetAspectratio() const {
	    return aspectratio_;
    }

    /** @brief get principal point, x in pixels relative to top left corner */
    inline int GetPrincipal(double& PrincipalX, double& PrincipalY) const {
	    PrincipalX = principalX_;
	    PrincipalY = principalY_;
	    return 0;
    }

    // ---- Setter methods  ------------------------------------------------

    
    /** @brief Set CCDaspect ratio = cellSizeX_ / cellSizeY_. 
        @author streckel */
    virtual inline void SetAspectratio(const double AspectRatio) {
	    aspectratio_ = AspectRatio;
    }

    /** @brief Set principal point in pixels relative to top left corner
        @author streckel */
    virtual inline void SetPrincipal(const double x, const double y) {
	    principalX_ = x;
	    principalY_ = y;
    }

    virtual inline void SetImageSize(const unsigned int w, 
                                     const unsigned int h) {
	    width_ = w; 
	    height_ = h; 
    }

    /** @brief adapt internal params to resampled image 
        @param ratio 2.0 = downsample, 0.5 = upsample
        \attention only delivers correct results if ratio
        is chosen to rescale image size to integral image dimensions 
        in both directions. Image size is rounded!
    */
    virtual inline void Rescale(float ratio) {
	    double widthTemp = (double)width_/ratio;
	    double heightTemp = (double)height_/ratio;
	    width_  = (int)rint(widthTemp);
	    height_ =  (int)rint(heightTemp);
	    if( (!BIAS::Equal(widthTemp, (double)width_)) || 
          (!BIAS::Equal(heightTemp, (double)height_))) { 
        //         BIASERR("image size has integer type."<<width_<<"x"<<height_
        //                 <<" Passed rescale factor "<< ratio <<" does not conform!\n");
	    }
	    principalX_ /= ratio;
	    principalY_ /= ratio;
    }


    /** @brief covariant virtual copy constructor for use in Projection */
    virtual ProjectionParametersBase* Clone() const = 0;

    virtual void SetQ(const BIAS::Quaternion<double>& Q) 
    { Pose_.SetQ(Q); QValid_ = true; }
    virtual void SetC(const BIAS::Vector3<double>& C)
    { Pose_.SetC(C); CValid_ = true; }
    virtual void SetQC(const BIAS::Quaternion<double>& Q,
                       const BIAS::Vector3<double>& C)
    { Pose_.Set(Q,C); QValid_ = CValid_ = true; }

    virtual void SetPose(const BIAS::Pose pose)
    { Pose_=pose; QValid_ = CValid_ = true; }
    
    virtual void SetR(const BIAS::RMatrix& R) 
    { Quaternion<double> Q;
    R.GetQuaternion(Q); 
    Pose_.SetQ(Q);
    QValid_ = true;
    }

    virtual void SetCov(const Matrix<POSE_TYPE>& Cov) 
    { Pose_.SetCovariance(Cov); }

    inline bool CValid() const {return CValid_;}

    inline bool QValid() const {return QValid_;}

    inline bool PoseValid() const
    { return CValid_ && QValid_; }

    inline void InvalidatePose()
    { CValid_ = QValid_ = false; }
 
    inline void ValidatePose(){
	    QValid_= true;
	    CValid_ = true;
    };

    /** @brief Returns a fake KMatrix for the camera.

    @attention Stolen from PPSpherical, since this is a general issue.
    THERE MAY BE PROBLEMS FOR OTHER THAN FISHEYE CAMERAS (UNTESTED)
    If this implementation is not correct for your derived projection 
    parameters you have to reimplement it in the derived class.

    The FoV of the KMatrix is specified by maxangle (PI*80/180 for 160deg)
    or the max angle, if this is <160deg.
    @param resolution The resolution parameter influences the 
    Resolution of the perspective image that is represented by the camera.
    0 - The perspective image has the same size as the this image
    1 - The perspective image has the same resolution as this
    image in the center (highest resolution)
    2 - The perspective image has the same resolution as this
    image near the border (lowest resolution)
    @param imgsize returns the image size
    @param maxangle maximum theta in rad, e.g. (PI*80/180 for 160deg fov)
    @author streckel 05/2006 (moved here by koeser)
    */
    virtual BIAS::KMatrix GetFakeKMatrix(double& imgsize, int resolution=0,
                                         const double& maxangle = 1.4) const;
    virtual BIAS::KMatrix GetFakeKMatrix(int resolution=0,
                                         const double& maxangle = 1.4) const;
   

    /** @brief returns difference of two projections like | C1-C2 | 

    This function can be used for inter camera distance and ordering purposes
    @author koeser 03/2006 */
    virtual double ViewDifference(const ProjectionParametersBase*) const;

#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int XMLGetClassName(std::string& TopLevelTag,
                                double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject); 
#endif


    /** unscented transform worker function 
        @author koeser */
    virtual int Transform_(const Vector<double>& src,
                           Vector<double>& dst) const;

    friend BIASGeometry_EXPORT std::ostream& 
    operator<<(std::ostream &os, 
               ProjectionParametersBase& p);


  protected:
    Pose Pose_;
    bool QValid_, CValid_;

    /// width of image in pixels
    unsigned int width_;
    /// height of image in pixels
    unsigned int height_;

    /// Principal Point Pixel Coordinates (one for all zoom settings)
    double principalX_, principalY_;
    
    /// aspect ratio of the camera CCD
    double aspectratio_;

    /// unscented transform worker variables
    bool ustTransformIntoImage_;
    bool ustIgnoreDistortion_;
    bool ustNormalize_;
  };
 

  BIASGeometry_EXPORT std::ostream& 
  operator<<(std::ostream &os, ProjectionParametersBase& p);

} // end namespace 


#include <Base/Common/BIASpragmaEnd.hh>

#endif 

