/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Geometry/ProjectionParametersPerspective.hh" 

#include "SphericalCoordinates.hh"

using namespace BIAS;
using namespace std;


std::ostream& BIAS::operator<<(std::ostream &os, 
                               ProjectionParametersPerspective& p) 
{
  if (p.PoseValid())
    os <<"R:"<<p.GetR()<<"C:"<<p.GetC()<<std::endl;
  else
    os <<" Pose is invalid, sorry."<<std::endl;
  os<<"ImageSize: "<<p.width_<<" x "<<p.height_<<std::endl;
  os<<"Principal: "<<p.principalX_<<" x "<<p.principalY_<<std::endl;
  os<<"Skew: "<<p.skew_<<std::endl;
  //operator<<(os, (ProjectionParametersBase&)p);
  os<<"K:"<<p.GetK()<<std::endl;
  bool brown = (p.distType_ == DISTYPE_BROWN);
  string dist_model = 
    (brown)?("Browns distortion model"):("Default distortion model");
  os<<dist_model<<": ";
  os<<  " kc1_: " <<p.kc1_<< " kc2_: "<<p.kc2_<<" kc3_: "
    <<p.kc3_<<" kc4_: "<<p.kc4_;
  if (brown)
    os <<" r0_: "<< p.r0_;
  os<<std::endl;
  return os;
}


bool ProjectionParametersPerspective::
GetCutOutParameters(const Vector2<int>& centerPoint,
                    const unsigned int halfWidth,
                    const unsigned int halfHeight, 
                    ProjectionParametersPerspective& cutOutParams) const
{
  cutOutParams = (*this);
  Vector2<int> UL(centerPoint[0]-halfWidth, centerPoint[1]-halfHeight);
  Vector2<int> LR(centerPoint[0]+halfWidth, centerPoint[1]+halfHeight);
  cutOutParams.width_ = 2 * halfWidth + 1;
  cutOutParams.height_ = 2 * halfHeight + 1;
  
  cutOutParams.SetPrincipal(principalX_ - UL[0], principalY_ - UL[1]);

  return ((UL[0]>=0) && (UL[0]<(int)width_)  && 
          (UL[1]>=0) && (UL[1]<(int)height_) && 
          (LR[0]>=0) && (LR[0]<(int)width_)  && 
          (LR[1]>=0) && (LR[1]<(int)height_));
}

int ProjectionParametersPerspective::
GetMinimalAngleInCorner_(const HomgPoint2D& corner, double& minAngle)
{
  HomgPoint2D neighbour[3];
  Vector3<double> ray = UnProjectLocal(corner);
  if(ray.NormL2()<1e-12) return -1;
  ray.Normalize();

  Vector3<double> neighbourRay[3];
  
  const int x = 0;
  const int y = 1;
  const int diag = 2;

  neighbour[x] = corner;
  neighbour[y] = corner;
  neighbour[diag] = corner;
  if(neighbour[x][0] == 0) {
    neighbour[x][0]+=1;
    neighbour[diag][0]+=1;
  } else {
    neighbour[x][0]-=1;
    neighbour[diag][0]-=1;
  }

  if(neighbour[y][1] == 0) {
    neighbour[y][1]+=1;
    neighbour[diag][1]+=1;
  } else {
    neighbour[y][1]-=1;
    neighbour[diag][1]-=1;
  }

  double angleDiff[3];
  for(unsigned int i=0; i<3; i++) {
    neighbourRay[i] = UnProjectLocal(neighbour[i]);
    if(neighbourRay[i].NormL2() < 1e-12) {
      return -1;
    }
    neighbourRay[i].Normalize();
    angleDiff[i] = acos(ray.ScalarProduct(neighbourRay[i]));
  }

  minAngle = (angleDiff[x] < angleDiff[y]) ? angleDiff[x] : angleDiff[y];
  if(minAngle > angleDiff[diag]) {
    minAngle = angleDiff[diag];
  }

  return 0;

}

// looks into the corners and calculates angle differences to extremal 
// optical rays
int ProjectionParametersPerspective::
GetMinimalAngularSamplingStep(double& minAngleStep)
{
  HomgPoint2D corner[4];
  
  corner[0][0] = 0;
  corner[0][1] = 0;
  corner[0][2] = 1;

  corner[1][0] = 0;
  corner[1][1] = height_-1;
  corner[1][2] = 1;

  corner[2][0] = width_-1;
  corner[2][1] = height_-1;
  corner[2][2] = 1;

  corner[3][0] = width_-1;
  corner[3][1] = 0;
  corner[3][2] = 1.0;

  double minAngle[4];
  for(unsigned int i=0; i<4; i++) {
    if(GetMinimalAngleInCorner_(corner[i], minAngle[i])!=0) {
      return -1;
    }
  }
  double minAngleStepTemp[2];
  minAngleStepTemp[0] = (minAngle[0]<minAngle[1]) ? minAngle[0] : minAngle[1];
  minAngleStepTemp[1] = (minAngle[2]<minAngle[3]) ? minAngle[2] : minAngle[3];

  minAngleStep = 
    (minAngleStepTemp[0]<minAngleStepTemp[1]) ? 
    minAngleStepTemp[0] : minAngleStepTemp[1];

  return 0;
}

const bool ProjectionParametersPerspective::
DoIntrinsicsDiffer(const ProjectionParametersBase* p) const
{
  if(ProjectionParametersBase::DoIntrinsicsDiffer(p))
    return true;
  
  const ProjectionParametersPerspective* ppp = 
    dynamic_cast<const ProjectionParametersPerspective*>(p);
  if(ppp == NULL)
    return true;

  if(focallength_ != ppp->focallength_)
    return true;

  if(kc1_ != ppp->kc1_ || kc2_ != ppp->kc2_ || 
     kc3_ != ppp->kc3_ || kc4_ != ppp->kc4_)
    return true;

  return false;
}

bool ProjectionParametersPerspective::
DoesPointProjectIntoImageLocal(const Vector3<double>& localX, 
                               HomgPoint2D& x,                       
                               bool IgnoreDistortion) const
{
  // check if in front of camera (max. allowed FoV 168 degree)
  // The FoV check is needed because near zero radial distortion
  // produces really huge numbers
  if(localX[2]/localX.NormL2() > minZlocal_) {
    // now compute coordinates
    x = ProjectLocal(localX, IgnoreDistortion);
    // check if in image
    return (x[0]>=-0.5 && x[1]>=-0.5 && 
            x[0]<=(double)(width_)-0.5 && x[1]<=(double)(height_)-0.5 );
  } 
  return false;
}


HomgPoint2D ProjectionParametersPerspective::
ProjectLocal(const Vector3<double>& point, bool IgnoreDistortion) const 
{
  // make sure we dont project points behind the camera. Also prevent radial
  // distortion from double overflow !
  if (point[2] <= minZlocal_) return HomgPoint2D(0,0,0);
 
  HomgPoint2D pos2(point), pos;
  if (!IgnoreDistortion) DistortNormalized(pos2);
  // in-plane affine coordinate transformation with KMatrix
  K_.Mult(pos2, pos);
 
  return pos.Homogenize();
}

HomgPoint3D ProjectionParametersPerspective::
UnProjectToImagePlane(const HomgPoint2D& pos, const double& depth,
                      bool IgnoreDistortion) const 
{
  Vector3<double> p = UnProjectLocal(pos, IgnoreDistortion);
  if(Equal(p[2], 0.0))
    return HomgPoint3D(0,0,0,0);
  p*= depth / p[2];
  if(Equal(p.NormL2(), 0.0))
    return HomgPoint3D(0,0,0,0);
  HomgPoint3D local(p);
  return Pose_.LocalToGlobal(local);
}


Vector3<double>  ProjectionParametersPerspective::
UnProjectLocal(const HomgPoint2D& cPos, bool IgnoreDistortion) const {
  HomgPoint2D mpos(cPos); // de-const pos
  // radial undistortion here 
  if (!IgnoreDistortion) Undistort(mpos);
  // in-plane affine coordinate transformation with KMatrix
  Vector3<double> point;
  invK_.Mult(mpos, point);
  const double n = point.NormL2();
  if (n>0.0) point /= n;
  return point;
}


void ProjectionParametersPerspective::
DistortNormalized(HomgPoint2D& point2d) const 
{
  // do nothing in case of ideal lense
  if (kc1_==0 && kc2_==0 && kc3_==0 && kc4_==0) return;
  point2d.Homogenize();

  const double x = point2d[0];
  const double y = point2d[1];
  const double xx = x*x;
  const double yy = y*y;
  const double rsquared = xx + yy;
     
  // radial distortion:
  const double raddist = 1.0 + kc1_*rsquared + kc2_*rsquared*rsquared;
  // tangential distortion:
  if (kc3_!=0.0 || kc4_!=0.0){
    const double twoxy = 2.0*x*y;
    // apply radial and tangential distortion:
    point2d[0] = raddist * x + kc3_*twoxy + kc4_ * (rsquared + 2.0*xx);
    point2d[1] = raddist * y + kc4_*twoxy + kc3_ * (rsquared + 2.0*yy);
  } else {
    // apply only radial distortion:
    point2d[0] = raddist * x;
    point2d[1] = raddist * y;
  }
}

void ProjectionParametersPerspective::
Distort(HomgPoint2D& point2d) const 
{
  switch(distType_)
  {  
    case DISTYPE_BROWN:
    {
      DistortBrown(point2d);
      break; 
    }
    case DISTYPE_DEF:
    {
      // do nothing in case of ideal lense
      if (kc1_==0 && kc2_==0 && kc3_==0 && kc4_==0) return;
      // copy of distorted point
      double distx,disty;
      double x,y;
      double twoxy;
      
      double rsquared;
      double raddist;
      double tandistx, tandisty;
      
      point2d.Homogenize();
      
      // this is effectively: (x,y,1) =  K^-1 * point2D  
      // Points must be Bougouet-normalized 
      // (subtract principal point, then divide by focal length)
      x = (point2d[0]-principalX_)/focallength_;
      y = (point2d[1]-principalY_)/(GetAspectratio()*focallength_);
          
      twoxy = 2 * x*y;
      rsquared = x*x + y*y;
        
      // radial distortion:
      raddist = 1 + kc1_*rsquared + kc2_*rsquared*rsquared;
      // tangential distortion:
      tandistx = kc3_*twoxy + kc4_ * (rsquared + 2*x*x);
      tandisty = kc4_*twoxy + kc3_ * (rsquared + 2*y*y);
    
      // apply distortion:
      distx = raddist * x + tandistx;
      disty = raddist * y + tandisty;
      
      // invert Bougouet-normalization
      point2d[0] = distx*focallength_ + principalX_;
      point2d[1] = disty*GetAspectratio()*focallength_ + principalY_;
      break;
    }
    default:
    {
      BIASERR("Unknown distortion type.");
    }
  }
}


void ProjectionParametersPerspective::
Undistort(HomgPoint2D& point2d) const 
{
  switch(distType_)
    {
    case DISTYPE_BROWN:
      {
        //cout << "using Browns undistortion model\n";
        UndistortBrown(point2d);
        break;
      }
    case DISTYPE_DEF:
      {
        //cout << "using Tsais? undistortion model\n";
        // do nothing in case of ideal lense
        if (kc1_==0 && kc2_==0 && kc3_==0 && kc4_==0) return;
    
        // copy of distorted point
        double distx,disty;
        double x,y;
        double twoxy;
      
        double rsquared;
        double raddist;
        double tandistx, tandisty;
      
        point2d.Homogenize();
    
        // Points must be Bougouet-normalized 
        // (subtract principal point, then divide by focal length)
        x = point2d[0] = (point2d[0]-principalX_)/focallength_;
        y = point2d[1] = (point2d[1]-principalY_)/(GetAspectratio()*focallength_);
      
        // 5 iterations are enough
        for (unsigned int i=0;i<5;i++){
            
          twoxy = 2 * x*y;
          rsquared = x*x + y*y;
        
          // radial distortion:
          raddist = 1.0 + (kc1_ + kc2_*rsquared)*rsquared;
          // tangential distortion:
          tandistx = kc3_*twoxy + kc4_ * (rsquared + 2*x*x);
          tandisty = kc4_*twoxy + kc3_ * (rsquared + 2*y*y);
        
          // apply distortion:
          distx = raddist * x + tandistx;
          disty = raddist * y + tandisty;
        
          // subtract distortion from distorted coordinates
          x = point2d[0] - (distx-x);
          y = point2d[1] - (disty-y);
    
          // prevent NaN in next iteration step
          if (fabs(x)>1e10 || fabs(y)>1e10) {
            point2d[0] = point2d[1] = point2d[2] = 0;
            return;
          }    
        }
    
        // invert Bougouet-normalization
        point2d[0] = x*focallength_ + principalX_;
        point2d[1] = y*GetAspectratio()*focallength_ + principalY_;
        break;
      }
    default:
      {
        BIASERR("unnknown distortion type!");
        BIASABORT;
        break;
      }
    }
}

void ProjectionParametersPerspective::
DistortBrown(HomgPoint2D& point2d) const 
{
  // do nothing in case of ideal lense
  if (kc1_==0 && kc2_==0 && kc3_==0 && kc4_==0 && r0_==0) return;

  // copy of distorted point
  double distx,disty;
  double x,y;
  double twoxy;
  double rc;
  
  double r0squared;
  double rsquared;
  double raddist;
  double tandistx, tandisty;
  
  point2d.Homogenize();
  
  // this is effectively: (x,y,1) =  K^-1 * point2D
  // Points must be Bougouet-normalized 
  // (subtract principal point, then divide by focal length)
  x = (point2d[0]-principalX_)/focallength_;
  y = (point2d[1]-principalY_)/(GetAspectratio()*focallength_);
    
  twoxy = 2 * x*y;
  rsquared = x*x + y*y;    

  // r0 is the distance of polynomial root from the distortion center
  // since the x and y are normalized r0 must be normalized too.
  r0squared = (r0_)*(r0_);  
  
  //"constant root of polynomial"
  rc = kc1_ * r0squared + kc2_* r0squared*r0squared;
  // radial distortion:
  
  raddist = 1 + kc1_*rsquared + kc2_*rsquared*rsquared - rc;
  // tangential distortion:
  tandistx = kc3_*twoxy + kc4_ * (rsquared + 2*x*x);
  tandisty = kc4_*twoxy + kc3_ * (rsquared + 2*y*y);

  // apply distortion:
  distx = raddist * x + tandistx;
  disty = raddist * y + tandisty;
   
  // invert Bougouet-normalization
  point2d[0] = distx*focallength_ + principalX_;
  point2d[1] = disty*GetAspectratio()*focallength_ + principalY_;
}

void ProjectionParametersPerspective::
UndistortBrown(HomgPoint2D& point2d) const {

  // do nothing in case of ideal lense
  if (kc1_==0 && kc2_==0 && kc3_==0 && kc4_==0) return;

// copy of distorted point 
  double x,y;
  double twoxy;
  double rc;
    
  double r0squared;  
  double rsquared;
  double raddist;
  double tandistx, tandisty;
  
  point2d.Homogenize();

  // Points must be Bougouet-normalized 
  // (subtract principal point, then divide by focal length)
  x = point2d[0] = (point2d[0]-principalX_)/focallength_;
  y = point2d[1] = (point2d[1]-principalY_)/(GetAspectratio()*focallength_);

  // r0 is the distance of polynomial root from the distortion center
  r0squared = (r0_)*(r0_);  

  //"constant root of polynomial"
  rc = kc1_ * r0squared + kc2_* r0squared* r0squared;
  

  // 5 iterations are enough
  for (unsigned int i=0;i<5;i++){
        
    twoxy = 2 * x*y;
    rsquared = x*x + y*y;
    
    // radial distortion:
    raddist = 1 + (kc1_ + kc2_*rsquared)*rsquared - rc;   
    // tangential distortion:
    tandistx = kc3_*twoxy + kc4_ * (rsquared + 2*x*x);
    tandisty = kc4_*twoxy + kc3_ * (rsquared + 2*y*y);
    
   // apply undistortion:
//     distx = (raddist) * x - tandistx;
//     disty = (raddist) * y - tandisty;
//     
//     // subtract distortion from distorted coordinates
//     x = point2d[0] - (distx-x);
//     y = point2d[1] - (disty-y);

    // subtract distortion from distorted coordinates
    x = (point2d[0] - tandistx)/raddist;
    y = (point2d[1] - tandisty)/raddist;

    // prevent NaN in next iteration step
    if (fabs(x)>1e10 || fabs(y)>1e10) {
      point2d[0] = point2d[1] = point2d[2] = 0;
      return;
    }
    
  }

  // invert Bougouet-normalization
  point2d[0] = x*focallength_ + principalX_;
  point2d[1] = y*GetAspectratio()*focallength_ + principalY_;
}

#ifdef BIAS_HAVE_XML2
 
int ProjectionParametersPerspective::
XMLGetClassName(std::string& TopLevelTag, double& Version) const {
  TopLevelTag = "ProjectionParametersPerspective";
  Version = 0.1;
  return 0;
}

int ProjectionParametersPerspective::
XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const {
  xmlNodePtr theNode;
  string TopLevelName;
  double Version;
  ProjectionParametersBase::XMLGetClassName(TopLevelName, Version);
  theNode = XMLObject.addChildNode(Node, TopLevelName);
  XMLObject.addAttribute(theNode, "Version", Version);
  ProjectionParametersBase::XMLOut(theNode, XMLObject);

  xmlNodePtr childNode;
  XMLObject.addAttribute(Node,"Focallength", focallength_);
  XMLObject.addAttribute(Node,"Skew", skew_);
  childNode = XMLObject.addChildNode(Node,string("Distortion"));

  //handling of advanced distortion types
  XMLObject.addAttribute(childNode, "k1", kc1_);
  XMLObject.addAttribute(childNode, "k2", kc2_);
  XMLObject.addAttribute(childNode, "p1", kc3_);
  XMLObject.addAttribute(childNode, "p2", kc4_);

  switch(distType_)
  {    
    case DISTYPE_BROWN:
    {
      XMLObject.addAttribute(childNode, "Type", "DISTYPE_BROWN");
      XMLObject.addAttribute(childNode, "r0", r0_);      
      break;
    }    
    default:
    {    
      //      BIASWARN("Unknown distortion type. only default parameters have been saved.");
    }
  }
  return 0; 
 }

 
int ProjectionParametersPerspective::
XMLIn(const xmlNodePtr Node, XMLIO& XMLObject) {

  string TopLevelName;
  double Version;
  ProjectionParametersBase::XMLGetClassName(TopLevelName, Version);

  xmlNodePtr childNode; 
  if ((childNode = XMLObject.getChild(Node, TopLevelName))==NULL) {
    BIASERR("Error in xml, no tag" << TopLevelName);
    return -1;
  }
  if (ProjectionParametersBase::XMLIn(childNode, XMLObject)!=0) return -1;

  focallength_  = XMLObject.getAttributeValueDouble(Node, "Focallength");
  if (XMLObject.getAttributeByName(Node, "Skew")!=NULL)
    skew_  = XMLObject.getAttributeValueDouble(Node, "Skew");
  if ((childNode = XMLObject.getChild(Node, "Distortion"))!=NULL) {
    kc1_ = XMLObject.getAttributeValueDouble(childNode, "k1");
    kc2_ = XMLObject.getAttributeValueDouble(childNode, "k2");
    kc3_ = XMLObject.getAttributeValueDouble(childNode, "p1");
    kc4_ = XMLObject.getAttributeValueDouble(childNode, "p2");
  
    //check if distortion is not default case
    xmlAttrPtr typeAttr = NULL;
    typeAttr = XMLObject.getAttributeByName(childNode, "Type");
    if(typeAttr!=NULL)
    {
       string typestr;
       typestr = XMLObject.getAttributeValueString(typeAttr);
       if(typestr == string("DISTYPE_BROWN"))
       {
          distType_ = DISTYPE_BROWN; //set type of distortion parameters
          r0_ = XMLObject.getAttributeValueDouble(childNode, "r0");
       }
       if(typestr == string("DISTYPE_DEF"))
       {
          distType_ = DISTYPE_DEF;
       }
    }
    else
    {
       distType_ = DISTYPE_DEF;
    }
  }
  ComputeK_();
  return 0;
}

#endif // BIAS_HAVE_XML2

void ProjectionParametersPerspective::
GetIdealK(KMatrix& K) const
{
  unsigned int imageWidth, imageHeight;
  GetImageSize(imageWidth, imageHeight);

  HomgPoint2D p(0, 0);
  Undistort(p);
  p = invK_ * p;
  p.Homogenize();
    
  double minX = p[0], maxX = p[0], minY = p[1], maxY = p[1];

  for (unsigned int x = 0; x < imageWidth; x++) {
    p = HomgPoint2D(x, 0);
    Undistort(p);
    p = invK_ * p;
    p.Homogenize();
    if (p[0] < minX) minX = p[0];
    if (p[0] > maxX) maxX = p[0];
    if (p[1] < minY) minY = p[1];
    if (p[1] > maxY) maxY = p[1];

    p = HomgPoint2D(x, imageHeight-1);
    Undistort(p);
    p = invK_ * p;
    p.Homogenize();
    if (p[0] < minX) minX = p[0];
    if (p[0] > maxX) maxX = p[0];
    if (p[1] < minY) minY = p[1];
    if (p[1] > maxY) maxY = p[1];
  }

  for (unsigned int y = 0; y < imageHeight; y++) {
    p = HomgPoint2D(0, y);
    Undistort(p);
    p = invK_ * p;
    p.Homogenize();
    if (p[0] < minX) minX = p[0];
    if (p[0] > maxX) maxX = p[0];
    if (p[1] < minY) minY = p[1];
    if (p[1] > maxY) maxY = p[1];

    p = HomgPoint2D(imageWidth-1, y);
    Undistort(p);
    p = invK_ * p;
    p.Homogenize();
    if (p[0] < minX) minX = p[0];
    if (p[0] > maxX) maxX = p[0];
    if (p[1] < minY) minY = p[1];
    if (p[1] > maxY) maxY = p[1];
  }

  double focalLengthX = idealImageWidth_ / (maxX - minX);
  double focalLengthY = idealImageHeight_ / (maxY - minY);
  double principalX = -focalLengthX * minX;
  double principalY = -focalLengthY * minY;

  K.SetIdentity();
  K[0][0] = focalLengthX;
  K[1][1] = focalLengthY;
  K[0][1] = skew_;
  K[0][2] = principalX;
  K[1][2] = principalY;
}

void ProjectionParametersPerspective::
SetIdealImageSize(unsigned int width, unsigned int height)
{
  idealImageWidth_ = width;
  idealImageHeight_ = height;
}

int ProjectionParametersPerspective::
SetIntrinsics(double minPhi, double maxPhi, 
              double minTheta, double maxTheta, 
              double angleStep)
{
  bool wrongAngles = false;
  wrongAngles = (minPhi <= -M_PI/2.0) || (maxPhi >= M_PI/2.0) || (minPhi>=maxPhi);
  wrongAngles = wrongAngles || (minTheta <= 0) || (maxTheta >= M_PI) 
    || (minTheta>=maxTheta);
  wrongAngles = wrongAngles || (angleStep<=0);
  if(wrongAngles) {
    BIASERR("invalid angle values");
    return -1;
  }

  //establish an spherical coordinate frame that interpretes
  //angles like described in the docu: 
  //  phi around the H-Axis relative to the A-Axis and 
  //  theta relative to the -H-Axis.
  //For this the internal base of the spherical frame 
  //    has to be rotated around the V-Axis by -90Deg.
  SphericalCoordinates sphericalCoord;
  CoordinateTransform3D sphericalBase; // will contain the canonical (cartesian) base,
  // since calculations are performed relative to the local camera frame 
  // only rotation mentioned above has to be applied:

  Quaternion<double> rot(0, -sin(M_PI/4.0), 0, cos(M_PI/4.0));
  sphericalBase.RotateLocalFrame(rot);
  sphericalCoord.SetAffineBase(sphericalBase);

  //first determine the minimal stepsize within the plane
  double minStepSize = 0;
  // there are three cases of interest:
  //  a. the principle point lies in the image
  //  b. the shortest distance of the image to the pp does not lie on
  //     an image corner (rather on a single border)
  //  c. the shortest distance of the image to the pp does lie on an
  //     image corner
  bool ppIsInPhiRange = (minPhi<0) && (0<maxPhi);
  bool ppIsInThetaRange = (minTheta<M_PI/2.0) && (M_PI/2.0<maxTheta);
  //case a
  if(ppIsInPhiRange && ppIsInThetaRange) {
    minStepSize = tan(angleStep);
  } else {
    BIASERR("Not completely implemented!");
    return -1;

    //determine angles closest to the pp
    double phiClose = (fabs(minPhi) < fabs(maxPhi)) ? minPhi : maxPhi;
    double thetaClose = (fabs(minTheta-M_PI/2.0) < fabs(maxTheta-M_PI/2.0)) ? minTheta : maxTheta;
    //case b
    if(ppIsInPhiRange) {
      //calculate point p on image plane 
      Vector3<double> p;
      HomgPoint3D ray;
      sphericalCoord.GetCartesianRayFromFullPhi(0, thetaClose, ray);
      ray.GetEuclidian(p); 
      p/=p[2]; //delivers point on plane since all is calculated in the local cam-frame.
      //since pp is in phi range p[1] should be close to zero
      //hence p[0] is the distance of the image border to the pp and can be used to determine
      // the sample distance:
      double distX = tan(thetaClose+angleStep); //distX = p[0]+minStepSize;
      minStepSize = distX - p[0];
    }
    if(ppIsInThetaRange) {
      //calculate point p on image plane 
      Vector3<double> p;
      HomgPoint3D ray;
      sphericalCoord.GetCartesianRayFromFullPhi(phiClose, 0, ray);
      ray.GetEuclidian(p); 
      p/=p[2]; //delivers point on plane since all is calculated in the local cam-frame.
      //since pp is in theta range p[0] should be close to zero
      //hence p[1] is the distance of the image border to the pp and can be used to determine
      // the sample distance:
      double distY = tan(phiClose+angleStep); //distY = p[1]+minStepSize;
      minStepSize = distY - p[1];
    }

    //case c
    if(!ppIsInPhiRange && !ppIsInThetaRange) {
      //calculate point p on image plane 
      Vector3<double> p;
      HomgPoint3D ray;
      sphericalCoord.GetCartesianRayFromFullPhi(phiClose, thetaClose, ray);
      ray.GetEuclidian(p); 
      p/=p[2]; //delivers point on plane since all is calculated in the local cam-frame.

    }
  }

  width_ = rint(ceil((maxTheta-minTheta))/minStepSize);
  height_ = rint(ceil((maxPhi-minPhi))/minStepSize);

  //upper left corner
  Vector3<double> pUL;
  HomgPoint3D rayUL;
  sphericalCoord.GetCartesianRayFromFullPhi(minPhi, minTheta, rayUL);
  rayUL.GetEuclidian(pUL); 
  pUL/=pUL[2];
  cerr<<pUL<<endl;
  
  //lower right corner
  Vector3<double> pLR;
  HomgPoint3D rayLR;
  sphericalCoord.GetCartesianRayFromFullPhi(maxPhi, maxTheta, rayLR);
  rayLR.GetEuclidian(pLR); 
  pLR/=pLR[2];
  cerr<<pLR<<endl;
  
  
  focallength_ = double(width_-1)/(pLR[0]-pUL[0]);
  double focallengthY = double(height_-1)/(pLR[1]-pUL[1]);
  aspectratio_ = focallengthY/focallength_;
  skew_ = 0;;
  principalX_ = -focallength_*pUL[0];
  principalY_ = -aspectratio_*focallength_*pUL[1];  
  ComputeK_();
 
  
  return 0;
}

void ProjectionParametersPerspective::
GetIdealImageSize(unsigned int& width, unsigned int& height) const
{
  width = idealImageWidth_;
  height = idealImageHeight_;
}

void ProjectionParametersPerspective::
ComputeK_() {
  K_.SetIdentity();
  K_[0][0] = focallength_;
  K_[1][1] = aspectratio_ * focallength_;
  K_[0][1] = skew_;
  K_[0][2] = principalX_;
  K_[1][2] = principalY_;
  invK_ = K_.Invert();
}

double ProjectionParametersPerspective::
ViewDifference(const ProjectionParametersBase* pPPB) const {
  const ProjectionParametersPerspective *pPPP = 
     dynamic_cast<const ProjectionParametersPerspective* >(pPPB);
  // different projection model has a very different view
  if (pPPP==NULL) {
    BIASERR("View difference for different projection models requested !");
    return DBL_MAX;
  }

  Vector3<double> C0 = pPPP->GetC(), A0, A;
  
  RMatrix R = GetR(), R0 = pPPP->GetR();
  
  A[0] = R[0][2]; A[1] = R[1][2]; A[2] = R[2][2];
  A.Normalize();
  
  A0[0] = R0[0][2]; A0[1] = R0[1][2]; A0[2] = R0[2][2];
  A0.Normalize();
  
  // compute field of view, assuming const K !
  KMatrix K0 =  GetK();
  double f = 0.5*(K0[0][0] + K0[1][1]);
  double halfw = double(width_ + height_) / 4.0;
  double coshalffieldofview = cos(atan(halfw/f));
  
  double newness = (GetC()-C0).NormL2();
  newness += 1.0/(1.0-coshalffieldofview)*(1.0-fabs(A.ScalarProduct(A0)));
  return newness;
}


void ProjectionParametersPerspective::UpdateMinZLocal() {

  minZlocal_ = 1e100; // 179.9... deg fov
     
  // run over all border pixels of image
  PixelIterator it;
  GetFirstBorderPixel(it);
  do {
    // compute ray in camera coordinate system
    Vector3<double> ray = UnProjectLocal(HomgPoint2D(it.x, it.y));
#ifdef BIAS_DEBUG
    if (!Equal( ray.NormL2(), 1.0)) {
      BIASERR("ray not unit length: " << ray);
      BIASABORT;
    }
#endif
    if (ray[2] < minZlocal_) {
      minZlocal_ = ray[2];
    }
  } while (GetNextBorderPixel(it));
  // add 3% safety margin
  if (minZlocal_<1e-10) minZlocal_=1e-10;
  else minZlocal_ *= 0.97;
  
  cout<<"maximum viewing angle of this camera is "
      <<acos(minZlocal_)*180.0/M_PI<<" degrees"
      <<endl;
}
