/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_ProjectionParametersPerspective_hh__
#define __BIAS_ProjectionParametersPerspective_hh__
#include "bias_config.h" 

#include <Base/Common/BIASpragmaStart.hh>
#include <Geometry/ProjectionParametersBase.hh>
#include <Base/Geometry/KMatrix.hh>
#include <Geometry/PMatrix.hh>

#define LUT_RES 10
namespace BIAS {

 /** @class ProjectionParametersPerspective
   *  @ingroup g_geometry
     @brief camera parameters which define the mapping between rays in the 
     camera coordinate system and pixels in the image as well as external pose
     @author koeser/evers/streckel
 */

  enum BIAS_ProjParaPersp_DISTORTION_TYPE
                         {DISTYPE_DEF = 0,   // standard distortion
                          DISTYPE_BROWN      // brown distortion
                         };

  class BIASGeometry_EXPORT ProjectionParametersPerspective
    : public  ProjectionParametersBase {
  public:

    ProjectionParametersPerspective()
      : ProjectionParametersBase() 
    {      
      K_.SetIdentity();
      invK_.SetIdentity();
      focallength_=1.0;
      skew_=0;
      kc1_=0;
      kc2_=0;
      kc3_=0;
      kc4_=0;
      r0_=0;
      distType_=DISTYPE_DEF;
      idealImageWidth_ = idealImageHeight_ = 512;
      minZlocal_ = 0.1;
    };

    ProjectionParametersPerspective(const ProjectionParametersPerspective & P)
    {
      *this = P;
    };

    ProjectionParametersPerspective& 
    operator=(const ProjectionParametersPerspective & P){
      ProjectionParametersBase::operator=(P);
      K_ = P.K_;
      invK_ = P.invK_;
      focallength_= P.focallength_;
      skew_=P.skew_;
      kc1_= P.kc1_;
      kc2_= P.kc2_;
      kc3_= P.kc3_;
      kc4_= P.kc4_;  
      r0_ = P.r0_;
      distType_ = P.distType_;
      idealImageWidth_ = P.idealImageWidth_;
      idealImageHeight_ = P.idealImageHeight_;
      minZlocal_ =  P.minZlocal_;
      return *this;
    };

    /** Method calculates K-Matrix and image size from specified angles.
    * Angle phi is determining the opening angle and the principle point shift in y direction,
    * while theta is determining the opening angle and the principle point shift in x direction.
    * angleStep is used to derive the resolution, hence the focallength and the aspectratio. 
    * The K-Matrix is free of skew.
    *
    * Valid values for phi are (-M_PI/2, M_PI/2), while theta must lie within (0,M_PI).
    * Phi's reference direction is the A-Vector, theta's is the -H-Vector (negated H-Vector).
    * The extrinsics are left unchanged (which you certainly guessed already).
    *
    * \returns -1 if angle values made no sense to the method.
    *
    * \author bartczak 10/2007
    **/
    int SetIntrinsics(double minPhi, double maxPhi, 
      double minTheta, double maxTheta, 
      double angleStep);

    virtual ~ProjectionParametersPerspective(){};

    virtual int GetMinimalAngularSamplingStep(double& minAngleStep);

    virtual const bool DoIntrinsicsDiffer(const ProjectionParametersBase* p) const;


    bool DoesPointProjectIntoImageLocal(const Vector3<double>& localX, 
                                        HomgPoint2D& x,                       
                                        bool IgnoreDistortion = false) const;
    

    /** @brief calculates the projection of a point 
        in the camera coordinate system  
        to a pixel in the image plane of the camera
        with lens distortion
        In the simplest case perspective pinhole projection x = K * y  
        where y is the projection of X using  y = (R^T | -R^T C) X 
    */
    virtual HomgPoint2D ProjectLocal(const Vector3<double>& point,
                                     bool IgnoreDistortion = false) const;

    /** @brief map points from image onto unit diameter image plane in 3D.    
     * Chosen is the image plane with distance of one from the image center.
     */
    HomgPoint3D UnProjectToImagePlane(const HomgPoint2D& pos,
                                      const double& depth = 1.0,
                                      bool IgnoreDistortion = false) const;
   

    /** @brief calculates the viewing ray from the camera center (in the 
        camera coordinate system) which belongs to the image position pos
        with lens distortion
        e.g. (0,0,1) means optical axis
    */
    virtual Vector3<double> UnProjectLocal(const HomgPoint2D& pos,
                                           bool IgnoreDistortion = false)const;
      
    /** @brief Using the Matlab camera calibration toolbox parameters 
        for an undistortion of distorted coordinates.
        TODO: faster implementation via Lookup-table
        @author streckel 06/2004 */
    virtual void Undistort(BIAS::HomgPoint2D& point2d) const;
 
    /** @brief Using the Matlab camera calibration toolbox parameters 
        for a distortion of undistorted coordinates. Inverse to the 
        undistort function.
        TODO: faster implementation via Lookup-table
        @author streckel 06/2004 */
    virtual void Distort(BIAS::HomgPoint2D& point2d) const;

    /** @brief Undistort distorted coordinates using the
         standard distortion parameters and the parameter describing the
         second constant root of the distortion polynomial of the
         radial symmetric distortion.
         In contrast to the the bouguet way of calculating the radial symmetric
         distortion by using the polynomial 
         raddist = 1 + kc1 * r^2 + kc2 * r^4
         the "brown" way uses a modyfied polynomial which has a second
         root. the second constant root of the polynomial is implemented by 
         subtracting the constant term rc from raddist with 
         rc = kc1 * r0^2 + kc2 * r0^4 
         r0 is the parameter describing the root of the polynomial.
         (For details see: Luhmann, Nahbereichsphotogrammetrie. page. 121 )        
       @author grauel 09/2007 */
    void UndistortBrown(HomgPoint2D& point2d) const;
    
    /** @brief Distort undistorted coordinates using the
        
        @author grauel 09/2007 */
    void DistortBrown(HomgPoint2D& point2d) const;
    
    /** @brief distortion of a normalized point
        @author woelk 11/2006 */
    void DistortNormalized(BIAS::HomgPoint2D& point2d) const;
 
    /** @brief covariant virtual copy constructor for use in Projection */
    virtual ProjectionParametersPerspective* Clone() const {
      return new ProjectionParametersPerspective(*this);
    } 

    /** Shifts the principle point so that imagePoint-(halfWidth, halfHeight)
     * is mapped to (0,0). Image size is adapted to halfWidth*2+1 and 
     * halfHeight*2+1. Distortion parameters are applied in normalized coords 
     * hence they keep their validity.
     * \returns true only if whole cutout lies in image.
     */
    bool GetCutOutParameters(const Vector2<int>& centerPoint,
                             const unsigned int halfWidth,
                             const unsigned int halfHeight, 
                             ProjectionParametersPerspective& cutOutParams) 
      const;

    /** @brief Get the current camera focal length.
        @author buck */
    inline int GetFocalLength(double& f) const {
      f = focallength_;
      return 0;
    }

    inline void GetUndistortion(double& kc1, double& kc2,
                                double& kc3, double& kc4) const 
    {
      kc1 = kc1_;
      kc2 = kc2_;
      kc3 = kc3_;
      kc4 = kc4_;
    }

    inline bool IsDistorted() const {
      return ((kc1_!=0.0) || (kc2_!=0.0) || (kc3_!=0.0) || (kc4_!=0.0));
    }

    /** @brief Set the current camera focal length in pixel and the a
        spect ratio */
    inline void SetFocalLengthAndAspect(double f, 
                                        double AspectRatio) {
      focallength_ = f;
      ProjectionParametersBase::SetAspectratio(AspectRatio);
      ComputeK_();
    }

    /** @brief Set principal point in pixels relative to top left corner,
        virtual overload to recalculate K_
        @author streckel */
    virtual void SetPrincipal(const double x, const double y) {
      ProjectionParametersBase::SetPrincipal(x, y);
      ComputeK_();      
    }

    /** @brief Set CCD cell-size in meter, virtual overload to 
        recalculate K_.
        @author streckel */
    virtual void SetAspectratio(const double AspectRatio) {
      ProjectionParametersBase::SetAspectratio(AspectRatio);
      ComputeK_();
    }

    /** @brief Set type of distortion parameters
        @author grauel */
    inline void SetDistortionType(BIAS_ProjParaPersp_DISTORTION_TYPE distype)
    {
      distType_ = distype;
    }
    
    /** @brief Get type of distortion parameters
        @author grauel */
    inline BIAS_ProjParaPersp_DISTORTION_TYPE GetDistortionType()
    {
      return(distType_);
    }

    /** @brief Set the lens undistortion parameters*/
    inline void SetUndistortion(double kc1, double kc2,
                                double kc3, double kc4) 
    {
      kc1_ = kc1;
      kc2_ = kc2;
      kc3_ = kc3;
      kc4_ = kc4;
      distType_ = DISTYPE_DEF; //reset to default distortion
    }
    
    /** @brief Set the lens undistortion parameters including the root of
        the polynomial.        
        @author grauel 9/2007 */
    inline void SetUndistortionBrown(double kc1, double kc2,
                                     double kc3, double kc4,
                                     double r0) 
    {
       kc1_ = kc1;
       kc2_ = kc2;
       kc3_ = kc3;
       kc4_ = kc4;
       r0_  = r0;
       distType_ = DISTYPE_BROWN;
    }
    
    /** @brief Get the lens undistortion parameters including the parameter
        describing root of the polynomial
        @author grauel 9/2007 */
    inline void GetUndistortionBrown(double& kc1, double& kc2,
                                     double& kc3, double& kc4,
                                     double& r0) const 
    {
       kc1 = kc1_;
       kc2 = kc2_;
       kc3 = kc3_;
       kc4 = kc4_;
       r0 = r0_;
       if(distType_ != DISTYPE_BROWN)
       {
          BIASWARN("Distortion type is not DISTTYPE_BROWN. the parameters might not be  valid.");
       }
    }
    
    /* @brief get the PMatrix composed from the projection object
     */
    inline virtual BIAS::PMatrix GetP() const {
      return PMatrix(K_, GetR(), GetC());
    }

    /* @brief get cached KMatrix
     */
    inline virtual BIAS::KMatrix GetK() const { 
      return K_;
    };

    /* @brief get cached inverted KMatrix
     */
    inline virtual BIAS::KMatrix GetKinv() const { 
      return invK_;
    };

    /** @brief sets the internal parameters from a given KMatrix
        and updates the cached K and its inverse */
    virtual void SetK(const KMatrix& K) {
      BIASASSERT(K[0][0]!=0);
      focallength_  = K[0][0];
      skew_ = K[0][1];
      ProjectionParametersBase::SetPrincipal(K[0][2],K[1][2]);
      double aspectratio = K[1][1]/K[0][0];
      ProjectionParametersBase::SetAspectratio(aspectratio);
      K_ = K; invK_ = K.Invert();
    };

    /** @brief set from P  */
    virtual void SetP(const PMatrix& P) {
      PMatrix p(P); // de-const
      SetK(p.GetK());
      SetC(p.GetC());
      SetR(p.GetR());
      ComputeK_(); 
    };


    /** @brief adapt internal params to resampled image 
        @param ratio 2.0 = downsample, 0.5 = upsample
    */
    virtual void Rescale(float ratio) {
      ProjectionParametersBase::Rescale(ratio);
      focallength_ /= ratio;
      ComputeK_();
      // dont have to update radial/tangential distortion coefficients
      // because we measure in normalize space
    }
    
    /** @brief compute scalar dissimilarity between two cameras based upon
        rough approximation of field of view overlap for pure rotation and
        penalty term for translation: 0=identical */
    virtual double ViewDifference(const ProjectionParametersBase* pPPB) const;

    void GetIdealK(KMatrix& K) const;
    void SetIdealImageSize(unsigned int width, unsigned int height);
    void GetIdealImageSize(unsigned int& width, unsigned int& height) const;


    /** @brief sets minimum z value of unit-ray in local CCS to be accepted
        as in front of camera (e.g. DoesPointProjectIntoImage) */
    void SetMinZLocal(const double& minz) {
      minZlocal_ = minz;
    }

    /** @get minimum z value of ray in image (if computed already) */
    double GetMinZLocal() const { return minZlocal_; }

    /** @brief run along image border and compute maximum field of view,
     save in minzlocal */
    void UpdateMinZLocal();

#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int XMLGetClassName(std::string& TopLevelTag,
                                double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject); 
#endif

    BIASGeometry_EXPORT friend std::ostream& operator<<(std::ostream &os, 
                                    ProjectionParametersPerspective& p);


    
  protected:
    /** @brief computes the cached KMatrix and its inverse
        call after every parameter change */
    void ComputeK_();

    /// focal length, aspect ratio, principal point in K_, inverse in invK_
    KMatrix K_, invK_;

    /// focal length is stored in pixel for cellSizeX
    double focallength_;
    /// lens undistortion parameters after Bouquet
    /// also used as parameters for distortion after Brown(?)
    double kc1_,kc2_,kc3_,kc4_;
    
    // additional lens undistortion parameter after Brown(?)
    // describing the constant root of the polynomial
    double r0_;
    
    // distortion type parameter
    BIAS_ProjParaPersp_DISTORTION_TYPE distType_;

    /// skew calibration parameter
    double skew_;

    /// ideal image size
    unsigned int idealImageWidth_, idealImageHeight_;

    /// rigorously clip unit rays in CCS with z smaller than this values
    double minZlocal_;

    /** Helper function for minimal angle step calculation. **/
   int GetMinimalAngleInCorner_(const HomgPoint2D& corner, double& minAngle);

  };

  BIASGeometry_EXPORT std::ostream& operator<<(std::ostream &os, 
                           ProjectionParametersPerspective& p);

 
} // end namespace 


#include <Base/Common/BIASpragmaEnd.hh>

#endif 

