/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_ProjectionParametersPerspectiveDepth_hh__
#define __BIAS_ProjectionParametersPerspectiveDepth_hh__
#include "bias_config.h" 

#include <Geometry/ProjectionParametersPerspective.hh>

#include <Base/Common/BIASpragmaStart.hh>

#define DEPTH_CALIB_PARAMS 6

namespace BIAS {

 /** @class ProjectionParametersPerspectiveDepth
   *  @ingroup g_geometry
     @brief additional depth calibration parameters for a perspective depth camera
 */
  class BIASGeometry_EXPORT ProjectionParametersPerspectiveDepth
    : public  ProjectionParametersPerspective {
  public:

    ProjectionParametersPerspectiveDepth()
      : ProjectionParametersPerspective() 
    {
      DepthCalibrationParameters_.resize(DEPTH_CALIB_PARAMS, 0.0);
    };

    ProjectionParametersPerspectiveDepth(const ProjectionParametersPerspectiveDepth & P)
    {
      *this = P;
    };

    ProjectionParametersPerspectiveDepth(const ProjectionParametersPerspective & P)
    {
      *this = P;
    };

    ProjectionParametersPerspectiveDepth& 
    operator=(const ProjectionParametersPerspectiveDepth & P){
      ProjectionParametersPerspective::operator=(P);
      DepthCalibrationParameters_ = P.DepthCalibrationParameters_;
      return *this;
    };

    ProjectionParametersPerspectiveDepth& 
    operator=(const ProjectionParametersPerspective & P){
      ProjectionParametersPerspective::operator=(P);
      DepthCalibrationParameters_.resize(DEPTH_CALIB_PARAMS, 0.0);
      return *this;
    };

    virtual ~ProjectionParametersPerspectiveDepth(){};

    
    /** @brief covariant virtual copy constructor for use in Projection */
    virtual ProjectionParametersPerspectiveDepth* Clone() const {
      return new ProjectionParametersPerspectiveDepth(*this);
    } 

    void SetDepthCalibrationParameters (std::vector<double> params) {
      BIASASSERT(params.size()==DEPTH_CALIB_PARAMS);
      DepthCalibrationParameters_ = params;
    }

  
    /* @brief Distort depth value d at image position pos */
    void DistortDepth (HomgPoint2D const &pos, float &d) const;

    /* @brief UnDistort depth value d at image position pos */
    void UnDistortDepth (HomgPoint2D const &pos, float &d) const;

#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int XMLGetClassName(std::string& TopLevelTag,
                                double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject); 
#endif

    friend std::ostream& operator<<(std::ostream &os, 
                                    ProjectionParametersPerspectiveDepth& p);

    
  protected:
    std::vector<double> DepthCalibrationParameters_;
  };

  inline std::ostream& operator<<(std::ostream &os, 
                                  ProjectionParametersPerspectiveDepth& p) 
  {
    ProjectionParametersPerspective ppp = p;
    os << ppp << std::endl;
    os << "Depth Calibration Parameters:" << std::endl;
    for (unsigned int i=0; i<p.DepthCalibrationParameters_.size(); i++) {
      os << p.DepthCalibrationParameters_[i] << std::endl;
    }
    return os;
  }


 
} // end namespace 

#include <Base/Common/BIASpragmaEnd.hh>

#endif 

