#include "Image/CylindricDepthPanorama.hh"
#include <Base/Image/ImageConvert.hh>
#include <Base/Image/ImageIO.hh>
#include "Utils/ThreeDOut.hh"

using namespace BIAS;
using namespace std;

/** Constructor */
CylindricDepthPanorama::CylindricDepthPanorama(){

  cylinderWidth_ = 0.0;
  cylinderHeight_ = 5.0;
  panoramaWidth_  = 1200;
  panoramaHeight_ =
    (unsigned int)(panoramaWidth_ / cylinderHeight_ * M_PI);
  
  bFirstPoseSet_ = false;
  pInitialProjection_ = NULL;
  pCylinderProj_ = NULL; 
  bTextureFilled_ = false;

  panorama_.Init(panoramaWidth_,panoramaHeight_,1);
  panoramaHitCount_.Init(panoramaWidth_,panoramaHeight_,1);
  panoramaHitCount_.FillImageWithConstValue(0.0);
  panoramaTexture_.Init(panoramaWidth_,panoramaHeight_,1);

 
  //Create the cylindrical Projection from initial Projection
  if(pInitialProjection_ == NULL){        
    const double xMin = -cylinderHeight_;
    const double xMax = cylinderHeight_;
    const double phiMin = -M_PI;
    const double phiMax = M_PI;
  
    pCylinderProj_ = 
      new ProjectionParametersCylindric(xMin, xMax, 
                                        phiMin, phiMax,
                                        panoramaWidth_, 
                                        panoramaHeight_);
 
    pCylinderProj_->Print();         
  }
  
}

/** Destructor */
CylindricDepthPanorama::~CylindricDepthPanorama(){
  if( pInitialProjection_ != NULL){
    delete pInitialProjection_;
    pInitialProjection_ = NULL;
  }
  if(pCylinderProj_){
    delete pCylinderProj_;
    pCylinderProj_ = NULL;
  }
}


int  CylindricDepthPanorama::
ProcessCylindricProjection(BIAS::Image<float> &depthMap,
                           BIAS::Projection & p,
                           BIAS::Image<unsigned char> *texture){

  int ret =0;
  float **ida = depthMap.GetImageDataArray();
  float **idaP = panorama_.GetImageDataArray();
  float **idaPHC = panoramaHitCount_.GetImageDataArray();
  unsigned char ** idaTex = panoramaTexture_.GetImageDataArray();

  int width = depthMap.GetWidth();
  int height = depthMap.GetHeight();
 
  float depthValue=0.0;
  BIAS::HomgPoint3D point3D;  
  BIAS::HomgPoint2D point2D,cylPoint;

  if(!bFirstPoseSet_){
    pInitialProjection_ = new BIAS::Projection(p);    
    //contruct cylindrical projection      
    pCylinderProj_->SetQC(pInitialProjection_->GetQ(),
                          pInitialProjection_->GetC());

    bFirstPoseSet_ = true;
  }
  

  unsigned int x1,y1;
  for(int x=0;x<width;x++){
    for(int y=0;y<height;y++){      
      point2D[0]=x;
      point2D[1]=y;
      point2D[2]=1;            
      depthValue = ida[y][x];   
      if (depthValue != 0.0) {       
        // Unproject to a 3Dpoint
        point3D = p.UnProjectToPoint(point2D,depthValue);            
        cylPoint = pCylinderProj_->Project(point3D);
        cylPoint.Homogenize();
        x1 = (unsigned int)cylPoint[0];
        y1 = (unsigned int)cylPoint[1];       
          
        //   cout<<"3DPoint: "<<point3D<<" Cylindercoords: "<<cylPoint<<endl;
                         
        if(x1>=0.0 && x1<panoramaWidth_ && 
           y1>=0.0 && y1<panoramaHeight_ ) {
          //fill texture panorama with values
          if(texture != NULL){
            idaTex[y1][x1] = texture->GetImageDataArray()[y][x];         
            bTextureFilled_ = true;
          }
          if(idaPHC[y1][x1] == 0.0){
            idaP[y1][x1]=depthValue;
            idaPHC[y1][x1]=1.0;
          }
          else{          
            idaP[y1][x1]= 
              (idaP[y1][x1] * idaPHC[y1][x1]+depthValue)/(idaPHC[y1][x1]+1);
            idaPHC[y1][x1]+=1.0;
          }
        }
      }
    }
  }
  return ret;



}


int  CylindricDepthPanorama::
GetPanorama(BIAS::Image<float> &image, 
            BIAS::Image<unsigned char> &texture){
  int ret =0;
  image = panorama_; 
  image.ScaleShiftBetween(0,255);
  //BIAS::ImageIO::Write("pan.mip",image);
  if(bTextureFilled_){
    texture = panoramaTexture_;
    //  BIAS::ImageIO::Write("tex.mip",panoramaTexture_);
  }
  else{
    BIAS::ImageConvert::ConvertST(panorama_,texture,
                                  BIAS::ImageBase::ST_unsignedchar);
    texture.ScaleShiftBetween(0,255);
  }
  return ret;
}


int CylindricDepthPanorama::
GetTriangleMesh(BIAS::TriangleMesh &mesh){
  int ret =-1; 
  BIAS::Image<unsigned char> tmpUC;
  if(bTextureFilled_){
    if(pCylinderProj_ != NULL){
      BIAS::Projection proj(*pCylinderProj_);
      mesh.GenerateDenseMesh(panorama_,proj,panoramaTexture_,0,0,
                             panoramaWidth_,panoramaHeight_);

//       mesh.GenerateSimplifiedMesh(panorama_,
//                                   proj,panoramaTexture_,0,0,
//                                   panoramaWidth_,panoramaHeight_);

      return 0;
    }  
    else
      return ret;

  }
  else{
    BIAS::ImageConvert::ConvertST(panorama_,tmpUC,
                                  BIAS::ImageBase::ST_unsignedchar);
    tmpUC.ScaleShiftBetween(0,255);
    if(pCylinderProj_ != NULL){
      BIAS::Projection proj(*pCylinderProj_);
      mesh.GenerateDenseMesh(panorama_,proj,tmpUC,0,0,
                             panoramaWidth_,panoramaHeight_);
//       mesh.GenerateSimplifiedMesh(panorama_,
//                                   proj,panoramaTexture_,0,0,
//                                   panoramaWidth_,panoramaHeight_);

      return 0;
    }
    else
      return ret;
    
  } 
}
