/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "bias_config.h"
#include "PyramidImage.hh"
#include "Base/Image/ImageIO.hh"
#include "Base/Image/ImageConvert.hh"

using namespace BIAS;
using namespace std;

//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////


template <class StorageType>
PyramidImage<StorageType>::
PyramidImage() 
  : std::vector<Image<StorageType>* >(), 
    _rescale()
{
  _PyramidSize=4;
  _RescaleFactor=2.0;
  _rescale.SetFactor(_RescaleFactor);
  _PositionOffset=0.0;
}

template <class StorageType>
PyramidImage<StorageType>::
PyramidImage(const PyramidImage<StorageType>& pim)
{
  _PyramidSize=4;
  _RescaleFactor=2.0;
  *this=pim;
}

template <class StorageType>
PyramidImage<StorageType>::~PyramidImage()
{
  clear();
}

template <class StorageType>
void PyramidImage<StorageType>::
Init(const Image<StorageType>& image)
{
#ifdef BIAS_DEBUG
  if (this->size()!=0){
    BIASERR("do not call init more than once");
    BIASABORT;
  }
#endif

  // this prepares that more pyramid levels may be added later
  // double smallerimagesize = image.GetWidth();
  // if (smallerimagesize>image.GetHeight()) 
  //   smallerimagesize = image.GetHeight();
  // reserve(int(log(smallerimagesize+1.0)/log(2.0))+1);
  
  const int size=_PyramidSize;
  if (size>0){
    Image<StorageType> *L0 = new Image<StorageType>(image);
    push_back(L0);
    for (int i=1; i<size; i++) this->push_back(NULL);
    _CreateLevels();
  }
}

template <class StorageType>
void PyramidImage<StorageType>::
InitFromImageBase(const ImageBase& image) {
#ifdef BIAS_DEBUG
  if (this->size()!=0){
    BIASERR("do not call init more than once");
    BIASABORT;
  }
#endif


  // this prepares that more pyramid levels may be added later
  // double smallerimagesize = image.GetWidth();
  // if (smallerimagesize>image.GetHeight()) 
  //   smallerimagesize = image.GetHeight();
  // reserve(int(log(smallerimagesize+1.0)/log(2.0))+1);

  const int size=_PyramidSize;
  if (size>0){
    Image<StorageType> *L0 = 
      new Image<StorageType>(image.GetWidth(), image.GetHeight(), 
                             image.GetChannelCount());
    ImageConvert::ConvertST(image, *L0, L0->GetStorageType());
    push_back(L0);
    for (int i=1; i<size; i++) this->push_back(NULL);
    _CreateLevels();
  }
}




template <class StorageType>
void PyramidImage<StorageType>::
Init(unsigned width, unsigned height, unsigned channelcount)
{
#ifdef BIAS_DEBUG
  if (this->size()!=0){
    BIASERR("do not call init more than once");
    BIASABORT;
  }
#endif
  unsigned int newwidth=0, newheight=0;
  double pyfac=1.0;
  const int size=_PyramidSize;
  const double factor=_RescaleFactor;

  for (int i=0; i<size; i++){
    newwidth = (unsigned int)rint((double)width / pyfac);
    newheight = (unsigned int)rint((double)height / pyfac);
    Image<StorageType> *L = 
      new Image<StorageType>(newwidth, newheight, channelcount);
    push_back(L);
    _vFactor.push_back(pyfac);
    pyfac*=factor;
  }
}

template <class StorageType>
void PyramidImage<StorageType>::clear()
{
  typename vector<Image<StorageType>* >::iterator it;
  int i=0;
  for (it=this->begin(); it!=this->end(); it++){
    if (!(*it)->IsEmpty()) {
      (*it)->Release();
    }
    if ((*it) != NULL) {
      delete (*it);
      (*it)=NULL;
    }
    i++;
  }
  std::vector<Image<StorageType>* >::clear();
  _vFactor.clear();
  _PositionOffset=0.0;
}

template <class StorageType> 
PyramidImage<StorageType>& 
PyramidImage<StorageType>::
operator=(const PyramidImage<StorageType>& pim)
{
#ifdef BIAS_DEBUG
  //if (this->size()!=0){
  //  BIASERR("possible memory leak, assigning to a non empty pyramid image");
  //}
#endif
  _rescale=pim._rescale;
  _PyramidSize=pim._PyramidSize;
  _RescaleFactor=pim._RescaleFactor;
  _vFactor = pim.GetFactors();

  if (this->size()!=0) clear();
  this->resize(pim.size());
  typename PyramidImage<StorageType>::iterator it;
  typename PyramidImage<StorageType>::const_iterator cit;

  for (it=this->begin(), cit=pim.begin(); 
       it!=this->end() && cit!=pim.end(); it++, cit++){
    *it = new Image<StorageType>(*(*cit));
  }
  _PositionOffset=pim.GetPositionOffset();

  return *this;
}

/////////////////////////////////////////////////////////////////////////////

template <class StorageType>
int PyramidImage<StorageType>::
Downsample()
{
  int res=0;
  typename vector<Image<StorageType>* >::iterator it, lastit;
  for (lastit=this->begin(), it=lastit+1; 
       it!=this->end() && lastit!=this->end(); it++, lastit++)
    res += _rescale.Downsample(*(*lastit), *(*it));

  _PositionOffset= _rescale.GetPositionOffset();
  return res;
}

template <class StorageType>
void PyramidImage<StorageType>::
SetUID(BIAS::UUID uid)
{
  typename PyramidImage<StorageType>::iterator it;
  for (it=this->begin(); it!=this->end(); it++)
    (*it)->SetUID(uid);
}


template <class StorageType>
int PyramidImage<StorageType>::
SetROI(unsigned minx, unsigned miny, unsigned maxx, unsigned maxy)
{
  typename PyramidImage<StorageType>::iterator it;
  for (it=this->begin(); it!=this->end(); it++){
    if ((*it)->GetROI()->SetCorners(minx, miny, maxx, maxy)!=0){
      return -1;
    }
    minx=(unsigned)floor(minx / _RescaleFactor);
    miny=(unsigned)floor(miny / _RescaleFactor);
    maxx=(unsigned)floor(maxx / _RescaleFactor);
    maxy=(unsigned)floor(maxy / _RescaleFactor);
  }
  return 0;
}

template <class StorageType>
void PyramidImage<StorageType>::
SetZero()
{
  typename PyramidImage<StorageType>::iterator it;
  for (it=this->begin(); it!=this->end(); it++){
    (*it)->SetZero();
  }
}

template <class StorageType>
void PyramidImage<StorageType>::
GetSingleImage(Image<StorageType>& im) const
{
  unsigned size=(*this).size();
  if (size==1){
    im=*(*this)[0];
  } else if (size>1) {
    //SetDebugLevel(GetDebugLevel() | D_PYRAMID_SINGLE_IM);
    register unsigned int i=0, j=0, h=0;
    unsigned int width=(*this)[0]->GetWidth()+
      (*this)[1]->GetWidth()+PYRAMID_BORDER;
    unsigned int height=(*this)[0]->GetHeight();
    unsigned int height2=0;
    
    StorageType ***pida=new StorageType**[size];
    unsigned *widthstep=new unsigned[size];
    unsigned *pwidth = new unsigned[size];
    unsigned *pheight=new unsigned[size];
    unsigned cc=(*this)[0]->GetChannelCount();

    for (i=0; i<size; i++){
      pida[i]=(*this)[i]->GetImageDataArray();
      widthstep[i]=(*this)[i]->GetWidthStep();
      pwidth[i]=(*this)[i]->GetWidth();
      pheight[i]=(*this)[i]->GetHeight();
      BIASCDOUT(D_PYRAMID_SINGLE_IM, "pida[i]: "<< pida[i] << "\t"
                <<(*this)[i]->GetImageDataArray()<<"\t"<<widthstep[i]
                <<"\t"<<pheight[i]<<endl);
    }
    // calculate the height of images 1- including border
    for (i=1; i<size; i++)
      height2+=(pheight[i]+PYRAMID_BORDER);
    if (height2>height) height=height2;
    // init image
    if (im.GetWidth()!=width || im.GetHeight()!=height ||
        im.GetChannelCount()!=(*this)[0]->GetChannelCount()) {
      im.Release();
      im.Init(width, height, (*this)[0]->GetChannelCount());
    }
    im.FillImageWithConstValue((StorageType)PYRAMID_BACKGROUND);
    StorageType **ida = im.GetImageDataArray();

    BIASCDOUT(D_PYRAMID_SINGLE_IM,"width: "<<width<<"\theight: "
              <<height<<endl);
  
    h=0;
    for (j=1; j<size; j++){
      for (i=0; i<pheight[j]; i++){
        // copy biggest image
        if (i+h<pheight[0])
          memcpy((void *)(ida[i+h]), (void *)(pida[0][i+h]), widthstep[0]);
        // copy j-biggest image
        memcpy((void *)(ida[i+h]+pwidth[0]+PYRAMID_BORDER*cc), 
               (void *)(pida[j][i]),widthstep[j]);
      }
      for (i=0; i< PYRAMID_BORDER; i++)
        if (i+pheight[j]+h<pheight[0])
          memcpy((void *)(ida[i+pheight[j]+h]), 
                 (void *)(pida[0][i+pheight[j]+h]), 
                 widthstep[0]);
      h+=pheight[j]+PYRAMID_BORDER;
    }
    for (i=h; i<(*this)[0]->GetHeight(); i++){
      memcpy((void *)ida[i], pida[0][i], widthstep[0]);
    }
    delete[] pida;
    delete[] widthstep;
    delete[] pwidth;
    delete[] pheight;
  }
}


template <class StorageType>
int PyramidImage<StorageType>::
CreateAdditionalLayer(unsigned int numextralayers,
                      unsigned int minImageWidth) {
  //BIASWARN("creating "<<numextralayers<<" new levels ");
  BIASASSERT(numextralayers>0);
  BIASASSERT(! this->empty());
  BIASASSERT(this->back() != NULL);
  int res = 0;
  double pyfac = _vFactor.back();
  this->reserve(this->size()+numextralayers);
  for (unsigned int i=0; i< numextralayers; i++) {
    // check if image would be too small:
    if (double(this->back()->GetWidth()/_RescaleFactor<double(minImageWidth)))
      return 1;

    // create new image and downsample from top of pyramid
    Image<StorageType> *pImage = new Image<StorageType>;
    if ((res=_rescale.Filter(*(this->back()), *pImage))!=0){
      BIASERR("filter did not succeed "<<res);
      BIASABORT;
    }

    // update vectors and statistics
    this->push_back(pImage);
    pyfac *= _RescaleFactor;
    _vFactor.push_back(pyfac);
    _PyramidSize++;
  }
  return res;
}



////////////////////////////////////////////////////////////////////////////
//                private functions
////////////////////////////////////////////////////////////////////////////

template <class StorageType>
void PyramidImage<StorageType>::
_CreateLevels()
{
  const double factor=_RescaleFactor;
  
  int res = 0;
  double pyfac=1.0;
  _vFactor.push_back(pyfac);
  typename vector<Image<StorageType>* >::iterator it, lastit;
  for (lastit=this->begin(), it=lastit+1; 
       it!=this->end() && lastit!=this->end(); it++, lastit++){
    if ((*it)==NULL) (*it) = new Image<StorageType>();
    else if (!(*it)->IsEmpty()) (*it)->Release();
    if ((res=_rescale.Filter(*(*lastit), *(*it)))!=0){
      BIASERR("filter did not succeed "<<res);
      BIASABORT;
    }
    pyfac*=factor;
    _vFactor.push_back(pyfac);
  }
  _PositionOffset = _rescale.GetPositionOffset();
}

template <class StorageType>
int PyramidImage<StorageType>::
WriteImages(const std::string& prefix)
{
  int res=0;
  const int psize=this->size();
  ostringstream os;
  for (int i=0; i<psize; i++){
    os.str("");
    os << prefix << "-pl"<<i<<".mip";
    if ((res=ImageIO::Write(os.str(), *((*this)[i])))!=0){
      BIASERR("error writing "<<os.str());
      break;
    }
  }
  return res;
}

template <class StorageType>
StorageType PyramidImage<StorageType>::
GetImageValue(const double& x, const double& y,
              unsigned int scale1, int channel) const {

  if (scale1>this->size()-1) scale1 = this->size()-1;
  
  // compute coordinates in smaller image
  double x1 = x, y1 = y;
  const double offset = GetPositionOffset();
  for (unsigned int s=0; s<scale1; s++) {
    // -0.5 necessary because of offset in 2^n downsampling
    x1 = (x1 - offset)/ _RescaleFactor; 
    y1 = (y1 - offset)/ _RescaleFactor;
  }
  
  unsigned int UpperLeftX, UpperLeftY, LowerRightX, LowerRightY;
  (*this)[scale1]->GetROICorners(UpperLeftX, UpperLeftY, 
                                 LowerRightX, LowerRightY);
  
  // compute value only in image (no pyramid)
  if (x1 < (double)LowerRightX-1.0 && 
      y1 < (double)LowerRightY-1.0 &&
      x1 > double(UpperLeftX) && y1 > double(UpperLeftY) ) {
    return StorageType((*this)[scale1]->
                       BilinearInterpolation(x1, y1, 
                                             (unsigned short int)channel));
  }
  return StorageType(0);
}


template <class StorageType>
StorageType PyramidImage<StorageType>::
GetTrilinearImageValue(const double& xsource,
                       const double& ysource,
                       const double& scale,
                       int channel) const {
  BIASASSERT(scale>=0.0);
  // compute a weight for the smaller and a weight for the larger pyramid level
  // then compute the colors at these levels and average them according
  // to their weights
  
  double value1 =  0, value2 = 0;
  double scale_error = scale - int(scale);
  
  int scale1 = int(scale);
  if (scale1>=(int)this->size()-1) {
    // we might get aliasing, pyramid is too small !!!
    scale1 = this->size()-1;
    scale_error = 0.0;
  } else if (scale1<0) {
    // use max resolution image, dont need to upsample in pyramid
    scale1 = 0;
    scale_error = 0.0;
  }

  // compute coordinates in smaller image
  double x1 = xsource, y1 = ysource;
  const double offset = GetPositionOffset();
  for (int s=0; s<scale1; s++) {
    // -0.5 necessary because of offset in 2^n downsampling
    x1 = (x1 - offset) / _RescaleFactor; 
    y1 = (y1 - offset) / _RescaleFactor;
  }
  // if scale 1 fits perfectly
  // we dont need scale2, set it to the same value to avoid access violations
  const int scale2 = (scale_error==0.0)? scale1 : scale1 + 1;
  const double x2  = (scale_error==0.0)? x1     : (x1 - offset)/_RescaleFactor; 
  const double y2  = (scale_error==0.0)? y1     : (y1 - offset)/_RescaleFactor;
  

  double weight1 = 1.0 - scale_error, weight2 = scale_error;

  unsigned int UpperLeftX, UpperLeftY, LowerRightX, LowerRightY;
  (*this)[scale1]->GetROICorners(UpperLeftX, UpperLeftY, 
                                 LowerRightX, LowerRightY);
  
  // compute value only in image (no pyramid)
  if (x1 < (double)LowerRightX-1.0 && 
      y1 < (double)LowerRightY-1.0 &&
      x1 > double(UpperLeftX) && y1 > double(UpperLeftY) ) {
    // we are near the border, bilinear
    value1 = 
      (*this)[scale1]->BilinearInterpolation(x1, y1, 
                                             (unsigned short int)channel);
  } else {
    weight1 = 0;
  } 

  // compute second scale only if necessary
  if (weight2>0.0) {
    (*this)[scale2]->GetROICorners(UpperLeftX, UpperLeftY, 
                                   LowerRightX, LowerRightY);
  
    // compute value only in image (no pyramid)
    if (x2 < (double)LowerRightX-1.0 && 
        y2 < (double)LowerRightY-1.0 &&
        x2 > double(UpperLeftX) && y2 > double(UpperLeftY) ) {
      // we are near the border, bilinear
      value2 = 
        (*this)[scale2]->BilinearInterpolation(x2, y2, 
                                               (unsigned short int)channel);
    } else {
      weight2 = 0;
    }
  }
  // check if we computed any value at all
  const double weightsum = weight1 + weight2;
  if ((weightsum)<=DBL_EPSILON) return StorageType(0);

  // this is the tri in trilinear:
  return StorageType((weight1 * value1 + weight2 * value2) / weightsum);

}



//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////

template class PyramidImage<unsigned char>;
template class PyramidImage<float>;

#ifdef BUILD_IMAGE_INT
template class PyramidImage<int>;
#endif
#ifdef BUILD_IMAGE_CHAR
template class PyramidImage<char>;
#endif
#ifdef BUILD_IMAGE_SHORT
template class PyramidImage<short>;
#endif
#ifdef BUILD_IMAGE_USHORT
template class PyramidImage<unsigned short>;
#endif
#ifdef BUILD_IMAGE_UINT
template class PyramidImage<unsigned int>;
#endif
#ifdef BUILD_IMAGE_DOUBLE
template class PyramidImage<double>;
#endif
