/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __TextureTransformHomography__hh__
#define __TextureTransformHomography__hh__

#include <Base/Common/BIASpragmaStart.hh>
#include <Image/TextureTransform.hh>

// should be smaller than 1 to account for small changes non-translational part:
#define PROJECTIVE_SENSITIVITY 1e-4
#define AFFINE_SENSITIVITY 1e-2

namespace BIAS {

  /** @class TextureTransformHomography
   *  @brief analytic properties of homography transformation of an image
   *
   *  See BIAS::TextureTransform for explanation of methods.
   *  @author koeser 01/2008 */
  class BIASImage_EXPORT TextureTransformHomography: public TextureTransform {
  public:
    TextureTransformHomography():H_(MatrixIdentity), Hinv_(MatrixIdentity) {
      P_.newsize(8);
      P_[0] = P_[1] = P_[2] = P_[3] = P_[4] = P_[5] = P_[6] = P_[7] = 0.0;
    }

    virtual ~TextureTransformHomography(){};

    int MapForward(const HomgPoint2D& src, HomgPoint2D& sink) const;

    int MapBackward(const HomgPoint2D& sink, HomgPoint2D& src) const; 

    int TextureJacobianForward(const HomgPoint2D& src, 
                               Matrix2x2<double>& Jac) const {
      H_.GetJacobian(src, Jac);
      return 0;
    }

    int TextureJacobianBackward(const HomgPoint2D& sink, 
                                Matrix2x2<double>& Jac) const {
      Hinv_.GetJacobian(sink, Jac);
      return 0;
    }

    virtual bool TextureJacobianIsConstant() const { return false; };

    int ParameterJacobianForward(Matrix<double>& Jac, const HomgPoint2D& src);
    
    int ParameterJacobianBackward(Matrix<double>& Jac, const HomgPoint2D& src);

    bool ParameterJacobianIsConstant() const { return false;};

    int ParameterInversionJacobian(Matrix<double>& Jac) const;

    /** @brief h11-1, h12, h13, h21, h22-1, h23, h31, h32 (h33==1 assumed) */
    void SetParameters(const Vector<double>& p);
    void ComposeWithInverseDeltaP(const Vector<double>& deltaP);
    void ComposeWithDeltaP(const Vector<double>& deltaP);

    virtual TextureTransformHomography* Clone() const {
      return new TextureTransformHomography(*this);
    }

    /** @brief given H, compute a parameter vector */
    Vector<double> ExtractParameters(const BIAS::HMatrix& H) const;

    /** @brief given parameters, compute H */
    BIAS::HMatrix ConstructHFromParameters(const Vector<double>&p) const;
  
  protected:
    HMatrix H_, Hinv_;
    
  }; 

 

} // end namespace BIAS

#endif
