/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "UndistortionMapping.hh"
#include <Geometry/RMatrix.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Geometry/HomgPoint2D.hh>


using namespace BIAS;
using namespace std;

template <class InputStorageType, class OutputStorageType>
UndistortionMapping<InputStorageType, OutputStorageType>::
~UndistortionMapping() 
{}


template <class InputStorageType, class OutputStorageType>
UndistortionMapping<InputStorageType, OutputStorageType>::
UndistortionMapping()
: log2inv_(1.0 / log(2.0)) 
{
  mapping_ = UNDEF;
}


// template <class InputStorageType, class OutputStorageType>
// UndistortionMapping<InputStorageType, OutputStorageType>::
// UndistortionMapping(const UndistortionMapping & src)
// : log2inv_(1.0 / log(2.0)) 
//   // : log2inv_(1.0) 
// {
//   (*this) = src;
// }


template <class InputStorageType, class OutputStorageType>
int UndistortionMapping<InputStorageType, OutputStorageType>::
CalculateSphericalCoord_(const HomgPoint2D& sink,
                         HomgPoint2D& source) const {
  double xSrc, ySrc;
  xSrc=ySrc=-1;
  HomgPoint2D hp;
  if (su_.ProjectCoordBack(destProjType_, destK_, sink, hp)<0) {
    return -1;
  } else {
    double cx, cy;
    su_.GetSphereCenter(cx, cy);
    if (destCenterX_!=cx || destCenterY_!=cy) {
      double t, p;
      if (su_.TransfCoordImage2Sphere (hp[0], hp[1], t, p) < 0) {
        return -1;
      } else {
        if (su_.TransfCoordRotateSphere2Image(t,p,destCenterX_,destCenterY_,
                                              xSrc, ySrc) < 0) {
          return -1;
        }
      }
    } else {
      xSrc=hp[0];
      ySrc=hp[1];
    }
    
  }
  source[0]=xSrc;
  source[1]=ySrc;
  source[2]=1;
  return 0;
}

template <class InputStorageType, class OutputStorageType>
int UndistortionMapping<InputStorageType, OutputStorageType>::
CalculateToSphereMapping_(const HomgPoint2D& sink,
                         HomgPoint2D& source) const {

  HomgPoint2D hp;
  double cx, cy;
  su_.GetSphereCenter(cx, cy);

  if (destCenterX_!=cx || destCenterY_!=cy) {
    double t, p;
    if (su_.TransfCoordImage2Sphere(sink[0], sink[1], t, p) < 0) {
      return -1;
    } else {
      if (su_.TransfCoordRotateSphere2Image(t,p,destCenterX_,destCenterY_,
                                            hp[0], hp[1]) < 0) {
        return -1;
      }
      hp[2]=1;
    }
  } else {
    hp = sink;
  }
  if (su_.ProjectCoord(destProjType_, destK_, hp, source)<0) {
    return -1;
  } 

  return 0;
}

template <class InputStorageType, class OutputStorageType>
void UndistortionMapping<InputStorageType, OutputStorageType>::
CalculatePerspDistorion_(const HomgPoint2D& sink,
                         HomgPoint2D& source) const {
  source = sink;
  if (pu_.GetAspectratio()<1-UNDIST_EPS) {
    source[0] = ((double)(source[0])*pu_.GetAspectratio());
  }
  if (pu_.GetAspectratio()>1+UNDIST_EPS) {
    source[1] = ((double)(source[1])/pu_.GetAspectratio());
  }
  pu_.Distort(source);
    
}

template <class InputStorageType, class OutputStorageType>
void UndistortionMapping<InputStorageType, OutputStorageType>::
CalculatePerspUndistorion_(const HomgPoint2D& sink,
                           HomgPoint2D& source) const {

  source = sink;
  pu_.Undistort(source);
  if (pu_.GetAspectratio()<1-UNDIST_EPS) {
    source[0] = ((double)(source[0])/pu_.GetAspectratio());
  }
  if (pu_.GetAspectratio()>1+UNDIST_EPS) {
    source[1] = ((double)(source[1])*pu_.GetAspectratio());
  }

}

template <class InputStorageType, class OutputStorageType>
int UndistortionMapping<InputStorageType, OutputStorageType>::
GetSourceCoordinates_(const HomgPoint2D& sink,
                      HomgPoint2D& source) const {

  // assert that CameraParam is set
  BIASASSERT(mapping_!=UNDEF);
 
  if (mapping_==PERSPECTIVE_UNDIST) {
    // correcting perspective lens distortion
    CalculatePerspDistorion_(sink,source);
  } else if (mapping_==PERSPECTIVE_DIST) {
    // introducing perspective lens distortion
    CalculatePerspUndistorion_(sink,source);
  } else if (mapping_==SPHERICAL) {
    return CalculateSphericalCoord_(sink, source);
  } else if (mapping_==TO_SPHERE){
    return CalculateToSphereMapping_(sink, source);
  }
  return 0;
}


template <class InputStorageType, class OutputStorageType>
int UndistortionMapping<InputStorageType, OutputStorageType>::
ProjectSphereImage(const Image<InputStorageType>& src, 
                   Image<OutputStorageType>& dst,
                   const int& SizeX, const int& SizeY, 
                   const double& FovX, const double& FovY,
                   const double& CenterX, const double& CenterY, 
                   const int projectionType, const double& zoomFactor,
                   KMatrix& KMatDst)
{
  mapping_ = SPHERICAL;
  
  destProjType_ = projectionType;

  destCenterX_ = CenterX;
  destCenterY_ = CenterY;
  if (destCenterX_==0 || destCenterY_==0) {
    su_.GetSphereCenter(destCenterX_, destCenterY_);
  }

  double destSizeX, destSizeY;
  destSizeX = SizeX;
  destSizeY = SizeY;
      
  double fovXrad, fovYrad;//, dum;
  if (FovX!=0) fovXrad = FovX/360.0*M_PI;
  else fovXrad = 0;
  
  if (FovY!=0) fovYrad = FovY/360.0*M_PI;
  else fovYrad = 0;

  //DrawCircles(src, 20.0);
  
  // either result image size or FoV is fixed
  KMatDst.SetIdentity();
  if (fovXrad!=0 && destSizeX!=0) {
    KMatDst[0][0] = (destSizeX*zoomFactor)/(tan(fovXrad)*2.0);
  }
  if (fovYrad!=0 && destSizeY!=0) {
    KMatDst[1][1] = (destSizeY*zoomFactor)/(tan(fovYrad)*2.0);
  }
  if (KMatDst[0][0]==1) {
    KMatDst[0][0]=KMatDst[1][1];
  }
  if (KMatDst[1][1]==1) {
    KMatDst[1][1]=KMatDst[0][0];
  }
  KMatDst[0][2] = (destSizeX*zoomFactor)/2.0;
  KMatDst[1][2] = (destSizeY*zoomFactor)/2.0;

  destK_ = KMatDst;

  destSizeX *= zoomFactor;
  destSizeY *= zoomFactor;
  if (!dst.IsEmpty()) dst.Release();
  dst.Init((int)destSizeX, (int)destSizeY, src.GetChannelCount());
  dst.Clear();
  dst.SetColorModel(src.GetColorModel());

  return Map(src, dst);
}

template <class InputStorageType, class OutputStorageType>
int UndistortionMapping<InputStorageType, OutputStorageType>::
MapPerspToSphere(const Image<InputStorageType>& src, 
                 Image<OutputStorageType>& dst,
                 const KMatrix Ksrc,
                 const double& CenterX, const double& CenterY, 
                 const int projectionType)
{
  mapping_ = TO_SPHERE;
    
  destProjType_ = projectionType;

  destCenterX_ = CenterX;
  destCenterY_ = CenterY;
  if (destCenterX_==0 || destCenterY_==0) {
    su_.GetSphereCenter(destCenterX_, destCenterY_);
  }
  
  destK_ = Ksrc;

  if (dst.IsEmpty()) {
    dst.Init(su_.GetWidth(), su_.GetHeight(), src.GetChannelCount());
    dst.SetColorModel(src.GetColorModel());
  } else {
    if (su_.GetWidth()!=dst.GetWidth() || su_.GetHeight()!=dst.GetHeight() ||
        src.GetColorModel()!=dst.GetColorModel()) {
      BIASERR("Camera parameters and dst-image do not match, " 
              << "can't process mapping!!");
      return -1;
    }
  }
  return Map(src, dst);
  
}


template <class InputStorageType, class OutputStorageType>
int UndistortionMapping<InputStorageType, OutputStorageType>::
UndistortPerspImage(const Image<InputStorageType>& src, 
                    Image<OutputStorageType>& dst,
                    KMatrix& K)
{
  int ret = UndistortPerspImage(src, dst);
  if (pu_.GetAspectratio()<1-UNDIST_EPS) {
    K.SetFx(K.GetFx()*pu_.GetAspectratio());
    K.SetHx(K.GetHx()*pu_.GetAspectratio());
  } 
  if (pu_.GetAspectratio()>1+UNDIST_EPS) {
    K.SetFy(K.GetFy()/pu_.GetAspectratio());
    K.SetHy(K.GetHy()/pu_.GetAspectratio());
  } 
  return ret;
}


template <class InputStorageType, class OutputStorageType>
int UndistortionMapping<InputStorageType, OutputStorageType>::
UndistortPerspImage(const Image<InputStorageType>& src, 
                    Image<OutputStorageType>& dst)
{
  mapping_ = PERSPECTIVE_UNDIST;
 
  // setup dst-image
  if (!dst.IsEmpty()) dst.Release();
  if (pu_.GetAspectratio()<1-UNDIST_EPS) {
    dst.Init((int)((double)(src.GetWidth())/pu_.GetAspectratio()), 
             (int)(src.GetHeight()), 
             src.GetChannelCount());
  } else if (pu_.GetAspectratio()>1+UNDIST_EPS) {
    dst.Init((int)(src.GetWidth()), 
             (int)((double)(src.GetHeight())*pu_.GetAspectratio()),
             src.GetChannelCount());
  } else {
    dst.Init((int)(src.GetWidth()), (int)(src.GetHeight()), 
             src.GetChannelCount());
  }
  dst.SetColorModel(src.GetColorModel());

  return Map(src, dst);
}


template <class InputStorageType, class OutputStorageType>
int UndistortionMapping<InputStorageType, OutputStorageType>::
DistortPerspImage(const Image<InputStorageType>& src, 
                  Image<OutputStorageType>& dst,
                  KMatrix& K)
{
  int ret = DistortPerspImage(src, dst);
  if (pu_.GetAspectratio()<1-UNDIST_EPS) {
    K.SetFx(K.GetFx()/pu_.GetAspectratio());
    K.SetHx(K.GetHx()/pu_.GetAspectratio());
  } 
  if (pu_.GetAspectratio()>1+UNDIST_EPS) {
    K.SetFy(K.GetFy()*pu_.GetAspectratio());
    K.SetHy(K.GetHy()*pu_.GetAspectratio());
  } 
  return ret;
}

template <class InputStorageType, class OutputStorageType>
int UndistortionMapping<InputStorageType, OutputStorageType>::
DistortPerspImage(const Image<InputStorageType>& src, 
                  Image<OutputStorageType>& dst)
{
  mapping_ = PERSPECTIVE_DIST;

  // setup dst-image
  if (!dst.IsEmpty()) dst.Release();
  if (pu_.GetAspectratio()<1-UNDIST_EPS) {
    dst.Init((int)((double)(src.GetWidth())*pu_.GetAspectratio()), 
             (int)(src.GetHeight()), 
             src.GetChannelCount());
  } else if (pu_.GetAspectratio()>1+UNDIST_EPS) {
    dst.Init((int)(src.GetWidth()), 
             (int)((double)(src.GetHeight())/pu_.GetAspectratio()), 
             src.GetChannelCount());
  } else {
    dst.Init((int)(src.GetWidth()), (int)(src.GetHeight()), 
             src.GetChannelCount());
  }
  dst.SetColorModel(src.GetColorModel());

  return Map(src, dst);
}


// TODO FIXME win32 template instantiation problem (JW)
template class UndistortionMapping<unsigned char, unsigned char>;
template class UndistortionMapping<float, float>;
template class UndistortionMapping<unsigned char, float>;
template class UndistortionMapping<float, unsigned char>;

#if defined(BUILD_IMAGE_CHAR)
// TODO FIXME win32 template instantiation problem (JW)
template class UndistortionMapping<unsigned char, char>;
template class UndistortionMapping<char, char>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class UndistortionMapping<unsigned short,unsigned short>;
#endif

#if defined(BUILD_IMAGE_SHORT)
template class UndistortionMapping<short, short>;
#endif

#if defined(BUILD_IMAGE_SHORT)&&defined(BUILD_IMAGE_USHORT)
template class UndistortionMapping<unsigned short, short>;
#endif

#if defined(BUILD_IMAGE_INT)
template class UndistortionMapping<int,int>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class UndistortionMapping<unsigned short,float>;
#endif

#if defined(BUILD_IMAGE_USHORT) && defined(BUILD_IMAGE_INT)
template class UndistortionMapping<unsigned short,int>;
#endif

#if defined(BUILD_IMAGE_DOUBLE)
template class UndistortionMapping<double,double>;
#endif
