/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Matcher2D/ImageAlignment.hh>
#include <MathAlgo/SVD.hh>
#include <Base/Image/ImageIO.hh> // only for debug image output
#include <Base/Common/FileHandling.hh>

using namespace BIAS;
using namespace std;

BIAS::ImageAlignment::ImageAlignment():
  MaxIterations_(20), MaxUpdate_(0.01), Dampening_(0.0),
  AffineBrightnessInvariance_(false),
  MinGradient_(1.0),
  LineSearch_(0.3), T_(NULL), mean1_(0.0),
  mean2_(0.0), dev1_(1.0), dev2_(1.0),iterationDebug_(0) {

}


int BIAS::ImageAlignment::
Align(const PyramidImage<float>& I1, const PyramidImage<float>& I2, 
      Vector<double>& params, Matrix<double>& Cov, int PyramidLevel) 
{
  // make sure the user has set a warp model
  BIASASSERT(T_!=NULL);

  int status = 0;
  BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,
           "compute template and gradients from warped I1 with prediction "
           <<params);
  T_->SetParameters(params);



  Matrix<double> PIJ;
  T_->ParameterInversionJacobian(PIJ);
  const Matrix<double> CovInv(PIJ * Cov * PIJ.Transpose());


  // check whether we have to re-evaluate for each pixel:
  const bool TextureJacobianIsConstant = T_->TextureJacobianIsConstant();
  const bool ParameterJacobianIsConstant = T_->ParameterJacobianIsConstant();
  if (PyramidLevel<0) {
    BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,
             "auto pyramid, determine required scale for each pixel");
    Matrix<double> JacParams;
    Matrix2x2<double> JacTex;  
    for (unsigned int p=0; p<TemplatePositions_.size(); p++) {
      // which pixel in the source image ?
      const HomgPoint2D& x(TemplatePositions_[p]);
      // where does the prediction map it ?
      T_->MapForward(x, ImagePositions_[p]);
      // need new jacobian ?
      if (p==0 || !ParameterJacobianIsConstant)
        T_->ParameterJacobianForward(JacParams, x);

      BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,
               "propagate parameter uncertainty "<<Cov);
      // approximate isotropic 2d uncertainty of this pixel's position
      double scale = 
        sqrt(BIAS::Matrix<double>(JacParams*Cov*JacParams.Transpose()).Trace());
      BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL, " to position uncertainty "<<scale);
      // compute best pyramid level to access and save grey value
      double pyrLevel = 0.0;
      if (scale<=0.5) scale = 0.5;
      else if (scale>I1.size()-1.0) {
        scale = I1.size()-1.0;
        pyrLevel = log2(scale) + 1.0;
      } else {
        pyrLevel = log2(scale) + 1.0;
      }
      TemplateGreyValues_[p] = I1.GetTrilinearImageValue(x[0], x[1], pyrLevel);
      BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,"Using "<<x<<" at pyramid level "
               <<pyrLevel<<" with grey value "<<TemplateGreyValues_[p]);

      // approximate gradient at that level
      Vector2<double> grad(I1.GetTrilinearImageValue(x[0]+0.5*scale, x[1],
                                                     pyrLevel) -
                           I1.GetTrilinearImageValue(x[0]-0.5*scale, x[1],
                                                     pyrLevel), 
                           I1.GetTrilinearImageValue(x[0], x[1]+0.5*scale, 
                                                     pyrLevel) - 
                           I1.GetTrilinearImageValue(x[0], x[1]-0.5*scale, 
                                                     pyrLevel));
      // approximate gradient at that level
      BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,"Rotate and scale gradient "
               <<grad[0]<<";"<<grad[1]);
                
      // rotate gradient with texture transformation and divide secant length
      if (p==0 || !TextureJacobianIsConstant) 
        T_->TextureJacobianForward(x, JacTex);
      TemplateGradients_[p] = JacTex * (grad * 1.0 / scale);
      active_[p] = TemplateGradients_[p].NormL2()*scale>MinGradient_;
      BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,"GRAD="<<TemplateGradients_[p]
               <<(active_[p]?" active":" skipped"));
    }
  } else {
    // use user specified pyramid level
    const double scale = pow(2.0,PyramidLevel);
    BIASDOUT(D_IMAGEALIGNMENT_PROGRESS, "Using pyramid level "
             <<PyramidLevel<<" for grey values and gradients (scale="<<scale
             <<")");
    Matrix2x2<double> JacTex;
    for (unsigned int p=0; p<TemplatePositions_.size(); p++) {
      const HomgPoint2D& x(TemplatePositions_[p]);
      T_->MapForward(x, ImagePositions_[p]);
      TemplateGreyValues_[p] = I1.GetTrilinearImageValue(x[0], x[1], 
                                                         PyramidLevel);
      BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL, "Pixel grey value at "<<x<<" is "
               <<TemplateGreyValues_[p]);


      // hier kann noch kraeftig optimiert werden, selbe gewichte fuer alle !
      // trilinear ist unnoetig, benutze naechste ganze pyramidenstufe



      // approximate gradient at that level
      Vector2<double> grad(I1.GetTrilinearImageValue(x[0]+0.5*scale, x[1],
                                                     double(PyramidLevel)) -
                           I1.GetTrilinearImageValue(x[0]-0.5*scale, x[1],
                                                     double(PyramidLevel)), 
                           I1.GetTrilinearImageValue(x[0], x[1]+0.5*scale, 
                                                     double(PyramidLevel)) - 
                           I1.GetTrilinearImageValue(x[0], x[1]-0.5*scale, 
                                                     double(PyramidLevel)));

      // compute jacobian only at first time or when not constant
      if (p==0 || !TextureJacobianIsConstant) 
        T_->TextureJacobianForward(x, JacTex);
      BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL, 
               "rotate gradient "<<grad[0]<<";"<<grad[1]
               <<" with local texture warp ["<<JacTex[0][0]<<" "<<JacTex[0][1]
               <<"; "<<JacTex[1][0]<<" "<<JacTex[1][1]<<"]");
      TemplateGradients_[p] = JacTex * (grad * 1.0 / scale);
      double gradnorm =  TemplateGradients_[p].NormL2();
      active_[p] = (gradnorm*scale>MinGradient_);
      BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,"GRAD="<<TemplateGradients_[p]
               <<(active_[p]?" active":" skipped"));
    }
  }

  // -------------  account for appearance / illumination changes ? ---------
  if (AffineBrightnessInvariance_) {
    // compute mean and scatter of patch
    double mean = 0;
    for (unsigned int i=0; i<TemplateGreyValues_.size(); i++) {
      mean += TemplateGreyValues_[i];
    }
    mean1_ = mean / TemplateGreyValues_.size();

    double scatter = 0;
    for (unsigned int i=0; i<TemplateGreyValues_.size(); i++) {
      scatter += (TemplateGreyValues_[i]-mean1_) * 
        (TemplateGreyValues_[i]-mean1_);
    }
    scatter = scatter/(TemplateGreyValues_.size()-1);
    if (scatter>0.0001) dev1_ = sqrt(scatter);
    else dev1_ = 0.01;

    // now rescale grey values such that mean=0 and sigma=1 afterwards
    for (unsigned int i=0; i<TemplateGreyValues_.size(); i++) {
      TemplateGreyValues_[i] = (TemplateGreyValues_[i]-mean1_)/dev1_;
      TemplateGradients_[i] /= dev1_;
    }
    BIASDOUT(D_IMAGEALIGNMENT_PROGRESS, 
             "Affine normalization of gradients and grey values yielded mean="
             <<mean1_<<" and dev="<<dev1_);
  }

  BIASDOUT(D_IMAGEALIGNMENT_PROGRESS, "compute jacobians + hessian");
  // prepare compositional inverse update (start with 0)
  TextureTransform *pUpdateWarp = T_->Clone();
  
#ifdef BIAS_DEBUG
  if (GetDebugLevel() & D_IMAGEALIGNMENT_IMAGES) {
    int numberOfPixelsPerRow = 
      int(ceil(sqrt(double(TemplateGreyValues_.size()))));
    Image<float> thetemplate(numberOfPixelsPerRow,numberOfPixelsPerRow,1);
    Image<float> thegradients(numberOfPixelsPerRow,numberOfPixelsPerRow,3);
    thegradients.SetZero();
    float *pF = thetemplate.GetImageData();
    float *pG = thegradients.GetImageData();
    for (unsigned int i=0; i<TemplateGreyValues_.size(); i++) {
      *pF++ = TemplateGreyValues_[i];
      if (active_[i]) {
        // write active into green
        pG++;*pG++ = TemplateGradients_[i].NormL2();pG++;
      } else {
        // write low into into red
        *pG++ = TemplateGradients_[i].NormL2();pG++;pG++;
      }
    }
    ImageIO::Write(string("template-")+toString(iterationDebug_)
                   +string(".mip"), thetemplate);
    ImageIO::Write(string("gradients-")+toString(iterationDebug_)
                   +string(".mip"), thegradients);
  }
#endif

  Vector<double> parameterUpdate;
  T_->GetParameters(parameterUpdate);
  for (unsigned int i=0; i<parameterUpdate.Size(); i++) 
    parameterUpdate[i] = 0;
  pUpdateWarp->SetParameters(parameterUpdate);

  Matrix<double> Jac;
  Matrix<double> Hessian(parameterUpdate.Size(), parameterUpdate.Size(), 
                         MatrixZero);

  unsigned int numMeasurements = 0;
  for (unsigned int i=0; i<TemplateGreyValues_.size(); i++) {
    // need new jacobian ?
    if (i==0 || !ParameterJacobianIsConstant)
      pUpdateWarp->ParameterJacobianBackward(Jac, ImagePositions_[i]);
   
    // no gradient, skip!
    if (!active_[i]) continue;

    Jac.MultLeft(TemplateGradients_[i], SteepestDescentImages_[i]);
    BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,"Steepest descent image is "<<
             SteepestDescentImages_[i]);
    // add to hessian"
    Hessian += 
      SteepestDescentImages_[i].OuterProduct(SteepestDescentImages_[i]);
    numMeasurements++;
  }

  // dampening factor
  if (Dampening_>0.0) {
    for (int i=0; i<Hessian.num_cols(); i++) {
      Hessian[i][i] += Dampening_;
    }
  }

  BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,  "Finished Hessian, "
           <<Hessian<<", inverting ...");
  // the hessian must have full rank here, otherwise we have a bad feature:
  // "highdimensional aperture problem"
  SVD Hsvd(Hessian);

  Vector<double> S = Hsvd.GetS();
  if (Dampening_<0.0) {
    for (int i=0; i<Hessian.num_cols(); i++) {
      Hessian[i][i] -= S[0] * Dampening_;
    }
    Hsvd.Compute(Hessian);
  } else if (S[S.Size()-1]/S[0]<1e-16) {
    // cannot solve !
    BIASDOUT(D_IMAGEALIGNMENT_PROGRESS, "Not enough structure for "
             <<" uniquely solving for all params in H: "<<Hessian
             <<", singular values are "<<S);
    return -1;
  }
  Hessian = Hsvd.Invert();
  
  //---------------------       iterate         --------------------
  BIASDOUT(D_IMAGEALIGNMENT_PROGRESS, "Starting iterations");
  
  int iter = 1;
  bool converged = false;
  Vector<double> errorImage(TemplateGreyValues_.size());
  vector<double> greyValuesI2(TemplateGreyValues_.size());
  vector<double> pyramidLevelsI2(TemplateGreyValues_.size());
  Vector<double> steepestDescentImageSum(parameterUpdate.Size());

  // remember the best-so-far residual
  double lastresidual = DBL_MAX;
  double residual = 0;
  double pyrLevel = 0;

  // remember best parameters so far
  Vector<double> lastGoodParameters;
  pUpdateWarp->GetParameters(lastGoodParameters);
  while (!converged) {
    BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Iteration "<<iter);
    // warp I2 with compositional parameters and compute intensity difference
    HomgPoint2D curPos(0,0);    
    steepestDescentImageSum.SetZero();
#ifdef BIAS_DEBUG
    if (GetDebugLevel() & D_IMAGEALIGNMENT_PARAMETER) {
      pUpdateWarp->GetParameters(parameterUpdate);
      BIASDOUT(D_IMAGEALIGNMENT_PARAMETER,
               "Current parameter update onto prediction is "
               <<parameterUpdate);
    }
#endif
    // fetch predicted, raw, grey value
    if (PyramidLevel<0) {
      BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL, "Cov is "<<Cov);
      // use resolution from covariance
      Matrix<double> JacParams;
      for (unsigned int i=0; i<TemplateGreyValues_.size(); i++) {
        HomgPoint2D &x(ImagePositions_[i]);
        pUpdateWarp->MapBackward(x, curPos);
        BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,"Looking up Pixel at "<< curPos);
        // first iteration: need parameter jacobian fpr pyrLevel calculation:
        if (iter==1) {
          if ((i==0 || !ParameterJacobianIsConstant)) {
            pUpdateWarp->ParameterJacobianBackward(JacParams, x);
            BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL, 
                     " PJBW="<<JacParams);
            double scale = 
              sqrt(BIAS::Matrix<double>
                   (JacParams*CovInv*JacParams.Transpose()).Trace());
            BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL, 
                     " position uncertainty "<<scale);
            // compute best pyramid level to access and save grey value"
            pyrLevel = 0.0;
            if (scale<=0.5) scale = 0.5;
            else if (scale>I1.size()-1.0) {
              scale = I1.size()-1.0;
              pyrLevel = log2(scale) + 1.0;
            } else {
              pyrLevel = log2(scale) + 1.0;
            }
          }
          pyramidLevelsI2[i] = pyrLevel;
#ifdef BIAS_DEBUG
          // save debug image with weights
          if ((GetDebugLevel() & D_IMAGEALIGNMENT_IMAGES) && 
              (i+1==TemplateGreyValues_.size())) {
            int numberOfPixelsPerRow = 
              int(ceil(sqrt(double(pyramidLevelsI2.size()))));
            Image<float> 
              theweights(numberOfPixelsPerRow,
                         numberOfPixelsPerRow,1);
            theweights.SetZero();
            float *pF = theweights.GetImageData();
            for (unsigned int j=0; j<pyramidLevelsI2.size(); j++) {
              *pF++ = pyramidLevelsI2[j];
            }
            ImageIO::Write(string("levels-")+toString(iterationDebug_)
                           +string(".mip"), theweights);
          }
#endif
        }
        greyValuesI2[i] = I2.GetTrilinearImageValue(curPos[0], curPos[1], 
                                                    pyramidLevelsI2[i]);

        BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,"Using "<<curPos
                 <<" at pyramid level "<<pyramidLevelsI2[i]
                 <<" with grey value(2) "<<greyValuesI2[i]);
      }
    } else {
      // user specified pyramid level
      for (unsigned int i=0; i<TemplateGreyValues_.size(); i++) {
        pUpdateWarp->MapBackward(ImagePositions_[i], curPos);
        BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,"Looking up Pixel at "<< curPos);
        greyValuesI2[i] = 
          I2.GetImageValue(curPos[0], curPos[1], PyramidLevel);
        BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,"Grey value is "<< greyValuesI2[i]);
      }
    }

    if (AffineBrightnessInvariance_) {
      // compute mean+scatter of patch
      double mean = 0;
      for (unsigned int i=0; i<TemplateGreyValues_.size(); i++) {
        mean += greyValuesI2[i];
      }
      mean2_ = mean / TemplateGreyValues_.size();
      double scatter = 0;
      for (unsigned int i=0; i<TemplateGreyValues_.size(); i++) {
        scatter += (greyValuesI2[i] - mean2_)*(greyValuesI2[i] - mean2_);
      }
      scatter /= TemplateGreyValues_.size()-1;
      if (scatter>0.0001) dev2_ = sqrt(scatter);
      else dev2_ = 0.01;
     
      BIASDOUT(D_IMAGEALIGNMENT_PROGRESS, 
               "Affine normalization of grey values(2) yielded mean="
               <<mean2_<<" and dev="<<dev2_);

      // compute normalized grey value residuum
      for (unsigned int i=0; i<TemplateGreyValues_.size(); i++) {
        greyValuesI2[i] = (greyValuesI2[i]-mean2_)/dev2_;
      }
    }
#ifdef BIAS_DEBUG
    if (GetDebugLevel() & D_IMAGEALIGNMENT_IMAGES) {
      int numberOfPixelsPerRow = int(ceil(sqrt(double(greyValuesI2.size()))));
      Image<float> theimage(numberOfPixelsPerRow,numberOfPixelsPerRow,1);
      float *pF = theimage.GetImageData();
      for (unsigned int g=0; g<greyValuesI2.size(); g++) {
        *pF++ = greyValuesI2[g];
      }
      ImageIO::Write(string("image-")+toString(iterationDebug_)
                     +string("-")+toString(iter)+string(".mip"),
                     theimage);
    }
#endif

    // now weight steepestDescent images by residuum
    residual = 0;
    for (unsigned int i=0; i<TemplateGreyValues_.size(); i++) {
      // no good gradient ...
      if (!active_[i]) continue;

      errorImage[i] = greyValuesI2[i] - TemplateGreyValues_[i];
      residual += errorImage[i]*errorImage[i];
      BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,"Grey value difference is "<<
               errorImage[i]);
      
      steepestDescentImageSum += errorImage[i] * SteepestDescentImages_[i];
    }
    residual = sqrt(residual/numMeasurements);
    BIASDOUT(D_IMAGEALIGNMENT_PROGRESS, "RMS residual is "<<residual);
    if (residual>lastresidual) {
      BIASDOUT(D_IMAGEALIGNMENT_PROGRESS, "Error increase from  "
               <<lastresidual<<". Stopping.");
      if (LineSearch_>0.0) {
        BIASDOUT(D_IMAGEALIGNMENT_PROGRESS, "Step did not improve, trying a "
                 <<LineSearch_<<" smaller step.");
        pUpdateWarp->SetParameters(lastGoodParameters);
        // reduce update
        parameterUpdate *= LineSearch_;
      } else {
        BIASDOUT(D_IMAGEALIGNMENT_PROGRESS, "No further improvement.Stopping.");
        // error did not improve ...
        pUpdateWarp->SetParameters(lastGoodParameters);
        status = 2;
        break;
      }
    } else {
      pUpdateWarp->GetParameters(lastGoodParameters);
      lastresidual = residual;
      // solve for parameter update
      Hessian.Mult(steepestDescentImageSum, parameterUpdate);
    }
    BIASDOUT(D_IMAGEALIGNMENT_PARAMETER,
             "inv.compositional parameter update is "
             <<parameterUpdate);
    // concatenate parameters
    pUpdateWarp->ComposeWithInverseDeltaP(parameterUpdate);
    BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Update of "
             <<parameterUpdate.NormL2()<<" thresh="<<MaxUpdate_);
    if (parameterUpdate.NormL2()<MaxUpdate_) {
      converged = true;
      BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Convergence ! Update = "
               <<parameterUpdate.NormL2());
    } else if (++iter>MaxIterations_) {
      BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Too many iterations: "<<iter);
      status = 1;
      break;
    }
  }
  pUpdateWarp->GetParameters(parameterUpdate);
  BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,
           "Finished. Total parameter update onto prediction is "
           <<parameterUpdate);
  // merge both direction parameters
  T_->ComposeWithInverseDeltaP(parameterUpdate);
  // and output parameters
  T_->GetParameters(params);
  BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Final warp parameters are "<<params);

  // update covariance
  if (numMeasurements>params.Size()) {
    double refvar = residual*residual*double(numMeasurements) / 
      double(numMeasurements-params.size());
    // assume parameterUpdate is nearly zero
    Matrix<double> upJac;
    pUpdateWarp->ParameterInversionJacobian(upJac);
    // combine jacobians 
    SVD covtransformSVD(upJac*PIJ);
    PIJ = covtransformSVD.Invert();
    Hessian *= refvar;
    Cov = PIJ * Hessian * PIJ.Transpose();
  } else {
    // no redundancy, perfect linear fit ...
    Cov.SetIdentity();
    Cov *= (DBL_MAX/8.0);
  }
  
  BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Covariance changed to "<<Cov);
  return status;
}


void BIAS::ImageAlignment::
SetAlignmentPixelsRectangular(const LocalAffineFrame& LAF,
                              int numberOfPixelsPerRow)
{
  // automatic resolution ?
  if (numberOfPixelsPerRow<1) {
    // use feature area ("all" pixels at full resolution
    numberOfPixelsPerRow = int(rint(sqrt(4.0*LAF.GetA().det())))+1;
    BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"NumOfPixelsPerRow automatically "
             <<" computed as "<<numberOfPixelsPerRow<<" with A="<<LAF.GetA());
  } 

  double pixeldistance = 2.0/(numberOfPixelsPerRow-1);
  if (numberOfPixelsPerRow<2) pixeldistance = 1.0;
  
  int numberOfPixels = numberOfPixelsPerRow * numberOfPixelsPerRow;
  TemplateGreyValues_.resize(numberOfPixels);
  TemplateGradients_.resize(numberOfPixels);
  TemplatePositions_.resize(numberOfPixels);
  ImagePositions_.resize(numberOfPixels);
  SteepestDescentImages_.resize(numberOfPixels);
  active_.resize(numberOfPixels);
  ControlPositions_.resize(4);
  const double offset = -1.0;
  for (int y=0; y<numberOfPixelsPerRow; y++) {
    for (int x=0; x<numberOfPixelsPerRow; x++) {
      int index = y*numberOfPixelsPerRow+x;
      TemplatePositions_[index] = 
        LAF.LocalToGlobal(HomgPoint2D(x*pixeldistance+offset,
                                      y*pixeldistance+offset));
      BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,
               "Using Position "<<TemplatePositions_[y*numberOfPixelsPerRow+x]);
    }
  }
  ControlPositions_[0] = TemplatePositions_[0];
  ControlPositions_[1] = TemplatePositions_[numberOfPixels-1];
  ControlPositions_[2] = 
    TemplatePositions_[TemplatePositions_.size()-numberOfPixels-1];
  ControlPositions_[3] = 
    TemplatePositions_[TemplatePositions_.size()-1];

  BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Have "<<numberOfPixels
           <<" pixels with distance "<<pixeldistance<<" in local affine frame");

}


  

int BIAS::ImageAlignment::AutoAlign(const PyramidImage<float>& I1, 
                                    const PyramidImage<float>& I2, 
                                    Vector<double>& params, 
                                    Matrix<double>& Cov) {
  int returnvalue = 0;
  double oldLineSearch = LineSearch_;
  double lasttrace = 1e200;
  double trace = 1e100;
 
  double desiredAccuracy = GetMaxUpdate();
  const unsigned int numparams = params.size();
  BIASASSERT(Cov.num_rows()==int(numparams));
  BIASASSERT(Cov.num_cols()==int(numparams));
  iterationDebug_ = 0;
  Matrix<double> lastCov(Cov);
  try {
    while (trace<lasttrace) {
      lasttrace = trace*0.95; // want 5% improvement to continue
      // make symmetric and slightly decrease
      for (unsigned int i=0; i<numparams; i++) {
        for (unsigned int j=i; j<numparams; j++) {
          Cov[j][i] = Cov[i][j] = 
            (Cov[i][j] + Cov[j][i] + 0.5*lastCov[j][i])/4.0;
        }
      }
      BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Aligning with Cov="<<Cov);
      // we expect uncertainty in the range of sqrt(trace), if update is
      // much smaller (e.g. only 10%) declare convergence
      double newthresh = 0.1*sqrt(trace);
      if (newthresh<desiredAccuracy) newthresh = desiredAccuracy;
      SetUpdateThreshold(newthresh);
      Vector<double> lastP = params;
      lastCov = Cov;

      // now call align with new resolution
      returnvalue = Align(I1, I2, params, Cov, -1);
      if (returnvalue<0) {
        BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Align returned "<<returnvalue
                 <<". Stopping iteration.");
        throw int(1);
      }
      if (iterationDebug_++ > MaxIterations_) {
        BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Too many iterations: "
                 <<iterationDebug_);
        returnvalue = 1;
        throw int(2);
      }
      trace = Cov.Trace();


      // finish not before best layer is used for each pixel !
      // use "control pixels"

      /*
        lastP -= params;
        if (lastP.NormL2()<desiredAccuracy) {
        BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Covariance reached desired accuracy");
        break;
        }
      */
      if (CheckConvergence_(params, Cov, 0.1)==0) {
        BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,
                 "Covariance reached desired accuracy");
        break;
      }
    }
  }
  catch (...) {
    SetUpdateThreshold(desiredAccuracy);
    iterationDebug_ = 0;
    LineSearch_ = oldLineSearch;
    return returnvalue;
  }
  BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Covariance did not improve, trace="<<trace
           <<" while last time it was "<<lasttrace);
 
  SetUpdateThreshold(desiredAccuracy);
  LineSearch_ = oldLineSearch;
  BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,"Aligning with full resolution.");
  returnvalue = Align(I1, I2, params, Cov, 0);
  iterationDebug_ = 0;
  return returnvalue;
}

int BIAS::ImageAlignment::
CheckConvergence_(const Vector<double>& params, const Matrix<double>& Cov, 
                  const double& pixelAccuracy) 
{
  // make sure the user has set a warp model
  BIASASSERT(T_!=NULL);

  BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,
           "compute template and gradients from warped I1 with prediction "
           <<params);
  T_->SetParameters(params);

  // check whether we have to re-evaluate for each pixel:
  const bool ParameterJacobianIsConstant = T_->ParameterJacobianIsConstant();
  
  BIASDOUT(D_IMAGEALIGNMENT_PROGRESS,
           "auto pyramid, determine required scale for each pixel");
  Matrix<double> JacParams;
  for (unsigned int p=0; p<ControlPositions_.size(); p++) {
    // which pixel in the source image ?
    const HomgPoint2D& x(ControlPositions_[p]);
    // need new jacobian ?
    if (p==0 || !ParameterJacobianIsConstant)
      T_->ParameterJacobianForward(JacParams, x);

    BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL,
             "propagate parameter uncertainty "<<Cov);
    // approximate isotropic 2d uncertainty of this pixel's position
    double scale = 
      sqrt(BIAS::Matrix<double>(JacParams*Cov*JacParams.Transpose()).Trace());
    BIASDOUT(D_IMAGEALIGNMENT_PERPIXEL, " to position uncertainty "<<scale);
    
    if (scale>pixelAccuracy) return 1;
  }
  return 0;
} 
