#include "ShowCamWxFrame.hh"
#include "ShowCamWxThreads.hh"

using namespace BIAS;
using namespace std;

#define PARAM_FILE_NAME "ShowCamWx.param"

BEGIN_EVENT_TABLE(ShowCamWxFrame, wxFrame)
  EVT_MENU      (ID_Quit,               ShowCamWxFrame::OnQuit)
  EVT_MENU      (ID_Grab,               ShowCamWxFrame::OnGrab)
  EVT_MENU      (ID_GrabContinuous,     ShowCamWxFrame::OnGrabContinuous) 
  EVT_MENU      (ID_Show1,              ShowCamWxFrame::OnShowImage1) 
  EVT_MENU      (ID_Show2,              ShowCamWxFrame::OnShowImage2) 
  EVT_MENU      (ID_Show3,              ShowCamWxFrame::OnShowImage3) 
  EVT_MENU      (ID_Show4,              ShowCamWxFrame::OnShowImage4) 
  EVT_MENU      (ID_Show5,              ShowCamWxFrame::OnShowImage5) 
  EVT_MENU      (ID_Show6,              ShowCamWxFrame::OnShowImage6) 
  EVT_MENU      (ID_Show7,              ShowCamWxFrame::OnShowImage7) 
  EVT_MENU      (ID_Show8,              ShowCamWxFrame::OnShowImage8)
  EVT_MENU      (ID_Show9,              ShowCamWxFrame::OnShowImage9)   
  EVT_MENU      (ID_Controller,         ShowCamWxFrame::OnShowController)
  EVT_TIMER     (ID_Timer,              ShowCamWxFrame::OnTimer)
  
  EVT_KEY_DOWN  (ShowCamWxFrame::OnKey) 
END_EVENT_TABLE()

ShowCamWxFrame::ShowCamWxFrame(const wxString& title,
                               const wxPoint& pos, const wxSize& size)
  : wxFrame(NULL, -1, title, pos, size) ,
    Param (true)   // disable param desctructor warning
    
{
  // set default value for class variables
  InitCameras_ = false;
  InitParams_  = false;
  BSaveImages_ = false;
  EmptyDevice_ = false;
  FrameCounter_ = 0;
  GrabCounter_  = 0;
  LastWidth_    = 0;
  LastHeight_   = 0;
  NumCamSel_    = 0;
  Win32DCAMNumber_ = 0;
#ifdef BIAS_HAVE_PanTilt
  panTiltControl_=NULL;
  panTiltControlDialog_=NULL;
#endif //BIAS_HAVE_PanTilt

  // set timer delay to 5 ms
  Delay_ = 5;

  // init images for cameras
  for (int i = 0; i < NUMBER_CAMERAS; i++)
  {
    CamImg_.push_back(new Camera<unsigned char>);
    CamImg2D_.push_back(new Camera<unsigned char>);
    CamImgPMDDepth_.push_back(new Camera<float>);
    CamImgPMDAmp_.push_back(new Camera<float>);
    ColorModels_.push_back(ImageBase::CM_invalid);
    VideoControllers_.push_back(NULL);
    
#ifdef BIAS_HAVE_PMD_PMDZess       
    VideoControllersUSBPMD_.push_back(NULL);
#endif //BIAS_HAVE_PMD_PMDZess
#ifdef BIAS_HAVE_SWISSRANGER
    VideoControllersSwissRanger_.push_back(NULL);
#endif // BIAS_HAVE_SWISSRANGER

  }

  // create status bar and add status text
  CreateStatusBar(2);
  wxString text;
  text.Printf(wxT("size: %4dx%4d"),0, 0);
  SetStatusText(text, 0);
  
  // init image canvas
  IC_ = new ImageCanvas(this,GetStatusBar(), 0,-1,wxDefaultPosition,                          
                        wxSize(640,480));
  
  // add image canvas to main window
  wxBoxSizer *vbox = new wxBoxSizer(wxVERTICAL);
  vbox->Add(IC_,1,  wxEXPAND |wxALL |wxFIXED_MINSIZE , 5);
  vbox->SetMinSize(640,480);
  SetSizer(vbox);
  SetAutoLayout(true);
  Layout();
  
  // init and start timer
  Timer_.SetOwner(this, ID_Timer);  
  cout<<"Setting timer to "<<Delay_<<" ms"<<endl;
  Timer_.Start(Delay_,true);
  StopWatch_.Start();  
}

void ShowCamWxFrame::InitParams(int &argc, char *argv[])
{
  // initialize camera names for params
  vector<string> cameras;
  cameras.push_back("NONE");
  cameras.push_back("DCAM");
  cameras.push_back("PMD");
  cameras.push_back("PMDUSB");
  cameras.push_back("PMDSHM");  
  cameras.push_back("UEYE");
  cameras.push_back("SHM");
  cameras.push_back("NET");
  cameras.push_back("V4L");
  cameras.push_back("SR");

  // initialize color models for cameras
  vector<string> colorModels;
  colorModels.push_back("CM_Default");  
  colorModels.push_back("CM_Grey");        
  colorModels.push_back("CM_RGB");        
  colorModels.push_back("CM_BGR");        
  colorModels.push_back("CM_YUYV422");   
  colorModels.push_back("CM_UYVY422");   
  colorModels.push_back("CM_YUV420P");     
  colorModels.push_back("CM_YUV444");     
  colorModels.push_back("CM_YUV411");     
  colorModels.push_back("CM_HSV");        
  colorModels.push_back("CM_HSI_OBS");  
  colorModels.push_back("CM_DV");         
  colorModels.push_back("CM_RGBA");      
  colorModels.push_back("CM_GreyA");      
  colorModels.push_back("CM_Bayer_RGGB");  
  colorModels.push_back("CM_Bayer_GBRG");  
  colorModels.push_back("CM_Bayer_GRBG");  
  colorModels.push_back("CM_Bayer_BGGR");  
  colorModels.push_back("CM_HSL");
  colorModels.push_back("CM_hsL");     
  colorModels.push_back("CM_SymTensor2x2");
  colorModels.push_back("CM_BGRA");        
  colorModels.push_back("CM_RGBE");         

	

  for (int i = 0; i < NUMBER_CAMERAS; i++)
  {
    // add nine different cameras to params and set group name
    stringstream number;
    number << i + 1;
    ParamCameras_[i] = AddParamEnum("CameraType" + number.str(),
                                    "Type of camera", cameras, 0, NULL, 0, i);
    ParamColorModel_[i] = AddParamEnum("CameraColorModel" + number.str(),
                                       "Color Model for camera", colorModels,
                                       0, NULL, 0, i);     
    ParamFPS_[i]    = AddParamDouble("CameraFPS" + number.str(),
                                     "fps possibilities: 1.875 3.75 7.5 15.0 30.0 60.0 120.0 240.0",
                                     0.0, 0.0, 240.0, 0, i);
    ParamWidth_[i]  = AddParamInt("CameraWidth" + number.str(),
                                  "Width of camera", 0, 0, 10000, 0, i);
    ParamHeight_[i] = AddParamInt("CameraHeight" + number.str(),
                                  "Height of camera", 0, 0, 10000, 0, i);
    ParamDown_[i]   = AddParamInt("CameraDownsampling" + number.str(),
                                  "Downsampling of camera", 0, 0, 10, 0, i);
    ParamPort_[i]   = AddParamInt("CameraPort" + number.str(),
                                  "Port of NET camera", 0, 0, 100000, 0, i);
    ParamIP_[i]     = AddParamString("CameraIP" + number.str(),
                                     "IP of NET camera", "", 0, i);
    ParamBitFile_[i]= AddParamString("CameraBitFile" + number.str(),
                                     "Bit file for usb pmd cameras", "", 0, i);
    ParamDevice_[i] = AddParamString("CameraDevice" + number.str(),
                                     "Device/Interface name of camera",
                                     "", 0, i);    
    // every camera gets own group number
    SetGroupName(i, number.str() + ". Camera"); 
  }

  bUsePanTiltUnit_ = AddParamBool("UsePanTiltUnit","Use the Pan-Tilt-Unit?",false,'p');

  // check if param file exists, else create one
  ifstream ifs(PARAM_FILE_NAME);
  if (!ifs)  
    WriteParameter(PARAM_FILE_NAME);
  else
  {
    // read param from file
    ReadParameter(PARAM_FILE_NAME);
    ifs.close();
  }

  // update parameters from file with arguments from command line
  if (argc > 1)
    UpdateParameter(argc, argv, PARAM_FILE_NAME);

  // get color models for cameras
  ParseColorModels();
  
  InitParams_ = true;
}


void ShowCamWxFrame::
InitPanTiltUnit(){
  if(*bUsePanTiltUnit_){
#ifdef BIAS_HAVE_PanTilt
    panTiltControl_= new DPPanTiltControl();
    bool bRet = false;
#ifdef WIN32	  
    bRet = panTiltControl_->InitPanTiltUnit("COM1");
#else
    bRet = panTiltControl_->InitPanTiltUnit("/dev/ttyS0");
#endif
    panTiltControlDialog_ = new DPPanTiltControlDialog(this, 7651,
                                  wxT("Pan-Tilt-Unit Control"));
    panTiltControlDialog_->SetDPControl(panTiltControl_);
    panTiltControlDialog_->Show(true);
#endif //BIAS_HAVE_PanTilt
  }
}


void ShowCamWxFrame::InitCameras()
{
  if (!InitParams_)
  {
    BIASERR("Initialize params first");
    BIASABORT;
  }

  bool foundCamera = false; 
  for (int i = 0; i < NUMBER_CAMERAS; i++)
  {
    if (*ParamCameras_[i] != 0)
    {
      foundCamera = true;
    }
  }
  if (!foundCamera)
  {
    BIASERR("No Camera specified in param file!");
  }

  // create cameras according to param file
  for (int i = 0; i < NUMBER_CAMERAS; i++)
  {
    switch(*ParamCameras_[i])
    {
    case 0: // None camera
      Cameras_.push_back(NULL);
      break;
    case 1: // DCAM
      
#ifdef BIAS_HAVE_DCAM
      Cameras_.push_back(new VideoSource_DCAM);      
      Cameras_[i]->SetSize(1024,768,2);        
      Cameras_[i]->SetDebugLevel(D_DCAM);
#else
      BIASERR("DCAM Camera in param file, please enable support in BIAS");
      BIASABORT;
#endif
      
      break;
    case 2: // PMD
      
#ifdef BIAS_HAVE_PMD_PMDTec     
      Cameras_.push_back(new VideoSource_DcamPMD);         
#else
      BIASERR("PMD Camera in param file, please enable support in BIAS");
      BIASABORT;
#endif
      
      break;
    case 3: // PMDUSB
      
#ifdef BIAS_HAVE_PMD_PMDZess
      Cameras_.push_back(new VideoSource_usbPMD);  
      if (*ParamBitFile_[i] != "")
        dynamic_cast<VideoSource_usbPMD*>(Cameras_[i])->SetBitFile(*ParamBitFile_[i]);
      else
      {
        BIASERR("Specify bit file for usb pmd camera");
        BIASABORT;
      }
#else
      BIASERR("PMDUSB Camera in param file, please enable support in BIAS");
      BIASABORT;
#endif
      
      break;
    case 4: // PMDSHM
      
#ifdef BIAS_HAVE_TIFF 
      Cameras_.push_back(new VideoSource_ShmPMD);               
#else
      BIASERR("SHM PMD Camera in param file, please enable support in BIAS"); 
      BIASABORT;
#endif
      
      break;      
    case 5: // UEYE
      
#ifdef BIAS_HAVE_UEYE
      Cameras_.push_back(new VideoSource_uEye);
      dynamic_cast<VideoSource_uEye*>(Cameras_[i])->SetLeftTop(56,0);
      dynamic_cast<VideoSource_uEye*>(Cameras_[i])->SetShutter(0.01f);
      dynamic_cast<VideoSource_uEye*>(Cameras_[i])->SetGain(0);     
#else
      BIASERR("UEYE Camera in param file, please enable support in BIAS");
      BIASABORT;
#endif
      
      break;
    case 6: // SHM
      Cameras_.push_back(new VideoSource_Shm);  
      break;
    case 7: // NET 
      Cameras_.push_back(new VideoSource_Net);
      break;
    case 8: // V4L

      // TODO: test support for v4l devices
      BIASERR("V4L support not yet implemented!");
      BIASABORT;
      
#ifdef BIAS_HAVE_V4L      
      Cameras_.push_back(new VideoSource_V4L);
#else
      BIASERR("V4L Camera in param file, please enable support in BIAS");
      BIASABORT;
#endif
      
      break;      
	case 9: // SR
#ifdef BIAS_HAVE_SWISSRANGER
      Cameras_.push_back(new VideoSource_SwissRanger);
#else
    BIASERR("SwissRanger in param file, please enable support in BIAS");
    BIASABORT;
#endif
      break;
    default: // None
      Cameras_.push_back(NULL);      
      break;      
    }   
  }
    
  // set size of cameras if size and type are defined
  for (int i = 0; i < NUMBER_CAMERAS; i++)
  {
    if (ParamHeight_[i] != 0 && ParamWidth_[i] && Cameras_[i] != NULL)
    {
      Cameras_[i]->SetSize(*ParamWidth_[i], *ParamHeight_[i]); 
    }
    if (*ParamFPS_[i] != 0.0)
    {
      cout<<"Setting "<<i<<". camera fps to "<<*ParamFPS_[i]<<endl;
      Cameras_[i]->SetFPS((float)*ParamFPS_[i]);
    }
  }

  InitCameras_ = true;
}

void ShowCamWxFrame::OpenDevices()
{
  if (!InitCameras_)
  {
    BIASERR("Initialize params first");
    BIASABORT;
  }
  
  for (int i = 0; i < NUMBER_CAMERAS; i++)
  {
    switch(*ParamCameras_[i])
    {
      
#ifdef BIAS_HAVE_DCAM      
    case 1: // DCAM
      if (*ParamDevice_[i] != "")
      {        
        if (Cameras_[i]->OpenDevice(atoi((*ParamDevice_[i]).c_str())) < 0)
        { 
          BIASERR("Unable to find camera number " << i);
          BIASABORT;
        } 
        else
          cout<<"Camera number " << i << " successfully opened"<<endl;
      }
      else
      {  
#ifdef WIN32
		    // this does not work if not all cameras are dcam!
        if (Cameras_[i]->OpenDevice(Win32DCAMNumber_++) < 0)
        {   
          BIASERR("Camera number " << i << " could not be openend");
          BIASABORT;   
        }
#else
        // check if there are multiple cameras with no device name
        if (EmptyDevice_)
        {
          BIASERR("Found more than one camera with no device name!"
                  << "Specify name for camera " << i << "!");
          BIASABORT;
        }
        EmptyDevice_ = true;
        if (Cameras_[i]->OpenDevice() < 0)
        {
          BIASERR("Camera number " << i << " could not be openend");
          BIASABORT;          
        }
#endif // WIN32
      }   
      break;
#endif // BIAS_HAVE_DCAM
      
#ifdef BIAS_HAVE_PMD_PMDTec  
    case 2: // PMD
      ((VideoSource_DcamPMD*)Cameras_[i])->SetIntegrationTime(20000);
      if (Cameras_[i]->OpenDevice()<0)
      {
        BIASERR("Failed to initialize PMD-Camera number " << i);
        BIASABORT;
      }
#endif
#ifdef BIAS_HAVE_PMD_PMDZess 
    case 3: // PMDUSB
      if (Cameras_[i]->OpenDevice()<0)
      {
        BIASERR("Failed to initialize PMDUSB-Camera number " << i);
        BIASABORT;
      }    
      //dynamic_cast<VideoSource_usbPMD*>(Cameras_[i])->SetShutter(10.0f);
      //dynamic_cast<VideoSource_usbPMD*>(Cameras_[i])->SetShutter2D(10.0f);
      break;
#endif
#ifdef BIAS_HAVE_PMD_PMDZess
    case 4: // PMDSHM
      if ((Cameras_[i])->OpenDevice((*ParamDevice_[i]).c_str()) < 0)
      {
        BIASERR("Failed to open Shm PMD-Camera number " << i);
        BIASABORT;
      }
      break;
#endif // BIAS_HAVE_PMD_PMDZess

#ifdef BIAS_HAVE_UEYE      
    case 5: // UEYE
#endif // BIAS_HAVE_UEYE
      
    case 6: // SHM
      
#ifdef BIAS_HAVE_V4L            
    case 8: // V4L
#endif // BIAS_HAVE_V4L
      if (*ParamDevice_[i] != "")
      {    
        if (Cameras_[i]->OpenDevice((*ParamDevice_[i]).c_str()) < 0)
        {
          BIASERR("Failed to initialize Camera number " << i);
          BIASABORT;
        }
      }
      else
      {
        if (Cameras_[i]->OpenDevice() < 0)
        {
          BIASERR("Failed to initialize Camera number " << i);
          BIASABORT;
        }
      }
      break;

    case 7: // NET
      // open on default port specified by class
      if (*ParamPort_[i] == 0)
      {
        if (((VideoSource_Net*)Cameras_[i])->OpenDevice(*ParamIP_[i]) < 0)
        {
          BIASERR("Failed to NET Camera number " << i);
          BIASABORT;       
        }        
      }
      // open on defined port by param file
      else
      {
        if (((VideoSource_Net*)Cameras_[i])->OpenDevice(*ParamIP_[i],
                                                        *ParamPort_[i]) < 0)
        {
          BIASERR("Failed to NET Camera number " << i);
          BIASABORT;       
        }
      }
      break;

#ifdef BIAS_HAVE_SWISSRANGER
    case 9: // SwissRanger
      if (Cameras_[i]->OpenDevice()<0)
      {
        BIASERR("Failed to initialize SwissRanger-Camera number " << i);
        BIASABORT;
      }    
      break;
#endif

    default: // NONE
      break;
    }

    // init image for camera camera if opened
    if (Cameras_[i] != NULL)
    {
      // try to set color model from param file for camera
      if (ColorModels_[i] != ImageBase::CM_invalid)
        Cameras_[i]->SetColorModel(ColorModels_[i]);

      // initialize image for camera and grab
      Cameras_[i]->InitImage(*CamImg_[i]);

#ifdef BIAS_HAVE_PMD_PMDZess
      // init 2D image for usb pmd cameras
      if (dynamic_cast<VideoSource_usbPMD*>(Cameras_[i]) != NULL)
      {
       // dynamic_cast<VideoSource_usbPMD*>(Cameras_[i])->Init2DImage(*CamImg2D_[i]);  
       //CamImgCombined is 3 times image in horizontal way
       CamImgCombined_.Init((CamImg_[NumCamSel_]->GetWidth()+1)*3,
                             CamImg_[NumCamSel_]->GetHeight()+1);
      }
#endif // BIAS_HAVE_PMD_PMDZess

      Cameras_[i]->PreGrab();
      Cameras_[i]->SetDebugLevel(0);     
    }
  }  
}

void ShowCamWxFrame::CreateMenu()
{
  // create file menu
  wxMenu *menuFile = new wxMenu;
  menuFile->Append(ID_Grab, wxT("Grab &single image") ); 
  menuFile->AppendCheckItem(ID_GrabContinuous, wxT("Grab &continuously (Start/Stop)") );  
  menuFile->Append(ID_Quit, wxT("E&xit") );

  // create menu for showing different cameras
  wxMenu *menuShow = new wxMenu;
  menuShow->Append(ID_Show1, wxT("Camera 1") );
  menuShow->Append(ID_Show2, wxT("Camera 2") );
  menuShow->Append(ID_Show3, wxT("Camera 3") );
  menuShow->Append(ID_Show4, wxT("Camera 4") );
  menuShow->Append(ID_Show5, wxT("Camera 5") );
  menuShow->Append(ID_Show6, wxT("Camera 6") );
  menuShow->Append(ID_Show7, wxT("Camera 7") );
  menuShow->Append(ID_Show8, wxT("Camera 8") );
  menuShow->Append(ID_Show9, wxT("Camera 9") );
  menuShow->AppendSeparator();
  menuShow->Append(ID_Controller, wxT("Controller") );  
  
  // add menues to menu bar
  wxMenuBar *menuBar = new wxMenuBar;
  menuBar->Append(menuFile, wxT("&File"));
  menuBar->Append(menuShow, wxT("&Show"));
  SetMenuBar(menuBar);

  // deactivate uninitialized cameras
  for (int i = 0; i < NUMBER_CAMERAS; i++)
  {
    if (Cameras_[i] == NULL) 
    {
      menuShow->Enable(ID_Show1 + i, false);
    }
  }
  
  // fit all components to size of window
  Fit();
}

void ShowCamWxFrame::GrabImage()
{
  //create a dummy Projection to save Pan Tilt angles in Image headers
  BIAS::Projection projHeader;
 
  if(*bUsePanTiltUnit_){
    double pan=0.0, tilt=0.0;

#ifdef BIAS_HAVE_PanTilt
    panTiltControl_->GetCurrentPositionDegreesThreaded(pan, tilt);
    std::cout<<"Current Position:"<<pan<<","<<tilt<<std::endl;
#endif //BIAS_HAVE_PanTilt

    // rotation matrix resulting from panning
    BIAS::RMatrix pan_matrix = BIAS::RMatrix(BIAS::Vector3<double>(0,1,0), pan);
	  // rotation matrix resulting from tilting
	  BIAS::RMatrix tilt_matrix = BIAS::RMatrix(BIAS::Vector3<double>(1,0,0), tilt);

	  // new modelling of ptu pose:
	  // first, pan ptu
	  BIAS::Matrix4x4<double> pan_ptu = BIAS::Matrix4x4<double>(pan_matrix, BIAS::Vector3<double>(0,0,0));

	  // tilt ptu
	  BIAS::Matrix4x4<double> tilt_ptu = BIAS::Matrix4x4<double>(tilt_matrix, BIAS::Vector3<double>(0,0,0));

    // identity matrix
	  BIAS::Matrix3x3<double> id33 = BIAS::Matrix3x3<double>(1,0,0, 0,1,0, 0,0,1);
	  // translation from rotation joint to camera center
    BIAS::Vector3<double> translation(0.0, 0.0, 0.0);
	  BIAS::Matrix4x4<double> trans_to_camera_center = BIAS::Matrix4x4<double>(id33, translation);

	  // calculate complete ptu transformation matrix
    BIAS::Matrix4x4<double> ptu_result = tilt_ptu * pan_ptu * trans_to_camera_center;

	  // do some hackish access to the translation part of resulting matrix
	  BIAS::Vector4<double> translation_rotated4 = ptu_result.GetCol(3);
	  BIAS::Vector3<double> translation_rotated3;
	  double *translation_rotated4_ptr = translation_rotated4.GetData();
	  for (int i=0;i<3;i++) 
		  translation_rotated3.GetData()[i] = translation_rotated4_ptr[i];

	  // store transformation in dynamic data pool entry
	  Pose p;
	  Quaternion<double> panq;
	  pan_matrix.GetQuaternion(panq);
	  Quaternion<double> tiltq;
	  tilt_matrix.GetQuaternion(tiltq);
	  Quaternion<double> temp = panq * tiltq;
	  p.SetC(translation_rotated3);
    p.SetQ(temp);
    KMatrix K;
    projHeader.CreatePerspective(p,K,640,480);
    cout<<"Projection:"<<projHeader<<endl;
  }
  for (int i = 0; i < NUMBER_CAMERAS; i++)
  {      
    wxString text;
    text.Printf(wxT("ShowCamWx-CAM%02d-%05d.mip"), i, GrabCounter_);
    
    // check if camera is opened
    if(Cameras_[i] != NULL)
    {
      if(*bUsePanTiltUnit_){
        (*CamImg_[i]).SetProj(projHeader);
        (*CamImg_[i]).UpdateMetaData();
      }

#ifdef BIAS_HAVE_PMD_PMDTec            
      // if it is a pmd camera save tiff image to disk
      if (dynamic_cast<VideoSource_ShmPMD*>(Cameras_[i]) != NULL ||
          dynamic_cast<VideoSource_DcamPMD*>(Cameras_[i]) != NULL)
      {
        text.Printf(wxT("ShowCamWX-CAM%02d-%05d.tiff"), i, GrabCounter_);
        PMDImageMetaData meta;
        meta.width = 64;
        meta.height = 48;
        meta.channels = 1;
        meta.PrincipalX = 32;
        meta.PrincipalY = 24;
        meta.FocallengthX = 100;
        meta.AspectRatio = 1;
        meta.UndistCoeff[0] = 0;
        meta.UndistCoeff[1] = 0;
        meta.UndistCoeff[2] = 0;
        meta.UndistCoeff[3] = 0;  
        meta.CameraCenter[0] = 0;  
        meta.CameraCenter[1] = 0;  
        meta.CameraCenter[2] = 0;  
        meta.CameraRotation[0] = 0;  
        meta.CameraRotation[1] = 0;  
        meta.CameraRotation[2] = 0; 
        if(*bUsePanTiltUnit_){
          (*CamImgPMDDepth_[i]).SetProj(projHeader);
          (*CamImgPMDAmp_[i]).SetProj(projHeader);
          (*CamImgPMDAmp_[i]).UpdateMetaData();
          (*CamImgPMDDepth_[i]).UpdateMetaData();
        }
        PMDImageProc::Write(WxToAscii(text), *CamImg_[i], 
                            *CamImgPMDDepth_[i], *CamImgPMDAmp_[i], 
                            meta, meta);
        cout <<"Saved image: "<<WxToAscii(text)<<endl;               
      }
      else
      {
#endif // BIAS_HAVE_PMD_PMDTec     

      
#ifdef BIAS_HAVE_PMD_PMDZess
        // grab all images from usb pmd camera
        if (dynamic_cast<VideoSource_usbPMD*>(Cameras_[i]) != NULL)
        {
          if(*bUsePanTiltUnit_){
            (*CamImgPMDDepth_[i]).SetProj(projHeader);
            (*CamImgPMDAmp_[i]).SetProj(projHeader);
            (*CamImgPMDAmp_[i]).UpdateMetaData();
            (*CamImgPMDDepth_[i]).UpdateMetaData();
          }
          text.Printf(wxT("ShowCamWx-CAM%02d-%05d-grey.mip"), i, GrabCounter_);
          ImageIO::Write(WxToAscii(text), *CamImg_[i]);
          cout <<"Saved image: "<<WxToAscii(text)<<endl;                       
          text.Printf(wxT("ShowCamWx-CAM%02d-%05d-depth.mip"), i, GrabCounter_);
          ImageIO::Write(WxToAscii(text), *CamImgPMDDepth_[i]);
          text.Printf(wxT("ShowCamWx-CAM%02d-%05d-amp.mip"), i, GrabCounter_);
          ImageIO::Write(WxToAscii(text), *CamImgPMDAmp_[i]);
          //text.Printf(wxT("ShowCamWx-CAM%02d-%05d-2D.mip"), i, GrabCounter_);
          //ImageIO::Write(WxToAscii(text), *CamImg2D_[i]);
        }
        else
        {
#endif // BIAS_HAVE_PMD_PMDZess

#ifdef BIAS_HAVE_SWISSRANGER
        // grab all images from usb pmd camera
        if (dynamic_cast<VideoSource_SwissRanger*>(Cameras_[i]) != NULL)
        {
          if(*bUsePanTiltUnit_){
            (*CamImgPMDDepth_[i]).SetProj(projHeader);
            (*CamImgPMDAmp_[i]).SetProj(projHeader);
            (*CamImgPMDAmp_[i]).UpdateMetaData();
            (*CamImgPMDDepth_[i]).UpdateMetaData();
          }
          text.Printf(wxT("ShowCamWx-CAM%02d-%05d-grey.mip"), i, GrabCounter_);
          ImageIO::Write(WxToAscii(text), *CamImg_[i]);
          cout <<"Saved image: "<<WxToAscii(text)<<endl;                       
          text.Printf(wxT("ShowCamWx-CAM%02d-%05d-depth.mip"), i, GrabCounter_);
          ImageIO::Write(WxToAscii(text), *CamImgPMDDepth_[i]);
          text.Printf(wxT("ShowCamWx-CAM%02d-%05d-amp.mip"), i, GrabCounter_);
          ImageIO::Write(WxToAscii(text), *CamImgPMDAmp_[i]);
        }
        else
        {
#endif // BIAS_HAVE_SWISSRANGER

          // try to create thread which saves image to disk
          ImageSaveThread *imsave;
          imsave = new ImageSaveThread(text, *CamImg_[i], *ParamDown_[i]);
          if ( imsave->Create() != wxTHREAD_NO_ERROR )
          {
            BIASERR("Cannot create ImageSaveThread, saving unthreaded");
            ImageIO::Write(WxToAscii(text), *CamImg_[i]);
          }
          else
          {
            imsave->Run();
          }
          cout <<"Saved image: "<<WxToAscii(text)<<endl;   

#ifdef BIAS_HAVE_SWISSRANGER
        }
#endif // BIAS_HAVE_SWISSRANGER

#ifdef BIAS_HAVE_PMD_PMDZess
        }
#endif // BIAS_HAVE_PMD_PMDZess

#ifdef BIAS_HAVE_PMD_PMDTec
      }
#endif // BIAS_HAVE_PMDU
        
    }
    
  }//end for all cameras
  GrabCounter_++;
}

void ShowCamWxFrame::ParseColorModels()
{
  for (int i = 0; i < NUMBER_CAMERAS; i++)
  {
    switch(*ParamColorModel_[i])
    {
    case 0:  ColorModels_[i] = ImageBase::CM_invalid; break;
    case 1:  ColorModels_[i] = ImageBase::CM_Grey; break;
    case 2:  ColorModels_[i] = ImageBase::CM_RGB; break;
    case 3:  ColorModels_[i] = ImageBase::CM_BGR; break;
    case 4:  ColorModels_[i] = ImageBase::CM_YUYV422; break;
    case 5:  ColorModels_[i] = ImageBase::CM_UYVY422; break;
    case 6:  ColorModels_[i] = ImageBase::CM_YUV420P; break;
    case 7:  ColorModels_[i] = ImageBase::CM_YUV444; break;
    case 8:  ColorModels_[i] = ImageBase::CM_YUV411; break;
    case 9:  ColorModels_[i] = ImageBase::CM_HSV; break;
    case 10: ColorModels_[i] = ImageBase::CM_HSI_OBS; break;
    case 11: ColorModels_[i] = ImageBase::CM_DV; break;
    case 12: ColorModels_[i] = ImageBase::CM_RGBA; break;
    case 13: ColorModels_[i] = ImageBase::CM_GreyA; break;
    case 14: ColorModels_[i] = ImageBase::CM_Bayer_RGGB; break;
    case 15: ColorModels_[i] = ImageBase::CM_Bayer_GBRG; break;
    case 16: ColorModels_[i] = ImageBase::CM_Bayer_GRBG; break;
    case 17: ColorModels_[i] = ImageBase::CM_Bayer_BGGR; break;
    case 18: ColorModels_[i] = ImageBase::CM_HSL; break;
    case 19: ColorModels_[i] = ImageBase::CM_hsL; break;
    case 20: ColorModels_[i] = ImageBase::CM_SymTensor2x2; break;
    case 21: ColorModels_[i] = ImageBase::CM_BGRA; break;
    case 22: ColorModels_[i] = ImageBase::CM_RGBE; break;
    default: ColorModels_[i] = ImageBase::CM_invalid; break;
    }
  }
}

void ShowCamWxFrame::InitCombinedImage()
{
  // initialize combined image from pmd usb camera images
  CamImgCombined_.SetROICorners(0, 0, 
                                CamImg_[NumCamSel_]->GetWidth(),
                                CamImg_[NumCamSel_]->GetHeight());
  CamImgCombined_.Paste2ROI(*CamImg_[NumCamSel_]);
  CamImgCombined_.SetROICorners(CamImg_[NumCamSel_]->GetWidth(),
                                0,
                                CamImg_[NumCamSel_]->GetWidth() * 2,
                                CamImg_[NumCamSel_]->GetHeight());


  // add Amp image from pmd usb camera
  BIAS::Camera<float> tmpImg;
  BIAS::Camera<unsigned char> tmpImgUC;  
  tmpImg = *CamImgPMDAmp_[NumCamSel_];

  if (InitUSBPMD_)
    tmpImg.ScaleShiftBetween(0,255);

  BIAS::ImageConvert::ConvertST(tmpImg, tmpImgUC, ImageBase::ST_unsignedchar);
  CamImgCombined_.Paste2ROI(tmpImgUC);
  CamImgCombined_.SetROICorners(CamImg_[NumCamSel_]->GetWidth() * 2,
                                0,
                                CamImg_[NumCamSel_]->GetWidth() * 3,
                                CamImg_[NumCamSel_]->GetHeight());

  // add Depth image from pmd usb camera
  tmpImg = *CamImgPMDDepth_[NumCamSel_];

  if (InitUSBPMD_)
    tmpImg.ScaleShiftBetween(0,255);

  BIAS::ImageConvert::ConvertST(tmpImg, tmpImgUC, ImageBase::ST_unsignedchar);
  CamImgCombined_.Paste2ROI(tmpImgUC);
}

void ShowCamWxFrame::OnQuit(wxCommandEvent& event)
{
  // stop timer
  Timer_.Stop();

  // shut down all opened cameras
  for (int i = 0; i < NUMBER_CAMERAS; i++)
  {
    if (Cameras_[i] != NULL)
    {
      Cameras_[i]->PostGrab();
      Cameras_[i]->CloseDevice();
    }
  }

#ifdef BIAS_HAVE_PanTilt
  if(*bUsePanTiltUnit_){
    if(panTiltControlDialog_->IsShown())
      panTiltControlDialog_->Hide();
    delete panTiltControlDialog_;

    if(panTiltControl_ != NULL){  
      panTiltControl_->SetPosition(0,0);
      panTiltControl_->ClosePanTiltUnit();
      delete panTiltControl_;
    }
  }
#endif //BIAS_HAVE_PanTilt

  Close(true);
}
 
void ShowCamWxFrame::OnGrab(wxCommandEvent& event)
{
  GrabImage();
}

void ShowCamWxFrame::OnGrabContinuous(wxCommandEvent& event)
{
  // reset counter and enable/disable continuous saving of images
  FrameCounter_ = 0;    
  BSaveImages_ = !BSaveImages_;
}

void ShowCamWxFrame::OnShowImage1(wxCommandEvent& event)
{
  NumCamSel_ = 0;  
  cout<<"Selected Camera: "<<NumCamSel_<<endl;
}

void ShowCamWxFrame::OnShowImage2(wxCommandEvent& event)
{
  NumCamSel_ = 1;  
  cout<<"Selected Camera: "<<NumCamSel_<<endl;  
}

void ShowCamWxFrame::OnShowImage3(wxCommandEvent& event)
{
  NumCamSel_ = 2;  
  cout<<"Selected Camera: "<<NumCamSel_<<endl;  
}

void ShowCamWxFrame::OnShowImage4(wxCommandEvent& event)
{
  NumCamSel_ = 3;  
  cout<<"Selected Camera: "<<NumCamSel_<<endl;  
}

void ShowCamWxFrame::OnShowImage5(wxCommandEvent& event)
{
  NumCamSel_ = 4;  
  cout<<"Selected Camera: "<<NumCamSel_<<endl;  
}

void ShowCamWxFrame::OnShowImage6(wxCommandEvent& event)
{
  NumCamSel_ = 5;  
  cout<<"Selected Camera: "<<NumCamSel_<<endl;  
}

void ShowCamWxFrame::OnShowImage7(wxCommandEvent& event)
{
  NumCamSel_ = 6;  
  cout<<"Selected Camera: "<<NumCamSel_<<endl;  
}

void ShowCamWxFrame::OnShowImage8(wxCommandEvent& event)
{
  NumCamSel_ = 7;  
  cout<<"Selected Camera: "<<NumCamSel_<<endl;  
}

void ShowCamWxFrame::OnShowImage9(wxCommandEvent& event)
{
  NumCamSel_ = 8;  
  cout<<"Selected Camera: "<<NumCamSel_<<endl;  
}

void ShowCamWxFrame::OnShowController(wxCommandEvent& event)
{
  // check if camera is initialized
  if (Cameras_[NumCamSel_] != NULL) {

#ifdef BIAS_HAVE_PMD_PMDZess
    // check if it is a usb pmd camera
    if (dynamic_cast<VideoSource_usbPMD*>(Cameras_[NumCamSel_]) != NULL) {
      // initialize controller if it was not before 
      if(VideoControllersUSBPMD_[NumCamSel_] == NULL) {
        VideoControllersUSBPMD_[NumCamSel_] = new VideoSource_Controller_usbPMD(
          dynamic_cast<VideoSource_usbPMD*>(Cameras_[NumCamSel_]),
          this,
          0,3000,1002,8000,1000,8000);
      }
      // show controller
      VideoControllersUSBPMD_[NumCamSel_]->Show(true);  
    } else 
#endif // BIAS_HAVE_PMD_PMDZess
#ifdef BIAS_HAVE_SWISSRANGER
    if (dynamic_cast<VideoSource_SwissRanger*>(Cameras_[NumCamSel_]) != NULL) {
      // initialize controller if it was not before 
      if(VideoControllersSwissRanger_[NumCamSel_] == NULL) {
        VideoControllersSwissRanger_[NumCamSel_] = new VideoSource_Controller_SwissRanger(
          dynamic_cast<VideoSource_SwissRanger*>(Cameras_[NumCamSel_]),
          this);
      }
      // show controller
      VideoControllersSwissRanger_[NumCamSel_]->Show(true);  
    } else 
#endif // BIAS_HAVE_SWISSRANGER

    // initialize controller if it was not before
    if (VideoControllers_[NumCamSel_] == NULL)        
    {       
      VideoControllers_[NumCamSel_] = new VideoSource_Controller(
        Cameras_[NumCamSel_],
        this,
        "VideoController",1.0,99.0);        
      // show controller
      VideoControllers_[NumCamSel_]->Show(true);
    } else {
      // show controller
      VideoControllers_[NumCamSel_]->Show(true);
    }

  }

  // camera was not initalized
  else
  {
    wxMessageDialog errorDialog(this,
      wxT("No correct camera was selected"),
      wxT("ShowCamWx - Error"),
      wxICON_ERROR|wxOK);
    errorDialog.ShowModal();    
  }      
}

void ShowCamWxFrame::OnTimer(wxTimerEvent& event)
{
  bool startGrabThread = false;
  std::vector<GrabSingleThread*> grabSingleThread;
  grabSingleThread.resize(9, NULL);  
  
  // update frames per second and restart timer and counters
  Timer_.Stop();
  StopWatch_.Stop();
  double fps = 1000000.0 / StopWatch_.GetRealTime();
  StopWatch_.Reset();
  StopWatch_.Start();
  FrameCounter_++;
  wxString text;
  text.Printf(wxT("%05d %2.1f fps "),FrameCounter_, fps);
  SetStatusText(text, 1);
  // get images from camera
  for (int i = 0; i < NUMBER_CAMERAS; i++)
  {    
    if (dynamic_cast<VideoSource_Shm*>(Cameras_[i]) != NULL)
    {
      startGrabThread = true;      
    }
    
#ifdef BIAS_HAVE_DCAM    
    else if (dynamic_cast<VideoSource_DCAM*>(Cameras_[i]) != NULL)
    {
      startGrabThread = true;
    }
#endif // BIAS_HAVE_DCAM

#ifdef BIAS_HAVE_PMD_PMDTec    
    else if (dynamic_cast<VideoSource_DcamPMD*>(Cameras_[i]) != NULL)
    {
      startGrabThread = true;      
      dynamic_cast<VideoSource_DcamPMD*>(Cameras_[i])->GrabSingleDepth(*CamImgPMDDepth_[i]);
      dynamic_cast<VideoSource_DcamPMD*>(Cameras_[i])->GrabSingleModCoeff(*CamImgPMDAmp_[i]);      
    }
    else if (dynamic_cast<VideoSource_ShmPMD*>(Cameras_[i]) != NULL)
    {
      startGrabThread = true;            
      dynamic_cast<VideoSource_ShmPMD*>(Cameras_[i])->GrabSingleDepth(*CamImgPMDDepth_[i]);
      dynamic_cast<VideoSource_ShmPMD*>(Cameras_[i])->GrabSingleModCoeff(*CamImgPMDAmp_[i]);   
    }
#endif // BIAS_HAVE_PMD_PMDTec

#ifdef BIAS_HAVE_PMD_PMDZess    
    else if (dynamic_cast<VideoSource_usbPMD*>(Cameras_[i]) != NULL)
    {
      startGrabThread = true;          
      dynamic_cast<VideoSource_usbPMD*>(Cameras_[i])->GrabSingleDepth(*CamImgPMDDepth_[i]);
      dynamic_cast<VideoSource_usbPMD*>(Cameras_[i])->GrabSingleModCoeff(*CamImgPMDAmp_[i]);         
      if (dynamic_cast<VideoSource_usbPMD*>(Cameras_[i])->GrabSingle2D(*CamImg2D_[i]) == -1)
        InitUSBPMD_ = false;
      else
        InitUSBPMD_ = true;
    }
#endif // BIAS_HAVE_PMD_PMDZess

#ifdef BIAS_HAVE_UEYE    
    else if (dynamic_cast<VideoSource_uEye*>(Cameras_[i]) != NULL)
    {
      startGrabThread = true;        
    }
#endif // BIAS_HAVE_UEYE

    else if (dynamic_cast<VideoSource_Net*>(Cameras_[i]) != NULL)
    {
      startGrabThread = true;        
    }

#ifdef BIAS_HAVE_V4L    
    else if (dynamic_cast<VideoSource_V4L*>(Cameras_[i]) != NULL)
    {
      BIASERR("V4L support not implemented!");
      BIASABORT;      
    }    
#endif // BIAS_HAVE_V4L

#ifdef BIAS_HAVE_SWISSRANGER
    else if (dynamic_cast<VideoSource_SwissRanger*>(Cameras_[i]) != NULL)
    {
      dynamic_cast<VideoSource_SwissRanger*>(Cameras_[i])->GrabSingleDepth(*CamImgPMDDepth_[i]);
      dynamic_cast<VideoSource_SwissRanger*>(Cameras_[i])->GrabSingleModCoeff(*CamImgPMDAmp_[i]);         
      startGrabThread = true;        
    }    
#endif // BIAS_HAVE_SWISSRANGER


    // check if an image should be grabbed
    if (startGrabThread == true && Cameras_[i] != NULL)
    {
//#ifdef WIN32
//      Cameras_[i]->GrabSingle(*CamImg_[i]);   
//#else 
      // try to create thread which grab single image from camera
      grabSingleThread[i] = new GrabSingleThread(Cameras_[i], CamImg_[i]);
 
      if (grabSingleThread[i]->Create() != wxTHREAD_NO_ERROR )
      {
        BIASERR("Cannot create GrabSingleThread, saving unthreaded");
        Cameras_[i]->GrabSingle(*CamImg_[i]);        
      }
      else
      {
        grabSingleThread[i]->Run();
      }
      startGrabThread = false;
//#endif // WIN32
    }    
  }

  // wait till all started threads are finished
  for (int j = 0; j < NUMBER_CAMERAS; j++)
  {
	  if (grabSingleThread[j] != NULL) 
#ifdef WIN32
      while(grabSingleThread[j]->IsRunning())
        usleep(1);
#else
      grabSingleThread[j]->Wait();
#endif // WIN32

		delete grabSingleThread[j];
  }
  
  // show camera image in main window if a camera was selected
  if (Cameras_[NumCamSel_] != NULL)
  {
    unsigned int w;
    unsigned int h;

#ifdef BIAS_HAVE_PMD_PMDZess    
      if (dynamic_cast<VideoSource_usbPMD*>(Cameras_[NumCamSel_]) != NULL)
      {
        // get width and height from combined image
        w = CamImgCombined_.GetWidth(); 
        h = CamImgCombined_.GetHeight();
      }
      else
      {
#endif // BIAS_HAVE_PMD_PMDZess

        // get width and height from image
        w = CamImg_[NumCamSel_]->GetWidth();
        h = CamImg_[NumCamSel_]->GetHeight();      
        if(w > 1024 || h >768){        
          w= 1024; h=768;       
        }
#ifdef BIAS_HAVE_PMD_PMDZess    
      }
#endif // BIAS_HAVE_PMD_PMDZess

    text.Printf(wxT("size: %4dx%4d"),w, h);
    SetStatusText(text, 0);

    // set new height or width if changed
    if (w != LastWidth_ || h != LastHeight_)
    {
      LastWidth_  = w;
      LastHeight_ = h;
      IC_->SetSize(w+50,h+80);
      SetMinSize(wxSize(w+50,h+80));
      Fit();
      cout <<"Refit"<<endl;  
    }

    if (CamImg_[NumCamSel_] != NULL)
    {

#ifdef BIAS_HAVE_PMD_PMDZess    
      if (dynamic_cast<VideoSource_usbPMD*>(Cameras_[NumCamSel_]) != NULL)
      {
        InitCombinedImage();
        ShowImage_ = CamImgCombined_;
      }
      else
      {
        // show grabbed image on main frame
        ShowImage_ = *CamImg_[NumCamSel_];
      }
#else // BIAS_HAVE_PMD_PMDZess

      ShowImage_ = *CamImg_[NumCamSel_];

#endif // BIAS_HAVE_PMD_PMDZess

      // downsampling of actual camera image if set by param file
      int downSample = *ParamDown_[NumCamSel_];
      if (downSample != 0)
      {
        Camera<unsigned char> rescaledImage;        
        Rescale<unsigned char, unsigned char> rescale;
        rescale.Downsample(ShowImage_, rescaledImage,
                           (int)(ShowImage_.GetWidth() / downSample),
                           (int)(ShowImage_.GetHeight() / downSample));
        ShowImage_ = rescaledImage;
      }

      unsigned width = ShowImage_.GetWidth();
      unsigned height = ShowImage_.GetHeight();
     
      if(width > 1024 || height >768){ 
        IC_->SetScaledImageSize ( 1024,  768);
        IC_->SetShowScaled(true);
      }
      else
        IC_->SetShowScaled(false);
      // show actual camera image on main frame
      IC_->Show(ShowImage_, "Camera Image");
    }
  }

  // grab image if continous saving is true
  if(BSaveImages_)
     GrabImage();
    
  Timer_.Start(Delay_,true);  
}

void ShowCamWxFrame::OnKey(wxKeyEvent& event)
{
  // no key events implemented yet
}
