#ifndef __ShowCamWxFrame_hh__
#define __ShowCamWxFrame_hh__

#include <Base/Common/BIASpragmaStart.hh>
#include <bias_config.h>

#include <Gui/StringConv.hh>
#include <Gui/ImageCanvas.hh>
#include <Gui/ImageCanvas.hh>
#include <Image/Camera.hh>
#include <Base/Image/ImageIO.hh>
#include <Gui/ParamGUI.hh>
#include <Utils/Param.hh>
#include <Base/Debug/TimeMeasure.hh>
#include <Filter/Rescale.hh>

#include <VideoSource/VideoSource_Base.hh>
#include <VideoSource/VideoSource_Net.hh>
#include <VideoSource/VideoSource_Shm.hh>

#ifdef BIAS_HAVE_PanTilt
#include <PanTilt/DPPanTiltControl.hh>
#include <PanTilt/DPPanTiltControlDialog.hh>
#endif

#ifdef BIAS_HAVE_TIFF
#ifdef BIAS_HAVE_XML2
#include <VideoSource/VideoSource_ShmPMD.hh>
#endif
#endif
#include <Gui/VideoSource_Controller.hh>

#ifdef BIAS_HAVE_V4L
#  include <VideoSource/VideoSource_V4L.hh>
#endif // BIAS_HAVE_V4L

#ifdef BIAS_HAVE_UEYE
#  include <VideoSource/VideoSource_uEye.hh>
#endif // BIAS_HAVE_UEYE

#ifdef BIAS_HAVE_DCAM
#  include <VideoSource/VideoSource_DCAM.hh>
#endif // BIAS_HAVE_DCAM

#ifdef BIAS_HAVE_PMD_PMDTec
#  include <Image/PMDImageProc.hh>
#  include <VideoSource/VideoSource_DcamPMD.hh>
#endif // BIAS_HAVE_PMD_PMDTec

#ifdef BIAS_HAVE_PMD_PMDZess
#  include <VideoSource/VideoSource_usbPMD.hh>
#  include <Gui/VideoSource_Controller_usbPMD.hh>
#endif // BIAS_HAVE_PMD_PMDZess

#ifdef BIAS_HAVE_SWISSRANGER
#  include <VideoSource/VideoSource_SwissRanger.hh>
#  include <Gui/VideoSource_Controller_SwissRanger.hh>
#endif

#include <wx/wx.h>
#include <wx/toolbar.h> 

/**
   @class  ShowCamWxFrame
   @brief  This class is the main frame of the ShowCamWx Application
   @author Christoffer Menk
*/

#define NUMBER_CAMERAS 9

///////////////////////////
// TODO: Add support for V4L (Video for Linux)
// 
/////////////

namespace BIAS {

  // event ids
  enum {
    ID_Quit = 42,
    ID_Timer,
    ID_Grab,
    ID_GrabContinuous,    
    ID_Key,
    ID_Show1, 
    ID_Show2,
    ID_Show3,
    ID_Show4,
    ID_Show5,
    ID_Show6,
    ID_Show7,
    ID_Show8,
    ID_Show9,    
    ID_Controller
  };  

  class ShowCamWxFrame : public wxFrame, public BIAS::Param
  {
  public:

    ShowCamWxFrame(const wxString& title, const wxPoint& pos,
                   const wxSize& size);
    virtual ~ShowCamWxFrame(){};

    /** Initializes all paramaters for param file
     */
    void InitParams(int &argc, char *argv[]);

    /** Initialize cameras with param file. Only call after InitParams().
     */
    void InitCameras();

     /** Initialize Pan Tilt Unit. Only call after InitParams().
     */
    void InitPanTiltUnit();

    /** Tries to open all specified camera devices. Only call after
        cameras are initialized.
     */
    void OpenDevices();

    /** Creates the menu of the main window.
     */
    void CreateMenu();

    /** Stores one image from all activated cameras
     */
    void GrabImage();
    
  protected:
  
    /** Parses the color models from param file to enum EColorModel
     */
    void ParseColorModels();

    /** Initializes combined image for selected camera (usb pmd)
    */
    void InitCombinedImage();

    // functions for event handling
    void OnQuit(wxCommandEvent& event);
    void OnGrab(wxCommandEvent& event);
    void OnGrabContinuous(wxCommandEvent& event); 
    void OnShowImage1(wxCommandEvent& event); 
    void OnShowImage2(wxCommandEvent& event); 
    void OnShowImage3(wxCommandEvent& event); 
    void OnShowImage4(wxCommandEvent& event); 
    void OnShowImage5(wxCommandEvent& event); 
    void OnShowImage6(wxCommandEvent& event); 
    void OnShowImage7(wxCommandEvent& event); 
    void OnShowImage8(wxCommandEvent& event);
    void OnShowImage9(wxCommandEvent& event);     
    void OnShowController(wxCommandEvent& event); 
    void OnTimer(wxTimerEvent& event);  
    void OnKey(wxKeyEvent& event);

    // variables for parameters
    int *ParamCameras_[NUMBER_CAMERAS];                 // camera model
    int *ParamWidth_[NUMBER_CAMERAS];                   // width of cameras
    int *ParamHeight_[NUMBER_CAMERAS];                  // height of camera
    int *ParamDown_[NUMBER_CAMERAS];                    // rescale factor
    int *ParamPort_[NUMBER_CAMERAS];                    // port of NET cameras
    std::string *ParamIP_[NUMBER_CAMERAS];              // ip of NET cameras
    std::string *ParamDevice_[NUMBER_CAMERAS];          // device name of camera
    std::string *ParamBitFile_[NUMBER_CAMERAS];         // bit file for pmd cameras
    int *ParamColorModel_[NUMBER_CAMERAS];              // color model of camera
    double *ParamFPS_[NUMBER_CAMERAS];                  // frames per second for camera
    
	// include the Pan-Tilt-Unit?
	bool *bUsePanTiltUnit_;

    // vector for all cameras read from param file
    std::vector<BIAS::VideoSource*> Cameras_;
    std::vector<BIAS::ImageBase::EColorModel> ColorModels_;
    std::vector<BIAS::VideoSource_Controller*> VideoControllers_;
    
#ifdef BIAS_HAVE_PMD_PMDZess
    std::vector<BIAS::VideoSource_Controller_usbPMD*> VideoControllersUSBPMD_;
#endif // BIAS_HAVE_PMD_PMDZess
#ifdef BIAS_HAVE_SWISSRANGER
    std::vector<BIAS::VideoSource_Controller_SwissRanger*> VideoControllersSwissRanger_;
#endif // BIAS_HAVE_SWISSRANGER

#ifdef BIAS_HAVE_PanTilt
    //Pan Tilt Control
    BIAS::DPPanTiltControl *panTiltControl_;
    BIAS::DPPanTiltControlDialog *panTiltControlDialog_;
#endif //BIAS_HAVE_PanTilt

    bool InitCameras_;
    bool InitUSBPMD_;
    bool InitParams_; 
    unsigned int FrameCounter_;
    unsigned int GrabCounter_;
    unsigned int LastWidth_, LastHeight_;    
    unsigned int Win32DCAMNumber_;
    int NumCamSel_;
    int Delay_;
    bool BSaveImages_;
    bool EmptyDevice_;

    // camera images
    std::vector<BIAS::Camera<unsigned char>*> CamImg_;
    std::vector<BIAS::Camera<float>*> CamImgPMDDepth_;
    std::vector<BIAS::Camera<float>*> CamImgPMDAmp_;    
    
    // images for usb pmd cameras
    std::vector<BIAS::Camera<unsigned char>*> CamImg2D_; 
    BIAS::Camera<unsigned char> CamImgCombined_;

    BIAS::Camera<unsigned char> ShowImage_;    

    BIAS::ImageCanvas* IC_;
    
    wxTimer Timer_;
    TimeMeasure StopWatch_; 
    
    DECLARE_EVENT_TABLE()           
  }; // end of class
    
} // end of namespace MIP

#include <Base/Common/BIASpragmaEnd.hh>

#endif // __ShowCamWxFrame_hh__
