/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Base/Image/ImageBase.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>
#include <Image/HomographyMapping.hh>
#include <stdio.h>


using namespace BIAS;
using namespace std;

void usage()
{
  cout << "Usage: biassetuid [options] files"<<endl;
  cout <<" Options: "<< endl
       <<" -r         : rescale output by 2"<<endl
       << endl;
  cout <<"at each pixel position average values of all images"<<endl<<endl;
}

int DetermineFiletype(const string &fn, ImageIO::TFileFormat &thetype)
{
  int EndOfPrefix;
  string suffix;
  
  EndOfPrefix = fn.rfind('.');
  
  suffix = fn.substr(EndOfPrefix+1,fn.length());
  if (suffix == "pnm") thetype =  ImageIO::FF_ppm;
  else if (suffix == "ppm") thetype =  ImageIO::FF_ppm;
  else if (suffix == "pgm") thetype =  ImageIO::FF_pgm;
  else if (suffix == "mip") thetype =  ImageIO::FF_mip;
  else return -1;

  return 0;
}


/** @class biassetuid
    @ingroup g_tools
    @brief load image given by the command line, assign a uuid and write back
    to disk
    @author koeser */
int main (int argc, char *argv[])
{
  bool rescale = false;
  int i = 1;
  if (argc>1) if (!strcmp("-r",argv[1])) {
    std::cout << "rescale"<<endl;
    rescale = true;
    i = 2;
  }
  if (i>=argc) { 
    usage();
    return 0; // for Dart testing 
  }
  
  bool ErrorFlag = false;
  int IOResult = 0;
  BIAS::Image<unsigned char> CurImage;
 
  BIAS::Image<float> averageImage;
  bool initialized = false;
  int countImages = 0;
  for (;i<argc;i++) {
    IOResult = ImageIO::Load(argv[i], CurImage);
    if (IOResult!=0) {
      ErrorFlag = true;
      BIASERR("Error reading "<< argv[i]);
      continue;
    }
    if (!initialized) {
      averageImage.Init(CurImage.GetWidth(), CurImage.GetHeight(), 
                        CurImage.GetChannelCount());
      averageImage.SetZero();
      initialized = true;
    }

    const unsigned int channels = CurImage.GetChannelCount();
    BIASASSERT(channels==1 || CurImage.IsInterleaved()); 
    BIASASSERT(averageImage.SamePixelAndChannelCount(CurImage));
    for (unsigned int y=0; y<CurImage.GetHeight(); y++) {

      for (unsigned int x=0; x<CurImage.GetWidth(); x++) {
        for (unsigned int c=0; c<channels; c++) {
          
          averageImage.GetImageDataArray()[y][x*channels+c] +=
            float(CurImage.GetImageDataArray()[y][x*channels+c]);
        }
      }
    }
    countImages++;
  }
  averageImage.ScaleShift(1.0/double(countImages),0);

  ImageIO::Write("average.mip", averageImage);


  Image<unsigned char> ucImg;
  HomographyMapping<float, unsigned char> converter;
  HMatrix H(MatrixIdentity);
  if (rescale) {
    H[0][0] = H[1][1] = 0.5;
    ucImg.Init(2*averageImage.GetWidth(), 2*averageImage.GetHeight(), 
               averageImage.GetChannelCount());
  } else {
    ucImg.Init(averageImage.GetWidth(), averageImage.GetHeight(), 
               averageImage.GetChannelCount());
  }
  converter.SetHomography(H);
  converter.Map(averageImage, ucImg, MapBicubic);
  
  ImageIO::ExportImage("average.pgm", ucImg);
  return 0;
    
}












