/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Utils/ThreeDOut.hh>
#include <Utils/TriangleMesh.hh>
#include <Base/Image/ImageIO.hh>
#include <Filter/Rescale.hh>
#include <Base/Common/FileHandling.hh>
#include <Image/Camera.hh>
#include <Utils/Param.hh>

#include <sstream>
#include <iostream>

using namespace std;
using namespace BIAS;


/** @class biastrianglemesh
@ingroup g_tools
@brief   generate triangle mesh from depth map
@author  mfranke */


#define TILE_WIDTH 129
#define TILE_HEIGHT 129


int main(int argc, char *argv[]) 
{
  Param params(true);

  /**** GENERAL PARAMS ****/
  int generalGroupID = params.GetFreeGroupID();

  bool* help = params.AddParamBool("help",
                                   "print out help", 
                                   false, 'h',
                                   generalGroupID);

  params.SetGroupName(generalGroupID, "general");

  
  /**** INPUT PARAMS ****/
  int inputParamGroupID = params.GetFreeGroupID();

  string* inputDepthMap = params.AddParamString("DepthMap",
                                                "specify depth map",
                                                "", 
                                                'D',
                                                inputParamGroupID);
  
  string* inputTexture = params.AddParamString("Texture",
                                               "specify texture, "
                                               "if not specified -> triangle mesh will be untextured",
                                               "", 
                                               'T',
                                               inputParamGroupID);

  string* inputProjection = params.AddParamString("Projection",
                                                  "specify projection, " 
                                                  "if not specified -> trying to parse meta data from (1.)texture or (2.)depth map",
                                                  "", 
                                                  'P',
                                                  inputParamGroupID);
 
  params.SetGroupName(inputParamGroupID, "input parameters");
  

  /**** CONFIGURATION PARAMS ****/
  int configParamGroupID = params.GetFreeGroupID();

  bool* TileMode = params.AddParamBool("TileMode",
                                       "set tile mode, "
                                       "default is: false",
                                       false, 't',
                                       configParamGroupID);  

  double* MinCornerAngleDEG = params.AddParamDouble("MinCornerAngleDEG",
                                                    "discard triangles containing a smaller angle (in degree), "
                                                    "default is: 3.0",
                                                    3.0, 0.0, 180.0, 
                                                    'a',
                                                    configParamGroupID);
 
  double* MaxAngleDEG = 
    params.AddParamDouble("MaxAngleDEG",
                          "maximum angle between normal and viewray in degree, "
                          "0 means frontal view, default is: 91.0 (disabled)",
                          91.0, 0.0, 180.0, 
                          'A',
                          configParamGroupID);

  double* LevelOfDetail = params.AddParamDouble("LevelOfDetail",
                                                "set level of detail, "
                                                "1: vertex for each pixel of full resolution, "
                                                "2: of half resolution",
                                                1.0, 0.0, 10.0,
                                                'l',
                                                configParamGroupID);


  params.SetGroupName(configParamGroupID, "configuration parameters");
  

  /**** OUTPUT PARAMS ****/
  int outputGroupID = params.GetFreeGroupID();
  
  string* Outfile = params.AddParamString("Outfile",
                                          "write results into this file",
                                          "DenseTriangleMesh.wrl", 
                                          'o',
                                          outputGroupID);

  params.SetGroupName(outputGroupID, "output");
  
  /*** end params ***/
  
  int fup = params.ParseCommandLine(argc, argv);

  if(*help) {
    params.Usage();
    return 0;
  }


  if (fup < 3) {
    cout <<"==> Too few arguments."<<endl;
    params.Usage();
    exit(1);
  }

  Camera<unsigned char> texture;
  
  if (string(*inputTexture) == "") {
    cout<<"no texture specified, drawing untextured triangle mesh!"<<endl;
  } else {
    if(ImageIO::Load(*inputTexture, texture)!=0) {
      cout<<"texture could not be loaded!\n";
      return -1;
    }
    texture.ParseMetaData();
  }

  Camera<float> depthMap;
  if(ImageIO::Load(*inputDepthMap, depthMap)!=0) {
    cout<<"depth map could not be loaded!\n";
    params.Usage();
    if (string(*inputDepthMap)==string("UNITSPHERE")) {
      depthMap.Init(texture.GetWidth(), texture.GetHeight(), 1);
      depthMap.FillImageWithConstValue(1.0f);
    }
    else return -1;
  }
  
  depthMap.ParseMetaData();

  Projection P;
  int pLoad = -1;

  if (string(*inputProjection) == "") {
    cout<<"no projection specified, trying to parse from meta data..."<<endl;   
  } else {
    pLoad = P.Load(*inputProjection);
  }
  if (pLoad != 0) {
    if (texture.IsProjValid()) {
      P = texture.GetProj();
    } else if (depthMap.IsProjValid()) {
      P = depthMap.GetProj();
    } else {
      cout<<"Projection could not be loaded from file or meta data!\n";
      params.Usage();
      return -1;
    }
  }

  cout<<"meshing..."<<flush<<endl;
  double minCornerAngle =  (*MinCornerAngleDEG)*M_PI/180.0;
  double maxViewingAngle = (*MaxAngleDEG)*M_PI/180.0;
  
  
  cout<<"minCornerAngle = "<<minCornerAngle<<endl;
  cout<<"maxViewingAngle = "<<maxViewingAngle<<endl;
  TriangleMesh mesh(minCornerAngle, maxViewingAngle);
  

  cout<<"Level of detail is "<< *LevelOfDetail<<endl;
  if (!(*LevelOfDetail == 1.0)) {
    //double maxlevel = log(double(depthMap.GetWidth()))/log(2.0)-6;
    //cout<<"maximum level is "<<maxlevel<<endl;
    
    for (double curLevel = 1.0;
         curLevel < *LevelOfDetail; curLevel += 1.0) {
      Gauss<float, float> DepthSmooth; 
      DepthSmooth.SetSigma(sqrt(1.0*1.0 - 0.5*0.5)); 
      Image<float> copyOfDepth = depthMap;
      DepthSmooth.Filter7x7GreyIgnoreBelowThreshold(depthMap, copyOfDepth, 
                                                    0.0f);
      depthMap.Release();
      depthMap.Init(copyOfDepth.GetWidth()/2, copyOfDepth.GetHeight()/2);
      for (unsigned int y=0; y<depthMap.GetHeight(); y++) {
        for (unsigned int x=0; x<depthMap.GetWidth(); x++) {
          depthMap.GetImageDataArray()[y][x] = 
            copyOfDepth.GetImageDataArray()[2*y][2*x];
        }
      }
      P.Rescale(2.0);
      ImageIO::Write("depth"+toString(curLevel), depthMap);
    }
    
  }

  string vrmlfilename = *Outfile;
  cout<<"vrml file name is now "<<vrmlfilename<<endl;
  ThreeDOut vrmlOut;
  
  cout<<boolalpha;
  cout<<"Tile Mode is set to "<<*TileMode<<endl;
  
  if(*TileMode) {
  
    // tile here into small image blocks, allows for writing of huge files
    ofstream vrmlfile(vrmlfilename.c_str());
    vrmlOut.VRMLOutWriteHeader(vrmlfile);
    
    unsigned int width, height;
    P.GetParameters()->GetImageSize(width, height);
    unsigned int tileWidth, tileHeight;
    bool xFinished = false; 
    bool yFinished = false;
    //  bool finished = false;
    unsigned int xStart, yStart;
    unsigned int tileNum = 0;
    yStart = 0;
    int numTiles = (height/(TILE_HEIGHT-1)+1)*(width/(TILE_WIDTH-1)+1);
    cout<<"Writing "<< numTiles
        <<" tiles: "<<flush;
    while(!yFinished) {
      if(yStart+TILE_HEIGHT >= height) {
        yFinished = true;
        tileHeight = height-yStart;
      } 
      else tileHeight = TILE_HEIGHT;
      
      cout<< double(tileNum)/double(numTiles)*100.0<<"% "<<flush;
      xFinished = false;
      xStart = 0; 
      while(!xFinished) {
        stringstream ss; 
        ss<<"mesh_from_"<<string(*inputDepthMap)<<"_t"<<tileNum;
        tileNum++;
        if(xStart+TILE_WIDTH >= width) {
          xFinished = true;
          tileWidth = width -xStart;
        }
        else 
          tileWidth = TILE_WIDTH;
        
        if(mesh.GenerateDenseMesh(depthMap, P, texture, 
                                  xStart, yStart, tileWidth, tileHeight)==0) {
          vrmlOut.AddTriangleMesh(mesh, ss.str(), 
                                  *inputTexture, false);
          vrmlOut.VRMLOutIndexedFaceSets(vrmlfile);
          vrmlOut.RemoveAll();
        }
        xStart+=(tileWidth-1);
      }
      yStart+=(tileHeight-1);
    }
    cout<< "  finished."<<endl<<flush;
  
  } else {   // TileMode == false

   
    mesh.GenerateSimplifiedMesh(depthMap, P, texture, 0, 0, 
                                depthMap.GetWidth(),
                                depthMap.GetHeight(), 0.7, 3.0);  
    cout<<"finished\n"<<flush;


    string texturefilename = *Outfile;
    int dotIndex = texturefilename.find_last_of(".");
    texturefilename.replace(dotIndex+1, 3, "png");
    
    cout<<"Writing Texture to "<<texturefilename<<flush<<endl;
    vrmlOut.AddTriangleMesh(mesh, "mesh_from_"+string(*inputDepthMap), 
                            texturefilename, true);
    cout<<"finished\n";
    
    cout<<"Writing VRML to "<<vrmlfilename<<flush<<endl; 
    vrmlOut.VRMLOut(vrmlfilename);
    cout<<"finished\n";
  }

  return 0;
}
