/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __GENERATE_GROUND_TRUTH_HH__
#define __GENERATE_GROUND_TRUTH_HH__

#include <Base/Common/BIASpragmaStart.hh>

#include <Geometry/PMatrix.hh>
#include <Base/Geometry/KMatrix.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Geometry/HomgPoint3D.hh>
#include <Base/Image/Image.hh>

#include <vector>

namespace BIAS {

  class BIASUtils_EXPORT GenGroundTruth
  {
  public:
    /** normalizes the matches by multiplying with K^-1 
        @author woelk 05 2003 */
    static void NormalizeMatches(std::vector<BIAS::HomgPoint2D>& match1, 
                                 std::vector<BIAS::HomgPoint2D>& match2,
                                 BIAS::KMatrix K);

    /** de-normalizes the matches by multiplying with K 
        @author woelk 05 2003 */
    static void DeNormalizeMatches(std::vector<BIAS::HomgPoint2D>& match1, 
                                   std::vector<BIAS::HomgPoint2D>& match2,
                                   BIAS::KMatrix K);

    /** draws the matches in image im with background 0, match in color mcolor
        and line defined by match in color lcolcor
        @authior woelk 05 2003 */
    static void DrawMatches(BIAS::Image<unsigned char>& im, 
                            std::vector<BIAS::HomgPoint2D>& match1,  
                            std::vector<BIAS::HomgPoint2D>& match2,
                            unsigned char lcolor, unsigned char mcolor);

    /** Generates ideal matches in cameras from pvec by randomly creating
        a 3D point in the cube between min[x,y,z] and max[x,y,z] and 
        than looking if it projects into all cameras
        @author woelk 05 2003 */
    static void CreateIdealMatches(std::vector<BIAS::PMatrix>& pvec, unsigned num,
                                   unsigned width, unsigned height,
                                   std::vector<std::vector<BIAS::HomgPoint2D> >& 
                                   matches, 
                                   std::vector<BIAS::HomgPoint3D>& points,
                                   double minx=-50.0, double maxx=50.0, 
                                   double miny=-50.0, double maxy=50.0,
                                   double minz=-50.0, double maxz=50.0);

    /** Generates ideal matches in cameras from pvec by randomly creating
        a 3D point in the cube between min[x,y,z] and max[x,y,z] and 
        than looking if it projects into all cameras
        Adds spatial noise with sigma spatial_noise to position of 2D
        points in every camera.
        The ideal matches are returned in i_matches, the noisy matches are 
        returned in n_matches.
        @author woelk 05 2003 */
    static void CreateMatches(std::vector<BIAS::PMatrix>& pvec, unsigned num,
                       unsigned width, unsigned height, double spatial_noise,
                       std::vector<std::vector<BIAS::HomgPoint2D> >& i_matches,
                       std::vector<std::vector<BIAS::HomgPoint2D> >& n_matches,
                       std::vector<BIAS::HomgPoint3D>& points,
                       double minx=-50.0, double maxx=50.0, 
                       double miny=-50.0, double maxy=50.0,
                       double minz=-50.0, double maxz=50.0);

    /** Generates ideal matches in cameras from pvec by randomly creating
        a 3D point in the cube between min[x,y,z] and max[x,y,z] and 
        than looking if it projects into all cameras.
        Adds spatial noise with a sigma of spatial_noise to position of 2D
        points in every camera.
        The ideal matches are returned in i_matches, the noisy matches are 
        returned in n_matches.
        An additional normal ditributed position error with sigma of c_noise 
        and a normal distributed rotational error with sigma of r_noise are 
        assumed for the camera matrices.
        The noisy camera matrices are returned in n_pvec.
        @author woelk 05 2003 */
    static void CreateMatches(std::vector<BIAS::PMatrix>& pvec, 
                       std::vector<BIAS::PMatrix>& n_pvec, unsigned num,
                       unsigned width, unsigned height, double spatial_noise,
                       double r_noise, double c_noise,
                       std::vector<std::vector<BIAS::HomgPoint2D> >& i_matches,
                       std::vector<std::vector<BIAS::HomgPoint2D> >& n_matches,
                       std::vector<BIAS::HomgPoint3D>& points,
                       double minx=-50.0, double maxx=50.0, 
                       double miny=-50.0, double maxy=50.0,
                       double minz=-50.0, double maxz=50.0, bool print=false);
    
    /** Generates ideal matches in cameras from pvec by randomly creating
        a 3D point in the cube between min[x,y,z] and max[x,y,z] and 
        than looking if it projects into all cameras.
        Adds spatial noise with a sigma of spatial_noise to position of 2D
        points in every camera.
        The ideal matches are returned in i_matches, the noisy matches are 
        returned in n_matches.
        An additional normal ditributed position error with sigma of c_noise 
        and a normal distributed rotational error with sigma of r_noise are 
        assumed for the camera matrices.
        The noisy camera matrices are returned in n_pvec.
        outlier_fraction of the matches are outlier. They are generated by 
        adding a random offset to each x and y of each point after image 0. 
        The offset is generated between 2*spatial_noise and 10*spatial_noise.
        @author woelk 05 2003 */
    static void 
    CreateMatchesOutlier(std::vector<BIAS::PMatrix>& pvec, 
                     std::vector<BIAS::PMatrix>& n_pvec, unsigned num,
                     unsigned width, unsigned height, double spatial_noise,
                     double r_noise, double c_noise, double outlier_fraction,
                     std::vector<std::vector<BIAS::HomgPoint2D> >& i_matches,
                     std::vector<std::vector<BIAS::HomgPoint2D> >& n_matches,
                     std::vector<BIAS::HomgPoint3D>& points,
                     double minx=-50.0, double maxx=50.0, 
                     double miny=-50.0, double maxy=50.0,
                     double minz=-50.0, double maxz=50.0, 
                     bool print=false);

    /** Generates 'n' points which are all located on a plane specified by
        space point coords aX, aY, aZ and the coords of two spanning vectors 
        alphaX, alphaY, alphaZ and betaX, betaY, betaZ, which are normalized
        internally.
        'alphamax' and 'betamax' define width and height of the plane.
        This method does not clear the result vector 'Points'!
        By default it generates 200 points on a rectangular 100x100 plane in
        the x-y-plane.
        This method implicitly calls GenerateRandomCube with the third
        dimension set to zero.
        @author Matthias Dunda 08 2003 */
    static void GenerateRandomPlane(std::vector<BIAS::HomgPoint3D> &Points,
                      int n=200,
                      double aX=-50.0, double aY=-50.0, double aZ=0.0,
                      double alphaX=1.0, double alphaY=0.0, double alphaZ=0.0,
                      double betaX=0.0, double betaY=1.0, double betaZ=0.0,
                      double alphamax=100.0,
                      double betamax=100.0);

    /** Generates 'n' points which are all located in a cube specified by
        space point coords aX, aY, aZ and the coords of three spanning vectors 
        alphaX, alphaY, alphaZ and betaX, betaY, betaZ and 
        gammaX, gammaY, gammaZ, which are normalized internally.
        'alphamax', 'betamax' and 'gammamax' define width, height and 
        depth of the cube.
        This method does not clear the result vector 'Points'!
        By default it generates 200 points on a rectangular 100x100x100 cube
        centered at 0 0 0.
        @author Matthias Dunda 08 2003 */
    static void GenerateRandomCube(std::vector<BIAS::HomgPoint3D> &Points,
                      int n=200,
                      double aX=-50.0, double aY=-50.0, double aZ=-50.0,
                      double alphaX=1.0, double alphaY=0.0, double alphaZ=0.0,
                      double betaX=0.0, double betaY=1.0, double betaZ=0.0,
                      double gammaX=0.0, double gammaY=0.0, double gammaZ=1.0,
                      double alphamax=100.0,
                      double betamax=100.0,
                      double gammamax=100.0);

    static int ReadCorrespondences(const char *Filename, 
                                   std::vector<BIAS::HomgPoint2D> &Points1,
                                   std::vector<BIAS::HomgPoint2D> &Points2);

    /** Store 2D picture correspondences in a file in binary format, so that
        you can keep the full accuracy of a double.
        @status tested
        @author mdunda */
    static int WriteCorrespondences(const char *Filename, 
                                    std::vector<BIAS::HomgPoint2D> &Points1,
                                    std::vector<BIAS::HomgPoint2D> &Points2);

    /** Read 3D world coordinates from a file in binary format.
        @status tested
        @author mdunda */
    static int ReadWorldPoints( const char *Filename, 
                                std::vector<BIAS::HomgPoint3D> & Points );

    /** Store 3D world coordinates in a file in binary format, so that
        you can keep the full accuracy of a double.
        @status tested
        @author mdunda */
    static int WriteWorldPoints( const char *Filename, 
                                 std::vector<BIAS::HomgPoint3D> & Points );

  }; /* class GenGroundTruth */

} /* namespace MIP */

#include <Base/Common/BIASpragmaEnd.hh>

#endif /* __GENERATE_GROUND_TRUTH_HH__ */

