/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <libdc1394/dc1394_control.h>
#include "VideoSource_DCAM.hh"
#include "sys/time.h"
#include "unistd.h"
#include <sys/stat.h>

using namespace BIAS;

using namespace std;

int VideoSource_DCAM::InstancesTotal_ = 0;
int VideoSource_DCAM::Instances_[MAX_1394_CARDS];
raw1394handle_t VideoSource_DCAM::IEEE1394Handle_[MAX_1394_CARDS];
int VideoSource_DCAM::BytesPerCyle_[MAX_1394_CARDS];



int GetCamIDs(int busid ,int &ncams, int *&camids )
{
  raw1394handle_t handle;
  nodeid_t  *dcam_nodes;
  dc1394_camerainfo dcaminfo;
  
  
  handle = dc1394_create_handle(busid);
  if(handle == NULL)
  {
     BIASERR("dc1394 handle returned NULL-Pointer" << endl << "is there a camera attached?" << endl);
     return -1;
  }
  dcam_nodes = dc1394_get_camera_nodes(handle,&ncams,0);

  camids = new int[ncams];
  

  for (int i=0; i<ncams; i++) {
    int res;      
    res = dc1394_get_camera_info(handle, dcam_nodes[i],  &dcaminfo);
    camids[i] = (dcaminfo.euid_64 & 0x00000FFF);
  }


  raw1394_destroy_handle(handle);
  free(dcam_nodes);

  return 0;
  

}
void VideoSource_DCAM::SetCam_ID(int manualcamid_){
    camid=manualcamid_;

}
int VideoSource_DCAM::OpenBus_()
{
  switch (IEEE1394Card_) {
  case 0:  DefaultDevice_ = "/dev/video1394-0"; break;
  case 1:  DefaultDevice_ = "/dev/video1394-1"; break;
  case 2:  DefaultDevice_ = "/dev/video1394-2"; break;
  default:;
  }
  struct stat statbuf;
  stat(DefaultDevice_,&statbuf);
  if ( ! S_ISCHR(statbuf.st_mode)) {
     BIASERR(string(DefaultDevice_)+" is not a character device ");
  }

  if (IEEE1394Card_ > MAX_1394_CARDS) {
    BIASERR("VideoSource_DCAM::VideoSource_DCAM(int Card): requested card #"<<IEEE1394Card_<< " exceeds maximum of "<<MAX_1394_CARDS);
    return -1 ; 
  }
  if ( Instances_[IEEE1394Card_] == 0) {
    // I am the first instance on this Card, so open now
    BIASDOUT(D_DCAM,"IEEE1394Card_: "<<IEEE1394Card_);
    IEEE1394Handle_[IEEE1394Card_] = dc1394_create_handle(IEEE1394Card_);
    if (IEEE1394Handle_[IEEE1394Card_] == NULL) {
      perror("dc1394 - couldn't get handle");
      BIASERR("This error usually means that the ieee1394 drivers are not loaded or that /dev/raw1394 or /dev/video1394-0 does not exist.\n");
      return -1 ;
    }
    BytesPerCyle_[IEEE1394Card_] = 0;
  }
  
  Instances_[IEEE1394Card_] ++;
  
  BIASDOUT(D_DCAM,"I am the "<<InstancesTotal_<<". Instance, and the "<<Instances_[IEEE1394Card_]<<". on card "<<IEEE1394Card_);

  return 0;
}


VideoSource_DCAM::VideoSource_DCAM(int Card): VideoSource(Card)
{ camid=0;
  IEEE1394Card_ = Card;
  //SetDebugLevel(D_DCAM);
  TransmissionTime_ = 0;
  UserTopLeft_ = false;
  UseBinning_ = false;
  Shift16_ = 0;

  if (InstancesTotal_ == 0) {
    BIASDOUT(D_DCAM,"Initializing Instances_[] and IEEE1394Handle_[]");
    for (int i=0; i<MAX_1394_CARDS; i++) {
      Instances_[i] = 0;
      IEEE1394Handle_[i] = NULL;
    }
  }
  
  InstancesTotal_ ++;

  IEEE1394Card_= Card;
  IsControllable_ = true;



  SelectedCam_ = 0;

  // New Default: Use DMA
  UseDMA_ = true;

  char *usedmaenv;
  usedmaenv = getenv("DC1394_DMA");
  if (usedmaenv != NULL) {
    if (strcmp(usedmaenv,"false") == 0) {
      BIASDOUT(D_DCAM,"Using Non-DMA");
      UseDMA_ = false;
    }
    else
      cerr<<"Warning: Found environment variable DC1394_DMA set to: "<<usedmaenv<<". Only \"false\" would be valid. Using DMA anyway."<<endl;
  }
  
  DcamMode_ = 0;
  DcamFormat_ = 0;
  DcamFps_ = FRAMERATE_30;
  FramesPerSecond_ = 0.0;
  ForceF7_ = false;
  DirectF7_ = false;
  ColorMode_ = ImageBase::CM_UYVY422;
  Width_ = 0;
  Height_ = 0;
  BytesPerPacket_ = 0;
  ColorChannels_ = 0;

  
}

VideoSource_DCAM::~VideoSource_DCAM()
{
  InstancesTotal_ --;

}


int VideoSource_DCAM::OpenDevice(int camid)
{
  int numNodes;
  BIASDOUT(D_DCAM,"VideoSource_DCAM::OpenDevice("<<camid<<")");

  if (OpenBus_() <0) return -1;

  // if constructor failed to initialize raw1394 then bail out
  if (IEEE1394Handle_[IEEE1394Card_] == NULL) return -1;


  numNodes = raw1394_get_nodecount(IEEE1394Handle_[IEEE1394Card_]);
  DcamNodes_ = dc1394_get_camera_nodes(IEEE1394Handle_[IEEE1394Card_],
    &NumCams_,0);
  BIASDOUT(D_DCAM,"Found "<<numNodes<<" nodes on the bus, thereof "<<NumCams_<<" DCAM-Cameras");

  if (NumCams_<1)	  {
    BIASERR(" no DCAM cameras found on device/bus "<<camid
      <<".\nDo you have multiple firewire buses?" );
    raw1394_destroy_handle(IEEE1394Handle_[IEEE1394Card_]);
    return -1;
  }

  if (camid == 0) {
    SelectedCam_ = 0;
    dc1394_get_camera_info(IEEE1394Handle_[IEEE1394Card_], 
      DcamNodes_[SelectedCam_],&DcamInfo_);
    ID_ = (DcamInfo_.euid_64 & 0x00000FFF);
    CamModel_ = GetModelForEuid(DcamInfo_.euid_64);
  } else {
    SelectedCam_ = -1;
    for (int i=0; i<NumCams_; i++) {
      int res;
      res = dc1394_get_camera_info(IEEE1394Handle_[IEEE1394Card_], 
        DcamNodes_[i],&DcamInfo_);
      ID_ = (DcamInfo_.euid_64 & 0x00000FFF);
      if (ID_ == camid) {
        SelectedCam_ = i;
        CamModel_ = GetModelForEuid(DcamInfo_.euid_64);
        break;
      }
    }
    if (SelectedCam_ < 0) {
      BIASERR("Selected Camera "<<camid<<" not found.");
      return -1;
    }
  }
  BIASDOUT(D_DCAM,"Setting up camera "<<SelectedCam_<<" UID:"<<ID_);

  QueryCam();
  if (!DirectF7_) {
    // check for usersize mode and try do set it or set the highest 
    // mode as default
    if ((GetWidth() != 0) || (GetHeight()!=0) || UseBinning_){
      if ( UserSizeToMode() < 0) {
        BIASERR("UserSizeToMode() failed");
        return -1;
      }
    }
    else{
     SelectDefault();
    }
  }
  else BIASDOUT(D_DCAM,"Direct F7 requested from User, mode: "<<DcamMode_ 
    - MODE_FORMAT7_MIN);
  return InitCam();
}



int VideoSource_DCAM::OpenDevice()
{
  return OpenDevice(camid);
}



int  VideoSource_DCAM::SetFormat7(int mode, int bpp, int colorchannels)
{
  DcamFormat_  = FORMAT_SCALABLE_IMAGE_SIZE;
  DcamMode_ = mode + MODE_FORMAT7_MIN;
  BytesPerPacket_ = bpp;
  DirectF7_ = true;
  ColorChannels_ = colorchannels;
  return 0;
}



int VideoSource_DCAM::CloseDevice()
{
  if (UseDMA_) { 
    dc1394_dma_unlisten(IEEE1394Handle_[IEEE1394Card_],&DcamCapture_);
    //   dc1394_dma_release_camera(IEEE1394Handle_[IEEE1394Card_],&DcamCapture_);
  }
  else
    dc1394_release_camera(IEEE1394Handle_[IEEE1394Card_],&DcamCapture_);
  // cleanup 
  Instances_[IEEE1394Card_] --;
  if (Instances_[IEEE1394Card_] == 0) {
    // I am the last instance on this card, so close the door 
    raw1394_destroy_handle(IEEE1394Handle_[IEEE1394Card_]);
    BIASDOUT(D_DCAM,"I am the last Instance on this bus, so shutting the lights.");
  }  
  return 0;
}

  

  
int VideoSource_DCAM::UsePort(int channel)
{
  IEEE1394Card_ = channel;
  return 0;
}


int VideoSource_DCAM::GrabSingle(Camera <unsigned char> &image)
{
  int res;
  BIASDOUT(D_DCAM,"VideoSource_DCAM::GrabSingle("<<SelectedCam_<<")");
  assert(! image.IsEmpty());
  if (!Active_) {
    BIASERR("VideoSource_DCAM::GrabSingle(): VideoSource Object not active, use PreGrab()");
    return -1;
  }
  if (Grabbing_) {
    BIASERR("VideoSource_DCAM::GrabSingle(): VideoSource Object is already grabbing()");
    return -1;
  }
  Grabbing_ = true;
  if (UseDMA_)
    //  res = dc1394_dma_multi_capture(&DcamCapture_,1);
    res = dc1394_dma_single_capture(&DcamCapture_);
  
  else
    res = dc1394_single_capture(IEEE1394Handle_[IEEE1394Card_],&DcamCapture_);
  
  if (res != DC1394_SUCCESS)    {
    BIASERR("VideoSource_DCAM::GrabSingle(): unable to capture one frame from camera");
    return -1;
  }
  BIASDOUT(D_DCAM,DcamCapture_.frame_width<<" x "<<DcamCapture_.frame_height);
  image.SetTime(DcamCapture_.filltime.tv_sec,DcamCapture_.filltime.tv_usec);

  image.CopyIn_NoInit((unsigned char *)DcamCapture_.capture_buffer);

  image.InvalidateUID();
  image.SetUID(UUID::GenerateUUID());
//   if (DirectToDisk_) 
//     cerr<<"DirectToDisk is on "<<endl;
//     WriteDirect_((unsigned char *)DcamCapture_.capture_buffer, image.GetUID(),
//             DcamCapture_.filltime.tv_sec,DcamCapture_.filltime.tv_usec);
  if (UseDMA_)
    dc1394_dma_done_with_buffer(&DcamCapture_);
  Grabbing_ = false;
  return 0;
}


#ifdef BUILD_IMAGE_USHORT
int VideoSource_DCAM::GrabSingle(BIAS::Camera <unsigned short> &image)
{
  int res;
  BIASDOUT(D_DCAM,"VideoSource_DCAM::GrabSingle("<<SelectedCam_<<")");
  assert(! image.IsEmpty());
  if (!Active_) {
    BIASERR("VideoSource_DCAM::GrabSingle(): VideoSource Object not active, use PreGrab()");
    return -1;
  }
  if (Grabbing_) {
    BIASERR("VideoSource_DCAM::GrabSingle(): VideoSource Object is already grabbing()");
    return -1;
  }
  Grabbing_ = true;
  if (UseDMA_)
//     res = dc1394_dma_multi_capture(&DcamCapture_,1);
     res = dc1394_dma_single_capture(&DcamCapture_);
  else
    res = dc1394_single_capture(IEEE1394Handle_[IEEE1394Card_],&DcamCapture_);
  if (res != DC1394_SUCCESS)    {
    BIASERR("VideoSource_DCAM::GrabSingle(): unable to capture one frame from camera");
    return -1;
  }
  image.SetTime(DcamCapture_.filltime.tv_sec,DcamCapture_.filltime.tv_usec);

  unsigned int numbytes = image.GetSize();
// byte-wise pointer
  unsigned char *idata = (unsigned char *)image.GetImageData(); 
  unsigned char *buf = (unsigned char *)DcamCapture_.capture_buffer;
  for (unsigned int i=0;i<numbytes; i+=2){
    idata[i] = buf[i+1];
    idata[i+1] = buf[i];
    // the significant bits used by point grey are the 16-Shift16_  upper bits
    // cast pixel pointer to unsigned short and apply shift
    *(unsigned short *)(&idata[i]) = (*(unsigned short *)(&idata[i])) >>Shift16_;
  }

  //   image.CopyIn_NoInit((unsigned char *)DcamCapture_.capture_buffer);
  if (UseDMA_)
    dc1394_dma_done_with_buffer(&DcamCapture_);
  Grabbing_ = false;
  image.InvalidateUID();
  return 0;
}
#endif  


void VideoSource_DCAM::ActivateScorpionTimestamp()
{
  quadlet_t r12F8 = 0x80000001;//activate only timestamp, read Tech Reference
  SetCameraControlRegister(IEEE1394Handle_[IEEE1394Card_],
 			   DcamNodes_[SelectedCam_], 0x12f8, r12F8);
}



int VideoSource_DCAM::PreGrab()
{
  BIASDOUT(D_DCAM,"VideoSource_DCAM::PreGrab("<<SelectedCam_<<")");
  
  if (! CompleteInitialized_ )
    if (InitCam() <0) return -1;

  if (BytesPerCyle_[IEEE1394Card_] + BytesPerPacket_ >4096) {
    BIASERR("unable to start iso transfer for camera "<<SelectedCam_
	   <<"\n\tBytesPerCyle_ exceeds maximum of 4096 for card #"
	   <<IEEE1394Card_);
    
    return -1;
  }

  BytesPerCyle_[IEEE1394Card_] += BytesPerPacket_;
  BIASDOUT(D_DCAM,"BytesPerCyle_["<<IEEE1394Card_<<"]: "<<BytesPerCyle_[IEEE1394Card_]);

//   quadlet_t r0614;

//   GetCameraControlRegister(IEEE1394Handle_[IEEE1394Card_],
// 			   DcamNodes_[SelectedCam_], 0x0614, &r0614);
//   cout <<"0x614 before StartIso(): "<<r0614<<endl;
    if (StartIso() <0) return -1;
  
//   GetCameraControlRegister(IEEE1394Handle_[IEEE1394Card_],
// 			   DcamNodes_[SelectedCam_], 0x0614, &r0614);
//   cout <<"0x614 after StartIso(): "<<(r0614)<<endl;

//   r0614 = r0614 | 1;
//   cout <<"0x614 to be set: "<<r0614<<endl;
//    SetCameraControlRegister(IEEE1394Handle_[IEEE1394Card_],
// 			   DcamNodes_[SelectedCam_], 0x0614, r0614);
//   GetCameraControlRegister(IEEE1394Handle_[IEEE1394Card_],
// 			   DcamNodes_[SelectedCam_], 0x0614, &r0614);
//   cout <<"0x614 after set: "<<(r0614)<<endl;

  Active_ = true;
  return 0;
}


int VideoSource_DCAM::PostGrab()
{
  int res;
  BIASDOUT(D_DCAM,"VideoSource_DCAM::PostGrab("<<SelectedCam_<<")");
  Active_ = false;
  
  res = StopIso();

  BytesPerCyle_[IEEE1394Card_] -= BytesPerPacket_;
  BIASDOUT(D_DCAM,"BytesPerCyle_["<<IEEE1394Card_<<"]: "<<BytesPerCyle_[IEEE1394Card_]);
  
  return res;
}


int VideoSource_DCAM::QueryCam()
{
  int res;
  quadlet_t SuppForm;
  quadlet_t SuppModes;
  
  quadlet_t Suppfps;
  
  BIASDOUT(D_DCAM,"VideoSource_DCAM::QueryCam("<<SelectedCam_<<")");
  
  dc1394_get_camera_feature_set(IEEE1394Handle_[IEEE1394Card_], DcamNodes_[SelectedCam_], 
				&DcamFeatures_);
  
  // First, query for supported formats,
  // then query each format for supported modes 
  // finally, query each format/mode for supported framerates

  res = dc1394_query_supported_formats(IEEE1394Handle_[IEEE1394Card_],DcamNodes_[SelectedCam_]
				       , &SuppForm);

  SupportedFormats_[0] = SuppForm & (0x1<<31);
  SupportedFormats_[1] = SuppForm & (0x1<<30);
  SupportedFormats_[2] = SuppForm & (0x1<<29);
  SupportedFormats_[3] = false;
  SupportedFormats_[4] = false;
  SupportedFormats_[5] = false;
  SupportedFormats_[6] = SuppForm & (0x1<<25);
  SupportedFormats_[7] = SuppForm & (0x1<<24);
  // now query the modes
  for (int format=0; format < NUM_FORMATS; format++) {
    if (SupportedFormats_[format]) {
      dc1394_query_supported_modes(IEEE1394Handle_[IEEE1394Card_],  
				   DcamNodes_[SelectedCam_],
				   FORMAT_VGA_NONCOMPRESSED+ format,
				   &SuppModes);
      for (int i=0; i<NUM_FORMAT1_MODES; i++)
	SupportedModes_[format][i] = SuppModes & (0x1<<(31-i));
      
    }
    else 
      for (int i=0; i<NUM_FORMAT1_MODES; i++)
	SupportedModes_[format][i] =false;
      
  }
  // dont query format7 for fps, it is specified via bytes_per_packet
  for (int format=0; format < NUM_FORMATS-1; format++) 
    for (int i=0; i<NUM_FORMAT1_MODES; i++){
      if (SupportedModes_[format][i]) {
	res = dc1394_query_supported_framerates(IEEE1394Handle_[IEEE1394Card_],
						DcamNodes_[SelectedCam_],
						FORMAT_VGA_NONCOMPRESSED+ format,
						64 +i,
						&Suppfps);
	
	if (res >=0){  
	  SupportedFramerates_[format][i] = Suppfps;
	}
	else{
	  SupportedFramerates_[format][i] = 0;
	}
      }
      else
	SupportedFramerates_[format][i] = 0;      
    }
 
  if ( (GetDebugLevel() & D_DCAM)>0)
    ShowCapabilities();
  
  if (  SupportedFormats_[7]) {
    dc1394_query_format7_max_image_size(IEEE1394Handle_[IEEE1394Card_], 
					DcamNodes_[SelectedCam_], 
					MODE_FORMAT7_0, &MaxWidth_, 
					&MaxHeight_);
    dc1394_query_format7_unit_size(IEEE1394Handle_[IEEE1394Card_], 
				   DcamNodes_[SelectedCam_], 
				   MODE_FORMAT7_0, 
				   &UnitWidth_,&UnitHeight_);

    BIASDOUT(D_DCAM_F7," Format7 max:  "<<MaxWidth_<<" x "<<MaxHeight_);
    BIASDOUT(D_DCAM_F7," Format7 unit: "<<UnitWidth_<<" x "<<UnitHeight_);
    BIASDOUT(D_DCAM_F7," Format7 Bytes per Packet: "<<MinBpP_<<" - "<<MaxBpP_);
  }

  dc1394bool_t ispresent = DC1394_FALSE;
  for (int i=FEATURE_MIN; i<FEATURE_MAX+1; i++) {
    dc1394_is_feature_present(IEEE1394Handle_[IEEE1394Card_], 
			      DcamNodes_[SelectedCam_], i,
			      &ispresent);
    if (ispresent) SupportedFeature_[i-FEATURE_MIN] = true;
    else SupportedFeature_[i-FEATURE_MIN] = false;
  }
  

  // to determine the version, this indicats >=0.9.3
#ifdef DC1394_NO_FRAME  

  if (CamModel_ == CAMERA_PTGREY_SCORPION_COLOR_1600 
      || CamModel_ == CAMERA_PTGREY_DRAGONFLY_COLOR){
    quadlet_t r1040;
    GetCameraControlRegister(IEEE1394Handle_[IEEE1394Card_],
			     DcamNodes_[SelectedCam_], 0x1040, &r1040);
//     unsigned char BayerSense_A = r1040  & 0xF;
//     cout <<"R1040h: "<<hex<<r1040<<dec<<endl;
//     cout <<"BayerA: "<<BayerSense_A<<endl;
  }
#endif


  return 0;
  
}




int VideoSource_DCAM::ShowCapabilities()
{
  cout <<"Camera: "<<SelectedCam_<<" UID:"
       <<(DcamInfo_.euid_64 & 0x00000FFF)<<endl;

  //    dc1394_print_feature_set(&DcamFeatures_);


  cout <<"Supported Formats: ";
  for (int i=0; i<NUM_FORMATS; i++)
    if (SupportedFormats_[i])
      cout <<"Format"<<i<<" ";
  cout <<endl;

  cout <<"Suported Modes:"<<endl;
  for (int i=0; i<NUM_FORMATS; i++)
    if (SupportedFormats_[i]) {
      cout <<"Format"<<i<<": ";
      for (int j=0; j<NUM_FORMAT1_MODES; j++)
        if (SupportedModes_[i][j])
          cout <<"Mode"<<j<<" ";
      
      cout <<endl;
      
    }
  
  cout <<"Supported Framerates:"<<endl;
  for (int format=0; format < NUM_FORMATS; format++) {
    int fps;
    
    cout <<" Format"<<format<<": ";
    for (int i=0; i<NUM_FORMAT1_MODES; i++)
      if ((fps = SupportedFramerates_[format][i]) >0) {
        cout <<"Mode"<<i<<":  ";
        for (int f=0; f<NUM_FRAMERATES; f++)
          if (fps & (0x1<< (31-f) )) 
            cout <<1.875 *(0x1<<f)<<"  ";
	
      }
    
    cout <<endl;
  }
  
  
  return 0;
  
}

  



int VideoSource_DCAM::SelectDefault()
{
  BIASDOUT(D_DCAM,"VideoSource_DCAM::SelectDefault("<<SelectedCam_<<")");
  
  int best_m= 0;
  int best_f = 0;
  int fps;
  float best_fps = -1.0;

  // For 8bpp only scan mode 0-5, because 6&7 are Mono16 modes

  int ModeLimit = (BytesPerPixel_ == 1.0) ? 6 : NUM_FORMAT1_MODES;
  
  // find the best mode (not in Format7)
  for (int f=0; f<3; f++)
    for (int m=0; m<ModeLimit; m++) 
      if (SupportedModes_[f][m]){
	best_f = f;
	best_m = m;
	if ((fps = SupportedFramerates_[f][m]) > 0) {
	  for (int f=0; f<NUM_FRAMERATES; f++)
	    if (fps & (0x1<< (31-f) )) 
	      best_fps = 1.875 *(0x1<<f);
	}
      }

  if(CamModel_ == CAMERA_PTGREY_DRAGONFLY2_COLOR_HIRES){
    cout << " Ignoring autodetected mode for Dragofly2. Setting to format2, mode4, 7.5fps." << endl;
    best_f = 2;
    best_m = 4;
    best_fps = 7.5;
  }
  if (CamModel_ == CAMERA_PTGREY_GRASSHOPPER_COLOR) {
    cout << " Ignoring autodetected mode for Grasshopper. Setting to format2, mode3, 7.5fps." << endl;
    best_f = 2;
    best_m = 3;
    best_fps = 7.5;
  }

  switch (best_f) {
  case 0:  // Format 0
    DcamFormat_ = FORMAT_VGA_NONCOMPRESSED;  
    switch (best_m) {
    case 0:  // F0 Mode 0
      DcamMode_ = MODE_160x120_YUV444;
      Width_ = 160; Height_ = 120; ColorMode_ = ImageBase::CM_YUV444;
      break;
    case 1:  // F0 Mode 1
      DcamMode_ = MODE_320x240_YUV422;
      Width_ = 320; Height_ = 240; ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 2:  // F0 Mode 2
      DcamMode_ = MODE_640x480_YUV411;
      Width_ = 640; Height_ = 480; ColorMode_ = ImageBase::CM_YUV411;
      break;
    case 3:  // F0 Mode 3
      DcamMode_ = MODE_640x480_YUV422;
      ColorChannels_ = 2;
      Width_ = 640; Height_ = 480; ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 4:  // F0 Mode 4
      DcamMode_ = MODE_640x480_RGB;
      ColorChannels_ = 3;
      Width_ = 640; Height_ = 480; ColorMode_ = ImageBase::CM_RGB;
      break;
    case 5:  // F0 Mode 5
      DcamMode_ = MODE_640x480_MONO;
      ColorChannels_ = 1;
      ColorMode_ = ImageBase::CM_Grey;
      Width_ = 640; Height_ = 480; 
      break;
    case 6:     // F0 Mode 6   
      DcamMode_ = MODE_640x480_MONO16;
      ColorChannels_ = 1;
      Width_ = 640; Height_ = 480; 
      break;
    }
    
    break;
  case 1:   // Format 1
    DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_1;  
    switch (best_m) {
    case 0: // F1 Mode 0
      DcamMode_ = MODE_800x600_YUV422;
      ColorChannels_ = 2;
      Width_ = 800; Height_ = 600;  ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 1: // F1 Mode 1
      DcamMode_ = MODE_800x600_RGB;
      ColorChannels_ = 3;
      Width_ = 800; Height_ = 600;  ColorMode_ = ImageBase::CM_RGB;
      break;
    case 2: // F1 Mode 2
      DcamMode_ = MODE_800x600_MONO;
      ColorChannels_ = 1;
      Width_ = 800; Height_ = 600;  
      break;
    case 3:   // F1 Mode 3
      DcamMode_ = MODE_1024x768_YUV422;
      ColorChannels_ = 2;
      Width_ = 1024; Height_ = 768;  ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 4: // F1 Mode 4
      DcamMode_ = MODE_1024x768_RGB;
      ColorChannels_ = 3;
      Width_ = 1024; Height_ = 768;  ColorMode_ = ImageBase::CM_RGB;
      break;
    case 5: // F1 Mode 5
      DcamMode_ = MODE_1024x768_MONO;
      ColorChannels_ = 1;
      Width_ = 1024; Height_ = 768;  
      break;
    case 6: // F1 Mode 6
      DcamMode_ = MODE_800x600_MONO16;
      Width_ = 800; Height_ = 600;  
      break;
    case 7: // F1 Mode 
      DcamMode_ = MODE_1024x768_MONO16;
      Width_ = 1024; Height_ = 768;  
      ColorChannels_ = 1;
      BytesPerPixel_ = 2.0;
      break;
    }
    
    break;
  case 2: // Format 2
    DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_2;  
    switch (best_m) {
    case 0:
      DcamMode_ = MODE_1280x960_YUV422;
      ColorChannels_ = 2;
      Width_ = 1280; Height_ = 960;  ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 1:
      DcamMode_ = MODE_1280x960_RGB;
      ColorChannels_ = 3;
      Width_ = 1280; Height_ = 960;  ColorMode_ = ImageBase::CM_RGB;
      break;
    case 2:
      DcamMode_ = MODE_1280x960_MONO;
      ColorChannels_ = 1;
      Width_ = 1280; Height_ = 960; 
      break;
    case 3:
      DcamMode_ = MODE_1600x1200_YUV422;
      ColorChannels_ = 2;
      Width_ = 1600; Height_ = 1200;  ColorMode_ = ImageBase::CM_UYVY422;
      break;
    case 4:
      DcamMode_ = MODE_1600x1200_RGB;
      ColorChannels_ = 3;
      Width_ = 1600; Height_ = 1200;  ColorMode_ = ImageBase::CM_RGB;
      break;
    case 5:
      DcamMode_ = MODE_1600x1200_MONO;
      ColorChannels_ = 1;
      Width_ = 1600; Height_ = 1200; 
      break;
    case 6:
      DcamMode_ = MODE_1280x960_MONO16;
      ColorChannels_ = 1;
      Width_ = 1280; Height_ = 960;  
      break;
    case 7:
      DcamMode_ = MODE_1600x1200_MONO16;
      ColorChannels_ = 1;
      Width_ = 1600; Height_ = 1200;  
      break;
    }
    break;
  }

  // check if the user has set a correct framerate
  if (FramesPerSecond_ == 0.0 || FramesPerSecond_ > best_fps)
    // if this is true then a correct framerate was detected
    if (best_fps != -1)
      FramesPerSecond_ = best_fps >0.0 ? best_fps : 7.5;

  // PtGrey Scorpion does not provide supported framerates, so take 7.5   
  if(CamModel_ == CAMERA_PTGREY_DRAGONFLY2_COLOR_HIRES){
    cout << " Setting best framerate right for the DragonFly 2.\n";
    FramesPerSecond_ = 7.5; //DEBUG TODO: Dragonfly2
  }  
  
  BIASDOUT(D_DCAM,"Best F:"<<best_f<<" m:"<<best_m<<" : "<<Width_<<"x"<<Height_
	   <<" at "<<FramesPerSecond_<<"fps"<<" Colormodel: "<<ColorMode_);
  return 0;
}

#define USE_NATIVE_XGA

int VideoSource_DCAM::UserSizeToMode()
{
  BIASDOUT(D_DCAM,"VideoSource_DCAM::UserSizeToMode("<<SelectedCam_<<"): "<<Width_<<"x"<<Height_<<" BytesPerPixel_: "<<BytesPerPixel_<<" ColorMode: "<<ColorMode_);     
  int res=0;
  
  if (ForceF7_ )  res = UserSizeToFormat7_();
  else {
    if (BytesPerPixel_ == 1.0)      res =  UserSizeToMode_8bit_();
    else if (BytesPerPixel_ == 2.0) res =  UserSizeToMode_16bit_();
    else {
      BIASERR("VideoSource_DCAM::UserSizeToMode(): "<<BytesPerPixel_*8
		<<" -bit not yet impl");
      return -1;
    }
    // -1 means: did not find a mode, try Format7
    if (res == -1) res = UserSizeToFormat7_();
  }
  //PrintFormatAndMode_(); cout<<endl;
  return res;
}


// try to find a matching mode or return -1 to indicate that Format7
// may be good
int VideoSource_DCAM::UserSizeToMode_8bit_() {
  int  res = 0;
  switch (ColorMode_) {
  case ImageBase::CM_UYVY422: 
    ColorChannels_ = 2;
    if (Width_ == 1600 && Height_ == 1200) {
      DcamMode_ = MODE_1600x1200_YUV422;
      DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_1; 
    }
    else if (Width_ == 1280 && Height_ == 960) {
      DcamMode_ = MODE_1280x960_YUV422;
      DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_1; 
    }
    else if (Width_ == 1024 && Height_ == 768) {
      DcamMode_ = MODE_1024x768_YUV422;
      DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_1; 
      BIASDOUT(D_DCAM,"Using native XGA");
    }
    else if (Width_ == 800 && Height_ == 600) {
      DcamMode_ = MODE_800x600_YUV422;
      DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_1; 
    }
    else if (Width_ == 640 && Height_ == 480) {
      DcamMode_ = MODE_640x480_YUV422;
      DcamFormat_ = FORMAT_VGA_NONCOMPRESSED; 
    }
    else if (Width_ == 320 && Height_ == 240) {
      DcamMode_ = MODE_320x240_YUV422;
      DcamFormat_ = FORMAT_VGA_NONCOMPRESSED; 
    }
    else   res = -1;// Format 7 required,
    break;
  case ImageBase::CM_YUV411: 
    ColorChannels_ = 2;
    //     if (Width_ == 1600 && Height_ == 1200) {
    //       DcamMode_ = MODE_1600x1200_YUV411;
    //       DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_1; 
    //     }
    //     else if (Width_ == 1280 && Height_ == 960) {
    //       DcamMode_ = MODE_1280x960_YUV41;
    //       DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_1; 
    //     }
    //     else if (Width_ == 1024 && Height_ == 768) {
    //       DcamMode_ = MODE_1024x768_YUV411;
    //       DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_1; 
    //       BIASDOUT(D_DCAM,"Using native XGA");
    //     }
    //     else if (Width_ == 800 && Height_ == 600) {
    //       DcamMode_ = MODE_800x600_YUV41;
    //       DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_1; 
    //     }
    //     else 
    if (Width_ == 640 && Height_ == 480) {
      DcamMode_ = MODE_640x480_YUV411;
      DcamFormat_ = FORMAT_VGA_NONCOMPRESSED; 
    }
    else   res = -1;// Format 7 required,
    break;
  case ImageBase::CM_Grey:
  case ImageBase::CM_Bayer_RGGB:
  case ImageBase::CM_Bayer_GBRG:
  case ImageBase::CM_Bayer_GRBG:
  case ImageBase::CM_Bayer_BGGR:
    ColorChannels_ = 1;
    if (Width_ == 1600 && Height_ == 1200) {
      DcamMode_ = MODE_1600x1200_MONO;
      DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_2; 
    }
    else if (Width_ == 1024 && Height_ == 768) {
      DcamMode_ = MODE_1024x768_MONO;
      DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_1; 
    }
    else if (Width_ == 1280 && Height_ == 960) {
      DcamMode_ = MODE_1280x960_MONO;
      DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_1; 
    }
	
    else if (Width_ == 640 && Height_ == 480 && 
             CamModel_!=CAMERA_PTGREY_DRAGONFLY_HIRES) {
      DcamMode_ = MODE_640x480_MONO;
      DcamFormat_ = FORMAT_VGA_NONCOMPRESSED; 
      BIASDOUT(D_DCAM,"Format "<<FORMAT_VGA_NONCOMPRESSED<<" Mode: "
               <<" 640x480_MONO");
    }
    else return -1;
    break;
  case ImageBase::CM_RGB:
    if (Width_ == 640 && Height_ == 480) {
      DcamMode_ = MODE_640x480_RGB;
      DcamFormat_ = FORMAT_VGA_NONCOMPRESSED;
    }
    else res =  -1;
    break;
  default:
    BIASERR("VideoSource_DCAM::UserSizeToMode(): Unsupported ColorModel_"
            <<ColorMode_);
    res =  -2;
  }
  return res;
}


int VideoSource_DCAM::UserSizeToMode_16bit_()
{
  int res = 0;
  switch (ColorMode_) {
  case ImageBase::CM_Grey:
  case ImageBase::CM_Bayer_RGGB:
  case ImageBase::CM_Bayer_GBRG:
  case ImageBase::CM_Bayer_GRBG:
  case ImageBase::CM_Bayer_BGGR :  
    if (Width_ == 640 && Height_ == 480) {
      ColorChannels_ = 1;
      DcamMode_ = MODE_640x480_MONO16;
      DcamFormat_ = FORMAT_VGA_NONCOMPRESSED; 
      BIASDOUT(D_DCAM,"Format "<<FORMAT_VGA_NONCOMPRESSED<<" Mode: "
	       <<" 640x480_MONO16");
    } else 
      if (Width_ == 1024 && Height_ == 768) {
	ColorChannels_ = 1;
	DcamMode_ = MODE_1024x768_MONO16;
	DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_1; 
	BIASDOUT(D_DCAM,"Format "<<FORMAT_VGA_NONCOMPRESSED<<" Mode: "
		 <<" 1024x768_MONO16");
      } 
      else if (Width_ == 1280 && Height_ == 960) {
	ColorChannels_ = 1;
	DcamMode_ = MODE_1280x960_MONO16;
	DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_2; 
      }
      else if (Width_ == 1600 && Height_ == 1200) {
	ColorChannels_ = 1;
	DcamMode_ = MODE_1600x1200_MONO16;
	DcamFormat_ = FORMAT_SVGA_NONCOMPRESSED_2; 
      }
      else res = -1;
    break;
  default: 
    BIASERR("VideoSource_DCAM::UserSizeToMode(): Unsupported ColoModel_ for 16bpp:"<<ColorMode_);
    res =  -2;
  }
  return res;
}



unsigned int VideoSource_DCAM::CalcBytesPerPacket()
{
  unsigned int BpP=0;
  unsigned int QuadletsPerPacket;
  
  int QuadletsPerFrame ;  
  double TransmissionTimePerFrame;
  double MaxCyclesPerFrame;
  unsigned int CyclesPerFrame;

  if (FramesPerSecond_ == 0.0){
    BIASERR("VideoSource_DCAM::CalcBytesPerPacket(): FramesPerSecond_ is 0!")
      return 0;
  }
  
  
  dc1394_query_format7_packet_para(IEEE1394Handle_[IEEE1394Card_], 
				   DcamNodes_[SelectedCam_], 
				   MODE_FORMAT7_0, 
				   &MinBpP_, &MaxBpP_);

  BIASDOUT(D_DCAM_F7,"MinBpP: "<<MinBpP_<<" MaxBpP: "<<MaxBpP_);
 
  QuadletsPerFrame = Width_ * Height_ * ColorChannels_ / 4;
  // time to transmit one frame to reach this fps in [ms]
  float fps = FramesPerSecond_;
  // evil hack to let 15 fps work
  if (fps>14.9 && fps < 20.0) fps += 2.0;
  TransmissionTimePerFrame = 1000.0/fps;
  // each 125usec one packet is transmitted
  MaxCyclesPerFrame = TransmissionTimePerFrame / 0.125;
  CyclesPerFrame = (int)MaxCyclesPerFrame;
  
  

  QuadletsPerPacket = QuadletsPerFrame / CyclesPerFrame;

  while ((QuadletsPerFrame % QuadletsPerPacket)!=0) {
    CyclesPerFrame--;
    QuadletsPerPacket = QuadletsPerFrame /CyclesPerFrame;
  }
  

   BIASDOUT(D_DCAM_F7,"FramesPerSecond: "<<FramesPerSecond_);
   BIASDOUT(D_DCAM_F7,"TransmissionTimePerFrame: "<<TransmissionTimePerFrame);
   BIASDOUT(D_DCAM_F7,"MaxCyclesPerFrame: "<<MaxCyclesPerFrame);
   BIASDOUT(D_DCAM_F7,"CyclesPerFrame: "<<CyclesPerFrame);
   BIASDOUT(D_DCAM_F7,"QuadletsPerFrame: "<<QuadletsPerFrame);
   BIASDOUT(D_DCAM_F7,"QuadletsPerPacket: "<<QuadletsPerPacket);
   BIASDOUT(D_DCAM_F7,"Resulting theoretical fps: "
	     <<1000/((float)CyclesPerFrame * 0.125));
  BpP = QuadletsPerPacket * 4;
  if ( BpP < MinBpP_)  BytesPerPacket_ = MinBpP_;
  else if (BpP> MaxBpP_) BytesPerPacket_ = MaxBpP_;
  else  BytesPerPacket_ = BpP;



  return 0;
}



int  VideoSource_DCAM::ApplyBytesPerPacket()
{
  unsigned int realbpp;  
  
  dc1394_set_format7_byte_per_packet(IEEE1394Handle_[IEEE1394Card_], 
				     DcamNodes_[SelectedCam_], 
				     MODE_FORMAT7_0,BytesPerPacket_);
  usleep(10000);
  dc1394_query_format7_byte_per_packet(IEEE1394Handle_[IEEE1394Card_],
				       DcamNodes_[SelectedCam_],
				       DcamMode_, &realbpp);
  
  TransmissionTime_ =(int) (((Width_*Height_*ColorChannels_)/realbpp) * 0.125);  
  BIASDOUT(D_DCAM_F7,"requested BpP: "<< BytesPerPacket_ <<" got BpP: "<<realbpp
	  <<" TransmissionTime_: "<<TransmissionTime_<<" ms")
    BytesPerPacket_ = realbpp;

  return 0;
}



 




int VideoSource_DCAM::InitCam()
{
  BIASDOUT(D_DCAM,"VideoSource_DCAM::InitCam("<<SelectedCam_<<")");
  if (DcamFormat_ == 0) {
    BIASERR("VideoSource_DCAM::InitCam() : no Format/mode selected");
    return -1;
  }
  StopIso();
  
  if (FramesPerSecond_ == 30.0) DcamFps_ = FRAMERATE_30;
  if (FramesPerSecond_ == 15.0) DcamFps_ = FRAMERATE_15;
  if (FramesPerSecond_ == 7.5) DcamFps_ = FRAMERATE_7_5;
  if (FramesPerSecond_ == 3.75) DcamFps_ = FRAMERATE_3_75;
  if (FramesPerSecond_ == 1.875) DcamFps_ = FRAMERATE_1_875;
  if (DcamFps_ == 0) {
    BIASERR("VideoSource_DCAM::InitCam(): Framerate "<<FramesPerSecond_
            <<" not supported");
    return -1;
  }
  BIASDOUT(D_DCAM,"VideoSource_DCAM::InitCam(): Framerate "<<FramesPerSecond_);
  DcamIsoChannel_ = SelectedCam_ + IEEE1394Card_*10;
  BIASDOUT(D_DCAM,"  Iso channel: "<<DcamIsoChannel_);
  
  int res;
  //  const int extar_buffering = 0;
  const int drop_frames = 1; 
  if (DcamFormat_ == FORMAT_SCALABLE_IMAGE_SIZE) {
    if (BytesPerPacket_ == 0 )
      if (CalcBytesPerPacket()<0)   return -1;
    if (UseDMA_) {
      BIASDOUT(D_DCAM,"Using DMA!");
      DcamDMABuffers_ = MAX_DMABUFFERS; 
      DcamCapture_.drop_frames = 1;
      res = dc1394_dma_setup_format7_capture(IEEE1394Handle_[IEEE1394Card_],
                                             DcamNodes_[SelectedCam_], 
                                             DcamIsoChannel_, DcamMode_, 
                                             SPEED_400, BytesPerPacket_,
                                             Left_, Top_, 
                                             Width_, Height_,
                                             DcamDMABuffers_,
                                             //extra_buffering,
                                             // to determine the version, this indicats >=0.9.3
#ifdef DC1394_NO_FRAME  
                                             drop_frames,
                                             DefaultDevice_,
#endif
                                             &DcamCapture_);

      if (res !=DC1394_SUCCESS) {
        BIASERR("Can not set DCAM Camera to requested parameters "<<res);
      }
      else {
        unsigned width, height;
        dc1394_query_format7_image_size(IEEE1394Handle_[IEEE1394Card_], 
                                        DcamNodes_[SelectedCam_], 
                                        DcamMode_, &width, &height);
        Width_=width;
        Height_=height;
        BIASDOUT(D_DCAM,"Format 7 result: "<<Width_<<"x"<<Height_);
      }
    } else {
      res = dc1394_setup_format7_capture(IEEE1394Handle_[IEEE1394Card_],
                                         DcamNodes_[SelectedCam_], 
                                         DcamIsoChannel_, DcamMode_, 
                                         SPEED_400, BytesPerPacket_,
                                         Left_, Top_, 
                                         Width_, Height_,
                                         &DcamCapture_);

      if (res !=DC1394_SUCCESS) {
        BIASERR("Can not set DCAM Camera to requested parameters "<<res);
      }
    }
    ApplyBytesPerPacket( );

  } else { // setup one standard mode
    if (UseDMA_) {   
      DcamDMABuffers_ = MAX_DMABUFFERS;   
      res = dc1394_dma_setup_capture(IEEE1394Handle_[IEEE1394Card_],
                                     DcamNodes_[SelectedCam_], 
                                     DcamIsoChannel_, 
                                     DcamFormat_, DcamMode_, SPEED_400, 
                                     DcamFps_, DcamDMABuffers_, 
                                     //extra_buffering,
                                     drop_frames, 
                                     DefaultDevice_,
                                     &DcamCapture_);
      //       cout <<"hier fehlt setbytesperpacket()"<<endl; 
      //       cout <<DcamCapture_.
      BytesPerPacket_ = 0;
      if (res !=DC1394_SUCCESS) 
        BIASERR("Can not set DCAM Camera to requested parameters ");
      
    } else {
      res = dc1394_setup_capture(IEEE1394Handle_[IEEE1394Card_],
                                 DcamNodes_[SelectedCam_], 
                                 DcamIsoChannel_, 
                                 DcamFormat_, DcamMode_, SPEED_400, 
                                 DcamFps_, &DcamCapture_);
      if (res !=DC1394_SUCCESS) 
        BIASERR("Can not set DCAM Camera to requested parameters ");
      
    }
  }
  // PrintFormatAndMode_(); cout <<endl;
  if (res !=DC1394_SUCCESS) {
    return -1;
  }
  BIASDOUT(D_DCAM,"Initialization successfull");
  CompleteInitialized_ = true;
  return 0;
  
}


int VideoSource_DCAM::StopIso() 
{
  int res;

  //  dc1394bool_t  iso_is_on; 
    
//      dc1394_get_iso_status(handle, dcam_cam[i].node, &iso_is_on); 
//      if (iso_is_on) cout<<"Iso is on"<<endl; 
//      else cout << "Iso is off"<<endl; 
    
  res =dc1394_stop_iso_transmission(IEEE1394Handle_[IEEE1394Card_],DcamNodes_[SelectedCam_]);
  
  if (res !=DC1394_SUCCESS){ 
    BIASERR("couldn't stop ISO "); 
    return -1;
  } 
  
//        dc1394_get_iso_status(handle, dcam_cam[i].node, &iso_is_on); 
//        if (iso_is_on) cout<<"Iso is on"<<endl; 
//        else cout << "Iso is off"<<endl; 
  
  return 0;
}



int VideoSource_DCAM::StartIso() 
{
  int res;
  int trial;
  dc1394bool_t  iso_is_on; 
  BIASDOUT(D_DCAM,"VideoSource_DCAM::StartIso("<<SelectedCam_<<")")
  dc1394_get_iso_status(IEEE1394Handle_[IEEE1394Card_],DcamNodes_[SelectedCam_], &iso_is_on);
  trial = 1;

  while (!iso_is_on) {
    res =dc1394_start_iso_transmission(IEEE1394Handle_[IEEE1394Card_],DcamNodes_[SelectedCam_]);
    if (res !=DC1394_SUCCESS){ 
      BIASERR("couldn't start ISO");
      return -1;
    } 
    dc1394_get_iso_status(IEEE1394Handle_[IEEE1394Card_],DcamNodes_[SelectedCam_],&iso_is_on); 
  }	
  BIASDOUT(D_DCAM,"Starting iso successfull");
  return 0;
}


int VideoSource_DCAM::AutoMode( int feature, bool on)
{
  int res;
//    cout <<"Cam "<<ID_<<" AutoMode: "<<feature<<" : "<<on<<endl;
    res = dc1394_auto_on_off(IEEE1394Handle_[IEEE1394Card_],
			     DcamNodes_[SelectedCam_],feature,
			     on );
    if (res <0) {
      BIASERR("VideoSource_DCAM::AutoMode(): "<<dc1394_feature_desc[feature]);
      return -1;
    } 
    
    return 0;
    
}
void VideoSource_DCAM::SetAutoShutter(bool b)
{
  AutoMode(FEATURE_SHUTTER,b);
}
bool VideoSource_DCAM::GetAutoShutter()
{
  int res;
  dc1394bool_t OnOrOff;
  res = dc1394_is_feature_auto(IEEE1394Handle_[IEEE1394Card_],
			       DcamNodes_[SelectedCam_],
			       FEATURE_SHUTTER, &OnOrOff);
  return (DC1394_TRUE==OnOrOff);
}

void VideoSource_DCAM::SetAutoGain(bool b)
{
  AutoMode(FEATURE_GAIN,b);
}

bool VideoSource_DCAM::GetAutoGain()
{
  int res;
  dc1394bool_t OnOrOff;
  res = dc1394_is_feature_auto(IEEE1394Handle_[IEEE1394Card_],
			       DcamNodes_[SelectedCam_],
			       FEATURE_GAIN, &OnOrOff);
  return (DC1394_TRUE==OnOrOff);
}


int VideoSource_DCAM::OnePushAuto( int feature)
{
  int res;
  if(CamModel_ == CAMERA_PTGREY_DRAGONFLY_COLOR 
     && feature == FEATURE_WHITE_BALANCE)
  {
    SoftwareWhiteBalance();
  }
  else
  {
    res = dc1394_start_one_push_operation(IEEE1394Handle_[IEEE1394Card_],
                                          DcamNodes_[SelectedCam_],feature);
    if (res <0) {
    BIASERR("VideoSource_DCAM::OnePushAuto(): "<<dc1394_feature_desc[feature]);
      return -1;
    }
  } 
  return 0;
}



bool VideoSource_DCAM::OnePushRunning(int feature)
{
  int res;
  dc1394bool_t isrunning = DC1394_TRUE;
  res = dc1394_is_one_push_in_operation(IEEE1394Handle_[IEEE1394Card_],
					DcamNodes_[SelectedCam_],feature,
					  &isrunning);
  if (res <0)
    return -1;
  else
    return isrunning;
}




int VideoSource_DCAM::ExternalTrigger(bool on)
{
  int res =0;
  if (on) {
    Trigger_ = TRIGGER_MODE_0;
    BIASDOUT(D_DCAM,"Triggermode: 0 (extern)");
    res = dc1394_set_trigger_on_off(IEEE1394Handle_[IEEE1394Card_], 
				    DcamNodes_[SelectedCam_],
				    DC1394_TRUE);
  }
  else {
    Trigger_ = TRIGGER_MODE_3;
    BIASDOUT(D_DCAM,"Triggermode: 3  (intern)");
//      res = dc1394_set_trigger_on_off(IEEE1394Handle_[IEEE1394Card_], 
//  				    DcamNodes_[SelectedCam_],
//  				    DC1394_FALSE);
  }

  
  res = dc1394_set_trigger_mode(IEEE1394Handle_[IEEE1394Card_], 
				DcamNodes_[SelectedCam_], Trigger_);

    
    if (res != DC1394_SUCCESS)	  {
      BIASERR("unable to set camera "<<ID_<<" to trigger mode ");
          return -1;
    }
    
  return 0;
  
}



unsigned int VideoSource_DCAM::GetFeature(int feature)
{
  unsigned int value;
  if (SupportedFeature_[feature-FEATURE_MIN])
    dc1394_get_feature_value(IEEE1394Handle_[IEEE1394Card_], DcamNodes_[SelectedCam_], 
			     feature,&value);
  else {
    BIASERR("Feature "<<feature<<" not supported by camera "<<ID_);
    return 0;
    
  }
  
  return value;
}


int VideoSource_DCAM::SetFeature(int feature, unsigned int value)
{
  if (SupportedFeature_[feature-FEATURE_MIN]) {
//      cout <<"Setting cam "<<ID_<<" Feature: "<<feature<<" to "<<value<<endl;
    dc1394_set_feature_value(IEEE1394Handle_[IEEE1394Card_], 
			     DcamNodes_[SelectedCam_], 
			     feature,value);
  } else {
    BIASERR("Feature "<<feature<<" not supported by camera "<<ID_);
//     BIASASSERT(false);
    return -1;
  }
  
  return 0;
}


int VideoSource_DCAM::SetLeftTop(unsigned int left, unsigned int top)
{
  if (left<=(MaxWidth_ - Width_) && top <=(MaxHeight_ - Height_)) {
    Top_ = top;
    Left_ = left;
    UserTopLeft_ = true;
  }
  else {
    BIASERR("VideoSource_DCAM::SetLeftTop() values too large:"<<left<<"x"<<top)
    return -1;
  }		    
  
  if (DcamMode_ == MODE_FORMAT7_0 ) {
     dc1394_set_format7_image_position(IEEE1394Handle_[IEEE1394Card_], 
					DcamNodes_[SelectedCam_],DcamMode_,
					Left_,  Top_);
    }
 
       BIASDOUT(D_DCAM_F7,"New F7 Pos: "<<Left_<<"x"<<Top_);

  return 0;
}



// exptime is given in seconds JW)
int VideoSource_DCAM::SetShutter(float exptime)
{
  int value=0;
  switch (CamModel_) {
  case CAMERA_SONY_DFW_X700:
    if ((exptime <= 16.05120063) && (exptime >= 0.13376))
      value = 2049 - int(round(exptime*28636363.0/(1197.0*800.0*2.0)));
    else
      if ((exptime <=  0.13376) && (exptime > 0.067)){
        value = 2048;
      } else if ((exptime <= 0.067) && (exptime >= 0.00012)) {
        value = (int) (2848.4 - round(28636363.0 * exptime / (1197.0*2.0)));
      } else if ((exptime <= 0.00012) && (exptime >= 0.00004)) {
        value = 2848;
      } else {
        BIASERR("Exposuretime out of bounds: "<<exptime);
        return -1;
      }
    break;
  case CAMERA_SONY_DFW_V500:
     if (exptime <=5.0 && exptime >=(1.0/30) )
       value =  0x800 - (int)(30 * exptime);
     else
      if (exptime <(1.0/30) && exptime >=(1.0/10000.0) ) 
	value = (int) (2573.6 - 15734.3 * exptime);
      else {
	BIASERR("Exposuretime out of bounds: "<<exptime);
	return -1;
      }
    break;
  case CAMERA_PTGREY_DRAGONFLY_BW:
  case CAMERA_PTGREY_SCORPION_COLOR_1600: 
  case CAMERA_PTGREY_DRAGONFLY_HIRES:
  case CAMERA_PTGREY_DRAGONFLY_COLOR:
  case CAMERA_PTGREY_DRAGONFLY2_COLOR_HIRES:
  case CAMERA_PTGREY_FIREFLY_BW: 
  case CAMERA_PTGREY_FIREFLY_COLOR:
    dc1394_absolute_setting_on_off(IEEE1394Handle_[IEEE1394Card_], 
				   DcamNodes_[SelectedCam_], FEATURE_SHUTTER,
				   1);
    dc1394_set_absolute_feature_value(IEEE1394Handle_[IEEE1394Card_], 
				      DcamNodes_[SelectedCam_], 
				      FEATURE_SHUTTER,exptime);
    return 0;
    break;
  default:
    BIASERR("Sorry your Camera Model is not supported.");
    value = -1;
  }
  if (value >0)
    return SetFeature(FEATURE_SHUTTER, value);
  else
    return -1;
  
}


float VideoSource_DCAM::GetShutter()
{
  float exptime=0.0;
  int shutter;

  shutter = GetFeature(FEATURE_SHUTTER);
  switch (CamModel_) {
    // see technical ,manual. p. 27 (JW)
  case CAMERA_SONY_DFW_X700:
    if (shutter<2048)
      exptime = ( (float)(2048-shutter+1) * (1197.0*800.0*2.0) 
                  / (28.636363*1000000.0) );
    else
      // OK, linear dependency to intensity (apetersen, jw 11/2004)
      if (shutter <2848)
        exptime =( (2848.4 - float(shutter)) * (1197.0*2.0)
                   / (28.636363*1000000.0) );
      else
        exptime= 50.0/1000000.0;
    break;
  case CAMERA_SONY_DFW_V500:
    if (shutter<0x800)
      exptime = ( (float)(shutter-2048) / 30.0 );
    else
      exptime =( (float)(shutter-2573.6)/-15734.3 );
    break;
  case CAMERA_PTGREY_DRAGONFLY_BW:
  case CAMERA_PTGREY_SCORPION_COLOR_1600:
  case CAMERA_PTGREY_DRAGONFLY_HIRES:
  case CAMERA_PTGREY_DRAGONFLY_COLOR: 
  case CAMERA_PTGREY_DRAGONFLY2_COLOR_HIRES:
  case CAMERA_PTGREY_FIREFLY_BW: 
  case CAMERA_PTGREY_FIREFLY_COLOR:
      dc1394_query_absolute_feature_value(IEEE1394Handle_[IEEE1394Card_], 
				DcamNodes_[SelectedCam_], FEATURE_SHUTTER, &exptime);

    break;
      default:
    exptime = -1.0;
  }
  
  return exptime;
}



int VideoSource_DCAM::SetGain(float gain) // gain is in dB
{
  int value;
  int min;
  int max;
  int res = 0;
  switch (CamModel_ ) {
  case CAMERA_SONY_DFW_X700:
    min = 0x800;
    max = 0x8b4;
    value =  min + (int) (gain*(float)(max-min));  
    res = SetFeature(FEATURE_GAIN,value);
    break;
  case CAMERA_SONY_DFW_V500:
    min = 0x00;
    max = 0x12;
    value =  min + (int) (gain*(float)(max-min));  
    res = SetFeature(FEATURE_GAIN,value); 
    break;
  case CAMERA_PTGREY_DRAGONFLY_BW:
  case CAMERA_PTGREY_SCORPION_COLOR_1600:
  case CAMERA_PTGREY_DRAGONFLY_HIRES:
  case CAMERA_PTGREY_DRAGONFLY_COLOR:  
  case CAMERA_PTGREY_DRAGONFLY2_COLOR_HIRES:
  case CAMERA_PTGREY_FIREFLY_BW: 
  case CAMERA_PTGREY_FIREFLY_COLOR:
    dc1394_absolute_setting_on_off(IEEE1394Handle_[IEEE1394Card_], 
                                   DcamNodes_[SelectedCam_], FEATURE_GAIN,
                                   1);
    dc1394_set_absolute_feature_value(IEEE1394Handle_[IEEE1394Card_], 
                                      DcamNodes_[SelectedCam_], FEATURE_GAIN,
                                      gain);
    break;
  default: 
    BIASERR("Sorry your Camera Model is not supported.");
    value = -1;
    break;
  }
  return res;
}



float VideoSource_DCAM::GetGain()
{
  int value;
  int min;
  int max;
  float gain=0.0; // gain is in dB

  switch (CamModel_ ) {
  case CAMERA_SONY_DFW_X700:
    min = 0x800;
    max = 0x8b4;
    value = GetFeature(FEATURE_GAIN);
    cout<<"Value in GetGain from GetFeature(...) :"<<value<<endl;;
    gain = (float)(value-min) / (float)(max-min);
    // translate gain to dB
    break;
  case CAMERA_SONY_DFW_V500:
    min = 0x00;
    max = 0x12;
    value = GetFeature(FEATURE_GAIN);
    cout<<"Value in GetGain from GetFeature(...) :"<<value<<endl;;
    gain = (float)(value-min) / (float)(max-min);
    // translate gain to dB
    break; 
  case CAMERA_PTGREY_DRAGONFLY_BW:
  case CAMERA_PTGREY_SCORPION_COLOR_1600:
  case CAMERA_PTGREY_DRAGONFLY_HIRES:
  case CAMERA_PTGREY_DRAGONFLY_COLOR:
  case CAMERA_PTGREY_DRAGONFLY2_COLOR_HIRES:
  case CAMERA_PTGREY_FIREFLY_BW: 
  case CAMERA_PTGREY_FIREFLY_COLOR:
        // to determine the version, this indicats >=0.9.3
#ifdef DC1394_NO_FRAME
      dc1394_query_absolute_feature_value(IEEE1394Handle_[IEEE1394Card_], 
					  DcamNodes_[SelectedCam_], 
					  FEATURE_GAIN, &gain);
#endif
    break;
  default:
    gain = -1.0;
    break;
  }
  return gain;
  
}

int VideoSource_DCAM::SetGammaMode(unsigned char g)
{
  int res;
  res =SetFeature(FEATURE_GAMMA,g +128);
  if (res <0) return res;
  return 0;
}

unsigned char VideoSource_DCAM::GetGammaMode()
{
  int value;
  value = GetFeature(FEATURE_GAMMA);
  return value - 128;
}



int VideoSource_DCAM::SetWhiteBalance(float rvalue,float bvalue)
{
  // rvalue and bvalue are in range -0.5 to 0.5  
  int ibval=0,irval=0;
  switch (CamModel_)
  {
  case CAMERA_SONY_DFW_X700:
  case CAMERA_SONY_DFW_V500:
    // the values for sony cameras are from 0x00 to 0xFF,
    // scale to 0-255 in integer (cmenk)
    ibval = (int)rint((bvalue + 0.5f ) * 255.0f);
    irval = (int)rint((rvalue + 0.5f ) * 255.0f);
    if (ibval > 255) ibval = 255;
    if (irval > 255) irval = 255;
    if (ibval < 0) ibval=0;
    if (irval < 0) irval=0; 
    break;
  case CAMERA_PTGREY_DRAGONFLY_COLOR:
  case CAMERA_PTGREY_DRAGONFLY2_COLOR_HIRES:
  case CAMERA_PTGREY_SCORPION_COLOR_1600: // incorrect! HACk jw
    
    ibval = (unsigned int)rint(bvalue*32)+32;
    irval = (unsigned int)rint(rvalue*32)+32;
    if (ibval >63) ibval=63;
    if (irval >63) irval=63;
    if (ibval <0) ibval=0;
    if (irval <0) irval=0;
    //cerr << irval <<", "<<ibval<<endl;
    break;
  default:
    BIASERR("unknown CamModel_: "<<CamModel_);
    return -5;
    break;
  }
  
  return (dc1394_set_white_balance(IEEE1394Handle_[IEEE1394Card_],
                                   DcamNodes_[SelectedCam_], ibval, irval)
          ==DC1394_SUCCESS)?(0):(-2);  
}




int  VideoSource_DCAM::SetCamera(CameraParameterSet camparam)
{
  int res=0;
  BIASDOUT(D_DCAM,"Entering SetCamera()");
  
  if (camparam.AutoShutterChanged) {
    res = AutoMode(FEATURE_SHUTTER, camparam.AutoShutter);
    if (res <0) return res;
  }
  if (camparam.AutoGainChanged) {
    res =AutoMode(FEATURE_GAIN, camparam.AutoGain);
    if (res <0) return res;
  }
  if (camparam.AutoWhiteBalanceChanged) {
    res =AutoMode(FEATURE_WHITE_BALANCE, camparam.AutoWhiteBalance);
    if (res <0) return res;
  }

  if (camparam.ShutterChanged) {
    res = SetShutter(camparam.Shutter);
    if (res <0) return res;
  }
  if (camparam.GainChanged)  {
    res = SetGain(camparam.Gain);
    if (res <0) return res;
  }
  if (camparam.WhiteBalanceChanged)  {
    res =SetFeature(FEATURE_WHITE_BALANCE,camparam.WhiteBalance);
    if (res <0) return res;
  }
  if (camparam.SharpenChanged)  {
    res =SetFeature(FEATURE_SHARPNESS,camparam.Sharpen);
    if (res <0) return res;
  }
  if (camparam.GammaModeChanged) { 
    res =SetGammaMode(camparam.GammaMode);
    if (res <0) return res;
  }
  if (camparam.OpticalFilterChanged) { 
    // double the call, to cope with ignorant cams
    res =SetFeature(FEATURE_OPTICAL_FILTER,camparam.OpticalFilter);
   res =SetFeature(FEATURE_OPTICAL_FILTER,camparam.OpticalFilter);
    if (res <0) return res;
  }
  if (camparam.WhiteBalanceOnePush) {
    res =OnePushAuto(FEATURE_WHITE_BALANCE);
    if (res <0) return res;
  }
  
  
  return res;
}



CameraModel
VideoSource_DCAM::GetModelForEuid(const u_int64_t euid)
{
  //  AddDebugLevel(D_DCAM);
  unsigned int vendor = euid>>40 ;
  unsigned int model = (euid >>12) & 0xFF;
  unsigned int model_lo = (euid) & 0xFFF;
  BIASDOUT(D_DCAM,"EUID:"<<hex<<euid<<" vendorid: "<<vendor<<" modelid: "<<model<<dec);
  
  switch (vendor) {
    // Sony
  case 0x080046: 
    { 
      BIASDOUT(D_DCAM,"Vendor: Sony")
        switch (model) {
        case 0x20: 
          BytesPerPixel_ = 1.0;
          ColorMode_ = ImageBase::CM_UYVY422;
          return CAMERA_SONY_DFW_V500;
        case 0x60:
          BytesPerPixel_ = 1.0;
          ColorMode_ = ImageBase::CM_UYVY422;
          return CAMERA_SONY_DFW_X700;
        default:
          return CAMERA_UNKNOWN;
        }
    }
  case 0x00b09d: // PtGey
    {
      BIASDOUT(D_DCAM,"Vendor: PointGrey Research")
        cout<<"Model is: "<<model<<endl;
        switch (model) {
        case 0xcc:
          BIASDOUT(D_DCAM,"Model: Dragonfly BW LowRes");
          ColorMode_ = ImageBase::CM_Grey;
          Shift16_ = 6; // 16 bits - 10 significant bits
          return CAMERA_PTGREY_DRAGONFLY_BW;
        case 0x4c:
	  BIASDOUT(D_DCAM,"Model: Dragonfly Color LowRes");
          ColorMode_ = ImageBase::CM_Bayer_RGGB;
          Shift16_ = 6; // 16 bits - 10 significant bits
          return CAMERA_PTGREY_DRAGONFLY_COLOR;
        case 0xff:
        case 0xfe:
          BIASDOUT(D_DCAM,"Model: Dragonfly B&W HiRes");
          ColorMode_ = ImageBase::CM_Grey;
          Shift16_ = 6; // 16 bits - 10 significant bits
          return CAMERA_PTGREY_DRAGONFLY_HIRES;
        case 0xf5:
        case 0x28:
          BIASDOUT(D_DCAM,"Model: Scorpion Color 1600");
          // 	ColorMode_ = ImageBase::CM_Bayer_GBRG; 
          // ok, RGGB seems to be right, but did it changed ? friso
          ColorMode_ = ImageBase::CM_Bayer_RGGB; 
          Shift16_ = 4; // 16 bits - 12 significant bits

          return CAMERA_PTGREY_SCORPION_COLOR_1600;
	case 0x18:
          BIASDOUT(D_DCAM,"Model: FireFly Color");
          ColorMode_ = ImageBase::CM_Bayer_GBRG; 
	  return CAMERA_PTGREY_FIREFLY_COLOR;
	case 0x16:
          BIASDOUT(D_DCAM,"Model: FireFly BW");
          ColorMode_ = ImageBase::CM_Grey; 
      	  return CAMERA_PTGREY_FIREFLY_BW;
        case 0xfb:
          if (model_lo == 0x355) {
            BIASDOUT(D_DCAM,"Model: GrassHopper Color");
            ColorChannels_ = 3;
            ColorMode_ = ImageBase::CM_RGB;
            return CAMERA_PTGREY_GRASSHOPPER_COLOR;
          } else {
            BIASDOUT(D_DCAM,"Model: DragonFly2 HIRES Color");
            ColorChannels_ = 3;
            ColorMode_ = ImageBase::CM_RGB;
            return CAMERA_PTGREY_DRAGONFLY2_COLOR_HIRES;
          }
        default:
          BIASDOUT(D_DCAM,"Model: Unknown");
          return CAMERA_UNKNOWN;
        }
    }
  case 0x081443: {// UniBrain
          BIASDOUT(D_DCAM,"Vendor: Unibrain");

	  switch (model) {
	  case 0x37:
	    BIASDOUT(D_DCAM,"Model: Fire-i");
	    return CAMERA_UNIBRAIN_FIREI;
	  }
  }
    break;
    // unknown vendorid
  default:
    return CAMERA_UNKNOWN;
  }
  return CAMERA_UNKNOWN;
} 

void VideoSource_DCAM::PrintFormatAndMode_()
{
//   cout <<"Internal Format: "<<DcamFormat_<<" Mode: "<< DcamMode_<<endl;
  int f = DcamFormat_ - FORMAT_VGA_NONCOMPRESSED;
  int m=0;
  switch (f) {
  case 0: m = DcamMode_ - MODE_FORMAT0_MIN; break;
  case 1: m = DcamMode_ - MODE_FORMAT1_MIN; break;
  case 2: m = DcamMode_ - MODE_FORMAT2_MIN; break;
  case 7: m = DcamMode_ - MODE_FORMAT7_MIN; break;
  }
  cout <<"Format: "<<f<<" Mode: "<<m;
}

int VideoSource_DCAM::UserSizeToFormat7_()
{
  if ((unsigned int)Width_ % UnitWidth_ >0) {
    BIASERR("VideoSource_DCAM::UserSizeToMode(): width of "<<Width_
	    <<" not multiple of: "<<UnitWidth_);
    return -1;
  }
  if ((unsigned int)Height_ % UnitHeight_ >0) {
    BIASERR("VideoSource_DCAM::UserSizeToMode(): height of "<<Height_
	    <<" not multiple of: "<<UnitHeight_);
    return -1;
  }
  if (!UseBinning_) {
    DcamMode_ = MODE_FORMAT7_0;
    DcamFormat_ = FORMAT_SCALABLE_IMAGE_SIZE;
    if (!UserTopLeft_) {
      Left_ = (MaxWidth_ - Width_) / 2;
      Top_ = (MaxHeight_ - Height_) / 2;
    }
    BIASDOUT(D_DCAM_F7,"Using format7: "<<Width_<<"x"<<Height_<<"+"<<Left_
	     <<"+"<<Top_);
  }
  else { // Binning for PtGrey Dragonfly
    BIASDOUT(D_DCAM_F7,"Using format7 Mode 1 (Binning): "<<Width_<<"x"<<Height_
	     <<"+"<<Left_<<"+"<<Top_);
    DcamMode_ = MODE_FORMAT7_1; 
    DcamFormat_ = FORMAT_SCALABLE_IMAGE_SIZE;
    ColorChannels_ = 1;
  }
  return 0;
}

