/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef _WIN32
#pragma comment( lib, "libusbSR.lib" )
#endif


//THE header
#include "VideoSource_SwissRanger.hh"

//std stuff
#include <string>
#include <sstream>
#include <iostream>

//BIAS stuff
#include <Image/Camera.hh>
#include <Base/Image/ImageConvert.hh>

//mesa swissranger stuff
#include <libusbSR.h>

using namespace BIAS;
using namespace std;


BIAS::VideoSource_SwissRanger::VideoSource_SwissRanger() 
: VideoSource()
{
  //for VideoSource-class
  VideoSource::SetSize(-1,-1,1);
  VideoSource::SetColorModel(BIAS::ImageBase::CM_Grey);
  Width_ = -1;
  Height_ = -1;

  bIsOpen_ = false;

  amplitudeThreshold_ = 0;
  saturationThreshold_ = 0;
  integrationTime_ = 70;
  distanceOffset_ = 0;
  modulationFrequency_ = MF_20MHz;

  depth_ = NULL;
  depthf_ = NULL;
}

BIAS::VideoSource_SwissRanger::~VideoSource_SwissRanger()
{
  if (bIsOpen_) CloseDevice();
}

int BIAS::VideoSource_SwissRanger::OpenDevice()
{
  if (bIsOpen_) {
    BIASERR("Device SwissRanger has been opened already.");
    return -1;
  }

#ifdef WIN32
  int ret = SR_OpenUSB(&srCam_);
#else
  int ret = SR_Open(&srCam_);
#endif
  if ( ret < 0) {
    BIASERR("Error initializing SwissRanger " <<ret);
    return -1;
  }

  Width_ = SR_GetCols(srCam_);
  Height_ = SR_GetRows(srCam_);
  ColorChannels_ = SR_GetBytePerPix(srCam_);

  bufSz_=SR_GetBufferSize(srCam_);

  depth_=(unsigned short*)malloc(bufSz_);
  memset(depth_,0xaf,bufSz_);
  SR_SetBuffer(srCam_,depth_,bufSz_);
  int num=Width_*Height_;
  bufFSz_ = num*3*sizeof(float);
  depthf_=(float*)malloc(bufFSz_);

  integrationTime_ = SR_GetIntegrationTime(srCam_);
  modulationFrequency_ = SR_GetModulationFrequency(srCam_);
  SetMaxDist_();

  DepthImg_.Init(Width_, Height_, 1, ImageBase::ST_unsignedchar);
  DepthImg_.SetColorModel(ImageBase::CM_Grey);
  DepthImgF_.Init(Width_, Height_, 1, ImageBase::ST_float);
  DepthImgF_.SetColorModel(ImageBase::CM_Grey);
  
  ModCoeffImg_.Init(Width_, Height_, 1, ImageBase::ST_float);
  ModCoeffImg_.SetColorModel(ImageBase::CM_Grey);
  
  for (int i=1;i<=10;i++) {
    ret = SR_Acquire(srCam_,AM_COR_FIX_PTRN|AM_COR_LED_NON_LIN);
    if (ret == 2*Width_*Height_*ColorChannels_)
      break;
    BIASERR("SwissRanger driver can not read from Camera, retry " << i << "/10");
    SR_Close(srCam_);
#ifdef WIN32
    ret = SR_OpenUSB(&srCam_);
#else
    ret = SR_Open(&srCam_);
#endif
    if ( ret < 0) {
      BIASERR("Error initializing SwissRanger " <<ret);
      return -1;
    }
  }

  if (ret != 2*Width_*Height_*ColorChannels_) {
    BIASERR("SwissRanger could not read from usb bus in 10 retries. \n"
        << "Please check cable connections and restart application.");
    return -1;
  }
  
  bIsOpen_ = true;

  return 0;
}

int BIAS::VideoSource_SwissRanger::OpenDevice(int camNr)
{
  return OpenDevice();
}

int BIAS::VideoSource_SwissRanger::CloseDevice()
{
  if (!bIsOpen_) {
    BIASERR("Cannot close device, it has not been opened successfully!!!");
    return -1;
  }

  if (SR_Close(srCam_) < 0) {
    BIASERR("Disonnection of Swiss Ranger via USB failed!");
    return -1;
  }

  free(depth_);
  free(depthf_);

  bIsOpen_ = false;

  return 0;
}

int BIAS::VideoSource_SwissRanger::PreGrab()
{
  return 0;
}


int BIAS::VideoSource_SwissRanger::PostGrab()
{
  return 0;
}

int BIAS::VideoSource_SwissRanger::InitImage(ImageBase &img) {
  if (!img.IsEmpty()){
    BIASERR("VideoSource::InitImage() should be called with uninitialized image");
    img.Release();
  }
  img.Init(Width_, Height_, ColorChannels_, ImageBase::ST_unsignedchar);
  img.SetColorModel(ImageBase::CM_Grey);
  return 0;
}

int BIAS::VideoSource_SwissRanger::InitDepthImage(ImageBase &img) {
  if (!img.IsEmpty()){
    BIASERR("VideoSource::InitImage() should be called with uninitialized image");
    img.Release();
  }
  img.Init(Width_, Height_, 1, ImageBase::ST_float);
  img.SetColorModel(ImageBase::CM_Grey);
  return 0;
}

int BIAS::VideoSource_SwissRanger::GrabSingle(Camera<unsigned char> &image)
{
  int res=0;

  if (!bIsOpen_) {
    BIASERR("GrabSingle called but device not opened!");
    return -1;
  }

  memset(depth_,0xaf,bufSz_);
  memset(depthf_,0xaf,bufFSz_);
  float *x=depthf_;
  float *y=&x[Width_*Height_];
  float *z=&y[Width_*Height_];
  int pichX=sizeof(float);
  int pichY=sizeof(float);
  int pichZ=sizeof(float);

  res=SR_Acquire(srCam_,AM_COR_FIX_PTRN|AM_COR_LED_NON_LIN);
  if (res==0) {
    BIASERR("Reading image from SwissRanger failed. Please restart application");
    return -1;
  }
  res=SR_CoordTrfFlt(srCam_, x, y, z, pichX, pichY, pichZ);

  ConvertImage_(depthf_, DepthImg_);
  ConvertFImage_(depthf_, DepthImgF_);

  ConvertRawImage_(depth_+Width_*Height_, ModCoeffImg_);

  image = DepthImg_;
  
  return 0;
}

void BIAS::VideoSource_SwissRanger::ConvertImage_(const float* data,
                                                  Image<unsigned char> &dst)
{ 
  if (dst.IsEmpty()) {
    dst.Init(Width_, Height_, ColorChannels_, ImageBase::ST_unsignedchar);
  }

  const float* srcDatap=data;
  const float* srcDepthZ = &srcDatap[2*Width_*Height_];
  unsigned char** dstData=dst.GetImageDataArray();
  int idx = 0;
  float step = (float)0xff/0xffff;
  for (unsigned int y=0; y<dst.GetHeight(); y++) {
    for (unsigned int x=0; x< dst.GetWidth(); x++) {
      float val = srcDepthZ[idx];
      // scale to 255
      val = val*(((float)0xffff)/max_dist_)*step;
      dstData[y][x] = (unsigned char)val;
      idx++;
    }
  }
}

void BIAS::VideoSource_SwissRanger::ConvertFImage_(const float* data,
                                                   Image<float> &dst)
{ 
  if (dst.IsEmpty()) {
    dst.Init(Width_, Height_, ColorChannels_, ImageBase::ST_float);
  }

  const float* srcDatap=data;
  const float* srcDepthP = &srcDatap[2*Width_*Height_];
  float** dstData=dst.GetImageDataArray();
  for (unsigned int y=0; y<dst.GetHeight(); y++) {
    for (unsigned int x=0; x< dst.GetWidth(); x++) {
      dstData[y][x] = (*srcDepthP) * 1000.0f; // for conversion m -> mm 
      srcDepthP++;
    }
  }
}

void BIAS::VideoSource_SwissRanger::ConvertRawImage_(const unsigned short* data,
                                                   Image<float> &dst)
{ 
  if (dst.IsEmpty()) {
    dst.Init(Width_, Height_, ColorChannels_, ImageBase::ST_float);
  }

  const unsigned short* srcDepthP=data;
  float** dstData=dst.GetImageDataArray();
  float step = (float)0xff / (float)0xffff;
  for (unsigned int y=0; y<dst.GetHeight(); y++) {
    for (unsigned int x=0; x< dst.GetWidth(); x++) {
      float cur = (*srcDepthP) * step;
      dstData[y][x] = cur;
      srcDepthP++;
    }
  }
}

